/* eval.c 33a */
#include "all.h"

/* eval helpers 37a */
static Valuelist*
evallist(Explist *el, Valenv *globals, Funenv *functions, Valenv *formals)
{   
    Value v;

    if (el == NULL) {
        return NULL;
    } else {
        v = eval(el->hd, globals, functions, formals);
        return mkVL(v, evallist(el->tl, globals, functions, formals));
    }
}
/* eval.c 33b */
Value
eval(Exp *e, Valenv *globals, Funenv *functions, Valenv *formals)
{
    switch (e->ty) {
    case LITERAL:
        /* evaluate [[e->u.literal]] and return the result 34a */
        return e->u.literal;
    case VAR:
        /* evaluate [[e->u.var]] and return the result 34b */
        if (isvalbound(e->u.var, formals))
            return fetchval(e->u.var, formals);
        else if (isvalbound(e->u.var, globals))
            return fetchval(e->u.var, globals);
        else
            error("unbound variable %n", e->u.var);
        assert(0);   /* not reached */
        return 0;   
    case SET:
        /* evaluate [[e->u.set]] and return the result 34c */
        {
            Value v;

            v = eval(e->u.set.exp, globals, functions, formals);

            if (isvalbound(e->u.set.name, formals))
                bindval(e->u.set.name, v, formals);
            else if (isvalbound(e->u.set.name, globals))
                bindval(e->u.set.name, v, globals);
            else
                error("set: unbound variable %n", e->u.set.name);
            return v;
        }
    case IFX:
        /* evaluate [[e->u.ifx]] and return the result 35a */
        if (eval(e->u.ifx.cond, globals, functions, formals) != 0)
            return eval(e->u.ifx.true, globals, functions, formals);
        else
            return eval(e->u.ifx.false, globals, functions, formals);
    case WHILEX:
        /* evaluate [[e->u.whilex]] and return the result 35b */
        while (eval(e->u.whilex.cond, globals, functions, formals) != 0)
            eval(e->u.whilex.exp, globals, functions, formals);
        return 0;
    case BEGIN:
        /* evaluate [[e->u.begin]] and return the result 36a */
        {
            Explist *el;
            Value v;

            v = 0;
            for (el=e->u.begin; el; el=el->tl)
                v = eval(el->hd, globals, functions, formals);
            return v;
        }
    case APPLY:
        /* evaluate [[e->u.apply]] and return the result 36b */
        {
            Fun f;

            /* make [[f]] the function denoted by [[e->u.apply.name]], or call
                                                                [[error]] 36c */
            if (!isfunbound(e->u.apply.name, functions))
                error("call to undefined function %n", e->u.apply.name);
            f = fetchfun(e->u.apply.name, functions);

            switch (f.ty) {
            case USERDEF:
                /* apply [[f.u.userdef]] and return the result 37b */
                {
                    Namelist *nl;
                    Valuelist *vl;

                    nl = f.u.userdef.formals;
                    vl = evallist(e->u.apply.actuals, globals, functions,
                                                                       formals);

                    checkargc(e, lengthNL(nl), lengthVL(vl));

                    return eval(f.u.userdef.body, globals, functions, mkValenv(
                                                                       nl, vl));
                }
            case PRIMITIVE:
                /* apply [[f.u.primitive]] and return the result 38a */
                {
                    Valuelist *vl;

                    vl = evallist(e->u.apply.actuals, globals, functions,
                                                                       formals);

                    if (f.u.primitive == strtoname("print"))
                        /* apply [[print]] to [[vl]] and return 38b */
                        {
                            Value v;

                            checkargc(e, 1, lengthVL(vl));
                            v = nthVL(vl, 0);
                            print("%v\n", v);
                            return v;
                        }
                    else
                        /* apply arithmetic primitive to [[vl]] and return 39a
                                                                              */
                        {
                            const char *s;
                            Value v, w;

                            checkargc(e, 2, lengthVL(vl));
                            v = nthVL(vl, 0);
                            w = nthVL(vl, 1);

                            s = nametostr(f.u.primitive);
                            assert(strlen(s) == 1);
                            switch (s[0]) {
                            case '<':
                                return v < w;
                            case '>':
                                return v > w;
                            case '=':
                                return v == w;
                            case '+':
                                return v + w;
                            case '-':
                                return v - w;
                            case '*':
                                return v * w;
                            case '/':
                                if (w == 0)
                                    error("division by zero in %e", e);
                                return v / w;
                            default:
                                assert(0);
                                return 0;   /* not reached */
                            }
                        }
                }
            default:
                assert(0);
                return 0;   /* not reached */
            }
        }
    default:
        assert(0);
        return 0;   /* not reached */
    }
}
/* eval.c 39b */
void
readevalprint(Topreader *reader, Valenv *globals, Funenv *functions, int echo)
{
    Toplevel *t;

    while ((t = readtop(reader)))
        topeval(t, globals, functions, echo);
}
/* eval.c 40a */
void
topeval(Toplevel *t, Valenv *globals, Funenv *functions, int echo)
{
    switch (t->ty) {
    case EXP:
        /* evaluate [[t->u.exp]] and possibly print the result 40b */
        {
            Value v;

            v = eval(t->u.exp, globals, functions, mkValenv(NULL, NULL));
            if (echo)
                print("%v\n", v);
        }
        break;
    case DEFINE:
        /* evaluate [[t->u.define]], modifying [[functions]] 41a */
        /* fail if [[t->u.define]] has duplicate formal parameters 41b */
        if (duplicatename(t->u.define.userfun.formals) != NULL)
            error(
         "Formal parameter named %n appears twice in definition of function %n",
                  duplicatename(t->u.define.userfun.formals), t->u.define.name);
        bindfun(t->u.define.name, mkUserdef(t->u.define.userfun), functions);
        if (echo)
            print("%n\n", t->u.define.name);
        break;
    case VAL:
        /* evaluate [[t->u.val]], modifying [[globals]] 41c */
        {
            Value v;

            v = eval(t->u.val.exp, globals, functions, mkValenv(NULL, NULL));
            bindval(t->u.val.name, v, globals);
            print("%v\n", v);
        }
        break;
    case USE:
        /* evaluate [[t->u.use]], possibly modifying [[globals]] and
                                                            [[functions]] 41d */
        {
            FILE *fin;
            const char *filename;

            filename = nametostr(t->u.use);
            fin = fopen(filename, "r");
            if (fin == NULL)
                error("cannot open file \"%s\"", filename);
            readevalprint(topreader(filereader(filename, fin), 0), globals,
                                                                  functions, 0);
            fclose(fin);
        }
        break;
    default:
        assert(0);
    }
}
