(* timpcore.sml 213d *)


(*****************************************************************)
(*                                                               *)
(*   ENVIRONMENTS                                                *)
(*                                                               *)
(*****************************************************************)

(* environments 186 *)
type name = string
type 'a env = (name * 'a) list
val emptyEnv = []

(* lookup and assignment of existing bindings *)
exception NotFound of name
fun find (name, []) = raise NotFound name
  | find (name, (n, v)::tail) = if name = n then v else find(name, tail)

(* adding new bindings *)
exception BindListLength
fun bind(name, v, rho) = (name, v) :: rho
fun bindList(n::vars, v::vals, rho) = bindList(vars, vals, bind(n, v, rho))
  | bindList([], [], rho) = rho
  | bindList _ = raise BindListLength
(* type declarations for consistency checking *)
val _ = op emptyEnv : 'a env
val _ = op find     : name * 'a env -> 'a
val _ = op bind     : name      * 'a      * 'a env -> 'a env
val _ = op bindList : name list * 'a list * 'a env -> 'a env
exception LeftAsExercise


(*****************************************************************)
(*                                                               *)
(*   TYPES FOR \TIMPCORE                                         *)
(*                                                               *)
(*****************************************************************)

(* types for \timpcore 202b *)
datatype ty    = INTTY | BOOLTY | UNITTY | ARRAYTY of ty
datatype funty = FUNTY of ty list * ty
(* types for \timpcore 202c *)
(* printing types for \timpcore 579a *)
fun typeString BOOLTY        = "bool"
  | typeString INTTY         = "int"
  | typeString UNITTY        = "unit"
  | typeString (ARRAYTY tau) = "(array " ^ typeString tau ^ ")"


(*****************************************************************)
(*                                                               *)
(*   LEXICAL ANALYSIS                                            *)
(*                                                               *)
(*****************************************************************)

(* lexical analysis ((mlscheme)) 571a *)
type srcloc  = string * int
fun srclocString (source, line) = source ^ ", line " ^ Int.toString line
(* lexical analysis ((mlscheme)) 571b *)
datatype par = NAME  of string
             | INT   of int
             | SHARP of char  (* #t or #f *)
             | LIST  of srcloc * par list
(* lexical analysis ((mlscheme)) 572 *)

exception Paren of char list * srcloc
   (* raise if found ) when reading input *)
exception PrematureEOF of string (* shows what we were reading ... *)
exception EOF                 (* raised by reader *)
fun read nextline curloc : string -> par list =
  let (* character-fiddling functions 573a *)
      fun isDelim c = List.exists (fn c' => c' = c) [#"(", #")", #" ", #"\t",
                                                                    #"\n", #";"]
      (* character-fiddling functions 573b *)
      fun stringReader []     = NONE
        | stringReader (h::t) = SOME (h, t)
      val strton = Int.scan StringCvt.DEC stringReader
      (* type declarations for consistency checking *)
      val _ = op strton : char list -> (int * char list) option
      (* character-fiddling functions 573c *)
      fun skip s =
        case StringCvt.skipWS stringReader s
          of #";" :: s => skip (StringCvt.dropl (fn c => c <> #"\n")
                                                                 stringReader s)
           | s => s
      (* character-fiddling functions 573d *)
      fun nonwhite s =
        case skip s
          of [] => nonwhite (explode (nextline()))
           | s  => s
      fun readmap f (result, unconsumed) = (f result, unconsumed)
      (* get is called only when s is guaranteed non-blank, non-comment, non-
                                                                        empty *)
      fun get (#"("::s)  = readmap (fn l => LIST l) (list (curloc(), s, []))
        | get (#")"::s)  = raise (Paren (s, curloc()))
        | get (#"'"::s)  = let val loc = curloc()
                               fun quote p = LIST (loc, [NAME "quote", p])
                           in  readmap quote (get (nonwhite s))
                           end
        | get s          = atom (s, [])
      and atom(c::s, r) = if isDelim c then (mkAtom(rev r), c::s) 
                          else atom(s, c::r)
        | atom([],   r) = (mkAtom(rev r), [])
      and mkAtom [#"#", #"f"] = SHARP #"f"
        | mkAtom [#"#", #"t"] = SHARP #"t"
        | mkAtom s = (case strton s of SOME (n, []) => INT n | _ => NAME (
                                                                     implode s))
      and list (loc, s, l) =
        let val next = nonwhite s
        in  let val (a, u) = get next
            in  list(loc, u, a::l)
            end handle (Paren (u, _)) => ((loc, rev l), skip u)
                     | EOF => raise PrematureEOF (implode next)
        end handle EOF => raise PrematureEOF (implode s)
      fun readline [] = []
        | readline s  = 
             let val(p, u) = get s
             in  p :: readline(skip u)
             end
(* type declarations for consistency checking *)
val _ = op read : (unit -> string) -> (unit -> srcloc) -> string -> par list
val _ = op get  : char list                      -> par                 * char
                                                                            list
val _ = op atom : char list * char list          -> par                 * char
                                                                            list
val _ = op list : srcloc * char list * par list  -> (srcloc * par list) * char
                                                                            list
  in  fn s => readline(skip (explode s)) 
  end


(*****************************************************************)
(*                                                               *)
(*   ABSTRACT SYNTAX AND VALUES FOR \TIMPCORE                    *)
(*                                                               *)
(*****************************************************************)

(* abstract syntax and values for \timpcore 202d *)
datatype value = NUM   of int
               | ARRAY of value array
(* abstract syntax and values for \timpcore 202e *)
datatype exp   = LITERAL of value
               | VAR     of name
               | SET     of name * exp
               | IFX     of exp * exp * exp
               | WHILEX  of exp * exp
               | BEGIN   of exp list
               | APPLY   of name * exp list
               (* array extensions to \timpcore's abstract syntax 215c *)
               | AGET  of exp * exp
               | ASET  of exp * exp * exp
               | AMAKE of exp * exp
               | ALEN  of exp
(* abstract syntax and values for \timpcore 202f *)
type userfun = { formals : (name * ty) list, body : exp, returns : ty }
datatype toplevel = EXP    of exp
                  | DEFINE of name * userfun
                  | VAL    of name * exp
                  | USE    of name
(* abstract syntax and values for \timpcore 203a *)
datatype func = USERDEF   of string list * exp
              | PRIMITIVE of value list -> value
(* abstract syntax and values for \timpcore 203b *)
(* printing values for \timpcore 579b *)
fun valueString (NUM n) = Int.toString n
  | valueString (ARRAY a) =
      if Array.length a = 0 then
          "[]"
      else
          let val elts = Array.foldr (fn (v, s) => " " :: valueString v :: s) [
                                                                          "]"] a
          in  String.concat ("[" :: tl elts)
          end
(* type declarations for consistency checking *)
val _ = op typeString : ty -> string
(* type declarations for consistency checking *)
val _ = op valueString : value -> string
(* abstract syntax and values for \timpcore 216a *)
exception BugInTypeChecking of string
fun toArray (ARRAY a) = a
  | toArray _         = raise BugInTypeChecking "non-array value"
fun toInt   (NUM n)   = n
  | toInt _           = raise BugInTypeChecking "non-integer value"
(* type declarations for consistency checking *)
val _ = op toArray : value -> value array
val _ = op toInt   : value -> int


(*****************************************************************)
(*                                                               *)
(*   PARSING FOR \TIMPCORE                                       *)
(*                                                               *)
(*****************************************************************)

(* parsing for \timpcore 580a *)
exception SyntaxError of string
fun syntaxErrorAt (source, line) msg =
  raise SyntaxError (source ^ ", line " ^ Int.toString line ^ ": " ^ msg)
fun parse (NAME s)  = VAR s
  | parse (INT n)   = LITERAL (NUM n)
  | parse (SHARP c) = raise SyntaxError "No # in Typed Impcore"
  | parse (LIST (loc, NAME "set"    :: args)) = parseSet loc args
  | parse (LIST (loc, NAME "if"     :: args)) = parse3 loc IFX "if" args
  | parse (LIST (loc, NAME "begin"  :: args)) = BEGIN (map parse args)
  | parse (LIST (loc, NAME "while"  :: args)) = parse2 loc WHILEX "while" args
  | parse (LIST (loc, NAME "array-get"  :: args)) = parse2 loc AGET "array-get"
                                                                            args
  | parse (LIST (loc, NAME "array-set"  :: args)) = parse3 loc ASET "array-set"
                                                                            args
  | parse (LIST (loc, NAME "array-make" :: args)) = parse2 loc AMAKE
                                                               "array-make" args
  | parse (LIST (loc, NAME "array-length" :: args)) = parse1 loc ALEN
                                                             "array-length" args
  | parse (LIST (loc, NAME h::t))  = APPLY (h, map parse t)
  | parse (LIST (loc, _ :: _)) = syntaxErrorAt loc
                                           "only named functions can be applied"
  | parse (LIST (loc, [])) = syntaxErrorAt loc "empty application ()"
and parse3 _ f name [e1, e2, e3] = f (parse e1, parse e2, parse e3)
  | parse3 loc _ name _ = syntaxErrorAt loc ("expected (" ^ name ^ " e1 e2 e3)")
and parse2 _ f name [e1, e2] = f (parse e1, parse e2)
  | parse2 loc _ name _ = syntaxErrorAt loc ("expected (" ^ name ^ " e1 e2)")
and parse1 _ f name [e1] = f (parse e1)
  | parse1 loc _ name _ = syntaxErrorAt loc ("expected (" ^ name ^ " e)")
and parseSet _ [NAME x, e] = SET (x, parse e)
  | parseSet loc _ = syntaxErrorAt loc "expected (set x e)"
(* parsing for \timpcore 580b *)
fun toplevel (LIST (loc, NAME "define" :: args)) = parseDefine loc args
  | toplevel (LIST (loc, NAME "val"    :: args)) = parseVal    loc args
  | toplevel (LIST (loc, NAME "use"    :: args)) = parseUse    loc args
  | toplevel e = EXP (parse e)
and parseDefine loc [ty, NAME name, LIST (floc, formals), body] =
      DEFINE (name, { returns = parseTy loc ty,
                      formals = map (parseFormal floc) formals, body = parse
                                                                          body})
  | parseDefine loc _ = syntaxErrorAt loc "expected (define ty f (args) body)"
and parseFormal _ (LIST (loc, [ty, NAME x])) = (x, parseTy loc ty)
  | parseFormal _ (LIST (loc, _)) = syntaxErrorAt loc "expected (ty argname)"
  | parseFormal loc _ = syntaxErrorAt loc "expected (ty argname)"
and parseTy loc (NAME "bool") = BOOLTY
  | parseTy loc (NAME "unit") = UNITTY
  | parseTy loc (NAME "int")  = INTTY
  | parseTy loc (NAME n)      =
      syntaxErrorAt loc ("Cannot recognize name " ^ n ^ " as a type")
  | parseTy loc (LIST (aloc, [NAME "array", ty])) = ARRAYTY (parseTy aloc ty)
  | parseTy loc _ = syntaxErrorAt loc "expected int, bool, unit, or (array ty)"
and parseVal _ [NAME v, e] = VAL (v, parse e)
  | parseVal loc _ = syntaxErrorAt loc "expected (val v e)"
and parseUse _ [NAME v] = USE v
  | parseUse loc _ = syntaxErrorAt loc "expected (use filename)"


(*****************************************************************)
(*                                                               *)
(*   READERS                                                     *)
(*                                                               *)
(*****************************************************************)

(* readers 576b *)
type reader = { firstline : unit -> string
              , nextline  : unit -> string
              , curloc    : unit -> string * int
              }
(* readers 576c *)
fun promptReader (firstPrompt, nextPrompt, reader) =
  let val { firstline = first, nextline = next, curloc = cur } = reader
      fun promptIn get prompt () = ( TextIO.output(TextIO.stdOut, prompt)
                                   ; TextIO.flushOut(TextIO.stdOut)
                                   ; get ()
                                   )
  in  { firstline = promptIn first firstPrompt
      , nextline  = promptIn next  nextPrompt
      , curloc    = cur
      }
  end
(* type declarations for consistency checking *)
val _ = op promptReader : string * string * reader -> reader
(* readers 577a *)
fun filereader (filename, fd)=
  let val lineNum = ref 0
      fun get () =
        let val line = TextIO.inputLine fd
        in  if size line = 0 then raise EOF else line before lineNum := !lineNum
                                                                             + 1
        end
  in  { firstline = get, nextline = get, curloc = (fn () => (filename, !lineNum)
                                                                             ) }
  end

fun stringsreader (what, l) = 
  let val buffer  = ref l
      val lineNum = ref 0
      fun get ()  = case !buffer
                      of [] => raise EOF
                       | h :: t => h before buffer := t before lineNum := !
                                                                     lineNum + 1
  in  { firstline = get, nextline = get, curloc = (fn () => (what, !lineNum)) }
  end              
(* type declarations for consistency checking *)
val _ = op filereader    : string * TextIO.instream -> reader
val _ = op stringsreader : string * string list     -> reader
(* readers 577b *)
type topreader = { buffer : toplevel list ref, reader : reader }
fun readtop (r as { buffer, reader = { nextline, firstline, curloc } }) =
  case !buffer
    of h::t => h before buffer := t
     | []   => ( buffer := map toplevel (read nextline curloc (firstline()))
               ; readtop r
               )
(* readers 577c *)
fun topreader (reader, prompt) = 
  let val reader = if prompt then promptReader ("-> ", "   ", reader) else
                                                                          reader
  in  { buffer = ref [], reader = reader }
  end
(* readers 578 *)
fun echoTag f x =
  let val line = f x
      val _ = if (String.substring (line, 0, 2) = ";#" handle _ => false) then
                print line
              else
                ()
  in  line
  end

fun echoReader { nextline=n, firstline=f, curloc=c } =
     { nextline=echoTag n, firstline=echoTag f, curloc=c } 

fun echoTop { buffer = b, reader = r } = { buffer = b, reader = echoReader r }
val topreader = fn args => echoTop (topreader args)
(* type declarations for consistency checking *)
val _ = op echoReader : reader -> reader 


(*****************************************************************)
(*                                                               *)
(*   IMPLEMENTATION OF [[USE]]                                   *)
(*                                                               *)
(*****************************************************************)

(* implementation of [[use]] 193c *)
fun use readEvalPrint filename rho =
      let val fd = TextIO.openIn filename
          val treader = topreader (filereader (filename, fd), false)
          fun writeln s = app print [s, "\n"]
      in  readEvalPrint (treader, writeln, writeln) rho
          before TextIO.closeIn fd
      end 


(*****************************************************************)
(*                                                               *)
(*   TYPE CHECKING FOR \TIMPCORE                                 *)
(*                                                               *)
(*****************************************************************)

(* type checking for \timpcore 207 *)
exception TypeError of string
fun typeof (e, globals, functions, formals) =
  let (* function [[ty]], to check the type of an expression, given $\itenvs$ 20
                                                                           8a *)
      fun ty (LITERAL v) = INTTY
      (* function [[ty]], to check the type of an expression, given $\itenvs$ 20
                                                                           8b *)
        | ty (VAR x)     = (find (x, formals) handle NotFound _ => find(x,
                                                                       globals))
      (* function [[ty]], to check the type of an expression, given $\itenvs$ 20
                                                                           8c *)
        | ty (SET (x, e)) =
               let val tau_x = ty (VAR x)
                   val tau_e = ty e
               in  if tau_x = tau_e then
                     tau_x
                   else
                     raise TypeError ("Set variable " ^ x ^ " of type " ^
                                                              typeString tau_x ^
                                      " to value of type " ^ typeString tau_e)
               end
      (* function [[ty]], to check the type of an expression, given $\itenvs$ 20
                                                                           9a *)
        | ty (IFX (e1, e2, e3)) =
            let val tau1 = ty e1
                val tau2 = ty e2
                val tau3 = ty e3
            in  if tau1 = BOOLTY then
                  if tau2 = tau3 then
                    tau2
                  else
                    raise TypeError ("In if expression, true branch has type " ^
                                     typeString tau2 ^
                                                 " but false branch has type " ^
                                     typeString tau3)
                else
                  raise TypeError ("Condition in if expression has type " ^
                                                               typeString tau1 ^
                                   ", which should be " ^ typeString BOOLTY)
            end
      (* function [[ty]], to check the type of an expression, given $\itenvs$ 20
                                                                           9b *)
        | ty (WHILEX (e1, e2)) =
            let val tau1 = ty e1
                val tau2 = ty e2
            in  if tau1 = BOOLTY then UNITTY
                else
                  raise TypeError ("Condition in while expression has type " ^
                                   typeString tau1 ^ ", which should be " ^
                                   typeString BOOLTY)
            end
      (* function [[ty]], to check the type of an expression, given $\itenvs$ 20
                                                                           9c *)
        | ty (BEGIN es) =
             let val bodytypes = map ty es
                 fun last tau []     = tau
                   | last tau (h::t) = last h t
             in  last UNITTY bodytypes
             end
      (* function [[ty]], to check the type of an expression, given $\itenvs$ 21
                                                                           0a *)
        | ty (APPLY ("=", actuals)) =
             (case map ty actuals
               of [t1, t2] =>
                   if t1 = t2 then
                     BOOLTY
                   else
                     raise TypeError (
                          "equality comparisons of values of different types " ^
                                      typeString t1 ^ " and " ^ typeString t2)
                | _ => raise TypeError "Wrong number of arguments to =")
      (* function [[ty]], to check the type of an expression, given $\itenvs$ 21
                                                                           0b *)
        | ty (APPLY ("print", [arg])) = (ty arg; UNITTY)
      (* function [[ty]], to check the type of an expression, given $\itenvs$ 21
                                                                           0c *)
        | ty (APPLY (f, actuals)) =
             let val actualtypes                     = map ty actuals
                 val FUNTY (formaltypes, resulttype) = find (f, functions)
             in  if actualtypes = formaltypes then
                   resulttype
                 else
                   raise TypeError (
                              "argument-type mismatch in call to function " ^ f)
                   (* appalling error message *)
             end
      (* function [[ty]], to check the type of an expression, given $\itenvs$ 21
                                                                            7 *)
      | ty (AGET (a, i))       = raise LeftAsExercise
      | ty (ASET (a, i, e))    = raise LeftAsExercise
      | ty (AMAKE (len, init)) = raise LeftAsExercise
      | ty (ALEN a)            = raise LeftAsExercise
(* type declarations for consistency checking *)
val _ = op typeof  : exp * ty env * funty env * ty env -> ty
val _ = op ty      : exp -> ty
  in  ty e
  end
(* type checking for \timpcore 210d *)
exception RuntimeError of string
fun topty (t, globals, functions) =
    case t
      of (* cases for type-checking type-level items in \timpcore 211a *)
           VAL (x, e) => (bind(x, typeof (e, globals, functions, emptyEnv),
                                                            globals), functions)
         (* cases for type-checking type-level items in \timpcore 211b *)
         | EXP e => (typeof (e, globals, functions, emptyEnv); (globals,
                                                                     functions))
         (* cases for type-checking type-level items in \timpcore 211c *)
         | DEFINE (name, {returns, formals, body}) =>
            let val (fnames, ftys) = ListPair.unzip formals
                val functions' = bind(name, FUNTY (ftys, returns), functions)
                val tau = typeof (body, globals, functions', bindList(fnames,
                                                                ftys, emptyEnv))
            in  if tau = returns then
                  (globals, functions')
                else
                  raise TypeError ("Body of function has type " ^ typeString tau
                                                                               ^
                                   ", which does not match declaration of " ^
                                   typeString returns)
            end
         (* cases for type-checking type-level items in \timpcore 211d *)
         | USE _ => raise BugInTypeChecking "`use' reached the type checker"
(* type declarations for consistency checking *)
val _ = op topty : toplevel * ty env * funty env -> ty env * funty env


(*****************************************************************)
(*                                                               *)
(*   CHECKING AND EVALUATION FOR \TIMPCORE                       *)
(*                                                               *)
(*****************************************************************)

(* checking and evaluation for \timpcore 212a *)
(* evaluation for \timpcore 581c *)
(* definition of [[eval]] for \timpcore 581a *)
fun eval(e, globals, functions, formals) =
  let fun toBool (NUM 0) = false
        | toBool _       = true
      fun findVar v = find (v, formals) handle NotFound _ => find (v, globals)
      fun ev(LITERAL n)          = n
        | ev(VAR x)              = !(findVar x)
        | ev(SET (x, e))         = let val v = ev e in v before findVar x := v
                                                                             end
        | ev(IFX (cond, t, f))   = if toBool (ev cond) then ev t else ev f
        | ev(WHILEX (cond, exp)) =
            if toBool (ev cond) then
                (ev exp; ev (WHILEX (cond, exp)))
            else
                NUM 0
        | ev(BEGIN es) =
            let fun b(e::es, lastval) = b(es, ev e)
                  | b(   [], lastval) = lastval
            in  b(es, NUM 0)
            end
        | ev(APPLY (f, args)) = 
            (case find (f, functions)
               of PRIMITIVE p  => p (map ev args)
                | USERDEF func => (* apply user-defined function [[func]] to
                                                                [[args]] 581b *)
                                  let val (formals, body) = func
                                      val actuals         = map (ref o ev) args
                                  (* type declarations for consistency checking
                                                                              *)
                                  val _ = op formals : name      list
                                  val _ = op actuals : value ref list
                                  in  eval(body, globals, functions, bindList(
                                                    formals, actuals, emptyEnv))
                                      handle BindListLength => 
                                          raise BugInTypeChecking
                                         "Wrong number of arguments to function"
                                  end)
        (* more alternatives for [[ev]] for \timpcore 216b *)
        | ev (AGET (a, i)) =
            (Array.sub (toArray (ev a), toInt (ev i))
             handle Subscript => raise RuntimeError
                                                "Array subscript out of bounds")
        | ev (ASET (e1, e2, e3)) =
            (let val (a, i, v) = (ev e1, ev e2, ev e3)
             in  Array.update (toArray a, toInt i, v);
                 v
             end handle Subscript => raise RuntimeError
                                                "Array subscript out of bounds")
        | ev (AMAKE (len, init)) = 
             (ARRAY (Array.array (toInt (ev len), ev init))
              handle Size => raise RuntimeError
                                         "array length too large (or negative)")
        | ev (ALEN a) = NUM (Array.length (toArray (ev a)))
        (* type declarations for consistency checking *)
        val _ = op ev : exp -> value
(* type declarations for consistency checking *)
val _ = op ev : exp -> value
  in  ev e
  end
(* type declarations for consistency checking *)
val _ = op eval : exp * value ref env * func env * value ref env -> value
(* evaluation for \timpcore 582a *)
fun topeval (t, globals, functions, echo) =
  case t
    of DEFINE (f, {body=e, formals=xs, returns=rt}) =>
          (globals, bind(f, USERDEF (map #1 xs, e), functions))
          before echo f
     | VAL (name, e) => (* evaluate [[e]] and bind to [[name]] 582b *)
                        let val v = eval(e, globals, functions, emptyEnv)
                            val _ = echo (valueString v)
                        in  (bind(name, ref v, globals), functions)
                        end
     | EXP e => (* evaluate [[e]] 582c *)
                let val v = eval(e, globals, functions, emptyEnv)
                    val _ = echo (valueString v)
                in  (globals, functions)
                end
     | USE _ => raise RuntimeError "Internal error - `use' was evaluated"
(* evaluation for \timpcore 582d *)
fun binaryOp f = (fn [a, b] => f(a, b) | _ => raise BugInTypeChecking "arity 2")
fun unaryOp  f = (fn [a]    => f a     | _ => raise BugInTypeChecking "arity 1")
(* type declarations for consistency checking *)
val _ = op topeval : toplevel * value ref env * func env * (string->unit) ->
                                                        value ref env * func env
(* type declarations for consistency checking *)
val _ = op unaryOp  : (value         -> value) -> (value list -> value)
val _ = op binaryOp : (value * value -> value) -> (value list -> value)
(* evaluation for \timpcore 582e *)
fun arithOp f =
      binaryOp (fn (NUM n1, NUM n2) => NUM (f(n1, n2)) 
                 | _ => raise BugInTypeChecking "arithmetic on non-numbers")
val arithtype = FUNTY ([INTTY, INTTY], INTTY)
(* type declarations for consistency checking *)
val _ = op arithOp   : (int * int -> int) -> (value list -> value)
val _ = op arithtype : funty
(* evaluation for \timpcore 583b *)
fun embedPredicate f args = NUM (if f args then 1 else 0)
fun comparison f = binaryOp (embedPredicate f)
fun intcompare f = 
      comparison (fn (NUM n1, NUM n2) => f(n1, n2)
                   | _ => raise BugInTypeChecking "comparing non-numbers")
val comptype = FUNTY ([INTTY, INTTY], BOOLTY)
(* type declarations for consistency checking *)
val _ = op comparison : (value * value -> bool) -> (value list -> value)
val _ = op intcompare : (int   * int   -> bool) -> (value list -> value)
val _ = op comptype   : funty
type env_bundle = ty env * funty env * value ref env * func env
fun topCheckEval (t, envs as (tglobals, tfuns, vglobals, vfuns), echo) =
  case t
    of USE filename => use readCheckEvalPrint filename envs
     | _ =>
        let val (tglobals, tfuns) = topty   (t, tglobals, tfuns)
            val (vglobals, vfuns) = topeval (t, vglobals, vfuns, echo)
(* type declarations for consistency checking *)
val _ = op topCheckEval : toplevel * env_bundle * (string->unit) -> env_bundle
        in  (tglobals, tfuns, vglobals, vfuns)
        end
(* checking and evaluation for \timpcore 212b *)
and readCheckEvalPrint (reader, echo, errmsg) envs =
  let fun loop envs =
        let fun continue msg = (errmsg msg; loop envs)
            fun finish() = envs
        in  loop (topCheckEval (readtop reader, envs, echo))
            handle EOF => finish()
                (* more read-eval-print handlers 213a *)
                | TypeError         msg => continue ("type error: " ^ msg)
                | BugInTypeChecking msg => continue ("bug in type checking: " ^
                                                                            msg)
                (* more read-eval-print handlers 194b *)
                | IO.Io {name, ...} => continue ("I/O error: " ^ name)
                | Paren (_, loc)    => continue ("syntax error: extra ) in " ^
                                                               srclocString loc)
                | PrematureEOF s    => (errmsg ("Missing )? Got EOF reading (" ^
                                                                   s); finish())
                | SyntaxError msg   => continue ("syntax error: " ^ msg)
                (* more read-eval-print handlers 194c *)
                | Div               => continue "Division by zero"
                | Overflow          => continue "Arithmetic overflow"
                | RuntimeError msg  => continue ("run-time error: " ^ msg)
                | NotFound n        => continue ("variable " ^ n ^ " not found")
        end
  in  loop envs
  end
(* type declarations for consistency checking *)
val _ = op readCheckEvalPrint : topreader * (string->unit) * (string->unit) ->
                                                        env_bundle -> env_bundle


(*****************************************************************)
(*                                                               *)
(*   INITIALIZATION FOR \TIMPCORE                                *)
(*                                                               *)
(*****************************************************************)

(* initialization for \timpcore 213b *)
val initialEnvs =
  let fun addPrim ((name, prim, funty), (tfuns, vfuns)) = 
        ( bind (name, funty, tfuns)
        , bind (name, PRIMITIVE prim, vfuns)
        )
      val (tfuns, vfuns)  = foldl addPrim (emptyEnv, emptyEnv)
                            ((* primops for \timpcore\ [[::]] 583a *)
                             ("+", arithOp op +,   arithtype) :: 
                             ("-", arithOp op -,   arithtype) :: 
                             ("*", arithOp op *,   arithtype) :: 
                             ("/", arithOp op div, arithtype) ::
                             (* primops for \timpcore\ [[::]] 583c *)
                             ("<", intcompare op <, comptype) :: 
                             (">", intcompare op >, comptype) ::
                             ("=", comparison (fn (NUM n1,   NUM n2)   => n1 =
                                                                              n2
                                                | (ARRAY a1, ARRAY a2) => a1 =
                                                                              a2
                                                |  _                   => false)
                                 , FUNTY ([], BOOLTY) (* bogus *) ) ::
                             (* primops for \timpcore\ [[::]] 583d *)
                             ("print", unaryOp (fn x => (print (valueString x^
                                                                     "\n"); x)),
                                          FUNTY ([INTTY], UNITTY)) :: nil)
      val envs    = (emptyEnv, tfuns, emptyEnv, vfuns)
      val basis   = (* ML representation of initial basis *)
                      [ "; typesys.nw 388"
                      , "(define bool and ((bool b) (bool c)) (if b c b))"
                      , "(define bool or  ((bool b) (bool c)) (if b b c))"
                      ,
                  "(define bool not ((bool b))          (if b (= 1 0) (= 0 0)))"
                      , "; typesys.nw 394"
                      , "(define bool <= ((int x) (int y)) (not (> x y)))"
                      , "(define bool >= ((int x) (int y)) (not (< x y)))"
                      , "(define bool != ((int x) (int y)) (not (= x y)))"
                      , "; typesys.nw 400"
                      , "(define int mod ((int m) (int n)) (- m (* n (/ m n))))"
                      ]
      val treader = topreader (stringsreader ("initial basis", basis), false)
  in  readCheckEvalPrint (treader, fn _ => (), fn _ => ()) envs
  end
(* initialization for \timpcore 213c *)
fun runInterpreter noisy = 
  let fun writeln s = app print [s, "\n"]
      val treader = topreader (filereader ("standard input", TextIO.stdIn),
                                                                          noisy)
  in  ignore (readCheckEvalPrint (treader, writeln, writeln) initialEnvs)
  end 


(*****************************************************************)
(*                                                               *)
(*   COMMAND LINE                                                *)
(*                                                               *)
(*****************************************************************)

(* command line *)
fun main ["-q"] = runInterpreter false
  | main []     = runInterpreter true
  | main _      =
      TextIO.output(TextIO.stdErr, "Usage: " ^ CommandLine.name() ^ " [-q]\n")
val _ = main (CommandLine.arguments())
