(* ml.sml 280c *)


(*****************************************************************)
(*                                                               *)
(*   ENVIRONMENTS                                                *)
(*                                                               *)
(*****************************************************************)

(* environments 186 *)
type name = string
type 'a env = (name * 'a) list
val emptyEnv = []

(* lookup and assignment of existing bindings *)
exception NotFound of name
fun find (name, []) = raise NotFound name
  | find (name, (n, v)::tail) = if name = n then v else find(name, tail)

(* adding new bindings *)
exception BindListLength
fun bind(name, v, rho) = (name, v) :: rho
fun bindList(n::vars, v::vals, rho) = bindList(vars, vals, bind(n, v, rho))
  | bindList([], [], rho) = rho
  | bindList _ = raise BindListLength
(* type declarations for consistency checking *)
val _ = op emptyEnv : 'a env
val _ = op find     : name * 'a env -> 'a
val _ = op bind     : name      * 'a      * 'a env -> 'a env
val _ = op bindList : name list * 'a list * 'a env -> 'a env


(*****************************************************************)
(*                                                               *)
(*   TYPES                                                       *)
(*                                                               *)
(*****************************************************************)

(* types 257 *)
datatype ty = TYVAR  of name                (* type variable alpha *)
            | TYCON  of name                (* type constructor mu *)
            | CONAPP of ty * ty list        (* type-level application *)

datatype type_scheme = FORALL of name list * ty
(* types 259a *)
fun tysubst (tau, varenv) =
  let fun subst (TYVAR a) = (find(a, varenv) handle NotFound _ => TYVAR a)
        | subst (TYCON c) = (TYCON c)
        | subst (CONAPP (tau, taus)) = CONAPP (subst tau, map subst taus)
(* type declarations for consistency checking *)
val _ = op tysubst : ty * ty env -> ty
val _ = op subst   : ty          -> ty
  in  subst tau
  end
(* types 259b *)
type subst = ty -> ty
infix 7 |-->
fun a |--> tau = 
  fn tau' => tysubst(tau', bind(a, tau, emptyEnv))
fun idsubst tau = tau
(* types 259c *)
fun instantiate (FORALL (formals, tau), actuals) =
  tysubst(tau, bindList (formals, actuals, emptyEnv))
          handle BindListLength =>
            let exception ThisCan'tHappen in raise ThisCan'tHappen end
(* types 260a *)
(* printing types 595 *)
(* definition of [[separate]] 190d *)
fun separate (zero, sep) =  (* useful for printing *)
  let fun s []     = zero
        | s [x]    = x
        | s (h::t) = h ^ sep ^ s t
  in  s
end
local
  (* precedences *)
  val CONp   = 3
  val STARp  = 2
  val ARROWp = 1
  val NONEp  = 0
  
  fun parens s = "(" ^ s ^ ")"
  fun bracket(s, context, prec) = if prec <= context then parens s else s
  fun p(context, CONAPP (TYCON "tuple", l)) = bracket(ptuple l, context, STARp)
    | p(context, CONAPP (TYCON "function", [arg, ret])) = 
          bracket(p(ARROWp, arg) ^ " -> " ^ p(ARROWp, ret), context, ARROWp)
    | p(context, CONAPP (n, []))  = p(context, n)
    | p(context, CONAPP (n, [t])) = p(CONp, t) ^ " " ^ p(CONp, n)
    | p(context, CONAPP (n, l)) = 
          "(" ^ separate("", ", ") (map typeString l) ^ ") " ^ p(CONp, n)
    | p(context, TYCON n) = n
    | p(context, TYVAR v) = "'" ^ v
  and ptuple l = separate("unit", " * ") (map (fn t => p(STARp, t)) l)
  and typeString ty = p(NONEp, ty)
in 
  val typeString = typeString
  val ptuple = ptuple
end
fun typeSchemeString (FORALL (_, ty)) = 
  typeString ty
(* types 260b *)
val inttype  = TYCON "int"
val booltype = TYCON "bool"
val symtype  = TYCON "sym"
val alpha    = TYVAR "a"
val beta     = TYVAR "b"
fun tupletype l     = CONAPP (TYCON "tuple", l)
fun pairtype (x, y) = tupletype [x, y]
val unittype        = tupletype []
fun listtype ty     = CONAPP (TYCON "list", [ty])
fun funtype (args, result) = 
  CONAPP (TYCON "function", [tupletype args, result])
(* type declarations for consistency checking *)
type subst = subst
val _ = op |-->    : name * ty -> subst
val _ = op idsubst : subst
(* type declarations for consistency checking *)
val _ = op instantiate : type_scheme * ty list -> ty
(* type declarations for consistency checking *)
val _ = op typeString       : ty          -> string
val _ = op typeSchemeString : type_scheme -> string
(* type declarations for consistency checking *)
val _ = op inttype   : ty
val _ = op booltype  : ty
val _ = op symtype   : ty
val _ = op unittype  : ty
val _ = op listtype  : ty -> ty
val _ = op tupletype : ty list -> ty
val _ = op pairtype  : ty * ty -> ty
val _ = op funtype   : ty list * ty -> ty
val _ = op alpha     : ty
val _ = op beta      : ty
(* types 266 *)
type 'a set = 'a list
val emptyset = []
fun member x = 
  List.exists (fn y => y = x)
fun insert (x, l) = 
  if member x l then l else x::l
fun diff  (s1, s2) = 
  List.filter (fn x => not (member x s2)) s1
fun union (s1, s2) = foldl insert s2 s1
(* type declarations for consistency checking *)
type 'a set = 'a set
val _ = op emptyset : 'a set
val _ = op member   : ''a -> ''a set -> bool
val _ = op insert   : ''a     * ''a set  -> ''a set
val _ = op union    : ''a set * ''a set  -> ''a set
val _ = op diff     : ''a set * ''a set  -> ''a set
(* types 267a *)
fun freevars t =
  let fun f(TYVAR v,          l) = insert (v, l)
        | f(TYCON _,          l) = l
        | f(CONAPP(ty, tys),  l) = foldl f (f (ty, l)) tys
  in  rev (f(t, emptyset))
  end  
(* type declarations for consistency checking *)
val _ = op freevars : ty -> name set
(* types 267b *)
local
  val n = ref 1
in
  fun freshtyvar _ = TYVAR ("t" ^ Int.toString (!n) before n := !n + 1)
(* type declarations for consistency checking *)
val _ = op freshtyvar : 'a -> ty
end
(* types 268a *)
fun canonicalize (FORALL (bound, ty)) =
  let fun canonicalTyvarName n =
        if n < 26 then str(chr (ord #"a" + n))
        else "v" ^ Int.toString (n - 25)
      val free = diff (freevars ty, bound)
      fun unusedIndex n =
        if member (canonicalTyvarName n) free then unusedIndex (n+1) else n
      fun newBoundVars (index, [])                = []
        | newBoundVars (index, oldvar :: oldvars) =
            let val n = unusedIndex index
            in  canonicalTyvarName n :: newBoundVars (n+1, oldvars)
            end
      val newBound = newBoundVars (0, bound)
(* type declarations for consistency checking *)
val _ = op canonicalize : type_scheme -> type_scheme
val _ = op newBoundVars : int * name list -> name list
  in  FORALL (newBound, tysubst(ty, bindList(bound, map TYVAR newBound, emptyEnv
                                                                             )))
  end
(* types 268b *)
exception TypeError of string

fun unificationError (t1, t2) =
  case canonicalize (FORALL (union (freevars t1, freevars t2), tupletype [t1, t2
                                                                             ]))
    of FORALL(_, CONAPP(TYCON "tuple", [t1, t2])) => 
         raise TypeError ("Cannot unify " ^ typeString t1 ^ " and " ^ typeString
                                                                             t2)
     | _ => let exception ThisCan'tHappen in raise ThisCan'tHappen end
(* types 268c *)
exception LeftAsExercise
fun unify     (t1, t2) = 
  raise LeftAsExercise
and unifyList (l1, l2) = 
  raise LeftAsExercise
(* type declarations for consistency checking *)
val _ = op unificationError : ty * ty -> 'a
(* type declarations for consistency checking *)
val _ = op unify     : ty      * ty      -> subst
val _ = op unifyList : ty list * ty list -> subst
(* types 270a *)
type type_env = type_scheme env * subst * name set
(* types 270b *)
val emptyTypeEnv = (emptyEnv, idsubst, emptyset)
(* type declarations for consistency checking *)
val _ = op emptyTypeEnv : type_env
(* types 270c *)
fun freshInstance (FORALL (bound, tau)) =
  instantiate (FORALL (bound, tau), map freshtyvar bound)
fun findtype(v, (Gamma, theta, free)) = theta (freshInstance (find(v, Gamma)))
(* type declarations for consistency checking *)
val _ = op freshInstance : type_scheme -> ty
val _ = op findtype      : name * type_env -> ty
(* types 271a *)
fun bindtype(v, sigma as FORALL (bound, tau), (Gamma, theta, free)) = 
  (bind(v, sigma, Gamma), theta, union(diff (freevars tau, bound), free))
(* types 271b *)
infixr 2 on
fun theta on (Gamma, oldtheta, free) = 
  let val free = foldl union emptyset (map (freevars o theta o TYVAR) free)
  in  (Gamma, theta o oldtheta, free)
  end
(* type declarations for consistency checking *)
val _ = op bindtype : name * type_scheme * type_env -> type_env
(* type declarations for consistency checking *)
val _ = op on : subst * type_env -> type_env
(* types 271c *)
fun freevarsGamma (_, _, free) = free
(* types 271d *)
fun generalize (tau, Gamma) =
  canonicalize (FORALL (diff (freevars tau, freevarsGamma Gamma), tau))
(* type declarations for consistency checking *)
val _ = op freevarsGamma : type_env -> name set
(* type declarations for consistency checking *)
val _ = op generalize : ty * type_env -> type_scheme


(*****************************************************************)
(*                                                               *)
(*   LEXICAL ANALYSIS                                            *)
(*                                                               *)
(*****************************************************************)

(* lexical analysis ((mlscheme)) 571a *)
type srcloc  = string * int
fun srclocString (source, line) = source ^ ", line " ^ Int.toString line
(* lexical analysis ((mlscheme)) 571b *)
datatype par = NAME  of string
             | INT   of int
             | SHARP of char  (* #t or #f *)
             | LIST  of srcloc * par list
(* lexical analysis ((mlscheme)) 572 *)

exception Paren of char list * srcloc
   (* raise if found ) when reading input *)
exception PrematureEOF of string (* shows what we were reading ... *)
exception EOF                 (* raised by reader *)
fun read nextline curloc : string -> par list =
  let (* character-fiddling functions 573a *)
      fun isDelim c = List.exists (fn c' => c' = c) [#"(", #")", #" ", #"\t",
                                                                    #"\n", #";"]
      (* character-fiddling functions 573b *)
      fun stringReader []     = NONE
        | stringReader (h::t) = SOME (h, t)
      val strton = Int.scan StringCvt.DEC stringReader
      (* type declarations for consistency checking *)
      val _ = op strton : char list -> (int * char list) option
      (* character-fiddling functions 573c *)
      fun skip s =
        case StringCvt.skipWS stringReader s
          of #";" :: s => skip (StringCvt.dropl (fn c => c <> #"\n")
                                                                 stringReader s)
           | s => s
      (* character-fiddling functions 573d *)
      fun nonwhite s =
        case skip s
          of [] => nonwhite (explode (nextline()))
           | s  => s
      fun readmap f (result, unconsumed) = (f result, unconsumed)
      (* get is called only when s is guaranteed non-blank, non-comment, non-
                                                                        empty *)
      fun get (#"("::s)  = readmap (fn l => LIST l) (list (curloc(), s, []))
        | get (#")"::s)  = raise (Paren (s, curloc()))
        | get (#"'"::s)  = let val loc = curloc()
                               fun quote p = LIST (loc, [NAME "quote", p])
                           in  readmap quote (get (nonwhite s))
                           end
        | get s          = atom (s, [])
      and atom(c::s, r) = if isDelim c then (mkAtom(rev r), c::s) 
                          else atom(s, c::r)
        | atom([],   r) = (mkAtom(rev r), [])
      and mkAtom [#"#", #"f"] = SHARP #"f"
        | mkAtom [#"#", #"t"] = SHARP #"t"
        | mkAtom s = (case strton s of SOME (n, []) => INT n | _ => NAME (
                                                                     implode s))
      and list (loc, s, l) =
        let val next = nonwhite s
        in  let val (a, u) = get next
            in  list(loc, u, a::l)
            end handle (Paren (u, _)) => ((loc, rev l), skip u)
                     | EOF => raise PrematureEOF (implode next)
        end handle EOF => raise PrematureEOF (implode s)
      fun readline [] = []
        | readline s  = 
             let val(p, u) = get s
             in  p :: readline(skip u)
             end
(* type declarations for consistency checking *)
val _ = op read : (unit -> string) -> (unit -> srcloc) -> string -> par list
val _ = op get  : char list                      -> par                 * char
                                                                            list
val _ = op atom : char list * char list          -> par                 * char
                                                                            list
val _ = op list : srcloc * char list * par list  -> (srcloc * par list) * char
                                                                            list
  in  fn s => readline(skip (explode s)) 
  end


(*****************************************************************)
(*                                                               *)
(*   ABSTRACT SYNTAX AND VALUES                                  *)
(*                                                               *)
(*****************************************************************)

(* abstract syntax and values 252 *)
datatype exp = LITERAL of value
             | VAR     of name
             | IFX     of exp * exp * exp
             | BEGIN   of exp list
             | LETX    of let_kind * (name * exp) list * exp
             | LAMBDA  of name list * exp
             | APPLY   of exp * exp list
and let_kind = LET | LETREC | LETSTAR
and (* definition of [[value]] 253b *)
    value = NIL
          | BOOL      of bool
          | NUM       of int
          | SYM       of name
          | PAIR      of value * value
          | CLOSURE   of lambda * (unit -> value env)
          | PRIMITIVE of primop
    withtype primop = value list -> value (* raises RuntimeError *)
         and lambda = name list * exp
    exception RuntimeError of string (* error message *)
(* abstract syntax and values 253a *)
datatype toplevel = EXP    of exp
                  | DEFINE of name * (name list * exp)
                  | VAL    of name * exp
                  | VALREC of name * exp
                  | USE    of name


(*****************************************************************)
(*                                                               *)
(*   VALUES                                                      *)
(*                                                               *)
(*****************************************************************)

(* values ((mlscheme)) 188b *)



fun embedList []     = NIL
  | embedList (h::t) = PAIR(h, embedList t)
fun embedPredicate f = fn x => BOOL (f x)
fun bool (BOOL b) = b
  | bool _        = true
(* type declarations for consistency checking *)
val _ = op embedList      : value list -> value
val _ = op embedPredicate : ('a -> bool) -> ('a -> value)
val _ = op bool           : value -> bool
(* values ((mlscheme)) 189a *)
fun valueString (NIL)    = "()"
  | valueString (BOOL b) = if b then "#t" else "#f"
  | valueString (NUM n)  = String.map (fn #"~" => #"-" | c => c) (Int.toString n
                                                                               )
  | valueString (SYM v)  = v
  | valueString (PAIR(car, cdr))  = 
      let fun tail (PAIR(car, cdr)) = " " ^ valueString car ^ tail cdr
            | tail NIL = ")"
            | tail v = " . " ^ valueString v ^ ")"
      in  "(" ^ valueString car ^ tail cdr
      end
  | valueString (CLOSURE   _) = "<procedure>"
  | valueString (PRIMITIVE _) = "<procedure>"
(* type declarations for consistency checking *)
val _ = op valueString : value -> string


(*****************************************************************)
(*                                                               *)
(*   PARSING                                                     *)
(*                                                               *)
(*****************************************************************)

(* parsing 596a *)
exception SyntaxError of string
fun syntaxErrorAt (source, line) msg =
  raise SyntaxError (source ^ ", line " ^ Int.toString line ^ ": " ^ msg)
fun parse (NAME s)  = VAR s
  | parse (INT n)   = LITERAL (NUM n)
  | parse (SHARP c) = LITERAL (BOOL (c = #"t"))
  | parse (LIST (loc, NAME "quote"  :: args)) = parseQuote loc args
  | parse (LIST (loc, NAME "if"     :: args)) = parseIf    loc args
  | parse (LIST (loc, NAME "begin"  :: args)) = BEGIN (map parse args)
  | parse (LIST (loc, NAME "lambda" :: args)) = parseLambda loc args
  | parse (LIST (loc, NAME "let"    :: args)) = parseLet loc LET     args
  | parse (LIST (loc, NAME "letrec" :: args)) = parseLet loc LETREC  args
  | parse (LIST (loc, NAME "let*"   :: args)) = parseLet loc LETSTAR args
  | parse (LIST (loc, h::t))  = APPLY (parse h, map parse t)
  | parse (LIST (loc, [])) = raise SyntaxError "empty application ()"
and parseQuote loc [e] = LITERAL (quote e)
  | parseQuote loc _ = syntaxErrorAt loc "expected (quote e)"
and parseIf _ [e1, e2, e3] = IFX (parse e1, parse e2, parse e3)
  | parseIf loc _ = syntaxErrorAt loc "expected (if e1 e2 e3)"
and parseLambda _ [LIST (loc, formals), body] =
      LAMBDA (map (unname loc) formals, parse body)
  | parseLambda loc _ = syntaxErrorAt loc "expected (lambda (names) body)"
and parseLet _ kind [LIST (loc, bindings), body] = 
      LETX (kind, map (unBinding loc) bindings, parse body)
  | parseLet loc _ _ = syntaxErrorAt loc "expected (let (bindings) body)"
and unname _ (NAME n) = n
  | unname loc _ = syntaxErrorAt loc
                                 "formal parameters to function must be symbols"
and unBinding _ (LIST (_, [NAME n, e])) = (n, parse e)
  | unBinding loc _ = syntaxErrorAt loc "expected (v e) in bindings"
and quote (INT   n)     = NUM n
  | quote (SHARP c)     = BOOL (c = #"t")
  | quote (NAME  n)     = SYM n
  | quote (LIST (_, l)) = embedList (map quote l)
(* parsing 596b *)
fun toplevel (LIST (loc, NAME "define" :: args)) = parseDefine loc args
  | toplevel (LIST (loc, NAME "val"    :: args)) = parseVal    loc "val"     VAL
                                                                            args
  | toplevel (LIST (loc, NAME "val-rec":: args)) = parseVal    loc "val-rec"
                                                                     VALREC args
  | toplevel (LIST (loc, NAME "use"    :: args)) = parseUse    loc args
  | toplevel e = EXP (parse e)
and parseDefine _ [NAME name, LIST (loc, formals), body] =
      DEFINE (name, (map (unname loc) formals, parse body))
  | parseDefine loc _ = syntaxErrorAt loc "expected (define f (args) body)"
and parseVal _ concrete abstract [NAME v, e] = abstract (v, parse e)
  | parseVal loc concrete _ _ = syntaxErrorAt loc ("expected (" ^ concrete ^
                                                                        " v e)")
and parseUse _ [NAME v] = USE v
  | parseUse loc _ = syntaxErrorAt loc "expected (use filename)"


(*****************************************************************)
(*                                                               *)
(*   READERS                                                     *)
(*                                                               *)
(*****************************************************************)

(* readers 576b *)
type reader = { firstline : unit -> string
              , nextline  : unit -> string
              , curloc    : unit -> string * int
              }
(* readers 576c *)
fun promptReader (firstPrompt, nextPrompt, reader) =
  let val { firstline = first, nextline = next, curloc = cur } = reader
      fun promptIn get prompt () = ( TextIO.output(TextIO.stdOut, prompt)
                                   ; TextIO.flushOut(TextIO.stdOut)
                                   ; get ()
                                   )
  in  { firstline = promptIn first firstPrompt
      , nextline  = promptIn next  nextPrompt
      , curloc    = cur
      }
  end
(* type declarations for consistency checking *)
val _ = op promptReader : string * string * reader -> reader
(* readers 577a *)
fun filereader (filename, fd)=
  let val lineNum = ref 0
      fun get () =
        let val line = TextIO.inputLine fd
        in  if size line = 0 then raise EOF else line before lineNum := !lineNum
                                                                             + 1
        end
  in  { firstline = get, nextline = get, curloc = (fn () => (filename, !lineNum)
                                                                             ) }
  end

fun stringsreader (what, l) = 
  let val buffer  = ref l
      val lineNum = ref 0
      fun get ()  = case !buffer
                      of [] => raise EOF
                       | h :: t => h before buffer := t before lineNum := !
                                                                     lineNum + 1
  in  { firstline = get, nextline = get, curloc = (fn () => (what, !lineNum)) }
  end              
(* type declarations for consistency checking *)
val _ = op filereader    : string * TextIO.instream -> reader
val _ = op stringsreader : string * string list     -> reader
(* readers 577b *)
type topreader = { buffer : toplevel list ref, reader : reader }
fun readtop (r as { buffer, reader = { nextline, firstline, curloc } }) =
  case !buffer
    of h::t => h before buffer := t
     | []   => ( buffer := map toplevel (read nextline curloc (firstline()))
               ; readtop r
               )
(* readers 577c *)
fun topreader (reader, prompt) = 
  let val reader = if prompt then promptReader ("-> ", "   ", reader) else
                                                                          reader
  in  { buffer = ref [], reader = reader }
  end
(* readers 578 *)
fun echoTag f x =
  let val line = f x
      val _ = if (String.substring (line, 0, 2) = ";#" handle _ => false) then
                print line
              else
                ()
  in  line
  end

fun echoReader { nextline=n, firstline=f, curloc=c } =
     { nextline=echoTag n, firstline=echoTag f, curloc=c } 

fun echoTop { buffer = b, reader = r } = { buffer = b, reader = echoReader r }
val topreader = fn args => echoTop (topreader args)
(* type declarations for consistency checking *)
val _ = op echoReader : reader -> reader 


(*****************************************************************)
(*                                                               *)
(*   IMPLEMENTATION OF [[USE]]                                   *)
(*                                                               *)
(*****************************************************************)

(* implementation of [[use]] 193c *)
fun use readEvalPrint filename rho =
      let val fd = TextIO.openIn filename
          val treader = topreader (filereader (filename, fd), false)
          fun writeln s = app print [s, "\n"]
      in  readEvalPrint (treader, writeln, writeln) rho
          before TextIO.closeIn fd
      end 


(*****************************************************************)
(*                                                               *)
(*   TYPE INFERENCE                                              *)
(*                                                               *)
(*****************************************************************)

(* type inference 271e *)
fun typeof (e, Gamma) =
  let (* function [[typesof]], to infer the types of a list of expressions 272a
                                                                              *)
      fun typesof([], Gamma : type_env) = ([], idsubst)
        | typesof(e::es, Gamma) =
            let val (tau,  theta)  = typeof (e,  Gamma)
                val (taus, theta') = typesof(es, theta on Gamma)
            in  (theta' tau :: taus, theta' o theta)
            end
      (* function [[literal]], to infer the type of a literal constant 272c *)
      fun literal _ = raise LeftAsExercise
      (* function [[ty]], to infer the type of an expression, given [[Gamma]] 27
                                                                           2b *)
      fun ty(LITERAL n) = literal n
        | ty(VAR v) = (findtype(v, Gamma), idsubst)
        (* more alternatives for [[ty]] 273a *)
        | ty(APPLY (f, actuals)) = 
             (case typesof (f :: actuals, Gamma)
                of ([], _) => let exception ThisCan'tHappen in raise
                                                             ThisCan'tHappen end
                 | (funty :: actualtypes, theta) =>
                      let val rettype = freshtyvar()
                          val theta' = unify(funty, funtype(actualtypes, rettype
                                                                              ))
                      in  (theta' rettype, theta' o theta)
                      end)
        (* more alternatives for [[ty]] 273b *)
        | ty(LETX (LETSTAR, [], body)) = ty body
        | ty(LETX (LETSTAR, (b :: bs), body)) = ty (LETX (LET, [b], LETX (
                                                            LETSTAR, bs, body)))
        (* more alternatives for [[ty]] 273c *)
        | ty(IFX (e1, e2, e3))        = raise LeftAsExercise
        | ty(BEGIN es)                = raise LeftAsExercise
        | ty(LAMBDA (formals, body))  = raise LeftAsExercise 
        | ty(LETX (LET, bs, body))    = raise LeftAsExercise 
        | ty(LETX (LETREC, bs, body)) = raise LeftAsExercise 
(* type declarations for consistency checking *)
val _ = op typeof  : exp      * type_env -> ty      * subst
val _ = op typesof : exp list * type_env -> ty list * subst
val _ = op literal : value -> ty * subst
val _ = op ty      : exp   -> ty * subst
  in  ty e
  end
(* type inference 273d *)
fun topty (t, Gamma) =
  case t
    of USE filename       => raise RuntimeError 
                                     "internal error -- `use' reached topty"
     | EXP e              => (* infer type for top-level expression [[e]] 274a
                                                                              *)
                             let val (tau, _) = typeof (e, Gamma)
                                 val sigma    = generalize (tau, Gamma)
                             in  (Gamma, typeSchemeString sigma)
                             end
     | DEFINE (x, lambda) => topty (VALREC (x, LAMBDA lambda), Gamma)
     | VAL    (x, e)      => (* infer and bind type for [[VAL    (x, e)]] 273e
                                                                              *)
                             let val (tau, _) = typeof (e, Gamma)
                                 val sigma    = generalize (tau, Gamma)
                             in  (bindtype(x, sigma, Gamma), typeSchemeString
                                                                          sigma)
                             end
     | VALREC (x, e)      => (* infer and bind type for [[VALREC (x, e)]] 274b
                                                                              *)
                             let val tau           = freshtyvar ()
                                 val Gamma'        = bindtype(x, FORALL([], tau)
                                                                        , Gamma)
                                 val (tau', theta) = typeof(e, Gamma')
                                 val theta'        = unify (tau', theta tau)
                                 val sigma         = generalize(theta' tau',
                                                                          Gamma)
                             in  (bindtype(x, sigma, Gamma), typeSchemeString
                                                                          sigma)
                             end


(*****************************************************************)
(*                                                               *)
(*   CHECKING AND EVALUATION                                     *)
(*                                                               *)
(*****************************************************************)

(* checking and evaluation 279a *)
exception BugInTypeInference of string
(* evaluation 275a *)
(* definition of [[separate]] 190d *)
fun separate (zero, sep) =  (* useful for printing *)
  let fun s []     = zero
        | s [x]    = x
        | s (h::t) = h ^ sep ^ s t
  in  s
end
fun eval(e, rho) =
  let fun ev(LITERAL n)        = n
        | ev(VAR v)            = find (v, rho)
        | ev(IFX (e1, e2, e3)) = ev (if bool (ev e1) then e2 else e3)
        | ev(LAMBDA l)         = CLOSURE(l, fn _ => rho)
        | ev(BEGIN es) =
            let fun b(e::es, lastval) = b(es, ev e)
                  | b(   [], lastval) = lastval
            in  b(es, BOOL false)
            end
        | ev(APPLY (f, args)) = 
           (case ev f
              of PRIMITIVE prim => prim (map ev args)
               | CLOSURE clo => (* apply closure [[clo]] to [[args]] 275b *)
                               let val ((formals, body), mkRho) = clo
                                   val actuals = map ev args
                               in  eval(body, bindList(formals, actuals, mkRho()
                                                                              ))
                                   handle BindListLength => 
                                       raise BugInTypeInference
                                          "Wrong number of arguments to closure"
                               end
               | _ => raise BugInTypeInference "Applied non-function"
               )
        (* more alternatives for [[ev]] 275c *)
        | ev(LETX (LET, bs, body)) =
            let val (names, values) = ListPair.unzip bs
            in  eval (body, bindList(names, map ev values, rho))
            end
        (* more alternatives for [[ev]] 275d *)
        | ev(LETX (LETSTAR, bs, body)) =
            let fun step ((n, e), rho) = bind(n, eval(e, rho), rho)
            in  eval (body, foldl step rho bs)
            end
        (* more alternatives for [[ev]] 276a *)
        | ev(LETX (LETREC, bs, body)) =
            let fun makeRho' () =
                  let fun step ((n, e), rho) =
                            (case e
                               of LAMBDA l => bind(n, CLOSURE (l, makeRho'), rho
                                                                               )
                                | _ => raise RuntimeError "non-lambda in letrec"
                                                                               )
                  in  foldl step rho bs
                  end
            in  eval(body, makeRho'())
            end
  in  ev e
  end
(* type declarations for consistency checking *)
val _ = op topty : toplevel * type_env -> type_env * string
(* type declarations for consistency checking *)
val _ = op eval : exp * value env -> value
(* evaluation 276b *)
fun topeval (t, rho) =
  case t
    of USE filename => raise RuntimeError
                                       "internal error -- `use' reached topeval"
     | EXP e        => (rho, valueString (eval(e, rho)))
     | DEFINE (name, lambda) => topeval (VALREC (name, LAMBDA lambda), rho)
     | VAL    (name, e)      =>
          let val v   = eval(e, rho)
              val rho = bind(name, v, rho)
          in  (rho, showVal name v)
          end
     | VALREC (name, LAMBDA lambda) => 
          let fun makeRho' () = bind(name, CLOSURE(lambda, makeRho'), rho)
              val v           = CLOSURE(lambda, makeRho')
          in  (makeRho'(), showVal name v)
          end
     | VALREC _ => raise RuntimeError "expression in val-rec must be lambda"
and showVal name v =
      case v
        of CLOSURE   _ => name
         | PRIMITIVE _ => name
         | _ => valueString v
(* type declarations for consistency checking *)
val _ = op topeval : toplevel * value env -> value env * string
(* evaluation 277a *)
fun binaryOp f = (fn [a, b] => f(a, b) | _ => raise BugInTypeInference "arity 2"
                                                                               )
fun unaryOp  f = (fn [a]    => f a     | _ => raise BugInTypeInference "arity 1"
                                                                               )
(* type declarations for consistency checking *)
val _ = op unaryOp  : (value         -> value) -> (value list -> value)
val _ = op binaryOp : (value * value -> value) -> (value list -> value)
(* evaluation 277b *)
fun arithOp f =
      binaryOp (fn (NUM n1, NUM n2) => NUM (f(n1, n2)) 
                 | _ => raise BugInTypeInference "arithmetic on non-numbers")
val arithtype = funtype([inttype, inttype], inttype)
(* type declarations for consistency checking *)
val _ = op arithOp   : (int * int -> int) -> (value list -> value)
val _ = op arithtype : ty
(* evaluation 277d *)
fun predOp f     = unaryOp  (embedPredicate f)
fun comparison f = binaryOp (embedPredicate f)
fun intcompare f = 
      comparison (fn (NUM n1, NUM n2) => f(n1, n2)
                   | _ => raise BugInTypeInference "comparing non-numbers")
fun predtype x = funtype([x],    booltype)
fun comptype x = funtype([x, x], booltype)
(* type declarations for consistency checking *)
val _ = op predOp     : (value         -> bool) -> (value list -> value)
val _ = op comparison : (value * value -> bool) -> (value list -> value)
val _ = op intcompare : (int   * int   -> bool) -> (value list -> value)
val _ = op predtype   : ty -> ty
val _ = op comptype   : ty -> ty
type env_bundle = type_env * value env
fun topCheckEval (t, envs as (Gamma, rho), echo) =
  case t
    of USE filename => use readCheckEvalPrint filename envs
     | _ =>
        let val (Gamma, tystring)  = topty   (t, Gamma)
            val (rho,   valstring) = topeval (t, rho)
            val _ = if size valstring > 0 then echo (valstring ^ " : " ^
                                                                       tystring)
                    else ()
(* type declarations for consistency checking *)
val _ = op topCheckEval : toplevel * env_bundle * (string->unit) -> env_bundle
        in  (Gamma, rho)
        end
(* checking and evaluation 279b *)
and readCheckEvalPrint (reader, echo, errmsg) envs =
  let fun loop envs =
        let fun continue msg = (errmsg msg; loop envs)
            fun finish() = envs
        in  loop (topCheckEval (readtop reader, envs, echo))
            handle EOF => finish()
                (* more read-eval-print handlers 279c *)
                | TypeError          msg => continue ("type error: " ^ msg)
                | BugInTypeInference msg => continue ("bug in type inference: "
                                                                          ^ msg)
                (* more read-eval-print handlers 194b *)
                | IO.Io {name, ...} => continue ("I/O error: " ^ name)
                | Paren (_, loc)    => continue ("syntax error: extra ) in " ^
                                                               srclocString loc)
                | PrematureEOF s    => (errmsg ("Missing )? Got EOF reading (" ^
                                                                   s); finish())
                | SyntaxError msg   => continue ("syntax error: " ^ msg)
                (* more read-eval-print handlers 194c *)
                | Div               => continue "Division by zero"
                | Overflow          => continue "Arithmetic overflow"
                | RuntimeError msg  => continue ("run-time error: " ^ msg)
                | NotFound n        => continue ("variable " ^ n ^ " not found")
        end
  in  loop envs
  end
(* type declarations for consistency checking *)
val _ = op readCheckEvalPrint : topreader * (string->unit) * (string->unit) ->
                                                        env_bundle -> env_bundle


(*****************************************************************)
(*                                                               *)
(*   INITIALIZATION                                              *)
(*                                                               *)
(*****************************************************************)

(* initialization 280a *)
val initialEnvs =
  let fun addPrim ((name, prim, tau), (Gamma, rho)) = 
        ( bindtype(name, generalize(tau, Gamma), Gamma)
        , bind (name, PRIMITIVE prim, rho)
        )
      val envs  = foldl addPrim (emptyTypeEnv, emptyEnv) ((* primops [[::]] 277c
                                                                              *)
                                                          ("+", arithOp op +,
                                                                   arithtype) ::
                                                          ("-", arithOp op -,
                                                                   arithtype) ::
                                                          ("*", arithOp op *,
                                                                   arithtype) ::
                                                          ("/", arithOp op div,
                                                                   arithtype) ::
                                                          (* primops [[::]] 278a
                                                                              *)
                                                          ("<", intcompare op <,
                                                            comptype inttype) ::
                                                          (">", intcompare op >,
                                                            comptype inttype) ::
                                                          ("=", comparison (fn (
                                                       NIL,     NIL)     => true
                                                                             | (
                                                    NUM n1,  NUM n2)  => n1 = n2
                                                                             | (
                                                    SYM v1,  SYM v2)  => v1 = v2
                                                                             | (
                                                    BOOL b1, BOOL b2) => b1 = b2
                                                                             |
                                                     _                 => false)
                                                               , comptype alpha)
                                                                              ::
                                                          ("null?", predOp (fn
                        NIL => true | _ => false), predtype (listtype alpha)) ::
                                                          (* primops [[::]] 278b
                                                                              *)
                                                          ("cons", binaryOp (fn
                                                          (a, b) => PAIR(a, b)),

                            funtype([alpha, listtype alpha], listtype alpha)) ::
                                                          ("car",  unaryOp  (fn
                                                           (PAIR(car, _)) => car
                                                                              |
                           NIL => raise RuntimeError "car applied to empty list"
                                                                              |
                     _   => raise BugInTypeInference "car applied to non-list"),

                                            funtype([listtype alpha], alpha)) ::
                                                          ("cdr",  unaryOp  (fn
                                                           (PAIR(_, cdr)) => cdr
                                                                              |
                           NIL => raise RuntimeError "cdr applied to empty list"
                                                                              |
                     _   => raise BugInTypeInference "cdr applied to non-list"),

                                   funtype([listtype alpha], listtype alpha)) ::
                                                          (* primops [[::]] 278c
                                                                              *)
                                                          ("print", unaryOp (fn
                                          x => (print (valueString x^"\n"); x)),
                                                                       funtype([
                                                           alpha], unittype)) ::
                                                          ("error", unaryOp (fn
                                       x => raise RuntimeError (valueString x)),
                                                                       funtype([
                                                          alpha], beta)) :: nil)
      val basis = (* ML representation of initial basis *)
                    [ "; ml.nw 2289"
                    , "(define list1 (x) (cons x '()))"
                    , "(define bind (x y alist)"
                    , "  (if (null? alist)"
                    , "    (list1 (pair x y))"
                    , "    (if (= x (fst (car alist)))"
                    , "      (cons (pair x y) (cdr alist))"
                    , "      (cons (car alist) (bind x y (cdr alist))))))"
                    , "; ml.nw 2301"
                    , "(define isbound? (x alist)"
                    , "  (if (null? alist) "
                    , "    #f"
                    , "    (if (= x (fst (car alist)))"
                    , "      #t"
                    , "      (isbound? x (cdr alist)))))"
                    , "(define find (x alist)"
                    , "  (if (null? alist) "
                    , "    (error 'not-found)"
                    , "    (if (= x (fst (car alist)))"
                    , "      (snd (car alist))"
                    , "      (find x (cdr alist)))))"
                    , "; mla.nw 93"
                    , "(define caar (l) (car (car l)))"
                    , "(define cadr (l) (car (cdr l)))"
                    , "(define cdar (l) (cdr (car l)))"
                    , "(define length (l)"
                    , "  (if (null? l) 0"
                    , "    (+ 1 (length (cdr l)))))"
                    , "(define and (b c) (if b  c  b))"
                    , "(define or  (b c) (if b  b  c))"
                    , "(define not (b)   (if b #f #t))"
                    , "; mla.nw 103"
                    , "(define append (l1 l2)"
                    , "  (if (null? l1)"
                    , "     l2"
                    , "     (cons (car l1) (append (cdr l1) l2))))"
                    , "(define revapp (l1 l2)"
                    , "  (if (null? l1)"
                    , "     l2"
                    , "     (revapp (cdr l1) (cons (car l1) l2))))"
                    , "; mla.nw 112"
                    , "(define o (f g) (lambda (x) (f (g x))))"
                    , "(define curry   (f) (lambda (x) (lambda (y) (f x y))))"
                    , "(define uncurry (f) (lambda (x y) ((f x) y)))"
                    , "; mla.nw 116"
                    , "(define filter (p? l)"
                    , "  (if (null? l)"
                    , "    '()"
                    , "    (if (p? (car l))"
                    , "      (cons (car l) (filter p? (cdr l)))"
                    , "      (filter p? (cdr l)))))"
                    , "; mla.nw 123"
                    , "(define map (f l)"
                    , "  (if (null? l)"
                    , "    '()"
                    , "    (cons (f (car l)) (map f (cdr l)))))"
                    , "; mla.nw 128"
                    , "(define exists? (p? l)"
                    , "  (if (null? l)"
                    , "    #f"
                    , "    (if (p? (car l)) "
                    , "      #t"
                    , "      (exists? p? (cdr l)))))"
                    , "(define all? (p? l)"
                    , "  (if (null? l)"
                    , "    #t"
                    , "    (if (p? (car l))"
                    , "      (all? p? (cdr l))"
                    , "      #f)))"
                    , "; mla.nw 141"
                    , "(define foldr (op zero l)"
                    , "  (if (null? l)"
                    , "    zero"
                    , "    (op (car l) (foldr op zero (cdr l)))))"
                    , "(define foldl (op zero l)"
                    , "  (if (null? l)"
                    , "    zero"
                    , "    (foldl op (op (car l) zero) (cdr l))))"
                    , "; mla.nw 150"
                    , "(define <= (x y) (not (> x y)))"
                    , "(define >= (x y) (not (< x y)))"
                    , "(define != (x y) (not (= x y)))"
                    , "; mla.nw 154"
                    , "(define max (x y) (if (> x y) x y))"
                    , "(define min (x y) (if (< x y) x y))"
                    , "(define mod (m n) (- m (* n (/ m n))))"
                    , "(define gcd (m n) (if (= n 0) m (gcd n (mod m n))))"
                    , "(define lcm (m n) (* m (/ n (gcd m n))))"
                    , "; mla.nw 160"
                    , "(define min* (l) (foldr min (car l) (cdr l)))"
                    , "(define max* (l) (foldr max (car l) (cdr l)))"
                    , "(define gcd* (l) (foldr gcd (car l) (cdr l)))"
                    , "(define lcm* (l) (foldr lcm (car l) (cdr l)))"
                    , "; mla.nw 165"
                    , "(define list1 (x)               (cons x '()))"
                    , "(define list2 (x y)             (cons x (list1 y)))"
                    , "(define list3 (x y z)           (cons x (list2 y z)))"
                    , "(define list4 (x y z a)         (cons x (list3 y z a)))"
                    ,
                     "(define list5 (x y z a b)       (cons x (list4 y z a b)))"
                    ,
                   "(define list6 (x y z a b c)     (cons x (list5 y z a b c)))"
                    ,
                 "(define list7 (x y z a b c d)   (cons x (list6 y z a b c d)))"
                    ,
               "(define list8 (x y z a b c d e) (cons x (list7 y z a b c d e)))"
                    , "; mla.nw 174"
                    , "(define takewhile (p? l)"
                    , "  (if (null? l)"
                    , "     '()"
                    , "     (if (p? (car l))"
                    , "         (cons (car l) (takewhile p? (cdr l)))"
                    , "         '())))"
                    , "(define dropwhile (p? l)"
                    , "  (if (null? l)"
                    , "     '()"
                    , "     (if (p? (car l))"
                    , "         (dropwhile p? (cdr l))"
                    , "         l)))"
                    ]
      val treader = topreader (stringsreader ("initial basis", basis), false)
  in  readCheckEvalPrint (treader, fn _ => (), fn _ => ()) envs
  end
(* initialization 280b *)
fun runInterpreter noisy = 
  let fun writeln s = app print [s, "\n"]
      val treader = topreader (filereader ("standard input", TextIO.stdIn),
                                                                          noisy)
  in  ignore (readCheckEvalPrint (treader, writeln, writeln) initialEnvs)
  end 


(*****************************************************************)
(*                                                               *)
(*   COMMAND LINE                                                *)
(*                                                               *)
(*****************************************************************)

(* command line *)
fun main ["-q"] = runInterpreter false
  | main []     = runInterpreter true
  | main _      =
      TextIO.output(TextIO.stdErr, "Usage: " ^ CommandLine.name() ^ " [-q]\n")
val _ = main (CommandLine.arguments())
