(* mlscheme.sml 195d *)


(*****************************************************************)
(*                                                               *)
(*   ENVIRONMENTS                                                *)
(*                                                               *)
(*****************************************************************)

(* environments 186 *)
type name = string
type 'a env = (name * 'a) list
val emptyEnv = []

(* lookup and assignment of existing bindings *)
exception NotFound of name
fun find (name, []) = raise NotFound name
  | find (name, (n, v)::tail) = if name = n then v else find(name, tail)

(* adding new bindings *)
exception BindListLength
fun bind(name, v, rho) = (name, v) :: rho
fun bindList(n::vars, v::vals, rho) = bindList(vars, vals, bind(n, v, rho))
  | bindList([], [], rho) = rho
  | bindList _ = raise BindListLength
(* type declarations for consistency checking *)
val _ = op emptyEnv : 'a env
val _ = op find     : name * 'a env -> 'a
val _ = op bind     : name      * 'a      * 'a env -> 'a env
val _ = op bindList : name list * 'a list * 'a env -> 'a env


(*****************************************************************)
(*                                                               *)
(*   LEXICAL ANALYSIS                                            *)
(*                                                               *)
(*****************************************************************)

(* lexical analysis 571a *)
type srcloc  = string * int
fun srclocString (source, line) = source ^ ", line " ^ Int.toString line
(* lexical analysis 571b *)
datatype par = NAME  of string
             | INT   of int
             | SHARP of char  (* #t or #f *)
             | LIST  of srcloc * par list
(* lexical analysis 572 *)

exception Paren of char list * srcloc
   (* raise if found ) when reading input *)
exception PrematureEOF of string (* shows what we were reading ... *)
exception EOF                 (* raised by reader *)
fun read nextline curloc : string -> par list =
  let (* character-fiddling functions 573a *)
      fun isDelim c = List.exists (fn c' => c' = c) [#"(", #")", #" ", #"\t",
                                                                    #"\n", #";"]
      (* character-fiddling functions 573b *)
      fun stringReader []     = NONE
        | stringReader (h::t) = SOME (h, t)
      val strton = Int.scan StringCvt.DEC stringReader
      (* type declarations for consistency checking *)
      val _ = op strton : char list -> (int * char list) option
      (* character-fiddling functions 573c *)
      fun skip s =
        case StringCvt.skipWS stringReader s
          of #";" :: s => skip (StringCvt.dropl (fn c => c <> #"\n")
                                                                 stringReader s)
           | s => s
      (* character-fiddling functions 573d *)
      fun nonwhite s =
        case skip s
          of [] => nonwhite (explode (nextline()))
           | s  => s
      fun readmap f (result, unconsumed) = (f result, unconsumed)
      (* get is called only when s is guaranteed non-blank, non-comment, non-
                                                                        empty *)
      fun get (#"("::s)  = readmap (fn l => LIST l) (list (curloc(), s, []))
        | get (#")"::s)  = raise (Paren (s, curloc()))
        | get (#"'"::s)  = let val loc = curloc()
                               fun quote p = LIST (loc, [NAME "quote", p])
                           in  readmap quote (get (nonwhite s))
                           end
        | get s          = atom (s, [])
      and atom(c::s, r) = if isDelim c then (mkAtom(rev r), c::s) 
                          else atom(s, c::r)
        | atom([],   r) = (mkAtom(rev r), [])
      and mkAtom [#"#", #"f"] = SHARP #"f"
        | mkAtom [#"#", #"t"] = SHARP #"t"
        | mkAtom s = (case strton s of SOME (n, []) => INT n | _ => NAME (
                                                                     implode s))
      and list (loc, s, l) =
        let val next = nonwhite s
        in  let val (a, u) = get next
            in  list(loc, u, a::l)
            end handle (Paren (u, _)) => ((loc, rev l), skip u)
                     | EOF => raise PrematureEOF (implode next)
        end handle EOF => raise PrematureEOF (implode s)
      fun readline [] = []
        | readline s  = 
             let val(p, u) = get s
             in  p :: readline(skip u)
             end
(* type declarations for consistency checking *)
val _ = op read : (unit -> string) -> (unit -> srcloc) -> string -> par list
val _ = op get  : char list                      -> par                 * char
                                                                            list
val _ = op atom : char list * char list          -> par                 * char
                                                                            list
val _ = op list : srcloc * char list * par list  -> (srcloc * par list) * char
                                                                            list
  in  fn s => readline(skip (explode s)) 
  end


(*****************************************************************)
(*                                                               *)
(*   ABSTRACT SYNTAX AND VALUES                                  *)
(*                                                               *)
(*****************************************************************)

(* abstract syntax and values 187 *)
datatype       exp = LITERAL of value
                   | VAR     of name
                   | SET     of name * exp
                   | IFX     of exp * exp * exp
                   | WHILEX  of exp * exp
                   | BEGIN   of exp list
                   | LETX    of let_kind * (name * exp) list * exp
                   | LAMBDA  of lambda
                   | APPLY   of exp * exp list
and       let_kind = LET | LETREC | LETSTAR
and          value = NIL
                   | BOOL      of bool   
                   | NUM       of int
                   | SYM       of name
                   | PAIR      of value * value
                   | CLOSURE   of lambda * value ref env
                   | PRIMITIVE of primitive
withtype primitive = value list -> value (* raises RuntimeError *)
     and lambda    = name list * exp

exception RuntimeError of string (* error message *)
(* abstract syntax and values 188a *)
datatype toplevel = EXP    of exp
                  | DEFINE of name * lambda
                  | VAL    of name * exp
                  | USE    of name


(*****************************************************************)
(*                                                               *)
(*   VALUES                                                      *)
(*                                                               *)
(*****************************************************************)

(* values 188b *)



fun embedList []     = NIL
  | embedList (h::t) = PAIR(h, embedList t)
fun embedPredicate f = fn x => BOOL (f x)
fun bool (BOOL b) = b
  | bool _        = true
(* type declarations for consistency checking *)
val _ = op embedList      : value list -> value
val _ = op embedPredicate : ('a -> bool) -> ('a -> value)
val _ = op bool           : value -> bool
(* values 189a *)
fun valueString (NIL)    = "()"
  | valueString (BOOL b) = if b then "#t" else "#f"
  | valueString (NUM n)  = String.map (fn #"~" => #"-" | c => c) (Int.toString n
                                                                               )
  | valueString (SYM v)  = v
  | valueString (PAIR(car, cdr))  = 
      let fun tail (PAIR(car, cdr)) = " " ^ valueString car ^ tail cdr
            | tail NIL = ")"
            | tail v = " . " ^ valueString v ^ ")"
      in  "(" ^ valueString car ^ tail cdr
      end
  | valueString (CLOSURE   _) = "<procedure>"
  | valueString (PRIMITIVE _) = "<procedure>"
(* type declarations for consistency checking *)
val _ = op valueString : value -> string


(*****************************************************************)
(*                                                               *)
(*   PARSING                                                     *)
(*                                                               *)
(*****************************************************************)

(* parsing 574 *)
exception SyntaxError of string
fun syntaxErrorAt loc msg =
  raise SyntaxError (srclocString loc ^ ": " ^ msg)
fun nodups msg loc l =
  let fun dup [] = l
        | dup (h::t) = if List.exists (fn x => x = h) t then
                         syntaxErrorAt loc (msg h)
                       else
                         dup t
  in  dup l
  end
val lambdaDups   =
  nodups (fn x => "formal parameter " ^ x ^ " appears twice in lambda")
fun defineDups loc f =
  nodups (fn x => "formal parameter " ^ x ^ " appears twice in " ^
                  "definition of function " ^ f) loc
fun letDups loc (LETSTAR, bindings) = bindings
  | letDups loc (kind,    bindings) =
      let val names    = map (fn (n, _) => n) bindings
          val kindName = case kind of LET => "let" | LETREC => "letrec" | _ =>
                                                                            "??"
          fun msg x = "name " ^ x ^ " appears twice in " ^ kindName
      in  nodups msg loc names;
          bindings
      end
(* parsing 575 *)
fun parse (NAME s)  = VAR s
  | parse (INT n)   = LITERAL (NUM n)
  | parse (SHARP c) = LITERAL (BOOL (c = #"t"))
  | parse (LIST (loc, NAME "quote"  :: args)) = parseQuote  loc args
  | parse (LIST (loc, NAME "if"     :: args)) = parseIf     loc args
  | parse (LIST (loc, NAME "while"  :: args)) = parseWhile  loc args
  | parse (LIST (loc, NAME "set"    :: args)) = parseSet    loc args
  | parse (LIST (loc, NAME "begin"  :: args)) = BEGIN (map parse args)
  | parse (LIST (loc, NAME "lambda" :: args)) = parseLambda loc args
  | parse (LIST (loc, NAME "let"    :: args)) = parseLet loc LET     args
  | parse (LIST (loc, NAME "letrec" :: args)) = parseLet loc LETREC  args
  | parse (LIST (loc, NAME "let*"   :: args)) = parseLet loc LETSTAR args
  | parse (LIST (loc, h::t))  = APPLY (parse h, map parse t)
  | parse (LIST (loc, [])) = raise SyntaxError "empty application ()"
and parseQuote loc [e] = LITERAL (quote loc e)
  | parseQuote loc _ = syntaxErrorAt loc "expected (quote e)"
and parseIf _ [e1, e2, e3] = IFX (parse e1, parse e2, parse e3)
  | parseIf loc _ = syntaxErrorAt loc "expected (if e1 e2 e3)"
and parseWhile _ [e1, e2] = WHILEX (parse e1, parse e2)
  | parseWhile loc _ = syntaxErrorAt loc "expected (while e1 e2)"
and parseSet _ [NAME v, e] = SET (v, parse e)
  | parseSet loc _ = syntaxErrorAt loc "expected (set v e)"
and parseLambda _ [LIST (loc, formals), body] =
      LAMBDA (lambdaDups loc (map (unname loc) formals), parse body)
  | parseLambda loc _ =
      syntaxErrorAt loc "expected (lambda (names) body)"
and parseLet _ kind [LIST (loc, bindings), body] = 
      LETX(kind, letDups loc (kind, map (unBinding loc) bindings), parse body)
  | parseLet loc _ _ = syntaxErrorAt loc "expected (let (bindings) body)"
and unname _ (NAME n) = n
  | unname loc _ = syntaxErrorAt loc
                                 "formal parameters to function must be symbols"
and unBinding _ (LIST (_, [NAME n, e])) = (n, parse e)
  | unBinding loc _ = syntaxErrorAt loc "expected (v e) in bindings"
and quote loc (INT   n) = NUM n
  | quote loc (SHARP c) = BOOL (c = #"t")
  | quote loc (NAME  ".") =
      syntaxErrorAt loc
                      "this interpreter cannot handle . in quoted S-expressions"
  | quote loc (NAME  n) = SYM n
  | quote _ (LIST (loc, l)) = embedList (map (quote loc) l)
(* parsing 576a *)
fun toplevel (LIST (loc, NAME "define" :: args)) = parseDefine loc args
  | toplevel (LIST (loc, NAME "val"    :: args)) = parseVal    loc args
  | toplevel (LIST (loc, NAME "global" :: args)) = parseVal    loc args (*
                                                                       legacy *)
  | toplevel (LIST (loc, NAME "use"    :: args)) = parseUse    loc args
  | toplevel e = EXP (parse e)
and parseDefine _ [NAME name, LIST (loc, formals), body] =
      DEFINE (name, (defineDups loc name (map (unname loc) formals), parse body)
                                                                               )
  | parseDefine loc _ = syntaxErrorAt loc "expected (define f (args) body)"
and parseVal _ [NAME v, e] = VAL (v, parse e)
  | parseVal loc _ = syntaxErrorAt loc "expected (val v e)"
and parseUse _ [NAME v] = USE v
  | parseUse loc _ = syntaxErrorAt loc "expected (use filename)"


(*****************************************************************)
(*                                                               *)
(*   READERS                                                     *)
(*                                                               *)
(*****************************************************************)

(* readers 576b *)
type reader = { firstline : unit -> string
              , nextline  : unit -> string
              , curloc    : unit -> string * int
              }
(* readers 576c *)
fun promptReader (firstPrompt, nextPrompt, reader) =
  let val { firstline = first, nextline = next, curloc = cur } = reader
      fun promptIn get prompt () = ( TextIO.output(TextIO.stdOut, prompt)
                                   ; TextIO.flushOut(TextIO.stdOut)
                                   ; get ()
                                   )
  in  { firstline = promptIn first firstPrompt
      , nextline  = promptIn next  nextPrompt
      , curloc    = cur
      }
  end
(* type declarations for consistency checking *)
val _ = op promptReader : string * string * reader -> reader
(* readers 577a *)
fun filereader (filename, fd)=
  let val lineNum = ref 0
      fun get () =
        let val line = TextIO.inputLine fd
        in  if size line = 0 then raise EOF else line before lineNum := !lineNum
                                                                             + 1
        end
  in  { firstline = get, nextline = get, curloc = (fn () => (filename, !lineNum)
                                                                             ) }
  end

fun stringsreader (what, l) = 
  let val buffer  = ref l
      val lineNum = ref 0
      fun get ()  = case !buffer
                      of [] => raise EOF
                       | h :: t => h before buffer := t before lineNum := !
                                                                     lineNum + 1
  in  { firstline = get, nextline = get, curloc = (fn () => (what, !lineNum)) }
  end              
(* type declarations for consistency checking *)
val _ = op filereader    : string * TextIO.instream -> reader
val _ = op stringsreader : string * string list     -> reader
(* readers 577b *)
type topreader = { buffer : toplevel list ref, reader : reader }
fun readtop (r as { buffer, reader = { nextline, firstline, curloc } }) =
  case !buffer
    of h::t => h before buffer := t
     | []   => ( buffer := map toplevel (read nextline curloc (firstline()))
               ; readtop r
               )
(* readers 577c *)
fun topreader (reader, prompt) = 
  let val reader = if prompt then promptReader ("-> ", "   ", reader) else
                                                                          reader
  in  { buffer = ref [], reader = reader }
  end
(* readers 578 *)
fun echoTag f x =
  let val line = f x
      val _ = if (String.substring (line, 0, 2) = ";#" handle _ => false) then
                print line
              else
                ()
  in  line
  end

fun echoReader { nextline=n, firstline=f, curloc=c } =
     { nextline=echoTag n, firstline=echoTag f, curloc=c } 

fun echoTop { buffer = b, reader = r } = { buffer = b, reader = echoReader r }
val topreader = fn args => echoTop (topreader args)
(* type declarations for consistency checking *)
val _ = op echoReader : reader -> reader 


(*****************************************************************)
(*                                                               *)
(*   IMPLEMENTATION OF [[USE]]                                   *)
(*                                                               *)
(*****************************************************************)

(* implementation of [[use]] 193c *)
fun use readEvalPrint filename rho =
      let val fd = TextIO.openIn filename
          val treader = topreader (filereader (filename, fd), false)
          fun writeln s = app print [s, "\n"]
      in  readEvalPrint (treader, writeln, writeln) rho
          before TextIO.closeIn fd
      end 


(*****************************************************************)
(*                                                               *)
(*   EVALUATION                                                  *)
(*                                                               *)
(*****************************************************************)

(* evaluation 189b *)
(* definition of [[separate]] 190d *)
fun separate (zero, sep) =  (* useful for printing *)
  let fun s []     = zero
        | s [x]    = x
        | s (h::t) = h ^ sep ^ s t
  in  s
end
fun eval(e, rho) =
  let fun ev(LITERAL n) = n
        (* more alternatives for [[ev]] 189c *)
        | ev(VAR v) = !(find(v, rho))
        | ev(SET (n, e)) = 
            let val v = ev e
            in  find (n, rho) := v;
                v
            end
        (* more alternatives for [[ev]] 190a *)
        | ev(IFX (e1, e2, e3)) = ev (if bool (ev e1) then e2 else e3)
        | ev(WHILEX (guard, body)) = 
            if bool (ev guard) then 
              (ev body; ev(WHILEX (guard, body)))
            else
              NIL
        | ev(BEGIN es) =
            let fun b(e::es, lastval) = b(es, ev e)
                  | b(   [], lastval) = lastval
            in  b(es, BOOL false)
            end
        (* more alternatives for [[ev]] 190b *)
        | ev(LAMBDA l) = CLOSURE(l, rho)
        | ev(APPLY (f, args))  = 
               (case ev f
                  of PRIMITIVE prim => prim (map ev args)
                   | CLOSURE clo    => (* apply closure [[clo]] to [[args]] 190c
                                                                              *)
                                       let val ((formals, body), savedrho) = clo
                                           val actuals = map ev args
                                       in  eval(body, bindList(formals, map ref
                                                             actuals, savedrho))
                                           handle BindListLength => 
                                               raise RuntimeError (
                                      "Wrong number of arguments to closure; " ^
                                                                   "expected ("
                                              ^ separate("", " ") formals ^ ")")
                                       end
                   | v => raise RuntimeError ("Applied non-function " ^
                                                                  valueString v)
               )
        (* more alternatives for [[ev]] 191a *)
        | ev(LETX (LET, bs, body)) =
            let val (names, values) = ListPair.unzip bs
        (* type declarations for consistency checking *)
        val _ = ListPair.unzip : ('a * 'b) list -> 'a list * 'b list
            in  eval (body, bindList(names, map (ref o ev) values, rho))
            end
        | ev(LETX (LETSTAR, bs, body)) =
            let fun step ((n, e), rho) = bind(n, ref (eval(e, rho)), rho)
            in  eval (body, foldl step rho bs)
            end
        | ev(LETX (LETREC, bs, body)) =
            let val (names, values) = ListPair.unzip bs
                val rho' = bindList(names, map (fn _ => ref NIL) values, rho)
                val bs = map (fn (n, e) => (n, eval(e, rho'))) bs
            in  List.app (fn (n, v) => find(n, rho') := v) bs; 
                eval(body, rho')
            end
  in  ev e
  end
(* type declarations for consistency checking *)
val _ = op eval : exp * value ref env -> value
(* evaluation 191b *)
fun arityError n args =
  raise RuntimeError ("primitive function expected " ^ Int.toString n ^
                      "arguments; got " ^ Int.toString (length args))
fun binaryOp f = (fn [a, b] => f(a, b) | args => arityError 2 args)
fun unaryOp  f = (fn [a]    => f a     | args => arityError 1 args)
(* type declarations for consistency checking *)
val _ = op unaryOp  : (value         -> value) -> (value list -> value)
val _ = op binaryOp : (value * value -> value) -> (value list -> value)
(* evaluation 191c *)
fun arithOp f = binaryOp (fn (NUM n1, NUM n2) => NUM (f(n1, n2)) 
                           | _ => raise RuntimeError "integers expected")
(* type declarations for consistency checking *)
val _ = op arithOp: (int * int -> int) -> (value list -> value)
(* evaluation 192b *)
fun predOp f     = unaryOp  (embedPredicate f)
fun comparison f = binaryOp (embedPredicate f)
fun intcompare f = comparison (fn (NUM n1, NUM n2) => f(n1, n2)
                                | _ => raise RuntimeError "integers expected")
(* type declarations for consistency checking *)
val _ = op predOp     : (value         -> bool) -> (value list -> value)
val _ = op comparison : (value * value -> bool) -> (value list -> value)
val _ = op intcompare : (int   * int   -> bool) -> (value list -> value)
(* evaluation 193b *)
fun topeval (t, rho, echo) =
  case t
    of USE filename => use readEvalPrint filename rho
     | EXP e        => (echo (valueString (eval(e, rho))); rho)
     | DEFINE (name, lambda) => 
         topeval (VAL (name, LAMBDA lambda), rho, echo)
     | VAL (name, e) =>
         let val rho = (find(name, rho); rho)
                       handle NotFound _ => bind (name, ref NIL, rho)
             val v   = eval(e, rho)
         in  find(name, rho) := v;
             echo (showVal name e v);
             rho
         end
and showVal name (LAMBDA _) _ = name
  | showVal name _          v = valueString v
(* evaluation 194a *)
and readEvalPrint (reader, echo, errmsg) rho =
  let fun loop rho =
        let fun continue msg = (errmsg msg; loop rho)
            fun finish () = rho
        in  loop (topeval (readtop reader, rho, echo))
            handle EOF => finish()
                   (* more read-eval-print handlers 194b *)
                   | IO.Io {name, ...} => continue ("I/O error: " ^ name)
                   | Paren (_, loc)    => continue ("syntax error: extra ) in "
                                                             ^ srclocString loc)
                   | PrematureEOF s    => (errmsg (
                                  "Missing )? Got EOF reading (" ^ s); finish())
                   | SyntaxError msg   => continue ("syntax error: " ^ msg)
                   (* more read-eval-print handlers 194c *)
                   | Div               => continue "Division by zero"
                   | Overflow          => continue "Arithmetic overflow"
                   | RuntimeError msg  => continue ("run-time error: " ^ msg)
                   | NotFound n        => continue ("variable " ^ n ^
                                                                   " not found")
        end
  in  loop rho
  end
(* type declarations for consistency checking *)
val _ = op topeval : toplevel * value ref env * (string->unit) -> value ref env
(* type declarations for consistency checking *)
val _ = op readEvalPrint : topreader * (string->unit) * (string->unit) -> value
                                                        ref env -> value ref env


(*****************************************************************)
(*                                                               *)
(*   INITIALIZATION                                              *)
(*                                                               *)
(*****************************************************************)

(* initialization 195a *)
fun initialEnv() =
  let val rho = foldl (fn ((name, prim), rho) => bind(name, ref (PRIMITIVE prim)
                                                                         , rho))
                      emptyEnv ((* primitives [[::]] 192a *)
                                ("+", arithOp op +  ) :: 
                                ("-", arithOp op -  ) :: 
                                ("*", arithOp op *  ) :: 
                                ("/", arithOp op div) ::
                                (* primitives [[::]] 192c *)
                                ("<", intcompare op <) :: 
                                (">", intcompare op >) ::
                                ("=", comparison (fn (NIL,     NIL    ) => true
                                                   | (NUM  n1, NUM  n2) => n1 =
                                                                              n2
                                                   | (SYM  v1, SYM  v2) => v1 =
                                                                              v2
                                                   | (BOOL b1, BOOL b2) => b1 =
                                                                              b2
                                                   |  _                 => false
                                                                           )) ::
                                ("null?",    predOp (fn (NIL   ) => true | _ =>
                                                                      false)) ::
                                ("boolean?", predOp (fn (BOOL _) => true | _ =>
                                                                      false)) ::
                                ("number?",  predOp (fn (NUM  _) => true | _ =>
                                                                      false)) ::
                                ("symbol?",  predOp (fn (SYM  _) => true | _ =>
                                                                      false)) ::
                                ("pair?",    predOp (fn (PAIR _) => true | _ =>
                                                                      false)) ::
                                ("procedure?",
                                             predOp (fn (PRIMITIVE _) => true |
                                          (CLOSURE  _) => true | _ => false)) ::
                                (* primitives [[::]] 192d *)
                                ("cons", binaryOp (fn (a, b) => PAIR(a, b))) ::
                                ("car",  unaryOp  (fn (PAIR(car, _)) => car 
                                                    | v => raise RuntimeError
                                                                   (
                                "car applied to non-list " ^ valueString v))) ::
                                ("cdr",  unaryOp  (fn (PAIR(_, cdr)) => cdr 
                                                    | v => raise RuntimeError
                                                                   (
                                "cdr applied to non-list " ^ valueString v))) ::
                                (* primitives [[::]] 193a *)
                                ("print", unaryOp (fn v => (print (valueString v
                                                              ^"\n"); v)))    ::
                                ("error", unaryOp (fn v => raise RuntimeError (
                                                        valueString v))) :: nil)
      val basis  = (* ML representation of initial basis *)
                     [ "; scheme.nw 444"
                     , "(define caar (l) (car (car l)))"
                     , "(define cadr (l) (car (cdr l)))"
                     , "(define cdar (l) (cdr (car l)))"
                     , "; scheme.nw 464"
                     , "(define list1 (x)     (cons x '()))"
                     , "(define list2 (x y)   (cons x (list1 y)))"
                     , "(define list3 (x y z) (cons x (list2 y z)))"
                     , "; scheme.nw 485"
                     , "(define length (l)"
                     , "  (if (null? l) 0"
                     , "    (+ 1 (length (cdr l)))))"
                     , "; scheme.nw 3712"
                     , "(define and (b c) (if b  c  b))"
                     , "(define or  (b c) (if b  b  c))"
                     , "(define not (b)   (if b #f #t))"
                     , "; scheme.nw 528"
                     ,
"(define atom? (x) (or (number? x) (or (symbol? x) (or (boolean? x) (null? x)))))"
                     , "(define equal? (s1 s2)"
                     , "  (if (or (atom? s1) (atom? s2))"
                     , "    (= s1 s2)"
                     ,
             "    (and (equal? (car s1) (car s2)) (equal? (cdr s1) (cdr s2)))))"
                     , "; scheme.nw 548"
                     , "(define append (l1 l2)"
                     , "  (if (null? l1)"
                     , "     l2"
                     , "     (cons (car l1) (append (cdr l1) l2))))"
                     , "; scheme.nw 580"
                     , "(define revapp (l1 l2)"
                     , "  (if (null? l1)"
                     , "     l2"
                     , "     (revapp (cdr l1) (cons (car l1) l2))))"
                     , "; scheme.nw 737"
                     , "(define bind (x y alist)"
                     , "  (if (null? alist)"
                     , "    (list1 (list2 x y))"
                     , "    (if (equal? x (caar alist))"
                     , "      (cons (list2 x y) (cdr alist))"
                     , "      (cons (car alist) (bind x y (cdr alist))))))"
                     , "(define find (x alist)"
                     , "  (if (null? alist) '()"
                     , "    (if (equal? x (caar alist))"
                     , "      (car (cdar alist))"
                     , "      (find x (cdr alist)))))"
                     , "; scheme.nw 1636"
                     , "(define o (f g) (lambda (x) (f (g x))))"
                     , "; scheme.nw 1706"
                     , "(define curry   (f) (lambda (x) (lambda (y) (f x y))))"
                     , "(define uncurry (f) (lambda (x y) ((f x) y)))"
                     , "; scheme.nw 2014"
                     , "(define filter (p? l)"
                     , "  (if (null? l)"
                     , "    '()"
                     , "    (if (p? (car l))"
                     , "      (cons (car l) (filter p? (cdr l)))"
                     , "      (filter p? (cdr l)))))"
                     , "; scheme.nw 2026"
                     , "(define map (f l)"
                     , "  (if (null? l)"
                     , "    '()"
                     , "    (cons (f (car l)) (map f (cdr l)))))"
                     , "; scheme.nw 2036"
                     , "(define exists? (p? l)"
                     , "  (if (null? l)"
                     , "    #f"
                     , "    (if (p? (car l)) "
                     , "      #t"
                     , "      (exists? p? (cdr l)))))"
                     , "(define all? (p? l)"
                     , "  (if (null? l)"
                     , "    #t"
                     , "    (if (p? (car l))"
                     , "      (all? p? (cdr l))"
                     , "      #f)))"
                     , "; scheme.nw 2067"
                     , "(define foldr (op zero l)"
                     , "  (if (null? l)"
                     , "    zero"
                     , "    (op (car l) (foldr op zero (cdr l)))))"
                     , "(define foldl (op zero l)"
                     , "  (if (null? l)"
                     , "    zero"
                     , "    (foldl op (op (car l) zero) (cdr l))))"
                     , "; scheme.nw 3730"
                     , "(define <= (x y) (not (> x y)))"
                     , "(define >= (x y) (not (< x y)))"
                     , "(define != (x y) (not (= x y)))"
                     , "; scheme.nw 3736"
                     , "(define max (x y) (if (> x y) x y))"
                     , "(define min (x y) (if (< x y) x y))"
                     , "; scheme.nw 3743"
                     , "(define mod (m n) (- m (* n (/ m n))))"
                     , "(define gcd (m n) (if (= n 0) m (gcd n (mod m n))))"
                     , "(define lcm (m n) (* m (/ n (gcd m n))))"
                     , "; scheme.nw 3755"
                     , "(define caar (x) (car (car x)))"
                     , "(define cdar (x) (cdr (car x)))"
                     , "(define cadr (x) (car (cdr x)))"
                     , "(define cddr (x) (cdr (cdr x)))"
                     , "(define caaar (x) (car (caar x)))"
                     , "(define cdaar (x) (cdr (caar x)))"
                     , "(define caadr (x) (car (cadr x)))"
                     , "(define cdadr (x) (cdr (cadr x)))"
                     , "(define cadar (x) (car (cdar x)))"
                     , "(define cddar (x) (cdr (cdar x)))"
                     , "(define caddr (x) (car (cddr x)))"
                     , "(define cdddr (x) (cdr (cddr x)))"
                     , "; scheme.nw 3771"
                     , "(define list1 (x)               (cons x '()))"
                     , "(define list2 (x y)             (cons x (list1 y)))"
                     , "(define list3 (x y z)           (cons x (list2 y z)))"
                     , "(define list4 (x y z a)         (cons x (list3 y z a)))"
                     ,
                     "(define list5 (x y z a b)       (cons x (list4 y z a b)))"
                     ,
                   "(define list6 (x y z a b c)     (cons x (list5 y z a b c)))"
                     ,
                 "(define list7 (x y z a b c d)   (cons x (list6 y z a b c d)))"
                     ,
               "(define list8 (x y z a b c d e) (cons x (list7 y z a b c d e)))"
                     ]
      val treader = topreader (stringsreader ("initial basis", basis), false)
  in  readEvalPrint (treader, fn _ => (), fn _ => ()) rho
  end
(* initialization 195b *)
fun runInterpreter noisy = 
  let val rho = initialEnv()
      fun writeln s = app print [s, "\n"]
      val treader = topreader (filereader ("standard input", TextIO.stdIn),
                                                                          noisy)
  in  ignore (readEvalPrint (treader, writeln, writeln) rho) handle EOF => ()
  end 
(* type declarations for consistency checking *)
val _ = op runInterpreter : bool -> unit


(*****************************************************************)
(*                                                               *)
(*   COMMAND LINE                                                *)
(*                                                               *)
(*****************************************************************)

(* command line *)
fun main ["-q"] = runInterpreter false
  | main []     = runInterpreter true
  | main _      =
      TextIO.output(TextIO.stdErr, "Usage: " ^ CommandLine.name() ^ " [-q]\n")
val _ = main (CommandLine.arguments())
