(* usm.sml 397 *)


(*****************************************************************)
(*                                                               *)
(*   ENVIRONMENTS                                                *)
(*                                                               *)
(*****************************************************************)

(* environments 186 *)
type name = string
type 'a env = (name * 'a) list
val emptyEnv = []

(* lookup and assignment of existing bindings *)
exception NotFound of name
fun find (name, []) = raise NotFound name
  | find (name, (n, v)::tail) = if name = n then v else find(name, tail)

(* adding new bindings *)
exception BindListLength
fun bind(name, v, rho) = (name, v) :: rho
fun bindList(n::vars, v::vals, rho) = bindList(vars, vals, bind(n, v, rho))
  | bindList([], [], rho) = rho
  | bindList _ = raise BindListLength
(* type declarations for consistency checking *)
val _ = op emptyEnv : 'a env
val _ = op find     : name * 'a env -> 'a
val _ = op bind     : name      * 'a      * 'a env -> 'a env
val _ = op bindList : name list * 'a list * 'a env -> 'a env


(*****************************************************************)
(*                                                               *)
(*   LEXICAL ANALYSIS                                            *)
(*                                                               *)
(*****************************************************************)

(* lexical analysis 599a *)
type srcloc = string * int
val nullsrc = ("internally generated SEND node", 1)
fun srclocString (source, line) = source ^ ", line " ^ Int.toString line
(* lexical analysis 599b *)
datatype par = INT   of int
             | NAME  of name
             | SHARP of par
             | LIST  of srcloc * par list
(* lexical analysis 600 *)

fun srclocString (source, line) = source ^ ", line " ^ Int.toString line
exception Paren of char list * srcloc
   (* raise if found ) when reading input *)
exception PrematureEOF of string (* shows what we were reading ... *)
exception EOF                    (* raised by reader *)
fun read nextline curloc : string -> par list =
  let (* character-fiddling functions 573a *)
      fun isDelim c = List.exists (fn c' => c' = c) [#"(", #")", #" ", #"\t",
                                                                    #"\n", #";"]
      (* character-fiddling functions 573b *)
      fun stringReader []     = NONE
        | stringReader (h::t) = SOME (h, t)
      val strton = Int.scan StringCvt.DEC stringReader
      (* type declarations for consistency checking *)
      val _ = op strton : char list -> (int * char list) option
      (* character-fiddling functions 573c *)
      fun skip s =
        case StringCvt.skipWS stringReader s
          of #";" :: s => skip (StringCvt.dropl (fn c => c <> #"\n")
                                                                 stringReader s)
           | s => s
      (* character-fiddling functions 573d *)
      fun nonwhite s =
        case skip s
          of [] => nonwhite (explode (nextline()))
           | s  => s
      val isDelim = fn c => isDelim c orelse c = #"[" orelse c = #"]"
      fun readmap f (result, unconsumed) = (f result, unconsumed)
      (* get called only when guaranteed non-blank, non-comment, non-empty *)
      fun get (#"("::s)  = readmap LIST (list (curloc(), s, []))
        | get (#")"::s)  = raise (Paren (s, curloc()))
        | get (#"["::s)  = let fun wrap (loc, l) =
                                 LIST (loc, NAME "block" :: LIST (loc, []) :: l)
                           in  readmap wrap (list (curloc(), s, []))
                           end
        | get (#"]"::s)  = raise (Paren (s, curloc()))
        | get (#"#"::s)  = readmap SHARP (get (nonwhite s))
        | get s          = atom (s, [])
      and atom(x::s, r) = if isDelim x then (mkAtom(rev r), x::s) 
                          else atom(s, x::r)
        | atom([],   r) = (mkAtom(rev r), [])
      and mkAtom s = (case strton s of SOME (n, []) => INT n | _ => NAME (
                                                                     implode s))
      and list (loc, s, l) =
        let val next = nonwhite s
        in  let val (x, u) = get next
            in  list(loc, u, x::l)
            end handle (Paren (u, _)) => ((loc, rev l), skip u)
                     | EOF => raise PrematureEOF (implode next)
        end handle EOF => raise PrematureEOF (implode s)
      fun readline [] = []
        | readline s  = 
             let val(p, u) = get s
             in  p :: readline(skip u)
             end
  in  fn s => readline(skip (explode s)) 
  end
(* type declarations for consistency checking *)
val _ = op read : (unit -> string) -> (unit -> string * int) -> string -> par
                                                                            list


(*****************************************************************)
(*                                                               *)
(*   ABSTRACT SYNTAX AND VALUES                                  *)
(*                                                               *)
(*****************************************************************)

(* abstract syntax and values 390c *)
datatype exp = VAR     of name
             | SET     of name * exp
             | SEND    of srcloc * name * exp * exp list
             | BEGIN   of exp list
             | BLOCK   of name list * exp list
             | LITERAL of rep
             | VALUE   of value
             | SUPER
(* abstract syntax and values 391a *)
and toplevel 
  = DEFINE of name * name list * exp
  | CLASSD of { name    : string
              , super   : string
              , ivars   : string list     (* instance variables *)
              , methods : (method_kind * name * method_impl) list
              }
  | EXP of exp
  | VAL of name * exp
  | USE of name
and method_kind = IMETHOD 
                | CMETHOD
and method_impl = USER_IMPL of name list * name list * exp
                | PRIM_IMPL of name
(* abstract syntax and values 391b *)
and    rep = USER     of value ref env (* collection of named instance variables
                                                                              *)
           | ARRAY    of value Array.array
           | NUM      of int
           | SYM      of name
           | CLOSURE  of name list * exp list * value ref env * class
           | CLASSREP of class
(* abstract syntax and values 391c *)
and class  = CLASS of { name    : name            (* name of the class *)
                      , super   : class option    (* superclass, if any *)
                      , ivars   : string list     (* instance variables *)
                      , methods : method env      (* both exported and private
                                                                              *)
                      , id      : int             (* uniquely identifies class
                                                                              *)
                      }
(* abstract syntax and values 392a *)
and method
  = PRIM_METHOD of name * (value * value list -> value)
  | USER_METHOD of { name : name, formals : name list, locals : name list, body
                                                                           : exp
                   , superclass : class (* used to send messages to super *)
                   }
(* abstract syntax and values 392b *)
withtype value = class * rep
(* abstract syntax and values 392c *)
exception RuntimeError  of string (* error caused by user *)
exception InternalError of string (* bug in the interpreter *)
(* abstract syntax and values 398a *)
fun className (CLASS {name, ...}) = name
fun classId   (CLASS {id,   ...}) = id
(* type declarations for consistency checking *)
val _ = op className : class -> name
val _ = op classId   : class -> int
(* abstract syntax and values 398b *)
fun methodName (PRIM_METHOD (n, _)) = n
  | methodName (USER_METHOD { name, ... }) = name
fun renameMethod (n, PRIM_METHOD (_, f)) = PRIM_METHOD (n, f)
  | renameMethod _ = raise InternalError "renamed user method"
fun methods l = foldl (fn (m, rho) => bind(methodName m, m, rho)) emptyEnv l
(* type declarations for consistency checking *)
val _ = op methodName   : method -> name
val _ = op methods      : method list -> method env
val _ = op renameMethod : name * method -> method
(* abstract syntax and values 398c *)
local 
  val n = ref 10 (* new classes start here *)
  fun uid () = !n before n := !n + 1
in
  fun mkClass name super ivars ms =
    CLASS { name = name, super = SOME super, ivars = ivars , methods = methods
                                                                             ms,
            id = uid () }
end
(* type declarations for consistency checking *)
val _ = op mkClass : name -> class -> name list -> method list -> class


(*****************************************************************)
(*                                                               *)
(*   BOOTSTRAPPING SUPPORT                                       *)
(*                                                               *)
(*****************************************************************)

(* bootstrapping support 399a *)
local 
  val intClass    = ref NONE : class option ref
  val symbolClass = ref NONE : class option ref
  val arrayClass  = ref NONE : class option ref
  fun badlit what = 
    raise InternalError
        ("(bootstrapping) -- cannot " ^ what ^ " anywhere in initial basis")
in
  fun mkInteger n = (valOf (!intClass), NUM n)
    handle Option => badlit "evaluate integer literal or use array literal"
  
  fun mkSymbol s = (valOf (!symbolClass), SYM s)
    handle Option => badlit "evaluate symbol literal or use array literal"
  
  fun mkArray a = (valOf (!arrayClass), ARRAY (Array.fromList a))
    handle Option => badlit "use array literal"
(* type declarations for consistency checking *)
val _ = op mkInteger : int        -> value
val _ = op mkSymbol  : string     -> value
val _ = op mkArray   : value list -> value
(* bootstrapping support 399b *)
  fun findInitialClass (name, xi) =
        case !(find(name, xi))
          of (_, CLASSREP c) => c
           | _ => raise InternalError (name ^
                                         " is not a class in the initial basis")
  fun closeLiteralsCycle xi =
    ( intClass    := SOME (findInitialClass ("SmallInteger", xi))
    ; symbolClass := SOME (findInitialClass ("Symbol",       xi))
    ; arrayClass  := SOME (findInitialClass ("Array",        xi))
    )
end
(* type declarations for consistency checking *)
val _ = op findInitialClass : string * value ref env -> class
(* bootstrapping support 400a *)
local
  val trueValue  = ref NONE : value option ref
  val falseValue = ref NONE : value option ref
in
  fun mkBoolean b = valOf(!(if b then trueValue else falseValue))
    handle Option => 
        raise InternalError 
            "Bad booleanClass; evaluated boolean expression in initial basis?"
  fun closeBooleansCycle xi =
    ( trueValue  := SOME (!(find ("true",  xi)))
    ; falseValue := SOME (!(find ("false", xi)))
    )
end
(* type declarations for consistency checking *)
val _ = op mkBoolean : bool -> value
(* bootstrapping support 400b *)
local
  val blockClass = ref NONE : class option ref
in
  fun mkBlock c = (valOf (!blockClass), CLOSURE c)
    handle Option => 
        raise InternalError 
            "Bad blockClass; evaluated block expression in initial basis?"
  fun closeBlocksCycle xi =
    blockClass := SOME (findInitialClass ("Block", xi))
end
(* type declarations for consistency checking *)
val _ = op mkBlock : name list * exp list * value ref env * class -> value


(*****************************************************************)
(*                                                               *)
(*   PARSING                                                     *)
(*                                                               *)
(*****************************************************************)

(* parsing 601a *)
exception SyntaxError of string
val unknown = ("unknown source", ~99)
fun syntaxErrorAt (source, line) msg =
      if (source, line) = unknown then
        raise SyntaxError msg
      else
        raise SyntaxError (source ^ ", line " ^ Int.toString line ^ ": " ^ msg)
(* parsing 601b *)
fun arity "if"    = 2
  | arity "while" = 1
  | arity name =
      let val cs = explode name
      in  if Char.isAlpha (hd cs) then
            length (List.filter (fn c => c = #":") cs)
          else
            1
      end

fun arityOk "value" _ = true
  | arityOk name args = arity name = length args

fun arityErrorAt loc what msgname args =
  let fun argn n = if n = 1 then "1 argument" else Int.toString n ^ " arguments"
  in  syntaxErrorAt loc ("in " ^ what ^ ", message " ^ msgname ^ " expects " ^
                         argn (arity msgname) ^ ", but gets " ^
                         argn (length args))
  end
(* parsing 602a *)
fun parse loc (INT n) = LITERAL (NUM n)
  | parse loc (NAME "super") = SUPER
  | parse loc (NAME n) = VAR n
  | parse loc (SHARP (NAME n)) = LITERAL (SYM n)
  | parse loc (SHARP (INT  n)) = LITERAL (SYM (Int.toString n))
  | parse loc (SHARP p) = VALUE (sharp loc p) (* better not happen in initial
                                                                        basis *)
  | parse loc (LIST (_, NAME "set"    :: args)) = parseSet   loc args
  | parse loc (LIST (_, NAME "begin"  :: args)) = BEGIN (map (parse loc) args)
  | parse loc (LIST (_, NAME "block"  :: args)) = parseBlock loc args
  | parse loc (LIST (_, NAME "locals" :: args)) =
      syntaxErrorAt loc "expected and expression but found '(locals ...)'"
  | parse _   (LIST (loc, msg::receiver::args)) =
     let val msgname = unname loc "method names" msg
     in  if arityOk msgname args then
           SEND (loc, msgname, parse loc receiver, map (parse loc) args)
         else
           arityErrorAt loc "message send" msgname args
     end
  | parse _ (LIST (loc, [NAME x])) =
      syntaxErrorAt loc ("sent message " ^ x ^ " to no object")
  | parse _ (LIST (loc, [x])) = syntaxErrorAt loc "message send to no object"
  | parse _ (LIST (loc, []))  = syntaxErrorAt loc "empty message send ()"
and parseSet loc [NAME v, e] = SET (canSet loc v, parse loc e)
  | parseSet loc _ = syntaxErrorAt loc "expected (set v e)"
and canSet loc n =
  if List.exists (fn n' => n' = n) ["true", "false", "nil", "self", "super"]
                                                                            then
    syntaxErrorAt loc ("you cannot set or val-bind pseudovariable " ^ n)
  else
    n
(* parsing 602b *)
and parseBlock loc (LIST (floc, formals) :: body) =
      BLOCK (map (unname floc "block parameters") formals, map (parse loc) body)
  | parseBlock loc _ = 
      syntaxErrorAt loc "block must have formal-parameter list"
and unname loc ? (NAME n) = if String.sub(n, 0) = #"#" then badName loc ? else n
  | unname loc ? _        = badName loc ?
and badName loc ? = syntaxErrorAt loc (? ^ " must be ordinary names")
and sharp loc (NAME n)        = mkSymbol n
  | sharp loc (INT n)         = mkInteger n
  | sharp _   (LIST (loc, l)) = mkArray (map (sharp loc) l)
  | sharp loc (SHARP _)       = syntaxErrorAt loc ("# within # is not legal")
(* parsing 603a *)
fun toplevel (LIST (loc, NAME "define"      :: args)) = parseDefine   loc args
  | toplevel (LIST (loc, NAME "class"       :: args)) = parseClassDef loc args
  | toplevel (LIST (loc, NAME "val"         :: args)) = parseVal      loc args
  | toplevel (LIST (loc, NAME "use"         :: args)) = parseUse      loc args
  | toplevel (LIST (loc, NAME "method"      :: args)) = methodError loc "method"
                                                                            args
  | toplevel (LIST (loc, NAME "classMethod" :: args)) =
                                                   methodError loc "classMethod"
                                                                            args
  | toplevel e = EXP (parse unknown e)
and methodError loc s args = 
  syntaxErrorAt loc
    ("unexpected `(" ^ s ^ " " ^ mName args ^ "...'; " ^
     "did a class definition end prematurely?")
and mName (NAME n :: _) = n ^ " "
  | mName _ = ""
(* parsing 603b *)
and parseDefine loc l = DEFINE (parseDefinition loc l)
and parseDefinition loc [NAME name, LIST (floc, formals), body] =
      (name, map (unname floc "formal parameters") formals, parse loc body)
  | parseDefinition loc _ = syntaxErrorAt loc "expected (define f (args) body)"
(* parsing 604a *)
and parseClassDef loc (NAME name :: NAME super :: LIST (rloc, rep) :: methods) =
      CLASSD    { name = name, super = super
                , ivars = map (unname rloc "instance variables of class") rep
                , methods = map (parseMethod loc) methods
                }
  | parseClassDef loc _ = syntaxErrorAt loc
                                     "expected (class name super (rep) methods)"
and parseMethod _ (LIST (loc, NAME keyword :: NAME name :: impl)) =
      (parseMethodKind loc keyword, name, parseMethodImpl loc keyword name impl)
  | parseMethod loc _ = syntaxErrorAt loc "expected (method f (args) body)"
and parseMethodKind loc "method" = IMETHOD
  | parseMethodKind loc "classMethod" = CMETHOD
  | parseMethodKind loc n = syntaxErrorAt loc ("keyword `" ^ n ^
                                                     "' cannot define a method")
and parseMethodImpl loc _ _ [NAME "primitive", NAME name] = PRIM_IMPL name
  | parseMethodImpl loc key name (LIST (_, formals) ::
                                  LIST (_, NAME "locals" :: locals) :: body) =
      if arityOk name formals then
        USER_IMPL (parms loc "parameters" key name formals,
                   parms loc "locals" key name locals, 
                   BEGIN (map (parse loc) body))
      else
        arityErrorAt loc (key ^ " definition") name formals
  | parseMethodImpl loc key name (LIST (_, formals) :: body) =
      if arityOk name formals then
        USER_IMPL (parms loc "parameters" key name formals, [],
                   BEGIN (map (parse loc) body))
      else
        arityErrorAt loc (key ^ " definition") name formals
  | parseMethodImpl loc key name _ =
      syntaxErrorAt loc ("expected method definition (" ^
                       key ^ " " ^ name ^ " (args) body)")
and parms loc what key name = map (unname loc (what ^ " of " ^ key ^ " " ^ name)
                                                                               )
(* parsing 604b *)
and parseVal loc [NAME v, e] = VAL (canSet loc v, parse loc e)
  | parseVal loc _ = syntaxErrorAt loc "expected (val v e)"
(* parsing 604c *)
and parseUse loc [NAME v] = USE v
  | parseUse loc _ = syntaxErrorAt loc "expected (use filename)"


(*****************************************************************)
(*                                                               *)
(*   READERS                                                     *)
(*                                                               *)
(*****************************************************************)

(* readers 576b *)
type reader = { firstline : unit -> string
              , nextline  : unit -> string
              , curloc    : unit -> string * int
              }
(* readers 576c *)
fun promptReader (firstPrompt, nextPrompt, reader) =
  let val { firstline = first, nextline = next, curloc = cur } = reader
      fun promptIn get prompt () = ( TextIO.output(TextIO.stdOut, prompt)
                                   ; TextIO.flushOut(TextIO.stdOut)
                                   ; get ()
                                   )
  in  { firstline = promptIn first firstPrompt
      , nextline  = promptIn next  nextPrompt
      , curloc    = cur
      }
  end
(* type declarations for consistency checking *)
val _ = op promptReader : string * string * reader -> reader
(* readers 577a *)
fun filereader (filename, fd)=
  let val lineNum = ref 0
      fun get () =
        let val line = TextIO.inputLine fd
        in  if size line = 0 then raise EOF else line before lineNum := !lineNum
                                                                             + 1
        end
  in  { firstline = get, nextline = get, curloc = (fn () => (filename, !lineNum)
                                                                             ) }
  end

fun stringsreader (what, l) = 
  let val buffer  = ref l
      val lineNum = ref 0
      fun get ()  = case !buffer
                      of [] => raise EOF
                       | h :: t => h before buffer := t before lineNum := !
                                                                     lineNum + 1
  in  { firstline = get, nextline = get, curloc = (fn () => (what, !lineNum)) }
  end              
(* type declarations for consistency checking *)
val _ = op filereader    : string * TextIO.instream -> reader
val _ = op stringsreader : string * string list     -> reader
(* readers 577b *)
type topreader = { buffer : toplevel list ref, reader : reader }
fun readtop (r as { buffer, reader = { nextline, firstline, curloc } }) =
  case !buffer
    of h::t => h before buffer := t
     | []   => ( buffer := map toplevel (read nextline curloc (firstline()))
               ; readtop r
               )
(* readers 577c *)
fun topreader (reader, prompt) = 
  let val reader = if prompt then promptReader ("-> ", "   ", reader) else
                                                                          reader
  in  { buffer = ref [], reader = reader }
  end
(* readers 578 *)
fun echoTag f x =
  let val line = f x
      val _ = if (String.substring (line, 0, 2) = ";#" handle _ => false) then
                print line
              else
                ()
  in  line
  end

fun echoReader { nextline=n, firstline=f, curloc=c } =
     { nextline=echoTag n, firstline=echoTag f, curloc=c } 

fun echoTop { buffer = b, reader = r } = { buffer = b, reader = echoReader r }
val topreader = fn args => echoTop (topreader args)
(* type declarations for consistency checking *)
val _ = op echoReader : reader -> reader 


(*****************************************************************)
(*                                                               *)
(*   PRIMITIVE BASICS                                            *)
(*                                                               *)
(*****************************************************************)

(* primitive basics 400c *)
type primitive = value * value list -> value
fun arityError n (receiver, args) =
  raise RuntimeError ("primitive method expected " ^ Int.toString n ^
                      " arguments; got " ^ Int.toString (length args))
fun unaryPrim  f = (fn (a, [])  => f a     | ra => arityError 0 ra)
fun binaryPrim f = (fn (a, [b]) => f(a, b) | ra => arityError 1 ra)
fun primMethod name f = PRIM_METHOD (name, f)
(* type declarations for consistency checking *)
val _ = op unaryPrim  : (value         -> value) -> primitive
val _ = op binaryPrim : (value * value -> value) -> primitive
val _ = op primMethod : name -> primitive -> method
(* primitive basics 401a *)
fun userMethod name formals locals body = 
  let val super = CLASS { name = "should never be used", super = NONE, ivars = [
                                                                              ],
                          methods = [], id = 0 }
  in  USER_METHOD { name = name, formals = formals, locals = locals, body = exp
                                                                           body,
                    superclass = super }
  end
and exp s = 
  let fun nextline () = raise PrematureEOF "parsing internal exp"
      fun curloc   () = ("(internal expression)", 1)
  in  parse (curloc()) (hd (read nextline curloc s))
  end
(* type declarations for consistency checking *)
val _ = op exp : string -> exp


(*****************************************************************)
(*                                                               *)
(*   PRIMITIVE METHODS FOR [[OBJECT]] AND [[UNDEFINEDOBJECT]]    *)
(*                                                               *)
(*****************************************************************)

(* primitive methods for [[Object]] and [[UndefinedObject]] 401b *)
fun eqRep ((cx, x), (cy, y)) = 
  classId cx = classId cy andalso
  case (x, y)
    of (ARRAY x,    ARRAY    y) => x = y
     | (NUM   x,    NUM      y) => x = y
     | (SYM   x,    SYM      y) => x = y
     | (USER  x,    USER     y) => find("self", x) = find("self", y)
     | (CLOSURE  x, CLOSURE  y) => false
     | (CLASSREP x, CLASSREP y) => classId x = classId y
     | _ => false
(* type declarations for consistency checking *)
val _ = op eqRep : value * value -> bool
(* primitive methods for [[Object]] and [[UndefinedObject]] 401d *)
fun defaultPrint (self as (c, _)) = ( app print ["<", className c, ">"]; self )
(* primitive methods for [[Object]] and [[UndefinedObject]] 402a *)
fun memberOf ((c, _), (_, CLASSREP c')) = mkBoolean (classId c = classId c')
  | memberOf _ = raise RuntimeError "argument of isMemberOf: must be a class"

fun kindOf ((c, _), (_, CLASSREP (CLASS {id=u', ...}))) =
      let fun subclassOfClassU' (CLASS {super, id=u, ... }) =
            u = u' orelse (case super of NONE => false | SOME c =>
                                                            subclassOfClassU' c)
      in  mkBoolean (subclassOfClassU' c)
      end
  | kindOf _ = raise RuntimeError "argument of isKindOf: must be a class"
(* primitive methods for [[Object]] and [[UndefinedObject]] 402b *)
fun error (_, (_, SYM s)) = raise RuntimeError s
  | error (_, (c, _    )) =
      raise RuntimeError ("error: got class " ^ className c ^
                                                            "; expected Symbol")


(*****************************************************************)
(*                                                               *)
(*   BUILT-IN CLASSES [[OBJECT]] AND [[UNDEFINEDOBJECT]]         *)
(*                                                               *)
(*****************************************************************)

(* built-in classes [[Object]] and [[UndefinedObject]] 406a *)
val objectClass = 
  CLASS { name = "Object", super = NONE, ivars = ["self"], id = 1
        , methods = methods
            [ primMethod "print"   (unaryPrim defaultPrint)
            , userMethod "println" [] []
                                     "(begin (print self) (print newline) self)"
            , primMethod "isNil"   (unaryPrim (fn _ => mkBoolean false))
            , primMethod "notNil"  (unaryPrim (fn _ => mkBoolean true))
            , primMethod "error:"  (binaryPrim error)
            , primMethod "="       (binaryPrim (mkBoolean o eqRep))
            , userMethod "!=" ["x"] [] "(not (= self x))"
            , primMethod "isKindOf:"    (binaryPrim kindOf)
            , primMethod "isMemberOf:"  (binaryPrim memberOf)
            , primMethod "subclassResponsibility"
               (unaryPrim
                  (fn _ => raise RuntimeError

           "subclass failed to implement a method that was its responsibility"))
            ]
        }
(* built-in classes [[Object]] and [[UndefinedObject]] 406b *)
val nilClass = 
  mkClass "UndefinedObject" objectClass []
    [ primMethod "isNil"  (unaryPrim (fn _ => mkBoolean true))
    , primMethod "notNil" (unaryPrim (fn _ => mkBoolean false))
    , primMethod "print"  (unaryPrim (fn x => (print "nil"; x)))
    ]
(* built-in classes [[Object]] and [[UndefinedObject]] 406c *)
val nilValue = 
  let val nilCell  = ref (nilClass, USER []) : value ref
      val nilValue = (nilClass, USER (bind("self", nilCell, emptyEnv)))
      val _        = nilCell := nilValue
  in  nilValue
  end


(*****************************************************************)
(*                                                               *)
(*   PRIMITIVE METHODS FOR REMAINING CLASSES                     *)
(*                                                               *)
(*****************************************************************)

(* primitive methods for remaining classes 402c *)
fun printInt (self as (_, NUM n)) =
      ( print (String.map (fn #"~" => #"-" | c => c) (Int.toString n))
      ; self
      )
  | printInt _ =
      raise RuntimeError ("cannot print when object inherits from Int")
(* primitive methods for remaining classes 402d *)
fun binaryInt mk $ ((_, NUM n), (_, NUM m)) = mk ($ (n, m))
  | binaryInt _ _  ((_, NUM n), (c, _)) =
      raise RuntimeError ("numeric primitive expected numeric argument, got <"
                          ^ className c ^ ">")
  | binaryInt _ _  ((c, _), _) =
      raise RuntimeError ("numeric primitive method defined on <" ^ className c
                                                                          ^ ">")
fun arithop    $ = binaryPrim (binaryInt mkInteger $)
fun intcompare $ = binaryPrim (binaryInt mkBoolean $)
(* primitive methods for remaining classes 402e *)
fun newInteger ((_, CLASSREP c), (_, NUM n)) = (c, NUM n)
  | newInteger _ = raise RuntimeError (
                                   "made new integer with non-int or non-class")
(* primitive methods for remaining classes 403b *)
fun printSymbol (self as (_, SYM s)) = (print s; self)
  | printSymbol _ = raise RuntimeError
                                 "cannot print when object inherits from Symbol"
(* primitive methods for remaining classes 403c *)
fun newSymbol ((_, CLASSREP c), (_, SYM s)) = (c, SYM s)
  | newSymbol _ = raise RuntimeError (
                                 "made new symbol with non-symbol or non-class")
(* primitive methods for remaining classes 403f *)
fun newArray ((_, CLASSREP c), (_, NUM n)) = (c, ARRAY (Array.array (n, nilValue
                                                                             )))
  | newArray _ = raise RuntimeError
                                "Array new sent to non-class or got non-integer"
(* primitive methods for remaining classes 403g *)
fun arrayPrimitive f ((_, ARRAY a), l) = f (a, l)
  | arrayPrimitive f _ = raise RuntimeError "Array primitive used on non-array"

fun arraySize (a, []) = mkInteger (Array.length a)
  | arraySize ra      = arityError 0 ra
(* primitive methods for remaining classes 404a *)
fun arrayAt (a, [(_, NUM n)]) = Array.sub(a, n - 1)  (* convert to 0-indexed *)
  | arrayAt (_, [_])  = raise RuntimeError "Non-integer used as array subscript"
  | arrayAt ra        = arityError 1 ra

fun arrayAtPut (a, [(_, NUM n), x]) = (Array.update(a, n-1, x); x)
  | arrayAtPut (_, [_, _]) = raise RuntimeError
                                           "Non-integer used as array subscript"
  | arrayAtPut ra          = arityError 2 ra
(* primitive methods for remaining classes 404d *)
fun classPrimitive f = 
  unaryPrim (fn (meta, CLASSREP c) => f (meta, c)
              | _ => raise RuntimeError "class primitive sent to non-class")
(* type declarations for consistency checking *)
val _ = op binaryInt  : ('a -> value) -> (int * int -> 'a)   -> value * value ->
                                                                           value
val _ = op arithop    :                  (int * int -> int)  -> primitive
val _ = op intcompare :                  (int * int -> bool) -> primitive
(* type declarations for consistency checking *)
val _ = op arrayPrimitive : (value array * value list -> value) -> primitive
(* type declarations for consistency checking *)
val _ = op classPrimitive : (class * class -> value) -> primitive
(* primitive methods for remaining classes 404e *)
local
  fun mkIvars (CLASS { ivars, super, ... }) =
    let val supervars = case super of NONE => emptyEnv | SOME c => mkIvars c
    in  foldl (fn (n, rho) => bind(n, ref nilValue, rho)) supervars ivars
    end
in
  fun newUserObject (_, c) =
        let val ivars = mkIvars c
            val self = (c, USER ivars)
        in  (find("self", ivars) := self; self)
        end
(* type declarations for consistency checking *)
val _ = op mkIvars       : class -> value ref env
val _ = op newUserObject : class * class -> value
end
(* primitive methods for remaining classes 405 *)
(* definition of [[separate]] 190d *)
fun separate (zero, sep) =  (* useful for printing *)
  let fun s []     = zero
        | s [x]    = x
        | s (h::t) = h ^ sep ^ s t
  in  s
end
local
  fun showProtocol doSuper kind c =
    let fun member x l = List.exists (fn x' => x' = x) l
        fun insert (x, []) = [x]
          | insert (x, (h::t)) =
              case compare x h
                of LESS    => x :: h :: t
                 | EQUAL   => x :: t (* replace *)
                 | GREATER => h :: insert (x, t)
        and compare (name, _) (name', _) = String.compare (name, name')
        fun methods (CLASS { super, methods = ms, name, ... }) =
              if not doSuper orelse (kind = "classMethod" andalso name = "Class"
                                                                          ) then
                foldl insert [] ms
              else
                foldl insert (case super of NONE => [] | SOME c => methods c) ms
        fun show (name, PRIM_METHOD _) =
              app print ["(", kind, " ", name, " primitive ...)\n"]
          | show (name, USER_METHOD { formals, ... }) =
              app print ["(", kind, " ", name,
                         " (", separate ("", " ") formals, ") ...)\n"]
    in  app show (methods c)
    end
in
  fun protocols all (meta, c) =
    ( showProtocol all "classMethod" meta
    ; showProtocol all "method" c
    ; (meta, CLASSREP c)
    )
end
(* primitive methods for remaining classes 428b *)
fun withOverflow binop ((_, NUM n), [(_, NUM m), ovflw]) = 
      (mkBlock ([], [VALUE (mkInteger (binop (n, m)))], emptyEnv, objectClass)
       handle Overflow => ovflw)
  | withOverflow _ (_, [_, _]) =
      raise RuntimeError "numeric primitive with overflow expects numbers"
  | withOverflow _ _ =
      raise RuntimeError
                     "numeric primitive with overflow expects receiver + 2 args"


(*****************************************************************)
(*                                                               *)
(*   REMAINING BUILT-IN CLASSES                                  *)
(*                                                               *)
(*****************************************************************)

(* remaining built-in classes 406d *)
val classClass =
  mkClass "Class" objectClass []
    [ primMethod "new"           (classPrimitive newUserObject) 
    , primMethod "protocol"      (classPrimitive (protocols true))
    , primMethod "localProtocol" (classPrimitive (protocols false))
    ]


(*****************************************************************)
(*                                                               *)
(*   IMPLEMENTATION OF [[USE]], WITH BOOLEAN [[ECHO]]            *)
(*                                                               *)
(*****************************************************************)

(* implementation of [[use]], with Boolean [[echo]] 606d *)
fun use readEvalPrint filename rho =
      let val fd = TextIO.openIn filename
          val treader = topreader (filereader (filename, fd), false)
          fun writeln s = app print [s, "\n"]
      in  readEvalPrint (treader, true, writeln) rho
          before TextIO.closeIn fd
      end 


(*****************************************************************)
(*                                                               *)
(*   EVALUATION                                                  *)
(*                                                               *)
(*****************************************************************)

(* evaluation 407a *)
(* tracing 605a *)
local
  (* private state and functions for printing indented traces 605b *)
  fun traceMe xi =
    let val count = find("&trace", xi)
    in  case !count
          of (c, NUM n) =>
              if n = 0 then false
              else ( count := (c, NUM (n - 1))
                   ; if n = 1 then (print "<trace ends>\n"; false) else true
                   )
           | _ => false
    end handle NotFound _ => false
  (* type declarations for consistency checking *)
  val _ = op traceMe : value ref env -> bool
  (* private state and functions for printing indented traces 605c *)
  val tindent = ref 0
  fun indent 0 = ()
    | indent n = (print "  "; indent (n-1))
  (* private state and functions for printing indented traces 605d *)
  datatype indentation = INDENT_AFTER | OUTDENT_BEFORE

  fun tracePrint direction xi f =
      if traceMe xi then
        let val msg = f () (* could change tindent *)
        in  ( if direction = OUTDENT_BEFORE then tindent := !tindent - 1 else ()
            ; indent (!tindent)
            ; app print msg
            ; print "\n"
            ; if direction = INDENT_AFTER   then tindent := !tindent + 1 else ()
            )
        end
      else
          ()    
  (* private state and functions for mainting a stack of source-code locations 6
                                                                          06a *)
  val locationStack = ref [] : (string * srcloc) list ref
  fun push srcloc = locationStack := srcloc :: !locationStack
  fun pop () = case !locationStack
                 of []     => raise InternalError "tracing stack underflows"
                  | h :: t => locationStack := t
in
  (* exposed tracing functions 606b *)
  fun resetTrace ()       = (locationStack := []; tindent := 0)
  fun traceIndent what xi = (push what; tracePrint INDENT_AFTER   xi)
  fun outdentTrace     xi = (pop ();    tracePrint OUTDENT_BEFORE xi)
  fun showStackTrace () =
    let fun show (msg, (file, n)) =
          app print ["  Sent '", msg, "' in ", file, ", line ", Int.toString n,
                                                                           "\n"]
    in  case !locationStack
          of [] => ()
           | l  => ( print "Method-stack traceback:\n"; app show (!locationStack
                                                                             ) )
    end
  (* type declarations for consistency checking *)
  val _ = op resetTrace     : unit -> unit
  val _ = op traceIndent    : string * srcloc -> value ref env -> (unit ->
                                                            string list) -> unit
  val _ = op outdentTrace   :                    value ref env -> (unit ->
                                                            string list) -> unit
  val _ = op showStackTrace : unit -> unit
end
(* tracing 606c *)
fun valueString (c, NUM n) = 
      String.map (fn #"~" => #"-" | c => c) (Int.toString n) ^
      valueString(c, USER [])
  | valueString (_, SYM v) = v
  | valueString (c, _) = "<" ^ className c ^ ">"
fun eval(e, rho, superclass, xi) =
  let (* local helper functions of [[eval]] 407b *)
      fun findMethod (name, class) =
        let fun fm (CLASS { methods, super, ...}) =
              find (name, methods)
              handle NotFound m =>
                case super
                  of SOME c => fm c
                   | NONE   => raise RuntimeError
                                 (className class ^
                                            " does not understand message " ^ m)
        in  fm class
        end
      (* type declarations for consistency checking *)
      val _ = op findMethod: name * class -> method
      (* local helper functions of [[eval]] 408a *)
      fun evalMethod (PRIM_METHOD (name, f), receiver, actuals) = f (receiver,
                                                                        actuals)
        | evalMethod (USER_METHOD { name, superclass, formals, locals, body },
                      receiver, actuals) =
            let val rho = instanceVars receiver
                val rho = bindList(formals, map ref actuals, rho)
                val rho = bindList(locals, map (fn _ => ref nilValue) locals,
                                                                            rho)
            in  eval(body, rho, superclass, xi)
            end
              handle BindListLength => 
                  raise RuntimeError
                      ("Wrong number of arguments to method " ^ name ^
                                                                "; expected (" ^
                       separate("", " ") (name :: "self" :: formals) ^ ")")
      and instanceVars (_, USER rep) = rep
        | instanceVars self = bind("self", ref self, emptyEnv)
      (* type declarations for consistency checking *)
      val _ = op evalMethod   : method * value * value list -> value
      val _ = op instanceVars : value -> value ref env
      (* local helper functions of [[eval]] 408b *)
      fun applyClosure ((formals, body, rho, superclass), actuals) =
        eval(BEGIN body, bindList(formals, map ref actuals, rho), superclass, xi
                                                                               )
        handle BindListLength => 
            raise RuntimeError ("Wrong number of arguments to block; expected "
                                                                               ^
                                "(value <block>" ^ separate("", " ") formals ^
                                                                            ")")
      (* type declarations for consistency checking *)
      val _ = op applyClosure : (name list * exp list * value ref env * class) *
                                                             value list -> value
      (* function [[ev]], the evaluator proper 408c *)
      fun ev(VALUE v) = v
        | ev(LITERAL c) = 
            (case c of NUM n => mkInteger n
                     | SYM n => mkSymbol n
                     | _ => raise InternalError "unexpected literal")
      (* function [[ev]], the evaluator proper 409a *)
        | ev(VAR v) = !(find(v, rho) handle NotFound _ => find(v, xi))
        | ev(SET (n, e)) =
            let val v = ev e
                val cell = find(n, rho) handle NotFound _ => find(n, xi)
            in  cell := v; v
            end 
      (* function [[ev]], the evaluator proper 409b *)
        | ev(SUPER) = ev (VAR "self")
      (* function [[ev]], the evaluator proper 409c *)
        | ev(BEGIN es) =
            let fun b(e::es, lastval) = b(es, ev e)
                  | b(   [], lastval) = lastval
            in  b(es, nilValue)
            end
      (* function [[ev]], the evaluator proper 409d *)
        | ev(BLOCK (formals, body)) = mkBlock (formals, body, rho, superclass)
      (* function [[ev]], the evaluator proper 409e *)
        | ev(SEND (srcloc, message, receiver, args))  =
            let val obj as (class, rep) = ev receiver
                val args = map ev args
                val dispatchingClass = case receiver of SUPER => superclass | _
                                                                        => class
                (* definitions of message-tracing procedures 410 *)
                fun traceSend (file, line) =
                  traceIndent (message, (file, line)) xi (fn () =>
                     [file, ", line ", Int.toString line, ": Sending message (",
                      separate ("", " ") (message :: map valueString args), ")",
                      " to object of class ", className dispatchingClass])
                fun traceAnswer (file, line) answer =
                  ( outdentTrace xi (fn () =>
                       [file, ", line ", Int.toString line, ": (",
                        separate  ("", " ") (message :: map valueString (obj ::
                                                                         args)),
                        ") = ", valueString answer])
                  ; answer
                  )
                fun performSend () =
                  case (message, rep)
                    of ("value", CLOSURE clo) => applyClosure(clo, args)
                     | _ => evalMethod(findMethod(message, dispatchingClass),
                                                                      obj, args)

            in  ( traceSend srcloc
                ; traceAnswer srcloc (performSend ())
                )
            end
      (* type declarations for consistency checking *)
      val _ = op ev : exp -> value
(* type declarations for consistency checking *)
val _ = op eval: exp * value ref env * class * value ref env -> value
val _ = op ev  : exp -> value
  in  ev e
  end
(* evaluation 411a *)
val primitiveMethods = methods ((* primitive methods [[::]] 401c *)
                                primMethod "eqObject" (binaryPrim (mkBoolean o
                                                                      eqRep)) ::
                                (* primitive methods [[::]] 401e *)
                                primMethod "print" (unaryPrim defaultPrint) ::
                                (* primitive methods [[::]] 403a *)
                                primMethod "printSmallInteger" (unaryPrim
                                                                    printInt) ::
                                primMethod "newSmallInteger:"  (binaryPrim
                                                                  newInteger) ::
                                primMethod "+"   (arithop op +  )  ::
                                primMethod "-"   (arithop op -  )  ::
                                primMethod "*"   (arithop op *  )  ::
                                primMethod "div" (arithop op div)  ::
                                primMethod "<"   (intcompare op <) ::
                                primMethod ">"   (intcompare op >) ::
                                (* primitive methods [[::]] 403d *)
                                primMethod "printSymbol" (unaryPrim  printSymbol
                                                                            ) ::
                                primMethod "newSymbol"   (binaryPrim newSymbol
                                                                            ) ::
                                (* primitive methods [[::]] 404b *)
                                primMethod "arrayNew:"    (binaryPrim newArray)
                                                                              ::
                                primMethod "arraySize"    (arrayPrimitive
                                                                   arraySize) ::
                                primMethod "arrayAt:"     (arrayPrimitive
                                                                     arrayAt) ::
                                primMethod "arrayAt:Put:" (arrayPrimitive
                                                                  arrayAtPut) ::
                                (* primitive methods [[::]] 404c *)
                                primMethod "value" (fn _ => raise InternalError
                                                    "got to primitive value") ::
                                (* primitive methods [[::]] 428c *)
                                primMethod "add:withOverflow:" (withOverflow op
                                                                          + ) ::
                                primMethod "sub:withOverflow:" (withOverflow op
                                                                          - ) ::
                                primMethod "mul:withOverflow:" (withOverflow op
                                                                     * ) :: nil)
fun newClassObject {name, super, ivars, methods = ms} xi =
  let val (superMeta, super) = findClass(super, xi)
        handle NotFound s => raise RuntimeError ("Superclass " ^ s ^
                                                                   " not found")
      fun method (kind, name, PRIM_IMPL prim) =
            renameMethod(name, primitiveMethod prim)
        | method (kind, name, USER_IMPL (formals, ls, body)) =
            USER_METHOD { name = name, formals = formals, body = body, locals =
                                                                              ls
                        , superclass = case kind of IMETHOD => super
                                                  | CMETHOD => superMeta
                        }
      fun addMethodDefn (m as (CMETHOD, _, _), (c's, i's)) = (method m :: c's,
                                                                            i's)
        | addMethodDefn (m as (IMETHOD, _, _), (c's, i's)) = (c's, method m ::
                                                                            i's)
      val (cmethods, imethods) = foldr addMethodDefn ([], []) ms
      val metaclassName = "class " ^ name
      val class     = mkClass name          super      ivars imethods
      val metaclass = mkClass metaclassName superMeta  []    cmethods
  in  (metaclass, CLASSREP class)
  end
and primitiveMethod name =
  find(name, primitiveMethods)
  handle NotFound n => raise RuntimeError ("There is no primitive method named "
                                                                            ^ n)
(* evaluation 411b *)
and findClass(supername, xi) =
  case !(find(supername, xi))
    of (meta, CLASSREP c) => (meta, c)
     | v => raise RuntimeError ("object " ^ supername ^ " = " ^ valueString v ^
                                " is not a class")
(* evaluation 412a *)
fun topeval (t, echo, xi)  =
  case t
    of USE filename => use readEvalPrint filename xi 
     | DEFINE (name, args, body) => 
         topeval (VAL (name, BLOCK (args, [body])), echo, xi)
     | VAL (name, e) => 
         doVal echo (name, eval(e, emptyEnv, objectClass, xi), xi)
     | CLASSD (d as {name, methods, ...}) =>
         doVal echo (name, newClassObject d xi, xi)
     | EXP e => 
         (printValue echo xi (eval(e, emptyEnv, objectClass, xi)); xi)
and doVal echo (name, value, xi)  =
      ((find(name, xi) := value; xi)
       handle NotFound _ => bind (name, ref value, xi)
      )
      before printValue echo xi value
(* evaluation 412b *)
and printValue echo xi v = 
  if echo then
    ( eval (SEND (nullsrc, "print", VALUE v, []), emptyEnv, objectClass, xi)
    ; print "\n"
    ; ()
    )
  else
    ()
(* evaluation 413a *)
and readEvalPrint (reader, echo, errmsg) xi =
  let fun loop xi =
        let fun continue msg = (errmsg msg; showStackTrace(); resetTrace(); loop
                                                                             xi)
            fun finish () = xi
            val _ = (closeLiteralsCycle xi; closeBlocksCycle xi)
                    handle NotFound _ => ()
        in  loop (topeval (readtop reader, echo, xi))
            handle EOF => finish () 
                 (* more read-eval-print handlers 413b *)
                 | InternalError s => continue ("internal interpreter error: " ^
                                                                              s)
                 | Subscript       => continue ("array subscript out of bounds")
                 | Size            => continue ("bad array size")
                 (* more read-eval-print handlers 194b *)
                 | IO.Io {name, ...} => continue ("I/O error: " ^ name)
                 | Paren (_, loc)    => continue ("syntax error: extra ) in " ^
                                                               srclocString loc)
                 | PrematureEOF s    => (errmsg ("Missing )? Got EOF reading ("
                                                                 ^ s); finish())
                 | SyntaxError msg   => continue ("syntax error: " ^ msg)
                 (* more read-eval-print handlers 194c *)
                 | Div               => continue "Division by zero"
                 | Overflow          => continue "Arithmetic overflow"
                 | RuntimeError msg  => continue ("run-time error: " ^ msg)
                 | NotFound n        => continue ("variable " ^ n ^ " not found"
                                                                               )
        end
  in  loop xi
  end
(* type declarations for consistency checking *)
val _ = op findClass : name * value ref env -> class * class
(* type declarations for consistency checking *)
val _ = op topeval : toplevel * bool * value ref env -> value ref env


(*****************************************************************)
(*                                                               *)
(*   INITIALIZATION                                              *)
(*                                                               *)
(*****************************************************************)

(* initialization 413c *)
val initialXi = emptyEnv

fun mkMeta c = mkClass ("class " ^ className c) classClass [] []
fun addClass (c, xi) = bind(className c, ref (mkMeta c, CLASSREP c), xi)
val initialXi = foldl addClass initialXi [ objectClass, nilClass, classClass ]
(* initialization 414a *)
val echoBasis = false
val initialXi =
  let val rdr = stringsreader ("initial basis", (* ML representation of initial
                                                                        basis *)
                                                  [ "; usm.nw 2929"
                                                  , "(class Boolean Object ()"
                                                  ,
"    (method ifTrue:ifFalse: (trueBlock falseBlock) (subclassResponsibility self))"
                                                  , "  "
                                                  ,
                           "    (method ifFalse:ifTrue: (falseBlock trueBlock) "
                                                  ,
                          "        (ifTrue:ifFalse: self trueBlock falseBlock))"
                                                  ,
        "    (method ifTrue:  (trueBlock)  (ifTrue:ifFalse: self trueBlock []))"
                                                  ,
       "    (method ifFalse: (falseBlock) (ifTrue:ifFalse: self [] falseBlock))"
                                                  , "    "
                                                  ,
   "    (method not ()          (ifTrue:ifFalse: self [false]          [true]))"
                                                  ,
"    (method eqv: (aBoolean) (ifTrue:ifFalse: self [aBoolean]       [(not aBoolean)]))"
                                                  ,
"    (method xor: (aBoolean) (ifTrue:ifFalse: self [(not aBoolean)] [aBoolean]))"
                                                  , ""
                                                  ,
            "    (method & (aBoolean) (ifTrue:ifFalse: self [aBoolean] [self]))"
                                                  ,
            "    (method | (aBoolean) (ifTrue:ifFalse: self [self] [aBoolean]))"
                                                  , "  "
                                                  ,
"    (method and: (alternativeBlock) (ifTrue:ifFalse: self alternativeBlock [self]))"
                                                  ,
"    (method or:  (alternativeBlock) (ifTrue:ifFalse: self [self] alternativeBlock))"
                                                  , ""
                                                  ,
"    (method if (trueBlock falseBlock) (ifTrue:ifFalse: self trueBlock falseBlock))"
                                                  , ")"
                                                  , "; usm.nw 2961"
                                                  , "(class True Boolean ()"
                                                  ,
           "  (method ifTrue:ifFalse: (trueBlock falseBlock) (value trueBlock))"
                                                  , ")"
                                                  , "(class False Boolean ()"
                                                  ,
          "  (method ifTrue:ifFalse: (trueBlock falseBlock) (value falseBlock))"
                                                  , ")"
                                                  , "; usm.nw 3124"
                                                  , "(class Block Object "
                                                  ,
                                              "    () ; internal representation"
                                                  ,
                                            "    (method value primitive value)"
                                                  ,
                                                 "    (method whileTrue: (body)"
                                                  ,
                                         "        (ifTrue:ifFalse: (value self)"
                                                  , "            [(value body)"
                                                  ,
                                          "             (whileTrue: self body)]"
                                                  , "            [nil]))"
                                                  ,
                              "    (method while (body) (whileTrue: self body))"
                                                  ,
                                               "    (method whileFalse: (body) "
                                                  ,
                                        "         (ifTrue:ifFalse: (value self)"
                                                  ,
                                                "             [nil]            "
                                                  ,
                                                   "             [(value body) "
                                                  ,
                                      "              (whileFalse: self body)]))"
                                                  , ")"
                                                  , "; usm.nw 5913"
                                                  , "(class Symbol Object"
                                                  ,
                                              "    () ; internal representation"
                                                  ,
              "    (classMethod new () (error: self #can't-send-new-to-Symbol))"
                                                  ,
                                  "    (classMethod new:   primitive newSymbol)"
                                                  ,
                                "    (method      print  primitive printSymbol)"
                                                  , ")"
                                                  , "; usm.nw 4106"
                                                  , "(class Magnitude Object "
                                                  , "    () ; abstract class"
                                                  ,
"    (method =  (x) (subclassResponsibility self)) ; may not inherit from Object"
                                                  ,
                             "    (method <  (x) (subclassResponsibility self))"
                                                  ,
                                                "    (method >  (y) (< y self))"
                                                  ,
                                          "    (method <= (x) (not (> self x)))"
                                                  ,
                                          "    (method >= (x) (not (< self x)))"
                                                  ,
   "    (method min: (aMagnitude) (if (< self aMagnitude) [self] [aMagnitude]))"
                                                  ,
   "    (method max: (aMagnitude) (if (> self aMagnitude) [self] [aMagnitude]))"
                                                  , ")"
                                                  , "; usm.nw 4120"
                                                  , "(class Number Magnitude"
                                                  , "    ()  ; abstract class"
                                                  ,
                                             "    ;;;;;;; basic Number protocol"
                                                  ,
                  "    (method +   (aNumber)     (subclassResponsibility self))"
                                                  ,
                  "    (method *   (aNumber)     (subclassResponsibility self))"
                                                  ,
                  "    (method negated    ()     (subclassResponsibility self))"
                                                  ,
                  "    (method reciprocal ()     (subclassResponsibility self))"
                                                  , "    "
                                                  ,
                  "    (method asInteger  ()     (subclassResponsibility self))"
                                                  ,
                  "    (method asFraction ()     (subclassResponsibility self))"
                                                  ,
                  "    (method asFloat    ()     (subclassResponsibility self))"
                                                  ,
                  "    (method coerce: (aNumber) (subclassResponsibility self))"
                                                  , "    "
                                                  , "; usm.nw 4138"
                                                  ,
                                        "(method -   (y) (+ self (negated  y)))"
                                                  ,
                "(method abs ()  (if (negative self) [(negated  self)] [self]))"
                                                  ,
                                      "(method /   (y) (* self (reciprocal y)))"
                                                  , "; usm.nw 4146"
                                                  ,
                       "(method negative         () (<  self (coerce: self 0)))"
                                                  ,
                       "(method positive         () (>= self (coerce: self 0)))"
                                                  ,
                       "(method strictlyPositive () (>  self (coerce: self 0)))"
                                                  , "; usm.nw 4154"
                                                  ,
                                             "(method squared () (* self self))"
                                                  ,
                                          "(method raisedToInteger: (anInteger)"
                                                  , "    (if (= anInteger 0)"
                                                  , "        [(coerce: self 1)]"
                                                  ,
                                           "        [(if (= anInteger 1) [self]"
                                                  ,
          "            [(* (squared (raisedToInteger: self (div: anInteger 2)))"
                                                  ,
              "                (raisedToInteger: self (mod: anInteger 2)))])]))"
                                                  , "; usm.nw 4181"
                                                  ,
                   "(method sqrt () (sqrtWithin: self (coerce: self (/ 1 10))))"
                                                  ,
                      "(method sqrtWithin: (epsilon) (locals two x_{i-1} x_{i})"
                                                  ,
                             "    ; find square root of receiver within epsilon"
                                                  ,
                                            "    (set two     (coerce: self 2))"
                                                  ,
                                            "    (set x_{i-1} (coerce: self 1))"
                                                  ,
                        "    (set x_{i}   (/ (+ x_{i-1} (/ self x_{i-1})) two))"
                                                  ,
                              "    (while [(> (abs (- x_{i-1} x_{i})) epsilon)]"
                                                  ,
                                               "           [(set x_{i-1} x_{i})"
                                                  ,
                "            (set x_{i} (/ (+ x_{i-1} (/ self x_{i-1})) two))])"
                                                  , "    x_{i})"
                                                  , "; usm.nw 4133"
                                                  , ")"
                                                  , "; usm.nw 4513"
                                                  , "(class Integer Number"
                                                  , "    () ; abstract class"
                                                  ,
                           "    (method div: (n) (subclassResponsibility self))"
                                                  ,
                            "    (method mod: (n) (- self (* n (div: self n))))"
                                                  ,
"    (method gcd: (n) (if (= n (coerce: self 0)) [self] [(gcd: n (mod: self n))]))"
                                                  ,
                         "    (method lcm: (n) (* self (div: n (gcd: self n))))"
                                                  , "    "
                                                  , "; usm.nw 4529"
                                                  ,
                            "(method asFraction () (num:den:  Fraction self 1))"
                                                  ,
                            "(method asFloat    () (mant:exp: Float    self 0))"
                                                  ,
                                                   "(method asInteger  () self)"
                                                  , "; usm.nw 4538"
                                                  ,
                                "(method coerce: (aNumber) (asInteger aNumber))"
                                                  , "; usm.nw 4543"
                                                  ,
                            "(method reciprocal () (num:den: Fraction 1 self)) "
                                                  ,
                            "(method / (aNumber) (/ (asFraction self) aNumber))"
                                                  , "; usm.nw 4549"
                                                  ,
                                  "(method timesRepeat: (aBlock) (locals count)"
                                                  ,
          "    (ifTrue: (negative self) [(error: self #negative-repeat-count)])"
                                                  , "    (set count self)"
                                                  , "    (while [(!= count 0)]"
                                                  , "         [(value aBlock)"
                                                  ,
                                          "          (set count (- count 1))]))"
                                                  , "; usm.nw 4520"
                                                  , ")"
                                                  , "; usm.nw 4576"
                                                  ,
                                                   "(class SmallInteger Integer"
                                                  ,
                                             "    () ; primitive representation"
                                                  ,
                             "    (classMethod new: primitive newSmallInteger:)"
                                                  ,
                                      "    (classMethod new   () (new: self 0))"
                                                  ,
                                         "    (method negated    () (- 0 self))"
                                                  ,
                                "    (method print primitive printSmallInteger)"
                                                  ,
                                                "    (method +     primitive +)"
                                                  ,
                                                "    (method -     primitive -)"
                                                  ,
                                                "    (method *     primitive *)"
                                                  ,
                                              "    (method div:  primitive div)"
                                                  ,
                                         "    (method =     primitive eqObject)"
                                                  ,
                                                "    (method <     primitive <)"
                                                  ,
                                                "    (method >     primitive >)"
                                                  , ")"
                                                  , "; usm.nw 4608"
                                                  , "(class Fraction Number"
                                                  , "    (num den)"
                                                  ,
                "    (classMethod num:den: (a b) (initNum:den: (new self) a b))"
                                                  ,
                                      "    (method initNum:den: (a b) ; private"
                                                  ,
                                                "        (setNum:den: self a b)"
                                                  , "        (signReduce self)"
                                                  , "        (divReduce self))"
                                                  ,
         "    (method setNum:den: (a b) (set num a) (set den b) self) ; private"
                                                  , "    "
                                                  , "; usm.nw 4641"
                                                  ,
                                               "(method signReduce () ; private"
                                                  ,
                                                   "    (ifTrue: (negative den)"
                                                  ,
                    "        [(set num (negated num)) (set den (negated den))])"
                                                  , "    self)"
                                                  , ""
                                                  ,
                                  "(method divReduce () (locals temp) ; private"
                                                  , "    (if (= num 0)"
                                                  , "        [(set den 1)]"
                                                  ,
                                      "        [(set temp (gcd: (abs num) den))"
                                                  ,
                                           "         (set num  (div: num temp))"
                                                  ,
                                         "         (set den  (div: den temp))])"
                                                  , "    self)"
                                                  , "; usm.nw 4660"
                                                  , "(method num () num)"
                                                  , "(method den () den)"
                                                  , "; usm.nw 4669"
                                                  , "(method = (f)"
                                                  ,
                                 "    (and: (= num (num f)) [(= den (den f))]))"
                                                  , "(method < (f)"
                                                  ,
                                      "    (< (* num (den f)) (* (num f) den)))"
                                                  , "; usm.nw 4678"
                                                  ,
            "(method negated () (setNum:den: (new Fraction) (negated num) den))"
                                                  , "; usm.nw 4692"
                                                  , "(method * (f)"
                                                  , "    (divReduce"
                                                  ,
                                           "        (setNum:den: (new Fraction)"
                                                  ,
                                       "                        (* num (num f))"
                                                  ,
                                    "                        (* den (den f)))))"
                                                  , "; usm.nw 4711"
                                                  ,
                                                   "(method + (f) (locals temp)"
                                                  ,
                                             "    (set temp (lcm: den (den f)))"
                                                  , "    (divReduce"
                                                  ,
                                           "        (setNum:den: (new Fraction)"
                                                  ,
"                     (+ (* num (div: temp den)) (* (num f) (div: temp (den f))))"
                                                  ,
                                                  "                     temp)))"
                                                  , "; usm.nw 4725"
                                                  , "(method reciprocal ()"
                                                  ,
                         "   (signReduce (setNum:den: (new Fraction) den num)))"
                                                  , "; usm.nw 4730"
                                                  ,
                     "(method print () (print num) (print #/) (print den) self)"
                                                  , "; usm.nw 4740"
                                                  ,
                                         "(method asInteger  () (div: num den))"
                                                  ,
                        "(method asFloat    () (/ (asFloat num) (asFloat den)))"
                                                  ,
                                                   "(method asFraction () self)"
                                                  ,
                               "(method coerce: (aNumber) (asFraction aNumber))"
                                                  , "; usm.nw 4749"
                                                  ,
                                   "(method negative         () (negative num))"
                                                  ,
                                   "(method positive         () (positive num))"
                                                  ,
                           "(method strictlyPositive () (strictlyPositive num))"
                                                  , "; usm.nw 4617"
                                                  , ")"
                                                  , "; usm.nw 4782"
                                                  , "(class Float Number"
                                                  , "    (mant exp)"
                                                  ,
              "    (classMethod mant:exp: (m e) (initMant:exp: (new self) m e))"
                                                  ,
                                     "    (method initMant:exp: (m e) ; private"
                                                  ,
                            "        (set mant m) (set exp e) (normalize self))"
                                                  ,
                                         "    (method normalize ()    ; private"
                                                  ,
                                         "        (while [(> (abs mant) 32767)]"
                                                  ,
                                     "               [(set mant (div: mant 10))"
                                                  ,
                                         "                (set exp (+ exp 1))])"
                                                  , "        self)"
                                                  , "    "
                                                  , "; usm.nw 4798"
                                                  ,
                                              "(method mant () mant)  ; private"
                                                  ,
                                              "(method exp  () exp)   ; private"
                                                  , "; usm.nw 4806"
                                                  ,
                                          "(method < (x) (negative (- self x)))"
                                                  ,
                                          "(method = (x) (isZero   (- self x)))"
                                                  ,
                                                 "(method isZero () (= mant 0))"
                                                  , "; usm.nw 4813"
                                                  ,
                      "(method negated () (mant:exp: Float (negated mant) exp))"
                                                  , "; usm.nw 4834"
                                                  , "(method + (prime) "
                                                  ,
                                                  "    (if (>= exp (exp prime))"
                                                  ,
"        [(mant:exp: Float (+ (* mant (raisedToInteger: 10 (- exp (exp prime))))"
                                                  ,
                                    "                             (mant prime))"
                                                  ,
                                       "                          (exp prime))]"
                                                  , "        [(+ prime self)]))"
                                                  , "; usm.nw 4846"
                                                  , "(method * (prime) "
                                                  ,
              "    (mant:exp: Float (* mant (mant prime)) (+ exp (exp prime))))"
                                                  , "; usm.nw 4855"
                                                  , "(method reciprocal ()"
                                                  ,
                      "    (mant:exp: Float (div: 1000000000 mant) (- -9 exp)))"
                                                  , "; usm.nw 4861"
                                                  ,
                                  "(method coerce: (aNumber) (asFloat aNumber))"
                                                  , "(method asFloat () self)"
                                                  , "; usm.nw 4867"
                                                  , "(method asInteger ()"
                                                  , "    (if (< exp 0)"
                                                  ,
                     "        [(div: mant (raisedToInteger: 10 (negated exp)))]"
                                                  ,
                             "        [(*    mant (raisedToInteger: 10 exp))]))"
                                                  , "; usm.nw 4875"
                                                  , "(method asFraction ()"
                                                  , "    (if (< exp 0)"
                                                  ,
        "        [(num:den: Fraction mant (raisedToInteger: 10 (negated exp)))]"
                                                  ,
          "        [(num:den: Fraction (* mant (raisedToInteger: 10 exp)) 1)]))"
                                                  , "; usm.nw 4896"
                                                  ,
                                  "(method negative         () (negative mant))"
                                                  ,
                                  "(method positive         () (positive mant))"
                                                  ,
                          "(method strictlyPositive () (strictlyPositive mant))"
                                                  , "; usm.nw 4904"
                                                  , "(method print () "
                                                  ,
                                                   "    (print-normalize self) "
                                                  ,
                                    "    (print mant) (print #x10^) (print exp)"
                                                  , "    self)"
                                                  , ""
                                                  , "(method print-normalize ()"
                                                  ,
                          "    (while [(and: (< exp 0) [(= (mod: mant 10) 0)])]"
                                                  ,
                     "        [(set exp (+ exp 1)) (set mant (div: mant 10))]))"
                                                  , "; usm.nw 4793"
                                                  , ")"
                                                  , "; usm.nw 3187"
                                                  , "(class Collection Object"
                                                  , "  () ; abstract"
                                                  ,
                  "  (method do:     (aBlock)    (subclassResponsibility self))"
                                                  ,
                  "  (method add:    (newObject) (subclassResponsibility self))"
                                                  ,
                               "  (method remove:ifAbsent: (oldObject exnBlock)"
                                                  ,
                  "                              (subclassResponsibility self))"
                                                  ,
                  "  (method species ()          (subclassResponsibility self))"
                                                  , "  "
                                                  , "; usm.nw 3199"
                                                  ,
                                   "(classMethod with: (anObject) (locals temp)"
                                                  , "    (set temp (new self))"
                                                  , "    (add: temp anObject)"
                                                  , "    temp)"
                                                  , "; usm.nw 3211"
                                                  ,
                                                  "(method remove: (oldObject) "
                                                  ,
"    (remove:ifAbsent: self oldObject [(error: self #tried-to-remove-absent-object)]))"
                                                  ,
                                                "(method addAll: (aCollection) "
                                                  ,
                               "    (do: aCollection (block (x) (add: self x)))"
                                                  , "    aCollection)"
                                                  ,
                                             "(method removeAll: (aCollection) "
                                                  ,
                            "    (do: aCollection (block (x) (remove: self x)))"
                                                  , "    aCollection)"
                                                  , "; usm.nw 3231"
                                                  ,
                                         "(method isEmpty () (= (size self) 0))"
                                                  ,
                                                 "(method size () (locals temp)"
                                                  , "    (set temp 0)"
                                                  ,
                              "    (do: self (block (_) (set temp (+ temp 1))))"
                                                  , "    temp)"
                                                  ,
                               "(method occurrencesOf: (anObject) (locals temp)"
                                                  , "    (set temp 0)"
                                                  , "    (do: self (block (x)"
                                                  ,
                     "       (ifTrue: (= x anObject) [(set temp (+ temp 1))])))"
                                                  , "    temp)"
                                                  ,
            "(method includes: (anObject) (< 0 (occurrencesOf: self anObject)))"
                                                  , "(method detect: (aBlock) "
                                                  ,
         "    (detect:ifNone: self aBlock [(error: self #no-object-detected)]))"
                                                  ,
            "(method detect:ifNone: (aBlock exnBlock) (locals answer searching)"
                                                  , "    (set searching true)"
                                                  , "    (do: self (block (x)"
                                                  ,
                          "        (ifTrue: (and: searching [(value aBlock x)])"
                                                  ,
                                           "             [(set searching false)"
                                                  ,
                                              "              (set answer x)])))"
                                                  ,
                                         "    (if searching exnBlock [answer]))"
                                                  , "; usm.nw 3263"
                                                  ,
                                  "(method inject:into: (thisValue binaryBlock)"
                                                  ,
     "   (do: self (block (x) (set thisValue (value binaryBlock x thisValue))))"
                                                  , "   thisValue)"
                                                  , "; usm.nw 3271"
                                                  ,
                                        "(method select: (aBlock) (locals temp)"
                                                  ,
                                            "   (set temp (new (species self)))"
                                                  ,
          "   (do: self (block (x) (ifTrue: (value aBlock x) [(add: temp x)])))"
                                                  , "   temp)"
                                                  , "(method reject: (aBlock)"
                                                  ,
                         "   (select: self (block (x) (not (value aBlock x)))))"
                                                  ,
                                       "(method collect: (aBlock) (locals temp)"
                                                  ,
                                            "   (set temp (new (species self)))"
                                                  ,
                        "   (do: self (block (x) (add: temp (value aBlock x))))"
                                                  , "   temp)"
                                                  , "; usm.nw 3284"
                                                  ,
                                                "(method asSet () (locals temp)"
                                                  , "     (set temp (new Set))"
                                                  ,
                                     "     (do: self (block (x) (add: temp x)))"
                                                  , "     temp)"
                                                  , "; usm.nw 3293"
                                                  , "(method print ()"
                                                  , "    (printName self)"
                                                  , "    (print lparen)"
                                                  ,
                            "    (do: self (block (x) (print space) (print x)))"
                                                  , "    (print space)"
                                                  , "    (print rparen)"
                                                  , "    self)"
                                                  ,
                                     "(method printName () (print #Collection))"
                                                  , "; usm.nw 3195"
                                                  , ")"
                                                  , "; usm.nw 3315"
                                                  , "(class Set Collection"
                                                  ,
                                             "    (members)  ; list of elements"
                                                  ,
                                "    (classMethod new () (initSet (new super)))"
                                                  ,
              "    (method initSet  () (set members (new List)) self) ; private"
                                                  ,
                                "    (method do: (aBlock) (do: members aBlock))"
                                                  , "    (method add: (item)"
                                                  ,
             "        (ifFalse: (includes: members item) [(add: members item)])"
                                                  , "        item)"
                                                  ,
                                 "    (method remove:ifAbsent: (item exnBlock) "
                                                  ,
                             "        (remove:ifAbsent: members item exnBlock))"
                                                  ,
                                                "    (method asSet     () self)"
                                                  ,
                                        "    (method printName () (print #Set))"
                                                  ,
                                                 "    (method species   () Set)"
                                                  , ")"
                                                  , "; usm.nw 3359"
                                                  ,
                                             "(class KeyedCollection Collection"
                                                  , "    ()  ; abstract class"
                                                  ,
          "    (method at:put: (key value)       (subclassResponsibility self))"
                                                  ,
          "    (method associationsDo: (aBlock)  (subclassResponsibility self))"
                                                  , "    "
                                                  , "; usm.nw 3370"
                                                  , "(method do: (aBlock) "
                                                  ,
              "    (associationsDo: self (block (x) (value aBlock (value x)))))"
                                                  , "(method at: (key)    "
                                                  ,
                   "    (at:ifAbsent: self key [(error: self #key-not-found])))"
                                                  ,
                                          "(method at:ifAbsent: (key exnBlock) "
                                                  ,
                                 "    (value (associationAt:ifAbsent: self key "
                                                  ,
             "               [(key:value: Association nil (value exnBlock))])))"
                                                  ,
                                                   "(method includesKey: (key) "
                                                  ,
            "    (isKindOf: (associationAt:ifAbsent: self key []) Association))"
                                                  ,
                                                 "(method associationAt: (key) "
                                                  ,
        "    (associationAt:ifAbsent: self key [(error: self #key-not-found])))"
                                                  ,
           "(method associationAt:ifAbsent: (key exnBlock) (locals finishBlock)"
                                                  ,
                                                "    (set finishBlock exnBlock)"
                                                  ,
                                         "    (associationsDo: self (block (x) "
                                                  ,
                   "        (ifTrue: (= (key x) key) [(set finishBlock [x])])))"
                                                  , "    (value finishBlock))"
                                                  , "; usm.nw 3392"
                                                  ,
                                                  "(method keyAtValue: (value) "
                                                  ,
       "    (keyAtValue:ifAbsent: self value [(error: self #value-not-found)]))"
                                                  ,
              "(method keyAtValue:ifAbsent: (value aBlock) (locals finishBlock)"
                                                  ,
                                                  "    (set finishBlock aBlock)"
                                                  ,
                                         "    (associationsDo: self (block (x) "
                                                  ,
         "        (ifTrue: (= (value x) value) [(set finishBlock [(key x)])])))"
                                                  , "    (value finishBlock))"
                                                  , "; usm.nw 3364"
                                                  , ")"
                                                  , "; usm.nw 3433"
                                                  , "(class Association Object "
                                                  , "   (key value)"
                                                  ,
              "   (classMethod key:value: (a b) (setKey:value: (new self) a b))"
                                                  ,
      "   (method setKey:value: (x y) (set key x) (set value y) self) ; private"
                                                  , "   (method key    ()  key)"
                                                  ,
                                                  "   (method value  ()  value)"
                                                  ,
                                          "   (method key:   (x) (set key   x))"
                                                  ,
                                          "   (method value: (y) (set value y))"
                                                  , ")"
                                                  , "; usm.nw 3446"
                                                  ,
                                             "(class Dictionary KeyedCollection"
                                                  ,
                                                   "    (table) ; list of pairs"
                                                  ,
                   "    (classMethod new ()       (initDictionary (new super)))"
                                                  ,
          "    (method initDictionary () (set table (new List)) self) ; private"
                                                  ,
                            "    (method printName ()      (print #Dictionary))"
                                                  ,
                                     "    (method species ()        Dictionary)"
                                                  , "    "
                                                  , "; usm.nw 3460"
                                                  ,
                          "(method associationsDo: (aBlock) (do: table aBlock))"
                                                  ,
                                "(method at:put: (key value) (locals tempassoc)"
                                                  ,
                     "    (set tempassoc (associationAt:ifAbsent: self key []))"
                                                  , "    (if (isNil tempassoc)"
                                                  ,
                    "         [(add: table (key:value: Association key value))]"
                                                  ,
                                          "         [(value: tempassoc value)])"
                                                  , "    value)"
                                                  , "; usm.nw 3471"
                                                  ,
              "(method add: (_) (error: self #can't-just-add:-to-a-Dictionary))"
                                                  , "; usm.nw 3453"
                                                  , ")"
                                                  , "; usm.nw 3484"
                                                  ,
                                 "(class SequenceableCollection KeyedCollection"
                                                  , "    () ; abstract class"
                                                  ,
                        "    (method firstKey () (subclassResponsibility self))"
                                                  ,
                        "    (method lastKey  () (subclassResponsibility self))"
                                                  ,
                           "    (method last     () (at: self (lastKey  self)))"
                                                  ,
                           "    (method first    () (at: self (firstKey self)))"
                                                  ,
        "    (method at:ifAbsent: (index exnBlock) (locals current resultBlock)"
                                                  ,
                                            "        (set resultBlock exnBlock)"
                                                  ,
                                         "        (set current (firstKey self))"
                                                  ,
                                                  "        (do: self (block (v)"
                                                  ,
               "            (ifTrue: (= current index) [(set resultBlock [v])])"
                                                  ,
                                     "            (set current (+ current 1))))"
                                                  ,
                                                  "        (value resultBlock))"
                                                  , ")"
                                                  , "; usm.nw 3906"
                                                  , "(class Cons Object"
                                                  , "    (car cdr)"
                                                  ,
                                             "    (method car ()           car)"
                                                  ,
                                             "    (method cdr ()           cdr)"
                                                  ,
                         "    (method car: (anObject)  (set car anObject) self)"
                                                  ,
                         "    (method cdr: (anObject)  (set cdr anObject) self)"
                                                  , "    "
                                                  , "; usm.nw 3921"
                                                  ,
                                        "(method deleteAfter () (locals answer)"
                                                  , "    (set answer (car cdr))"
                                                  , "    (set cdr    (cdr cdr))"
                                                  , "    (pred: cdr self)"
                                                  , "    answer)"
                                                  ,
                                               "(method insertAfter: (anObject)"
                                                  ,
                           "    (set cdr (car: (cdr: (new Cons) cdr) anObject))"
                                                  , "    (pred: (cdr cdr) cdr)"
                                                  , "    anObject)"
                                                  , "; usm.nw 3942"
                                                  , "(method do: (aBlock)"
                                                  , "    (value aBlock car)"
                                                  , "    (do: cdr aBlock))"
                                                  , "; usm.nw 3952"
                                                  ,
                   "(method rejectOne:ifAbsent:withPred: (aBlock exnBlock pred)"
                                                  ,
                                                   "    (if (value aBlock self)"
                                                  ,
                                                  "        [(deleteAfter pred)]"
                                                  ,
           "        [(rejectOne:ifAbsent:withPred: cdr aBlock exnBlock self)]))"
                                                  , "; usm.nw 3913"
                                                  ,
                                             "    (method pred: (aCons)    nil)"
                                                  , ")"
                                                  , "; usm.nw 3963"
                                                  , "(class ListSentinel Cons"
                                                  , "    (pred)"
                                                  ,
                                              "    (method do:   (aBlock)  nil)"
                                                  ,
               "    (method rejectOne:ifAbsent:withPred: (aBlock exnBlock pred)"
                                                  ,
                                 "                            (value exnBlock))"
                                                  ,
                                 "    (method pred: (aCons)   (set pred aCons))"
                                                  ,
                                             "    (method pred  ()        pred)"
                                                  ,
                                      "    (classMethod new ()     (locals tmp)"
                                                  ,
                                                 "        (set tmp (new super))"
                                                  , "        (pred: tmp tmp)"
                                                  , "        (cdr:  tmp tmp)"
                                                  , "        tmp))"
                                                  , "; usm.nw 3833"
                                                  ,
                                            "(class List SequenceableCollection"
                                                  , "    (sentinel)"
                                                  ,
   "    (classMethod new  ()        (sentinel: (new super) (new ListSentinel)))"
                                                  ,
              "    (method sentinel: (s)       (set sentinel s) self) ; private"
                                                  ,
                  "    (method isEmpty   ()        (= sentinel (cdr sentinel)))"
                                                  , "    "
                                                  , "; usm.nw 3852"
                                                  ,
               "(method addLast:  (item)   (insertAfter: (pred sentinel) item))"
                                                  ,
                      "(method addFirst: (item)   (insertAfter: sentinel item))"
                                                  ,
                              "(method add: (item)        (addLast: self item))"
                                                  , "; usm.nw 3859"
                                                  ,
                            "(method removeFirst ()     (deleteAfter sentinel))"
                                                  , "; usm.nw 3873"
                                                  ,
                                 "(method remove:ifAbsent: (oldObject exnBlock)"
                                                  ,
                                             "    (rejectOne:ifAbsent:withPred:"
                                                  , "        (cdr sentinel)"
                                                  ,
                                     "        (block (x) (= oldObject (car x)))"
                                                  , "        exnBlock"
                                                  , "        sentinel))"
                                                  , "; usm.nw 3886"
                                                  , "(method firstKey () 1)"
                                                  ,
                                              "(method lastKey  () (size self))"
                                                  ,
                                        "(method at:put: (n value) (locals tmp)"
                                                  ,
                                                  "    (set tmp (cdr sentinel))"
                                                  , "    (whileFalse: [(= n 1)]"
                                                  , "       [(set n (- n 1))"
                                                  ,
                                                 "        (set tmp (cdr tmp))])"
                                                  , "    (car: tmp value))"
                                                  , "; usm.nw 3839"
                                                  ,
                        "    (method last      ()        (car (pred sentinel)))"
                                                  ,
                  "    (method do:       (aBlock)  (do: (cdr sentinel) aBlock))"
                                                  ,
                                         "    (method species   ()        List)"
                                                  ,
                                "    (method printName ()        (print #List))"
                                                  , ")"
                                                  , "; usm.nw 4040"
                                                  ,
                                           "(class Array SequenceableCollection"
                                                  ,
                                          "    () ; representation is primitive"
                                                  ,
                                    "    (classMethod new: primitive arrayNew:)"
                                                  ,
       "    (classMethod new () (error: self #size-of-Array-must-be-specified))"
                                                  ,
                                     "    (method size     primitive arraySize)"
                                                  ,
                                      "    (method at:      primitive arrayAt:)"
                                                  ,
                                  "    (method at:put:  primitive arrayAt:Put:)"
                                                  , "    "
                                                  , "; usm.nw 4057"
                                                  ,
                         "(method add:             (x)   (fixedSizeError self))"
                                                  ,
                         "(method remove:ifAbsent: (x b) (fixedSizeError self))"
                                                  ,
         "(method fixedSizeError   ()    (error: self #arrays-have-fixed-size))"
                                                  , "; usm.nw 4064"
                                                  ,
         "(method select:  (_) (error: self #select-on-arrays-not-implemented))"
                                                  ,
        "(method collect: (_) (error: self #collect-on-arrays-not-implemented))"
                                                  , "; usm.nw 4070"
                                                  , "(method firstKey () 1)"
                                                  ,
                                              "(method lastKey  () (size self))"
                                                  ,
                                           "(method do: (aBlock) (locals index)"
                                                  ,
                                               "    (set index (firstKey self))"
                                                  ,
                                                 "    (timesRepeat: (size self)"
                                                  ,
                                       "       [(value aBlock (at: self index))"
                                                  ,
                                            "        (set index (+ index 1))]))"
                                                  , "; usm.nw 4048"
                                                  ,
                                               "    (method species   () Array)"
                                                  ,
                "    (method printName () nil) ; names of arrays aren't printed"
                                                  , ")"
                                                  ])
  in  readEvalPrint
        ( topreader (rdr, false)
        , echoBasis
        , fn s => app print ["error in initial basis: ", s, "\n"]
        )
        initialXi
  end
(* initialization 414b *)
val initialXi =
      topeval (VAL ("true",  SEND(nullsrc, "new", VAR "True",  [])), false,
                                                                      initialXi)
val initialXi =
      topeval (VAL ("false", SEND(nullsrc, "new", VAR "False", [])), false,
                                                                      initialXi)
(* initialization 414c *)
val _ =
  ( closeLiteralsCycle initialXi
  ; closeBooleansCycle initialXi
  ; closeBlocksCycle   initialXi
  ) handle NotFound n =>
      ( app print ["Fatal error: ", n, " is not defined in the initial basis\n"]
      ; raise InternalError "this can't happen"
      )
  | e => ( print "Error binding basis classes into interpreter\n"; raise e)
(* initialization 414e *)
fun addVal ((name, v), xi) = topeval (VAL (name, VALUE v), false, xi)
val initialXi = foldl addVal initialXi
  [ ("nil", nilValue)
  , ("lparen", mkSymbol "("), ("rparen", mkSymbol ")")
  , ("lbrack", mkSymbol "["), ("rbrack", mkSymbol "]")
  , ("newline", mkSymbol "\n"), ("space",  mkSymbol " ")
  ]
(* initialization 415 *)
fun runInterpreter noisy = 
  let val treader = topreader (filereader ("standard input", TextIO.stdIn),
                                                                          noisy)
      fun writeln s = app print [s, "\n"]
  in  ignore (readEvalPrint (treader, true, writeln) initialXi) handle EOF => ()
  end 
(* type declarations for consistency checking *)
val _ = op runInterpreter : bool -> unit


(*****************************************************************)
(*                                                               *)
(*   COMMAND LINE                                                *)
(*                                                               *)
(*****************************************************************)

(* command line *)
fun main ["-q"] = runInterpreter false
  | main []     = runInterpreter true
  | main _      =
      TextIO.output(TextIO.stdErr, "Usage: " ^ CommandLine.name() ^ " [-q]\n")
val _ = main (CommandLine.arguments())
