/* The header file is split into three parts:   */
/* type definitions, then structure             */
/* definitions, then function prototypes. These */
/* splits make it easy for functions or         */
/* structures declared in one interface to use  */
/* types defined in another; because            */
/* declarations and definitions of types always */
/* precede the function prototypes that use     */
/* those types, we need not worry about getting */
/* things in the right order. So we can reuse   */
/* some of the code in this interpreter in      */
/* later interpreters, we distinguish between   */
/* shared and unshared definitions.             */
/* <{\Tt all.h} for \impcore>=                  */
/*UNIX -- FAS gcc barfs if stdio.h precedes stdarg.h */
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <setjmp.h>
#include <ctype.h>

/* For real implementations, it is convenient   */
/* to build names from strings. Unlike          */
/* C strings, names are immutable, and they can */
/* be compared using pointer equality.          */
/* <shared type definitions>=                   */
typedef struct Name Name;
typedef struct Namelist Namelist;   /* list of Name* */
/*    Reading characters                        */
/*                                              */
/* A [[Reader]] encapsulates a source of        */
/* characters: a string or a file. The          */
/* [[readline]] function prints a prompt, reads */
/* the next line of input from the source, and  */
/* returns a pointer to the line. The caller    */
/* needn't worry about how long the line is;    */
/* [[readline]] always allocates enough memory  */
/* to hold it. Because [[readline]] reuses the  */
/* same memory to hold successive lines, it is  */
/* an unchecked run-time error to retain a      */
/* pointer returned by [[readline]] after a     */
/* subsequent call to [[readline]]. A client    */
/* that needs to save input characters must     */
/* copy the result of [[readline]] before       */
/* calling [[readline]] again.                  */
/* <shared type definitions>=                   */
typedef struct Reader Reader;
/*    Reading top-level items                   */
/*                                              */
/* A [[Topreader]] encapsulates a source of     */
/* top-level items. The [[readtop]] function    */
/* returns the next top-level item. It may also */
/* return [[NULL]], if there are no more        */
/* top-level items in the source, or call       */
/* [[error]] (page [->]), if there is some      */
/* problem converting the source to abstract    */
/* syntax.                                      */
/* <shared type definitions>=                   */
typedef struct Topreader Topreader;
/* <shared type definitions>=                   */
typedef void Printer(va_list*);
/* <shared type definitions>=                   */
typedef struct Parlist Parlist; /* list of Par* */
/* <shared type definitions>=                   */
typedef struct Par Par;
/* We also define a type for lists of [[Exp]]s. */
/*                                              */
/* <type definitions for \impcore>=             */
typedef struct Explist Explist; /* list of Exp* */
/* [[Explist]] is list of pointers of type      */
/* [[Exp*]]. We use this naming convention      */
/* throughout our C code. The definitions of    */
/* our list types are in the lists interface,   */
/* in chunk [->].                               */

/*  Values                                      */
/*                                              */
/* The value interface defines the type of      */
/* value that our expressions evaluate to.      */
/* Impcore supports only integers. We use       */
/* [[Value]]s rather than [[Value*]]s           */
/* throughout, since we want to pass them by    */
/* value, not by reference. A [[Valuelist]] is  */
/* a list of [[Value]]s.                        */
/* <type definitions for \impcore>=             */
typedef int Value;
typedef struct Valuelist Valuelist;     /* list of Value */
/*  Functions                                   */
/*                                              */
/* In the Impcore interpreter, the type         */
/* ``function'' is another discriminated-union  */
/* type; the two alternatives are user-defined  */
/* functions and primitive functions. Just like */
/* the operational semantics, which represents  */
/* a user-defined function as \user(<x_1, ...,  */
/* x_n>, e), the interpreter represents a       */
/* user-defined function as a pair containing   */
/* formals and body. The interpreter represents */
/* each primitive by its name.                  */
/* <type definitions for \impcore>=             */
typedef struct Funlist Funlist; /* list of Fun */
/*  Environments                                */
/*                                              */
/* In the operational semantics, the            */
/* environments rho and xi hold values, and the */
/* environment phi holds functions. To          */
/* represent these two kinds of environments,   */
/* C offers these choices:                      */
/*   * We can define one C type for             */
/*     environments that hold a [[Value]] and   */
/*     another for environments that hold a     */
/*     [[Fun]], and we can define two versions  */
/*     of each function. This choice guarantees */
/*     type safety, but requires duplication of */
/*     code.                                    */
/*   * We can define a single C type for        */
/*     environments that hold a [[void*]]       */
/*     pointer, define a single version of each */
/*     function, and use type casting to        */
/*     convert a [[void*]] to a [[Value*]] or   */
/*     [[Fun*]] as needed. This choice          */
/*     duplicates no code, but it is unsafe; if */
/*     we accidentally put a [[Value*]] in an   */
/*     environment intended to hold a [[Fun*]], */
/*     it is an error that neither the          */
/*     C compiler nor the run-time system can   */
/*     detect.                                  */
/*                                              */
/* In the interests of safety, we have chosen   */
/* to duplicate code. Chapter [->] shows how in */
/* another implementation language, ML, we can  */
/* use polymorphism to achieve type safety      */
/* without duplicating code.                    */
/* <type definitions for \impcore>=             */
typedef struct Valenv Valenv;
typedef struct Funenv Funenv;
/* <type definitions for \impcore>=             */
typedef struct Exp Exp;
/* <type definitions for \impcore>=             */
typedef struct Userfun Userfun;
typedef struct Toplevel Toplevel;
/* <type definitions for \impcore>=             */
typedef struct Fun Fun;
/* An input reader is just a line-reading       */
/* function and some extra state.               */
/* <shared structure definitions>=              */
struct Reader {
    char *buf;               /* holds the last line read */
    int nbuf;                /* size of buf */
    int line;                /* current line number */
    const char *readername;  /* identifies this reader */

    FILE *fin;               /* filereader */
    const char *s;           /* stringreader */
};
/* <shared structure definitions>=              */
typedef enum { ATOM, LIST } Party;

struct Par {
    Party ty;
    union {
        Name *atom;
        Parlist *list;
    } u;
};

/* Lists are recursive data types. Either a     */
/* list is empty, or it is a pointer to a pair  */
/* (hd,tl), where [[hd]] is the first element   */
/* of the list and [[tl]] is the rest of the    */
/* list. We use the null pointer to represent   */
/* the empty list. [*]                          */
/* <structure definitions for \impcore>=        */
/* <structure definitions for \impcore>=        */
typedef enum { LITERAL, VAR, SET, IFX, WHILEX, BEGIN, APPLY } Expty;

struct Exp {
    Expty ty;
    union {
        Value literal;
        Name *var;
        struct {  Name *name; Exp *exp; } set;
        struct {  Exp *cond; Exp *true; Exp *false; } ifx;
        struct {  Exp *cond; Exp *exp; } whilex;
        Explist *begin;
        struct {  Name *name; Explist *actuals; } apply;
    } u;
};

/* <structure definitions for \impcore>=        */
struct Userfun {
    Namelist *formals;
    Exp *body;
};
typedef enum { EXP, VAL, DEFINE, USE } Toplevelty;

struct Toplevel {
    Toplevelty ty;
    union {
        Exp *exp;
        struct {  Name *name; Exp *exp; } val;
        struct {  Name *name; Userfun userfun; } define;
        Name *use;
    } u;
};

/* <structure definitions for \impcore>=        */
typedef enum { USERDEF, PRIMITIVE } Funty;

struct Fun {
    Funty ty;
    union {
        Userfun userdef;
        Name *primitive;
    } u;
};

/* <structure definitions for \impcore>=        */
struct Explist {
    Exp *hd;
    Explist *tl;
};

struct Parlist {
    Par *hd;
    Parlist *tl;
};

struct Valuelist {
    Value hd;
    Valuelist *tl;
};

struct Funlist {
    Fun hd;
    Funlist *tl;
};

struct Namelist {
    Name *hd;
    Namelist *tl;
};

/* Pointer comparison is built into C, but we   */
/* provide two other operations on names.       */
/* <shared function prototypes>=                */
Name *strtoname(const char *s);
const char *nametostr(Name *n);
/* [[]]                                         */
/*   [[strcmp(s, t) == 0]] if and only if       */
/*   [[strtoname(s) == strtoname(t)]]           */
/*                                              */
/* Informally, the first law says if you build  */
/* a name from a string, [[nametostr]] returns  */
/* a copy of your original string. The second   */
/* law says you can compare names using pointer */
/* equality.                                    */

/* Function [[printname]] prints names.         */
/*                                              */
/* <shared function prototypes>=                */
Printer printname;
/* <shared function prototypes>=                */
char *readline(Reader *r, char *prompt);
/* To create a reader, we need a string or a    */
/* file. A reader also gets a name, which we    */
/* can use in error messages. [*] [*]           */
/* <shared function prototypes>=                */
Reader *stringreader(const char *stringname, const char *s);
Reader *filereader  (const char *filename,   FILE *fin);
/* <shared function prototypes>=                */
Toplevel *readtop(Topreader *r);
/* <shared function prototypes>=                */
Topreader *topreader(Reader *r, int doprompt);
/* [[doprompt]] flag controls whether the       */
/* resulting [[Topreader]] prompts when reading */
/* input.                                       */

/*  Printing                                    */
/*                                              */
/* [*] The standard C function [[printf]] and   */
/* its siblings [[fprintf]] and [[vfprintf]]    */
/* convert internal C values to sequences of    */
/* characters in an output file. Unfortunately, */
/* they convert only numbers, characters, and   */
/* strings. It would be a lot more convenient   */
/* if [[printf]] also converted [[Name]]s,      */
/* [[Exp]]s, and so on. We can't change         */
/* [[printf]], so instead we define two new     */
/* functions: [[print]] and [[vprint]], with    */
/* similar interfaces.                          */
/* <shared function prototypes>=                */
void print (const char *fmt, ...);
void vprint(const char *fmt, va_list va);
/* [[print]] and                                */
/* [[vprint]] is that they are extensible; we   */
/* can add new printing functions to print new  */
/* kinds of values. Such functions have type    */
/* [[Printer]], and we add them using           */
/* [[installprinter]]. [*]                      */

/* <shared function prototypes>=                */
void installprinter(unsigned char c, Printer *fmt);
/* The printer interface provides printers to   */
/* format percent signs, strings, and decimal   */
/* integers.                                    */
/* <shared function prototypes>=                */
Printer printpercent, printstring, printdecimal;
/*  Error handling                              */
/*                                              */
/* Any module can signal an error by calling    */
/* [[error]]. When called, [[error]] acts just  */
/* like [[print]] except that it does not       */
/* return after printing; instead it            */
/* [[longjmp]]s to [[errorjmp]]. It is an       */
/* unchecked run-time error to call [[error]]   */
/* except while a suitable [[setjmp]] is active */
/* on the C call stack. [*]                     */
/* <shared function prototypes>=                */
void error(const char *fmt, ...);
extern jmp_buf errorjmp;        /* longjmp here on error */
/* Function [[checkargc]] is used to check the  */
/* number of arguments to both user-defined and */
/* primitive functions. The first argument is   */
/* an abstract-syntax tree representing the     */
/* application being checked; if                */
/* [[expected]] != [[actual]], [[checkargc]]    */
/* calls [[error]], passing a message that      */
/* contains [[e]].                              */
/* <shared function prototypes>=                */
void checkargc(Exp *e, int expected, int actual);
/* The [[duplicatename]] function finds a       */
/* duplicate name on a [[Namelist]] if such a   */
/* name exists. It is used to check that formal */
/* parameters to user-defined functions all     */
/* have different names. If the name list       */
/* [[names]] contains a duplicate occurrence of */
/* any name, the function returns such a name;  */
/* otherwise it returns [[NULL]].               */
/* <shared function prototypes>=                */
Name *duplicatename(Namelist *names);
/* <shared function prototypes>=                */
/* <shared function prototypes>=                */
Printer printpar;
/* The [[readparlist]] function reads and       */
/* returns a list of parenthesized expressions  */
/* from the given [[Reader]], stopping when an  */
/* end-of-expression and end-of-line coincide.  */
/* The [[doquote]] flag specifies that an       */
/* expression like [['(1 2 3)]] should be       */
/* turned into [[(quote (1 2 3))]]; it is used  */
/* to implement micro-Scheme. The [[doprompt]]  */
/* flag specifies that prompts should be        */
/* printed when reading input lines.            */
/* <shared function prototypes>=                */
Parlist *readparlist(Reader *r, int doquote, int doprompt);
/* <shared function prototypes>=                */
Par *mkAtom(Name*);
Par *mkList(Parlist*);
/* To make these structures easy to create, we  */
/* define a creator function for each           */
/* alternative in the sum, as well as for       */
/* [[Userfun]].                                 */
/* <function prototypes for \impcore>=          */
/* For debugging, it's convenient to be able to */
/* print abstract-syntax trees, and we provide  */
/* functions to do so. These functions are set  */
/* up to work with our extensible replacement   */
/* for [[printf]] (Section [->]); the type      */
/* [[Printer]] is defined in chunk [->].        */
/* <function prototypes for \impcore>=          */
Printer printexp, printtoplevel;
/* <function prototypes for \impcore>=          */
Printer printvalue;
/* We also generate prototypes for creator      */
/* functions.                                   */
/* <function prototypes for \impcore>=          */
/* Finally, the module exports a print          */
/* function.                                    */
/* <function prototypes for \impcore>=          */
Printer printfun;
/* A new environment may be created by passing  */
/* a list of names and a list of associated     */
/* values or function definitions. For example, */
/* mkValenv(<x_1, ..., x_n>, <v_1, ..., v_n>)   */
/* returns the environment {x_1 |->v_1, ...,    */
/* x_n |->v_n}. It is a checked runtime error   */
/* if the two lists are not the same length.    */
/* <function prototypes for \impcore>=          */
Valenv *mkValenv(Namelist *vars, Valuelist *vals);
Funenv *mkFunenv(Namelist *vars, Funlist   *defs);
/* To retrieve a value or function definition,  */
/* we use [[fetchval]] or [[fetchfun]]. In the  */
/* operational semantics, we write the lookup   */
/* fetchval(x, rho) simply as rho(x).           */
/* <function prototypes for \impcore>=          */
Value fetchval(Name *name, Valenv *env);
Fun   fetchfun(Name *name, Funenv *env);
/* [[]] or [[isfunbound]]; they                 */
/* return [[1]] if the given name is in the     */
/* environment, and [[0]] otherwise. Formally,  */
/* isvalbound(x, rho) is written x in dom rho.  */

/* <function prototypes for \impcore>=          */
int isvalbound(Name *name, Valenv *env);
int isfunbound(Name *name, Funenv *env);
/* [[]] and [[bindfun]]. Unlike previous        */
/* operations on environments, [[bindval]] and  */
/* [[bindfun]] cannot be specified as pure      */
/* functions. Instead, [[bindval]] and          */
/* [[bindfun]] mutate their environments,       */
/* replacing the old bindings with new ones.    */
/* Calling bindval(x, v, rho) is equivalent to  */
/* performing the assignment rho := rho{x       */
/* |->v}. Because rho is passed by reference,   */
/* the caller can see the modifications to the  */
/* environment.                                 */

/* <function prototypes for \impcore>=          */
void bindval(Name *name, Value val, Valenv *env);
void bindfun(Name *name, Fun   fun, Funenv *env);
/*  Evaluation                                  */
/*                                              */
/* The evaluator's interface is not very        */
/* interesting: it is the implementation, which */
/* starts on page [->], that is interesting.    */
/* The function [[topeval]] corresponds to the  */
/* --> relation in our operational semantics,   */
/* while [[eval]] corresponds to the ==>        */
/* relation. For example, eval(e, xi, phi, rho) */
/* finds a v, xi', and rho' such that \evale    */
/* ==>\eval[']v, assigns rho := rho' and xi :=  */
/* xi', and returns v.                          */
/* <function prototypes for \impcore>=          */
void topeval(Toplevel *t, Valenv *globals, Funenv *functions, int echo);
Value eval(Exp *e, Valenv *globals, Funenv *functions, Valenv *formals);
/* [[echo]] parameter to [[topeval]] is         */
/* nonzero, [[topeval]] prints the values and   */
/* names of top-level expressions and           */
/* functions.                                   */

/* The function [[readevalprint]] consumes all  */
/* of an input source, evaluating each          */
/* expression and printing the results. As for  */
/* [[topeval]], the [[echo]] parameter controls */
/* whether [[readevalprint]] prints the values  */
/* and names of top-level expressions and       */
/* functions.                                   */
/* <function prototypes for \impcore>=          */
void readevalprint(Topreader *r, Valenv *globals, Funenv *functions, int echo);
/* Here are the prototypes:                     */
/*                                              */
/* <function prototypes for \impcore>=          */
/* <function prototypes for \impcore>=          */
Exp *mkLiteral(Value);
Exp *mkVar(Name*);
Exp *mkSet(Name *name, Exp *exp);
Exp *mkIfx(Exp *cond, Exp *true, Exp *false);
Exp *mkWhilex(Exp *cond, Exp *exp);
Exp *mkBegin(Explist*);
Exp *mkApply(Name *name, Explist *actuals);
/* <function prototypes for \impcore>=          */
Userfun mkUserfun(Namelist *formals, Exp *body);
Toplevel *mkExp(Exp*);
Toplevel *mkVal(Name *name, Exp *exp);
Toplevel *mkDefine(Name *name, Userfun userfun);
Toplevel *mkUse(Name*);
/* <function prototypes for \impcore>=          */
Fun mkUserdef(Userfun);
Fun mkPrimitive(Name*);
/* <function prototypes for \impcore>=          */
int      lengthEL (Explist *l);
Exp*     nthEL    (Explist *l, unsigned n);
Explist *mkEL     (Exp *hd, Explist *tl);
Printer  printexplist;
/* <function prototypes for \impcore>=          */
int      lengthPL (Parlist *l);
Par*     nthPL    (Parlist *l, unsigned n);
Parlist *mkPL     (Par *hd, Parlist *tl);
Printer  printparlist;
/* <function prototypes for \impcore>=          */
int        lengthVL (Valuelist *l);
Value      nthVL    (Valuelist *l, unsigned n);
Valuelist *mkVL     (Value hd, Valuelist *tl);
Printer    printvaluelist;
/* <function prototypes for \impcore>=          */
int      lengthFL (Funlist *l);
Fun      nthFL    (Funlist *l, unsigned n);
Funlist *mkFL     (Fun hd, Funlist *tl);
Printer  printfunlist;
/* <function prototypes for \impcore>=          */
int       lengthNL (Namelist *l);
Name*     nthNL    (Namelist *l, unsigned n);
Namelist *mkNL     (Name *hd, Namelist *tl);
Printer   printnamelist;
