(*   Putting all the pieces together            *)
(*                                              *)
(* We stitch together the parts of the          *)
(* implementation in this order:                *)
(* <timpcore.sml>=                              *)


(*****************************************************************)
(*                                                               *)
(*   ENVIRONMENTS                                                *)
(*                                                               *)
(*****************************************************************)

(* <environments>=                              *)
type name = string
type 'a env = (name * 'a) list
val emptyEnv = []

(* lookup and assignment of existing bindings *)
exception NotFound of name
fun find (name, []) = raise NotFound name
  | find (name, (n, v)::tail) = if name = n then v else find(name, tail)

(* adding new bindings *)
exception BindListLength
fun bind(name, v, rho) = (name, v) :: rho
fun bindList(n::vars, v::vals, rho) = bindList(vars, vals, bind(n, v, rho))
  | bindList([], [], rho) = rho
  | bindList _ = raise BindListLength
(* We represent an environment of type [['a     *)
(* env]] as a list of ([[name]], [['a]]) pairs. *)
(* The declarations in the box give the         *)
(* interface to our implementation; through     *)
(* some Noweb hackery, they are checked by the  *)
(* ML compiler.                                 *)
(* <boxed values 14>=                           *)
val _ = op emptyEnv : 'a env
val _ = op find     : name * 'a env -> 'a
val _ = op bind     : name      * 'a      * 'a env -> 'a env
val _ = op bindList : name list * 'a list * 'a env -> 'a env
(* [[error]] procedure,                         *)
(* to indicate when things have gone wrong. The *)
(* exceptions we use are listed in Table [->].  *)

exception LeftAsExercise


(*****************************************************************)
(*                                                               *)
(*   TYPES FOR \TIMPCORE                                         *)
(*                                                               *)
(*****************************************************************)

(*   Abstract syntax, types, and values of Typed Impcore *)
(*                                              *)
(* As in Chapter [->], we define the abstract   *)
(* syntax of Typed Impcore by presenting the    *)
(* representations we use in our                *)
(* implementation. The type system is so simple *)
(* that we use [[ty]] not only as the abstract  *)
(* syntax for types but also as the internal    *)
(* representation of types. The type [[funty]]  *)
(* represents the type of a function in Typed   *)
(* Impcore; there is no corresponding abstract  *)
(* syntax. [*]                                  *)
(* <types for \timpcore>=                       *)
datatype ty    = INTTY | BOOLTY | UNITTY | ARRAYTY of ty
datatype funty = FUNTY of ty list * ty
(* <types for \timpcore>=                       *)
(* Printing types and values                    *)
(*                                              *)
(* This code prints types.                      *)
(* <printing types for \timpcore>=              *)
fun typeString BOOLTY        = "bool"
  | typeString INTTY         = "int"
  | typeString UNITTY        = "unit"
  | typeString (ARRAYTY tau) = "(array " ^ typeString tau ^ ")"


(*****************************************************************)
(*                                                               *)
(*   LEXICAL ANALYSIS                                            *)
(*                                                               *)
(*****************************************************************)

(* A value of type [[par]] represents a list of *)
(* tokens with balanced parentheses. In order   *)
(* to issue better error messages, we keep the  *)
(* source-code location of the each             *)
(* parenthesized list. A source names the       *)
(* source and gives a line number.              *)
(* <lexical analysis ((mlscheme))>=             *)
type srcloc  = string * int
fun srclocString (source, line) = source ^ ", line " ^ Int.toString line
(* <lexical analysis ((mlscheme))>=             *)
datatype par = NAME  of string
             | INT   of int
             | SHARP of char  (* #t or #f *)
             | LIST  of srcloc * par list
(* <lexical analysis ((mlscheme))>=             *)

exception Paren of char list * srcloc
   (* raise if found ) when reading input *)
exception PrematureEOF of string (* shows what we were reading ... *)
exception EOF                 (* raised by reader *)
fun read nextline curloc : string -> par list =
  let (* To help with reading, we have functions to   *)
      (* deal with lexical details. We begin by       *)
      (* defining tests for the delimiters of         *)
      (* symbols:                                     *)
      (* <character-fiddling functions>=              *)
      fun isDelim c = List.exists (fn c' => c' = c) [#"(", #")", #" ", #"\t",
                                                                    #"\n", #";"]
      (* [[]] is like the micro-Scheme                *)
      (* [[exists?]]. We needn't write its            *)
      (* implementation because it is in the initial  *)
      (* basis for Standard ML.                       *)

      (* <character-fiddling functions>=              *)
      fun stringReader []     = NONE
        | stringReader (h::t) = SOME (h, t)
      val strton = Int.scan StringCvt.DEC stringReader
      (* We continue with integer conversion;         *)
      (* [[strton s]] returns an integer, plus        *)
      (* leftover characters from [[s]], or [[NONE]]  *)
      (* if no prefix of [[s]] is an integer literal. *)
      (* It uses [[Int.scan]] from the Standard Basis *)
      (* Library.                                     *)
      (* <boxed values 25>=                           *)
      val _ = op strton : char list -> (int * char list) option
      (* We need to be able to skip whitespace and    *)
      (* comments. [[StringCvt.skipWS]] skips white   *)
      (* space; if we encounter a semicolon, we use   *)
      (* [[StringCvt.dropl]] to skip everything up to *)
      (* the next newline, then resume skipping white *)
      (* space.                                       *)
      (* <character-fiddling functions>=              *)
      fun skip s =
        case StringCvt.skipWS stringReader s
          of #";" :: s => skip (StringCvt.dropl (fn c => c <> #"\n")
                                                                 stringReader s)
           | s => s
      (* [[nonwhite]], which                          *)
      (* summons more input lines as needed.          *)

      (* <character-fiddling functions>=              *)
      fun nonwhite s =
        case skip s
          of [] => nonwhite (explode (nextline()))
           | s  => s
      fun readmap f (result, unconsumed) = (f result, unconsumed)
      (* get is called only when s is guaranteed non-blank, non-comment, non-
                                                                        empty *)
      fun get (#"("::s)  = readmap (fn l => LIST l) (list (curloc(), s, []))
        | get (#")"::s)  = raise (Paren (s, curloc()))
        | get (#"'"::s)  = let val loc = curloc()
                               fun quote p = LIST (loc, [NAME "quote", p])
                           in  readmap quote (get (nonwhite s))
                           end
        | get s          = atom (s, [])
      and atom(c::s, r) = if isDelim c then (mkAtom(rev r), c::s) 
                          else atom(s, c::r)
        | atom([],   r) = (mkAtom(rev r), [])
      and mkAtom [#"#", #"f"] = SHARP #"f"
        | mkAtom [#"#", #"t"] = SHARP #"t"
        | mkAtom s = (case strton s of SOME (n, []) => INT n | _ => NAME (
                                                                     implode s))
      and list (loc, s, l) =
        let val next = nonwhite s
        in  let val (a, u) = get next
            in  list(loc, u, a::l)
            end handle (Paren (u, _)) => ((loc, rev l), skip u)
                     | EOF => raise PrematureEOF (implode next)
        end handle EOF => raise PrematureEOF (implode s)
      fun readline [] = []
        | readline s  = 
             let val(p, u) = get s
             in  p :: readline(skip u)
             end
(* The interface to the [[read]] function       *)
(* includes a [[curloc]] parameter, which can   *)
(* be used to keep track of the source-code     *)
(* location of any token. Of the interpreters   *)
(* in this book, only the uSmalltalk interpter  *)
(* (Appendix [->]) takes advantage of this      *)
(* capability.                                  *)
(*                                              *)
(* Function [[read]] is implemented by a group  *)
(* of local lexing functions. Each such         *)
(* function takes a list of unconsumed          *)
(* characters and returns a pair consisting of  *)
(* the [[par]] read plus unconsumed characters. *)
(* The functions are:                           *)
(*                                              *)
(*   [[get]] --- Read the next input            *)
(*   [[atom]] --- Read a name, integer, or      *)
(*   [[#]] literal                              *)
(*   [[list]] --- Read a list of inputs in      *)
(*   parentheses                                *)
(*                                              *)
(* The [[list]] function takes as an additional *)
(* parameter the source-code location of the    *)
(* start of the list. Internal function         *)
(* [[readline]] puts it all together by reading *)
(* as many inputs are as on the line, until all *)
(* the line is consumed.                        *)
(* <boxed values 24>=                           *)
val _ = op read : (unit -> string) -> (unit -> srcloc) -> string -> par list
val _ = op get  : char list                      -> par                 * char
                                                                            list
val _ = op atom : char list * char list          -> par                 * char
                                                                            list
val _ = op list : srcloc * char list * par list  -> (srcloc * par list) * char
                                                                            list
  in  fn s => readline(skip (explode s)) 
  end


(*****************************************************************)
(*                                                               *)
(*   ABSTRACT SYNTAX AND VALUES FOR \TIMPCORE                    *)
(*                                                               *)
(*****************************************************************)

(* There are two kinds of value: [[NUM]], which *)
(* represents both integers and Booleans, and   *)
(* [[ARRAY]], which represents arrays.          *)
(* <abstract syntax and values for \timpcore>=  *)
datatype value = NUM   of int
               | ARRAY of value array
(* The abstract syntax of expressions is what   *)
(* you would expect for Impcore, except that as *)
(* described in Section [->], array operations  *)
(* require special abstract syntax.             *)
(* <abstract syntax and values for \timpcore>=  *)
datatype exp   = LITERAL of value
               | VAR     of name
               | SET     of name * exp
               | IFX     of exp * exp * exp
               | WHILEX  of exp * exp
               | BEGIN   of exp list
               | APPLY   of name * exp list
               (* As required by the specification, the        *)
               (* [[array-get]] operation is quite flexible.   *)
               (* In the last example, the [[array-get]]       *)
               (* operation takes an argument of type [[(array *)
               (* int)]] and returns a result of type [[int]], *)
               (* but in the example before that, it takes an  *)
               (* argument of type [[(array (array int))]] and *)
               (* returns a result of type [[(array int)]]. No *)
               (* mere function in Typed Impcore can do such   *)
               (* things; this behavior is called polymorphic. *)
               (* Typed Impcore is monomorphic, which means    *)
               (* that each function can be used for arguments *)
               (* and results of one and only one type. This   *)
               (* means we can't just put [[array-get]] into   *)
               (* the initial basis for Typed Impcore; we have *)
               (* to make it a special operation in the        *)
               (* abstract syntax. In fact, all of the array   *)
               (* operations are polymorphic, so we have to    *)
               (* add these expressions to the abstract        *)
               (* syntax:                                      *)
               (* <array extensions to \timpcore's abstract syntax>= *)
               | AGET  of exp * exp
               | ASET  of exp * exp * exp
               | AMAKE of exp * exp
               | ALEN  of exp
(* The abstract syntax for a function           *)
(* definition requires that the result type and *)
(* the types of the formal parameters be        *)
(* identified explicitly. Although the concrete *)
(* syntax puts types on the left, by analogy    *)
(* with C, the abstract syntax puts types on    *)
(* the right, as is customary in the formal     *)
(* semantics and in Pascal-like and ML-like     *)
(* languages.                                   *)
(* <abstract syntax and values for \timpcore>=  *)
type userfun = { formals : (name * ty) list, body : exp, returns : ty }
datatype toplevel = EXP    of exp
                  | DEFINE of name * userfun
                  | VAL    of name * exp
                  | USE    of name
(* In Typed Impcore, the function name in an    *)
(* application is not an ordinary value, so it  *)
(* can't stand for something of type [[value]]  *)
(* in our implementation. In Chapter [->], we   *)
(* represent Impcore functions using the C type *)
(* [[Fun]], but [[fun]] is a reserved word      *)
(* in ML, so we use the name [[func]] instead.  *)
(* As in our interpreter for Impcore, a         *)
(* primitive function is represented as an ML   *)
(* function.                                    *)
(* <abstract syntax and values for \timpcore>=  *)
datatype func = USERDEF   of string list * exp
              | PRIMITIVE of value list -> value
(* [[]]; types are needed only during type      *)
(* checking, and the [[func]] representation is *)
(* used at run time, after all types have been  *)
(* checked.                                     *)

(* <abstract syntax and values for \timpcore>=  *)
(* It would be good to figure out how to use    *)
(* [[separate]] in this code.                   *)
(* <printing values for \timpcore>=             *)
fun valueString (NUM n) = Int.toString n
  | valueString (ARRAY a) =
      if Array.length a = 0 then
          "[]"
      else
          let val elts = Array.foldr (fn (v, s) => " " :: valueString v :: s) [
                                                                          "]"] a
          in  String.concat ("[" :: tl elts)
          end
(* Appendix [->] defines a function used to     *)
(* print types.                                 *)
(* <boxed values 1>=                            *)
val _ = op typeString : ty -> string
(* Just like types, values can be printed. The  *)
(* code appears in Appendix [->].               *)
(* <boxed values 1>=                            *)
val _ = op valueString : value -> string
(* <abstract syntax and values for \timpcore>=  *)
exception BugInTypeChecking of string
fun toArray (ARRAY a) = a
  | toArray _         = raise BugInTypeChecking "non-array value"
fun toInt   (NUM n)   = n
  | toInt _           = raise BugInTypeChecking "non-integer value"
(* To interpret any array operation, we need to *)
(* convert at least one argument from a         *)
(* [[value]] to an array or to an integer. If   *)
(* the program type checks, the conversion      *)
(* should always succeed; if a conversion       *)
(* fails, there is a bug in the type checker.   *)
(* The [[toArray]] and [[toInt]] functions do   *)
(* the job.                                     *)
(* <boxed values 6>=                            *)
val _ = op toArray : value -> value array
val _ = op toInt   : value -> int


(*****************************************************************)
(*                                                               *)
(*   PARSING FOR \TIMPCORE                                       *)
(*                                                               *)
(*****************************************************************)

(* Parsing                                      *)
(*                                              *)
(* Typed Impcore can use micro-Scheme's lexical *)
(* analysis, so all we have here is a parser.   *)
(* <parsing for \timpcore>=                     *)
exception SyntaxError of string
fun syntaxErrorAt (source, line) msg =
  raise SyntaxError (source ^ ", line " ^ Int.toString line ^ ": " ^ msg)
fun parse (NAME s)  = VAR s
  | parse (INT n)   = LITERAL (NUM n)
  | parse (SHARP c) = raise SyntaxError "No # in Typed Impcore"
  | parse (LIST (loc, NAME "set"    :: args)) = parseSet loc args
  | parse (LIST (loc, NAME "if"     :: args)) = parse3 loc IFX "if" args
  | parse (LIST (loc, NAME "begin"  :: args)) = BEGIN (map parse args)
  | parse (LIST (loc, NAME "while"  :: args)) = parse2 loc WHILEX "while" args
  | parse (LIST (loc, NAME "array-get"  :: args)) = parse2 loc AGET "array-get"
                                                                            args
  | parse (LIST (loc, NAME "array-set"  :: args)) = parse3 loc ASET "array-set"
                                                                            args
  | parse (LIST (loc, NAME "array-make" :: args)) = parse2 loc AMAKE
                                                               "array-make" args
  | parse (LIST (loc, NAME "array-length" :: args)) = parse1 loc ALEN
                                                             "array-length" args
  | parse (LIST (loc, NAME h::t))  = APPLY (h, map parse t)
  | parse (LIST (loc, _ :: _)) = syntaxErrorAt loc
                                           "only named functions can be applied"
  | parse (LIST (loc, [])) = syntaxErrorAt loc "empty application ()"
and parse3 _ f name [e1, e2, e3] = f (parse e1, parse e2, parse e3)
  | parse3 loc _ name _ = syntaxErrorAt loc ("expected (" ^ name ^ " e1 e2 e3)")
and parse2 _ f name [e1, e2] = f (parse e1, parse e2)
  | parse2 loc _ name _ = syntaxErrorAt loc ("expected (" ^ name ^ " e1 e2)")
and parse1 _ f name [e1] = f (parse e1)
  | parse1 loc _ name _ = syntaxErrorAt loc ("expected (" ^ name ^ " e)")
and parseSet _ [NAME x, e] = SET (x, parse e)
  | parseSet loc _ = syntaxErrorAt loc "expected (set x e)"
(* <parsing for \timpcore>=                     *)
fun toplevel (LIST (loc, NAME "define" :: args)) = parseDefine loc args
  | toplevel (LIST (loc, NAME "val"    :: args)) = parseVal    loc args
  | toplevel (LIST (loc, NAME "use"    :: args)) = parseUse    loc args
  | toplevel e = EXP (parse e)
and parseDefine loc [ty, NAME name, LIST (floc, formals), body] =
      DEFINE (name, { returns = parseTy loc ty,
                      formals = map (parseFormal floc) formals, body = parse
                                                                          body})
  | parseDefine loc _ = syntaxErrorAt loc "expected (define ty f (args) body)"
and parseFormal _ (LIST (loc, [ty, NAME x])) = (x, parseTy loc ty)
  | parseFormal _ (LIST (loc, _)) = syntaxErrorAt loc "expected (ty argname)"
  | parseFormal loc _ = syntaxErrorAt loc "expected (ty argname)"
and parseTy loc (NAME "bool") = BOOLTY
  | parseTy loc (NAME "unit") = UNITTY
  | parseTy loc (NAME "int")  = INTTY
  | parseTy loc (NAME n)      =
      syntaxErrorAt loc ("Cannot recognize name " ^ n ^ " as a type")
  | parseTy loc (LIST (aloc, [NAME "array", ty])) = ARRAYTY (parseTy aloc ty)
  | parseTy loc _ = syntaxErrorAt loc "expected int, bool, unit, or (array ty)"
and parseVal _ [NAME v, e] = VAL (v, parse e)
  | parseVal loc _ = syntaxErrorAt loc "expected (val v e)"
and parseUse _ [NAME v] = USE v
  | parseUse loc _ = syntaxErrorAt loc "expected (use filename)"


(*****************************************************************)
(*                                                               *)
(*   READERS                                                     *)
(*                                                               *)
(*****************************************************************)

(* Readers                                      *)
(*                                              *)
(*  Line readers                                *)
(*                                              *)
(* Our readers are structured slightly          *)
(* differently from those in C. A reader is a   *)
(* bundle containing three related functions:   *)
(* [[firstline]], which returns the first line  *)
(* read; [[nextline]], which returns all        *)
(* subsequent lines; and [[curloc]], which      *)
(* returns a pair identifying the source        *)
(* location of the most recently returned line. *)
(* Both [[firstline]] and [[nextline]] can      *)
(* raise [[EOF]], (We distinguish [[firstline]] *)
(* from [[nextline]] because in an interactive  *)
(* reader, they may use different prompts.)     *)
(* <readers>=                                   *)
type reader = { firstline : unit -> string
              , nextline  : unit -> string
              , curloc    : unit -> string * int
              }
(* <readers>=                                   *)
fun promptReader (firstPrompt, nextPrompt, reader) =
  let val { firstline = first, nextline = next, curloc = cur } = reader
      fun promptIn get prompt () = ( TextIO.output(TextIO.stdOut, prompt)
                                   ; TextIO.flushOut(TextIO.stdOut)
                                   ; get ()
                                   )
  in  { firstline = promptIn first firstPrompt
      , nextline  = promptIn next  nextPrompt
      , curloc    = cur
      }
  end
(* Given a silent reader, we can make it issue  *)
(* a prompt every time it reads a new line. We  *)
(* use two prompts: one for the first line of   *)
(* an expression and one for later lines.       *)
(* <boxed values 26>=                           *)
val _ = op promptReader : string * string * reader -> reader
(* <readers>=                                   *)
fun filereader (filename, fd)=
  let val lineNum = ref 0
      fun get () =
        let val line = TextIO.inputLine fd
        in  if size line = 0 then raise EOF else line before lineNum := !lineNum
                                                                             + 1
        end
  in  { firstline = get, nextline = get, curloc = (fn () => (filename, !lineNum)
                                                                             ) }
  end

fun stringsreader (what, l) = 
  let val buffer  = ref l
      val lineNum = ref 0
      fun get ()  = case !buffer
                      of [] => raise EOF
                       | h :: t => h before buffer := t before lineNum := !
                                                                     lineNum + 1
  in  { firstline = get, nextline = get, curloc = (fn () => (what, !lineNum)) }
  end              
(* We can read from files or lists of strings.  *)
(* File I/O is provide by the standard module   *)
(* [[TextIO]].                                  *)
(* <boxed values 27>=                           *)
val _ = op filereader    : string * TextIO.instream -> reader
val _ = op stringsreader : string * string list     -> reader
(* [[]], a private mutable cell. When all       *)
(* the strings in the buffer are used up, there *)
(* aren't any more.                             *)

(*  Top-level readers                           *)
(*                                              *)
(* A top-level reader is a buffer of top-level  *)
(* items together with a reader. [*]            *)
(* <readers>=                                   *)
type topreader = { buffer : toplevel list ref, reader : reader }
fun readtop (r as { buffer, reader = { nextline, firstline, curloc } }) =
  case !buffer
    of h::t => h before buffer := t
     | []   => ( buffer := map toplevel (read nextline curloc (firstline()))
               ; readtop r
               )
(* We decide whether to prompt when building a  *)
(* top-level reader.                            *)
(* <readers>=                                   *)
fun topreader (reader, prompt) = 
  let val reader = if prompt then promptReader ("-> ", "   ", reader) else
                                                                          reader
  in  { buffer = ref [], reader = reader }
  end
(* <readers>=                                   *)
fun echoTag f x =
  let val line = f x
      val _ = if (String.substring (line, 0, 2) = ";#" handle _ => false) then
                print line
              else
                ()
  in  line
  end

fun echoReader { nextline=n, firstline=f, curloc=c } =
     { nextline=echoTag n, firstline=echoTag f, curloc=c } 

fun echoTop { buffer = b, reader = r } = { buffer = b, reader = echoReader r }
val topreader = fn args => echoTop (topreader args)
(* As in the C code, we want to print out any   *)
(* line read if it begins with the special      *)
(* string [[;#]]. This string is a formal       *)
(* comment that helps us test the chunks marked *)
(* [[]]. In our ML code, we                     *)
(* arrange it using the higher-order function   *)
(* [[echoReader]].                              *)
(* <boxed values 28>=                           *)
val _ = op echoReader : reader -> reader 


(*****************************************************************)
(*                                                               *)
(*   IMPLEMENTATION OF [[USE]]                                   *)
(*                                                               *)
(*****************************************************************)

(* The implementation of [[use]] is             *)
(* parameterized by [[readEvalPrint]] so we can *)
(* share it with other interpreters. Function   *)
(* [[use]] creates a top-level reader that does *)
(* not prompt, but it uses [[writeln]] to be    *)
(* sure that responses are printed.             *)
(* <implementation of [[use]]>=                 *)
fun use readEvalPrint filename rho =
      let val fd = TextIO.openIn filename
          val treader = topreader (filereader (filename, fd), false)
          fun writeln s = app print [s, "\n"]
      in  readEvalPrint (treader, writeln, writeln) rho
          before TextIO.closeIn fd
      end 
(* [[topreader]] is relegated                   *)
(* to chunk [->] in Appendix [->].              *)



(*****************************************************************)
(*                                                               *)
(*   TYPE CHECKING FOR \TIMPCORE                                 *)
(*                                                               *)
(*****************************************************************)

(* <type checking for \timpcore>=               *)
exception TypeError of string
fun typeof (e, globals, functions, formals) =
  let (* All literals are integers. \usetyLiteral     *)
      (*                                              *)
      (* <function [[ty]], to check the type of an expression, given $\itenvs$>=
                                                                              *)
      fun ty (LITERAL v) = INTTY
      (* To type a variable, we must try two          *)
      (* environments. \usetyFormalVar                *)
      (* \usetyGlobalVar The code is shorter than the *)
      (* rules!                                       *)
      (* <function [[ty]], to check the type of an expression, given $\itenvs$>=
                                                                              *)
        | ty (VAR x)     = (find (x, formals) handle NotFound _ => find(x,
                                                                       globals))
      (* We also need two environments to check       *)
      (* assignments. \usetyFormalAssign              *)
      (* \usetyGlobalAssign To implement one of these *)
      (* rules, we determine the types of the         *)
      (* variable and the expression, then compare    *)
      (* for equality. (The recursive call [[ty (VAR  *)
      (* x)]] is a short cut that avoids duplicating  *)
      (* code.) Giving an error message that is even  *)
      (* remotely helpful requires as much code as    *)
      (* checking the types.                          *)
      (* <function [[ty]], to check the type of an expression, given $\itenvs$>=
                                                                              *)
        | ty (SET (x, e)) =
               let val tau_x = ty (VAR x)
                   val tau_e = ty e
               in  if tau_x = tau_e then
                     tau_x
                   else
                     raise TypeError ("Set variable " ^ x ^ " of type " ^
                                                              typeString tau_x ^
                                      " to value of type " ^ typeString tau_e)
               end
      (* The premises of the \rulenameIf rule require *)
      (* both branches to have the same type.         *)
      (* \usetyIf The implementation computes the     *)
      (* types of the conditions and both branches.   *)
      (* Again, most of the code is devoted to error  *)
      (* messages.                                    *)
      (* <function [[ty]], to check the type of an expression, given $\itenvs$>=
                                                                              *)
        | ty (IFX (e1, e2, e3)) =
            let val tau1 = ty e1
                val tau2 = ty e2
                val tau3 = ty e3
            in  if tau1 = BOOLTY then
                  if tau2 = tau3 then
                    tau2
                  else
                    raise TypeError ("In if expression, true branch has type " ^
                                     typeString tau2 ^
                                                 " but false branch has type " ^
                                     typeString tau3)
                else
                  raise TypeError ("Condition in if expression has type " ^
                                                               typeString tau1 ^
                                   ", which should be " ^ typeString BOOLTY)
            end
      (* The checking of a [[while]] loop is very     *)
      (* similar to that of an [[if]], except that we *)
      (* don't care what the type of the body is.     *)
      (* \usetyWhile                                  *)
      (* <function [[ty]], to check the type of an expression, given $\itenvs$>=
                                                                              *)
        | ty (WHILEX (e1, e2)) =
            let val tau1 = ty e1
                val tau2 = ty e2
            in  if tau1 = BOOLTY then UNITTY
                else
                  raise TypeError ("Condition in while expression has type " ^
                                   typeString tau1 ^ ", which should be " ^
                                   typeString BOOLTY)
            end
      (* The code for [[begin]] checks the types of   *)
      (* all sub-expressions. \usetyBegin             *)
      (* \usetyEmptyBegin We handle the               *)
      (* \rulenameEmptyBegin rule by using [[UNITTY]] *)
      (* as the initial value of [[tau]] that is      *)
      (* passed to [[last]].                          *)
      (* <function [[ty]], to check the type of an expression, given $\itenvs$>=
                                                                              *)
        | ty (BEGIN es) =
             let val bodytypes = map ty es
                 fun last tau []     = tau
                   | last tau (h::t) = last h t
             in  last UNITTY bodytypes
             end
      (* Because the two polymorphic primitives have  *)
      (* special rules, they require special code in  *)
      (* the type checker. \usetyApplyEq Because      *)
      (* functions in Typed Impcore are called by     *)
      (* name, we can identify the equality primitive *)
      (* by name.                                     *)
      (* <function [[ty]], to check the type of an expression, given $\itenvs$>=
                                                                              *)
        | ty (APPLY ("=", actuals)) =
             (case map ty actuals
               of [t1, t2] =>
                   if t1 = t2 then
                     BOOLTY
                   else
                     raise TypeError (
                          "equality comparisons of values of different types " ^
                                      typeString t1 ^ " and " ^ typeString t2)
                | _ => raise TypeError "Wrong number of arguments to =")
      (* The [[print]] primitive is similar.          *)
      (* \usetyApplyPrint We must check the type of   *)
      (* the argument even though it doesn't affect   *)
      (* the type of the result.                      *)
      (* <function [[ty]], to check the type of an expression, given $\itenvs$>=
                                                                              *)
        | ty (APPLY ("print", [arg])) = (ty arg; UNITTY)
      (* For the general case of function             *)
      (* application, we look up the function in the  *)
      (* appropriate environment. \usetyApply         *)
      (* <function [[ty]], to check the type of an expression, given $\itenvs$>=
                                                                              *)
        | ty (APPLY (f, actuals)) =
             let val actualtypes                     = map ty actuals
                 val FUNTY (formaltypes, resulttype) = find (f, functions)
             in  if actualtypes = formaltypes then
                   resulttype
                 else
                   raise TypeError (
                              "argument-type mismatch in call to function " ^ f)
                   (* appalling error message *)
             end
      (* We don't show you how to turn these rules    *)
      (* into code for the type checker; that problem *)
      (* is left for Exercise [->]. [*]               *)
      (* <function [[ty]], to check the type of an expression, given $\itenvs$>=
                                                                              *)
      | ty (AGET (a, i))       = raise LeftAsExercise
      | ty (ASET (a, i, e))    = raise LeftAsExercise
      | ty (AMAKE (len, init)) = raise LeftAsExercise
      | ty (ALEN a)            = raise LeftAsExercise
(*   Type checking                              *)
(*                                              *)
(* Given an expression e and a collection of    *)
(* type environments \xgam, \fgam, and \vgam,   *)
(* calling typeof(e, \xgam, \fgam, \rgam)       *)
(* returns a type tau such that \itypeise tau.  *)
(* If no such type exists, [[typeof]] raises    *)
(* the [[TypeError]] exception (or possibly     *)
(* [[NotFound]]). We use an auxiliary, nested   *)
(* function [[ty]], which doesn't pass          *)
(* environments explicitly.                     *)
(* <boxed values 2>=                            *)
val _ = op typeof  : exp * ty env * funty env * ty env -> ty
val _ = op ty      : exp -> ty
  in  ty e
  end
(* [[]] by examining the rules of               *)
(* operational semantics, we can derive an      *)
(* implementation of [[ty]] by examining the    *)
(* rules of the type system. To help show the   *)
(* connection between the type system and the   *)
(* type checker, we show the relevant rules     *)
(* before each case of the function [[ty]].     *)

(* <type checking for \timpcore>=               *)
exception RuntimeError of string
fun topty (t, globals, functions) =
    case t
      of (* Type-checking a [[val]] binding requires     *)
         (* extending the global-variable environment.   *)
         (* \usetyVal                                    *)
         (* <cases for type-checking type-level items in \timpcore>= *)
           VAL (x, e) => (bind(x, typeof (e, globals, functions, emptyEnv),
                                                            globals), functions)
         (* A top-level expression must be type-checked, *)
         (* but it doesn't affect any environments.      *)
         (* \usetyExp                                    *)
         (* <cases for type-checking type-level items in \timpcore>= *)
         | EXP e => (typeof (e, globals, functions, emptyEnv); (globals,
                                                                     functions))
         (* A function definition requires significant   *)
         (* manipulation of the function environment.    *)
         (* \usetyDefine                                 *)
         (* <cases for type-checking type-level items in \timpcore>= *)
         | DEFINE (name, {returns, formals, body}) =>
            let val (fnames, ftys) = ListPair.unzip formals
                val functions' = bind(name, FUNTY (ftys, returns), functions)
                val tau = typeof (body, globals, functions', bindList(fnames,
                                                                ftys, emptyEnv))
            in  if tau = returns then
                  (globals, functions')
                else
                  raise TypeError ("Body of function has type " ^ typeString tau
                                                                               ^
                                   ", which does not match declaration of " ^
                                   typeString returns)
            end
         (* A [[use]] directive should never get this    *)
         (* far.                                         *)
         (* <cases for type-checking type-level items in \timpcore>= *)
         | USE _ => raise BugInTypeChecking "`use' reached the type checker"
(*   Type-checking top-level items              *)
(*                                              *)
(* The form of the typing judgment for a        *)
(* top-level item t is \itoptt                  *)
(* -->\stwo\fgamp\vgamp.                        *)
(* <boxed values 3>=                            *)
val _ = op topty : toplevel * ty env * funty env -> ty env * funty env


(*****************************************************************)
(*                                                               *)
(*   CHECKING AND EVALUATION FOR \TIMPCORE                       *)
(*                                                               *)
(*****************************************************************)

(* <checking and evaluation for \timpcore>=     *)
(* <evaluation for \timpcore>=                  *)
(* <definition of [[eval]] for \timpcore>=      *)
fun eval(e, globals, functions, formals) =
  let fun toBool (NUM 0) = false
        | toBool _       = true
      fun findVar v = find (v, formals) handle NotFound _ => find (v, globals)
      fun ev(LITERAL n)          = n
        | ev(VAR x)              = !(findVar x)
        | ev(SET (x, e))         = let val v = ev e in v before findVar x := v
                                                                             end
        | ev(IFX (cond, t, f))   = if toBool (ev cond) then ev t else ev f
        | ev(WHILEX (cond, exp)) =
            if toBool (ev cond) then
                (ev exp; ev (WHILEX (cond, exp)))
            else
                NUM 0
        | ev(BEGIN es) =
            let fun b(e::es, lastval) = b(es, ev e)
                  | b(   [], lastval) = lastval
            in  b(es, NUM 0)
            end
        | ev(APPLY (f, args)) = 
            (case find (f, functions)
               of PRIMITIVE p  => p (map ev args)
                | USERDEF func => (* <apply user-defined function [[func]] to
                                                                   [[args]]>= *)
                                  let val (formals, body) = func
                                      val actuals         = map (ref o ev) args
                                  (* To apply a function, we have to build an
                                                                              *)
                                  (* evaluation environment. We strip the types
                                                                              *)
                                  (* off the formals and we put the actuals in
                                                                              *)
                                  (* mutable ref cells.
                                                                              *)
                                  (* <boxed values 9>=
                                                                              *)
                                  val _ = op formals : name      list
                                  val _ = op actuals : value ref list
                                  in  eval(body, globals, functions, bindList(
                                                    formals, actuals, emptyEnv))
                                      handle BindListLength => 
                                          raise BugInTypeChecking
                                         "Wrong number of arguments to function"
                                  end)
        (* <more alternatives for [[ev]] for \timpcore>= *)
        | ev (AGET (a, i)) =
            (Array.sub (toArray (ev a), toInt (ev i))
             handle Subscript => raise RuntimeError
                                                "Array subscript out of bounds")
        | ev (ASET (e1, e2, e3)) =
            (let val (a, i, v) = (ev e1, ev e2, ev e3)
             in  Array.update (toArray a, toInt i, v);
                 v
             end handle Subscript => raise RuntimeError
                                                "Array subscript out of bounds")
        | ev (AMAKE (len, init)) = 
             (ARRAY (Array.array (toInt (ev len), ev init))
              handle Size => raise RuntimeError
                                         "array length too large (or negative)")
        | ev (ALEN a) = NUM (Array.length (toArray (ev a)))
        (* Once we have [[toArray]] and [[toInt]],      *)
        (* interpreting the array operations is easy.   *)
        (* Everything we need is in the [[Array]]       *)
        (* module from ML's Standard Basis Library. The *)
        (* library includes run-time checks for bad     *)
        (* subscripts or array sizes; we need these     *)
        (* checks because Typed Impcore's type system   *)
        (* is not powerful enough to preclude such      *)
        (* errors.                                      *)
        (* <boxed values 7>=                            *)
        val _ = op ev : exp -> value
(* Evaluation                                   *)
(*                                              *)
(* <boxed values 8>=                            *)
val _ = op ev : exp -> value
  in  ev e
  end
(* The implementation of the evaluator uses the *)
(* same techniques we used to implement         *)
(* micro-Scheme in Chapter [->]. Because of     *)
(* Typed Impcore's many environments, the       *)
(* evaluator does more bookkeeping. Another     *)
(* difference is that many potential run-time   *)
(* errors should be impossible because the      *)
(* relevant code would be rejected by the type  *)
(* checker. If one of those errors occurs       *)
(* anyway, we raise the exception               *)
(* [[BugInTypeChecking]].                       *)
(* <boxed values 10>=                           *)
val _ = op eval : exp * value ref env * func env * value ref env -> value
(* <evaluation for \timpcore>=                  *)
fun topeval (t, globals, functions, echo) =
  case t
    of DEFINE (f, {body=e, formals=xs, returns=rt}) =>
          (globals, bind(f, USERDEF (map #1 xs, e), functions))
          before echo f
     | VAL (name, e) => (* We check the type or every expression before *)
                        (* evaluating it. If [[e]] is not well typed,   *)
                        (* calling [[typeof(e, Gamma)]] raises the      *)
                        (* [[TypeError]] exception, and we never call   *)
                        (* [[eval]].                                    *)
                        (* <evaluate [[e]] and bind to [[name]]>=       *)
                        let val v = eval(e, globals, functions, emptyEnv)
                            val _ = echo (valueString v)
                        in  (bind(name, ref v, globals), functions)
                        end
     | EXP e => (* <evaluate [[e]]>=                            *)
                let val v = eval(e, globals, functions, emptyEnv)
                    val _ = echo (valueString v)
                in  (globals, functions)
                end
     | USE _ => raise RuntimeError "Internal error - `use' was evaluated"
(* <evaluation for \timpcore>=                  *)
fun binaryOp f = (fn [a, b] => f(a, b) | _ => raise BugInTypeChecking "arity 2")
fun unaryOp  f = (fn [a]    => f a     | _ => raise BugInTypeChecking "arity 1")
(* The definition of [[eval]] appears in        *)
(* Appendix [->].                               *)
(*                                              *)
(* Where other interpreters use a single        *)
(* environment, the uML interpreter uses a      *)
(* bundle of environments. One provides the rho *)
(* used for evaluation; the other provides      *)
(* the Gamma used for type checking. In the     *)
(* code, we use [[valenv]] to stand for rho, to *)
(* avoid confusion with explicit substitutions. *)
(* <boxed values 11>=                           *)
val _ = op topeval : toplevel * value ref env * func env * (string->unit) ->
                                                        value ref env * func env
(* Here are the primitives. As in Chapter [->], *)
(* all are either binary or unary operators.    *)
(* Type checking should guarantee that          *)
(* operators are used with the correct arity.   *)
(* <boxed values 11>=                           *)
val _ = op unaryOp  : (value         -> value) -> (value list -> value)
val _ = op binaryOp : (value * value -> value) -> (value list -> value)
(* <evaluation for \timpcore>=                  *)
fun arithOp f =
      binaryOp (fn (NUM n1, NUM n2) => NUM (f(n1, n2)) 
                 | _ => raise BugInTypeChecking "arithmetic on non-numbers")
val arithtype = FUNTY ([INTTY, INTTY], INTTY)
(* Arithmetic primitives expect and return      *)
(* integers.                                    *)
(* <boxed values 12>=                           *)
val _ = op arithOp   : (int * int -> int) -> (value list -> value)
val _ = op arithtype : funty
(* [[]]                                         *)
(* cons up all the primitives into one giant    *)
(* list, and we use that list to build the      *)
(* initial environment for the read-eval-print  *)
(* loop. The big difference is that in Typed    *)
(* Impcore, each primitive has a type as well   *)
(* as a value.                                  *)

(* <evaluation for \timpcore>=                  *)
fun embedPredicate f args = NUM (if f args then 1 else 0)
fun comparison f = binaryOp (embedPredicate f)
fun intcompare f = 
      comparison (fn (NUM n1, NUM n2) => f(n1, n2)
                   | _ => raise BugInTypeChecking "comparing non-numbers")
val comptype = FUNTY ([INTTY, INTTY], BOOLTY)
(* Comparisons take two arguments. Most         *)
(* comparisons (except for equality) apply only *)
(* to integers.                                 *)
(* <boxed values 13>=                           *)
val _ = op comparison : (value * value -> bool) -> (value list -> value)
val _ = op intcompare : (int   * int   -> bool) -> (value list -> value)
val _ = op comptype   : funty
type env_bundle = ty env * funty env * value ref env * func env
fun topCheckEval (t, envs as (tglobals, tfuns, vglobals, vfuns), echo) =
  case t
    of USE filename => use readCheckEvalPrint filename envs
     | _ =>
        let val (tglobals, tfuns) = topty   (t, tglobals, tfuns)
            val (vglobals, vfuns) = topeval (t, vglobals, vfuns, echo)
(*   Top-level processing: type checking and evaluation *)
(*                                              *)
(* In an interpreter for a dynamically typed    *)
(* language such as Impcore or micro-Scheme, we *)
(* can evaluate a top-level item as soon as it  *)
(* is parsed. In an interpreter for a           *)
(* statically typed language such as Typed      *)
(* Impcore, we require a type-checking step     *)
(* between parsing and evaluation. A simple way *)
(* to introduce this step is to define a        *)
(* function [[topCheckEval]] which first        *)
(* type-checks a top-level item, then evaluates *)
(* it. This function manipulates not only the   *)
(* top-level type environments \fgam and \vgam  *)
(* but also the top-level value and function    *)
(* environments phi and xi. We define type      *)
(* [[env_bundle]] to refer to this group of     *)
(* four environments. The value environment xi  *)
(* is the only one that can be mutated during   *)
(* evaluation, so it is the only one that has a *)
(* [[ref]] in its type.                         *)
(* <boxed values 4>=                            *)
val _ = op topCheckEval : toplevel * env_bundle * (string->unit) -> env_bundle
        in  (tglobals, tfuns, vglobals, vfuns)
        end
(* [[topty]], and                               *)
(* ``run time,'' where we run the evaluator     *)
(* [[topeval]], is sometimes called the phase   *)
(* distinction. It's easy to overlook the phase *)
(* distinction when you're working with an      *)
(* interactive interpreter, but the code shows  *)
(* the phase distinction is real.               *)

(* <checking and evaluation for \timpcore>=     *)
and readCheckEvalPrint (reader, echo, errmsg) envs =
  let fun loop envs =
        let fun continue msg = (errmsg msg; loop envs)
            fun finish() = envs
        in  loop (topCheckEval (readtop reader, envs, echo))
            handle EOF => finish()
                (* <more read-eval-print handlers>=             *)
                | TypeError         msg => continue ("type error: " ^ msg)
                | BugInTypeChecking msg => continue ("bug in type checking: " ^
                                                                            msg)
                (* The handlers in the first group deal with    *)
                (* problems that arise during I/O, lexical      *)
                (* analysis, and parsing.                       *)
                (* <more read-eval-print handlers>=             *)
                | IO.Io {name, ...} => continue ("I/O error: " ^ name)
                | Paren (_, loc)    => continue ("syntax error: extra ) in " ^
                                                               srclocString loc)
                | PrematureEOF s    => (errmsg ("Missing )? Got EOF reading (" ^
                                                                   s); finish())
                | SyntaxError msg   => continue ("syntax error: " ^ msg)
                (* [[IO.Io]] is part of the                     *)
                (* Standard Basis Library.                      *)

                (* The handlers in the second group deal with   *)
                (* problems that arise during evaluation. They  *)
                (* always continue with the next top-level      *)
                (* item.                                        *)
                (* <more read-eval-print handlers>=             *)
                | Div               => continue "Division by zero"
                | Overflow          => continue "Arithmetic overflow"
                | RuntimeError msg  => continue ("run-time error: " ^ msg)
                | NotFound n        => continue ("variable " ^ n ^ " not found")
                (* [[Div]] and [[Overflow]] are                 *)
                (* predefined.                                  *)

        end
  in  loop envs
  end
(*   The read-eval-print loop                   *)
(*                                              *)
(* As in micro-Scheme, the read-eval-print loop *)
(* is higher-order. There are two differences:  *)
(* the environment parameter's name is [[envs]] *)
(* instead of [[rho]], and we have handlers for *)
(* new exceptions: [[TypeError]] and            *)
(* [[BugInTypeChecking]].                       *)
(* <boxed values 5>=                            *)
val _ = op readCheckEvalPrint : topreader * (string->unit) * (string->unit) ->
                                                        env_bundle -> env_bundle
(* [[]]                                         *)
(* We also have handlers for two new            *)
(* exceptions. [[TypeError]] is raised not at   *)
(* parsing time, and not at evaluation time,    *)
(* but at type-checking time.                   *)
(* [[BugInTypeChecking]] should never be        *)
(* raised.                                      *)



(*****************************************************************)
(*                                                               *)
(*   INITIALIZATION FOR \TIMPCORE                                *)
(*                                                               *)
(*****************************************************************)

(*   Initializing the interpreter               *)
(*                                              *)
(* We're ready to put everything together into  *)
(* a working interpreter.                       *)
(* <initialization for \timpcore>=              *)
val initialEnvs =
  let fun addPrim ((name, prim, funty), (tfuns, vfuns)) = 
        ( bind (name, funty, tfuns)
        , bind (name, PRIMITIVE prim, vfuns)
        )
      val (tfuns, vfuns)  = foldl addPrim (emptyEnv, emptyEnv)
                            ((* <primops for \timpcore\ [[::]]>=             *)
                             ("+", arithOp op +,   arithtype) :: 
                             ("-", arithOp op -,   arithtype) :: 
                             ("*", arithOp op *,   arithtype) :: 
                             ("/", arithOp op div, arithtype) ::
                             (* <primops for \timpcore\ [[::]]>=             *)
                             ("<", intcompare op <, comptype) :: 
                             (">", intcompare op >, comptype) ::
                             ("=", comparison (fn (NUM n1,   NUM n2)   => n1 =
                                                                              n2
                                                | (ARRAY a1, ARRAY a2) => a1 =
                                                                              a2
                                                |  _                   => false)
                                 , FUNTY ([], BOOLTY) (* bogus *) ) ::
                             (* The last primitive is [[print]].             *)
                             (*                                              *)
                             (* <primops for \timpcore\ [[::]]>=             *)
                             ("print", unaryOp (fn x => (print (valueString x^
                                                                     "\n"); x)),
                                          FUNTY ([INTTY], UNITTY)) :: nil)
      val envs    = (emptyEnv, tfuns, emptyEnv, vfuns)
      val basis   = (* <ML representation of initial basis>=        *)
                      [ "; typesys.nw 388"
                      , "(define bool and ((bool b) (bool c)) (if b c b))"
                      , "(define bool or  ((bool b) (bool c)) (if b b c))"
                      ,
                  "(define bool not ((bool b))          (if b (= 1 0) (= 0 0)))"
                      , "; typesys.nw 394"
                      , "(define bool <= ((int x) (int y)) (not (> x y)))"
                      , "(define bool >= ((int x) (int y)) (not (< x y)))"
                      , "(define bool != ((int x) (int y)) (not (= x y)))"
                      , "; typesys.nw 400"
                      , "(define int mod ((int m) (int n)) (- m (* n (/ m n))))"
                      ]
      val treader = topreader (stringsreader ("initial basis", basis), false)
  in  readCheckEvalPrint (treader, fn _ => (), fn _ => ()) envs
  end
(* The function [[runInterpreter]] takes one    *)
(* argument, which tells it whether to prompt.  *)
(* <initialization for \timpcore>=              *)
fun runInterpreter noisy = 
  let fun writeln s = app print [s, "\n"]
      val treader = topreader (filereader ("standard input", TextIO.stdIn),
                                                                          noisy)
  in  ignore (readCheckEvalPrint (treader, writeln, writeln) initialEnvs)
  end 


(*****************************************************************)
(*                                                               *)
(*   COMMAND LINE                                                *)
(*                                                               *)
(*****************************************************************)

(* Building and exporting a program             *)
(*                                              *)
(* The final step in implementing the           *)
(* interpreter is to create a function that     *)
(* looks at the command line and calls          *)
(* [[runInterpreter]]. With a compiler like     *)
(* Moscow ML or MLton, the module isn't         *)
(* elaborated until run time, so we can simply  *)
(* insert an irrelevant [[val]] binding, the    *)
(* elaboration of which has the side effect of  *)
(* calling [[main]].                            *)
(* [[CommandLine.arguments()]] returns an       *)
(* argument list.                               *)
(* <command line>=                              *)
fun main ["-q"] = runInterpreter false
  | main []     = runInterpreter true
  | main _      =
      TextIO.output(TextIO.stdErr, "Usage: " ^ CommandLine.name() ^ " [-q]\n")
val _ = main (CommandLine.arguments())
