(*  Putting all the pieces together             *)
(*                                              *)
(* We stitch together the parts of the          *)
(* implementation in this order:                *)
(* <tuscheme.sml>=                              *)


(*****************************************************************)
(*                                                               *)
(*   ENVIRONMENTS                                                *)
(*                                                               *)
(*****************************************************************)

(* <environments>=                              *)
type name = string
type 'a env = (name * 'a) list
val emptyEnv = []

(* lookup and assignment of existing bindings *)
exception NotFound of name
fun find (name, []) = raise NotFound name
  | find (name, (n, v)::tail) = if name = n then v else find(name, tail)

(* adding new bindings *)
exception BindListLength
fun bind(name, v, rho) = (name, v) :: rho
fun bindList(n::vars, v::vals, rho) = bindList(vars, vals, bind(n, v, rho))
  | bindList([], [], rho) = rho
  | bindList _ = raise BindListLength
(* We represent an environment of type [['a     *)
(* env]] as a list of ([[name]], [['a]]) pairs. *)
(* The declarations in the box give the         *)
(* interface to our implementation; through     *)
(* some Noweb hackery, they are checked by the  *)
(* ML compiler.                                 *)
(* <boxed values 19>=                           *)
val _ = op emptyEnv : 'a env
val _ = op find     : name * 'a env -> 'a
val _ = op bind     : name      * 'a      * 'a env -> 'a env
val _ = op bindList : name list * 'a list * 'a env -> 'a env
(* [[error]] procedure,                         *)
(* to indicate when things have gone wrong. The *)
(* exceptions we use are listed in Table [->].  *)

exception LeftAsExercise
exception TypeError of string


(*****************************************************************)
(*                                                               *)
(*   DEFINITION OF [[SEPARATE]]                                  *)
(*                                                               *)
(*****************************************************************)

(* Function [[separate]] helps print a readable *)
(* list of formals for the error message.       *)
(* <definition of [[separate]]>=                *)
fun separate (zero, sep) =  (* useful for printing *)
  let fun s []     = zero
        | s [x]    = x
        | s (h::t) = h ^ sep ^ s t
  in  s
end


(*****************************************************************)
(*                                                               *)
(*   TYPES FOR {\TUSCHEME}                                       *)
(*                                                               *)
(*****************************************************************)

(* In our ML code, we represent the abstract    *)
(* syntax of a kind using the datatype          *)
(* [[kind]].                                    *)
(* <types for {\tuscheme}>=                     *)
datatype kind = TYPE                          (* kind of all types *)
              | ARROW of kind list * kind     (* kind of many constructors *)
(* To build types in Typed uScheme, we use not  *)
(* only type variables and the \astforall       *)
(* quantifier but also type constructors and    *)
(* constructor application. In our ML code, we  *)
(* represent the abstract syntax of types using *)
(* the type [[tyex]].                           *)
(* <types for {\tuscheme}>=                     *)
datatype tyex = TYCON  of name                (* type constructor *)
              | TYVAR  of name                (* type variable *)
              | CONAPP of tyex * tyex list    (* type-level application *)
              | FORALL of name list * tyex
(* [[forall]] is                                *)
(* actually a type (e.g., [[array array]] is    *)
(* not a type), it may be best to call such     *)
(* phrases ``type-level expressions.'' When     *)
(* we're not being careful, however, we refer   *)
(* to them simply as ``types.''                 *)

(* <types for {\tuscheme}>=                     *)
(* Printing types and values                    *)
(*                                              *)
(* This code prints types. It might be          *)
(* desirable to print them using a more ML-like *)
(* syntax.                                      *)
(* <printing types for {\tuscheme}>=            *)
val spaceSep = separate("", " ")
fun typeString (TYCON c) = c
  | typeString (TYVAR a) = tyvarString a
  | typeString (CONAPP (TYCON "function", [CONAPP (TYCON "tuple", args), result]
                                                                            )) =
      "(function (" ^ spaceSep (map typeString args) ^ ") " ^ typeString result
                                                                           ^ ")"
  | typeString (CONAPP (tau, [])) = "(" ^ typeString tau ^ ")"
  | typeString (CONAPP (tau, l)) =
      "(" ^ typeString tau ^ " " ^ spaceSep (map typeString l) ^ ")"
  | typeString (FORALL (l, tau)) =
      "(forall (" ^ spaceSep (map tyvarString l) ^ ") " ^ typeString tau ^ ")"
and tyvarString a = "'" ^ a
(* Throughout the interpreter, we print types   *)
(* using the [[typeString]] function from       *)
(* Appendix [->].                               *)
(* <boxed values 1>=                            *)
val _ = op typeString : tyex -> string
(* <types for {\tuscheme}>=                     *)
val inttype  = TYCON "int"
val booltype = TYCON "bool"
val symtype  = TYCON "sym"
val unittype = TYCON "unit"
val tyvarA   = TYVAR "a"
fun tupletype l = CONAPP (TYCON "tuple", l)
fun listtype ty = CONAPP (TYCON "list",[ty])
fun funtype (args, result) = CONAPP (TYCON "function", [tupletype args, result])
(*  Standard type constructors of Typed uScheme *)
(*                                              *)
(* The code above gives representations of and  *)
(* operations on types, but it doesn't make it  *)
(* easy to write types. For example, inside the *)
(* interpreter, the type of [[cons]] has to be  *)
(* written using this enormous phrase:          *)
(* {smallverbatim} FORALL(["a"], CONAPP(TYCON   *)
(* "function", [CONAPP(TYCON "tuple", [TYVAR    *)
(* "a", CONAPP(TYCON "list", [TYVAR "a"])]),    *)
(* CONAPP(TYCON "list", [TYVAR "a"])]))         *)
(* {smallverbatim} To make it easier to define  *)
(* the primitive operations of Typed uScheme,   *)
(* we provide convenience functions.            *)
(* <boxed values 7>=                            *)
val _ = op inttype   : tyex
val _ = op booltype  : tyex
val _ = op symtype   : tyex
val _ = op unittype  : tyex
val _ = op tyvarA    : tyex
val _ = op tupletype : tyex list -> tyex
val _ = op listtype  : tyex -> tyex
val _ = op funtype   : tyex list * tyex -> tyex


(*****************************************************************)
(*                                                               *)
(*   LEXICAL ANALYSIS                                            *)
(*                                                               *)
(*****************************************************************)

(* A value of type [[par]] represents a list of *)
(* tokens with balanced parentheses. In order   *)
(* to issue better error messages, we keep the  *)
(* source-code location of the each             *)
(* parenthesized list. A source names the       *)
(* source and gives a line number.              *)
(* <lexical analysis ((mlscheme))>=             *)
type srcloc  = string * int
fun srclocString (source, line) = source ^ ", line " ^ Int.toString line
(* <lexical analysis ((mlscheme))>=             *)
datatype par = NAME  of string
             | INT   of int
             | SHARP of char  (* #t or #f *)
             | LIST  of srcloc * par list
(* <lexical analysis ((mlscheme))>=             *)

exception Paren of char list * srcloc
   (* raise if found ) when reading input *)
exception PrematureEOF of string (* shows what we were reading ... *)
exception EOF                 (* raised by reader *)
fun read nextline curloc : string -> par list =
  let (* To help with reading, we have functions to   *)
      (* deal with lexical details. We begin by       *)
      (* defining tests for the delimiters of         *)
      (* symbols:                                     *)
      (* <character-fiddling functions>=              *)
      fun isDelim c = List.exists (fn c' => c' = c) [#"(", #")", #" ", #"\t",
                                                                    #"\n", #";"]
      (* [[]] is like the micro-Scheme                *)
      (* [[exists?]]. We needn't write its            *)
      (* implementation because it is in the initial  *)
      (* basis for Standard ML.                       *)

      (* <character-fiddling functions>=              *)
      fun stringReader []     = NONE
        | stringReader (h::t) = SOME (h, t)
      val strton = Int.scan StringCvt.DEC stringReader
      (* We continue with integer conversion;         *)
      (* [[strton s]] returns an integer, plus        *)
      (* leftover characters from [[s]], or [[NONE]]  *)
      (* if no prefix of [[s]] is an integer literal. *)
      (* It uses [[Int.scan]] from the Standard Basis *)
      (* Library.                                     *)
      (* <boxed values 30>=                           *)
      val _ = op strton : char list -> (int * char list) option
      (* We need to be able to skip whitespace and    *)
      (* comments. [[StringCvt.skipWS]] skips white   *)
      (* space; if we encounter a semicolon, we use   *)
      (* [[StringCvt.dropl]] to skip everything up to *)
      (* the next newline, then resume skipping white *)
      (* space.                                       *)
      (* <character-fiddling functions>=              *)
      fun skip s =
        case StringCvt.skipWS stringReader s
          of #";" :: s => skip (StringCvt.dropl (fn c => c <> #"\n")
                                                                 stringReader s)
           | s => s
      (* [[nonwhite]], which                          *)
      (* summons more input lines as needed.          *)

      (* <character-fiddling functions>=              *)
      fun nonwhite s =
        case skip s
          of [] => nonwhite (explode (nextline()))
           | s  => s
      fun readmap f (result, unconsumed) = (f result, unconsumed)
      (* get is called only when s is guaranteed non-blank, non-comment, non-
                                                                        empty *)
      fun get (#"("::s)  = readmap (fn l => LIST l) (list (curloc(), s, []))
        | get (#")"::s)  = raise (Paren (s, curloc()))
        | get (#"'"::s)  = let val loc = curloc()
                               fun quote p = LIST (loc, [NAME "quote", p])
                           in  readmap quote (get (nonwhite s))
                           end
        | get s          = atom (s, [])
      and atom(c::s, r) = if isDelim c then (mkAtom(rev r), c::s) 
                          else atom(s, c::r)
        | atom([],   r) = (mkAtom(rev r), [])
      and mkAtom [#"#", #"f"] = SHARP #"f"
        | mkAtom [#"#", #"t"] = SHARP #"t"
        | mkAtom s = (case strton s of SOME (n, []) => INT n | _ => NAME (
                                                                     implode s))
      and list (loc, s, l) =
        let val next = nonwhite s
        in  let val (a, u) = get next
            in  list(loc, u, a::l)
            end handle (Paren (u, _)) => ((loc, rev l), skip u)
                     | EOF => raise PrematureEOF (implode next)
        end handle EOF => raise PrematureEOF (implode s)
      fun readline [] = []
        | readline s  = 
             let val(p, u) = get s
             in  p :: readline(skip u)
             end
(* The interface to the [[read]] function       *)
(* includes a [[curloc]] parameter, which can   *)
(* be used to keep track of the source-code     *)
(* location of any token. Of the interpreters   *)
(* in this book, only the uSmalltalk interpter  *)
(* (Appendix [->]) takes advantage of this      *)
(* capability.                                  *)
(*                                              *)
(* Function [[read]] is implemented by a group  *)
(* of local lexing functions. Each such         *)
(* function takes a list of unconsumed          *)
(* characters and returns a pair consisting of  *)
(* the [[par]] read plus unconsumed characters. *)
(* The functions are:                           *)
(*                                              *)
(*   [[get]] --- Read the next input            *)
(*   [[atom]] --- Read a name, integer, or      *)
(*   [[#]] literal                              *)
(*   [[list]] --- Read a list of inputs in      *)
(*   parentheses                                *)
(*                                              *)
(* The [[list]] function takes as an additional *)
(* parameter the source-code location of the    *)
(* start of the list. Internal function         *)
(* [[readline]] puts it all together by reading *)
(* as many inputs are as on the line, until all *)
(* the line is consumed.                        *)
(* <boxed values 29>=                           *)
val _ = op read : (unit -> string) -> (unit -> srcloc) -> string -> par list
val _ = op get  : char list                      -> par                 * char
                                                                            list
val _ = op atom : char list * char list          -> par                 * char
                                                                            list
val _ = op list : srcloc * char list * par list  -> (srcloc * par list) * char
                                                                            list
  in  fn s => readline(skip (explode s)) 
  end


(*****************************************************************)
(*                                                               *)
(*   ABSTRACT SYNTAX AND VALUES FOR {\TUSCHEME}                  *)
(*                                                               *)
(*****************************************************************)

(*   Abstract syntax, values, and evaluation of Typed *)
(*   uScheme                                    *)
(*                                              *)
(* As with the concrete syntax, there are only  *)
(* minor differences between the abstract       *)
(* syntax of micro-Scheme and the abstract      *)
(* syntax of Typed uScheme. We add two new      *)
(* expressions, \asttylambda and \asttyapply,   *)
(* which introduce and eliminate polymorphic    *)
(* types. We drop \astletrec. We also require   *)
(* that the parameters of functions have        *)
(* explicit types.                              *)
(* <abstract syntax and values for {\tuscheme}>= *)
datatype       exp = LITERAL  of value
                   | VAR      of name
                   | SET      of name * exp
                   | IFX      of exp * exp * exp
                   | WHILEX   of exp * exp
                   | BEGIN    of exp list
                   | LETX     of let_kind * (name * exp) list * exp
                   | LAMBDA   of lambda_exp
                   | APPLY    of exp * exp list
                   | TYLAMBDA of name list * exp
                   | TYAPPLY  of exp * tyex list
and       let_kind = LET | LETSTAR
and          value = NIL
                   | BOOL      of bool   
                   | NUM       of int
                   | SYM       of name
                   | PAIR      of value * value
                   | CLOSURE   of lambda_value * value ref env
                   | PRIMITIVE of primitive
withtype primitive    = value list -> value (* raises RuntimeError *)
     and lambda_exp   = (name * tyex) list * exp
     and lambda_value = name          list * exp

exception RuntimeError of string
(* The top-level items of Typed uScheme are     *)
(* like those of Typed Impcore, with one        *)
(* addition. So we can define functions that    *)
(* are both recursive and polymorphic, we       *)
(* provide [[VALREC]].                          *)
(* <abstract syntax and values for {\tuscheme}>= *)
datatype toplevel = EXP    of exp
                  | DEFINE of name * tyex * lambda_exp
                  | VAL    of name * exp
                  | VALREC of name * tyex * exp
                  | USE    of name


(*****************************************************************)
(*                                                               *)
(*   VALUES                                                      *)
(*                                                               *)
(*****************************************************************)

(* <values ((mlscheme))>=                       *)



fun embedList []     = NIL
  | embedList (h::t) = PAIR(h, embedList t)
fun embedPredicate f = fn x => BOOL (f x)
fun bool (BOOL b) = b
  | bool _        = true
(*    Embedding and projection                  *)
(*                                              *)
(* An S-expression can represent an integer,    *)
(* Boolean, name, function, list, etc. We may   *)
(* sometimes have an ML boolean, list, or       *)
(* function that we wish to represent as an     *)
(* S-expression, or similarly, an S-expression  *)
(* that we wish to represent as a value of      *)
(* type [[bool]]. Here we define mappings       *)
(* between type [[value]] and some other        *)
(* ML types. Because the set of values          *)
(* representable by an ML value of type         *)
(* [[value]] strictly contains each of the sets *)
(* of values representable by these ML types,   *)
(* these mappings are called embedding and      *)
(* projection. Because the [[value]] type is    *)
(* strictly larger than these ML types, no      *)
(* embedding operation ever fails, but a        *)
(* projection operation might. [This property   *)
(* is a general characteristic of any           *)
(* embedding/projection pair. Mathematical      *)
(* terminology may clarify; an embedding e of   *)
(* S into S' is an injection from S-->S'. The   *)
(* corresponding projection pi_e is a left      *)
(* inverse of the embedding; that is pi_e oe is *)
(* the identity function on S. There is no      *)
(* corresponding guarantee for e opi_e; for     *)
(* example, pi_e may be undefined (_|_) on some *)
(* elements of S', or e(pi_e(x)) may not        *)
(* equal x.] For example, although any ML       *)
(* function of type [[value -> bool]] can be    *)
(* embedded into [[value]] by using the         *)
(* [[PRIMITIVE]] constructor, there are values  *)
(* of type [[value]] that cannot be projected   *)
(* into an ML function of type [[value ->       *)
(* bool]].                                      *)
(*                                              *)
(* Lists and Booleans are straightforward.      *)
(* Function [[embedPredicate]] is not a true    *)
(* embedding; it takes any function returning   *)
(* [[bool]] and returns a corresponding         *)
(* function returning [[value]]. It really      *)
(* embeds the function's result, not the        *)
(* function itself.                             *)
(* <boxed values 20>=                           *)
val _ = op embedList      : value list -> value
val _ = op embedPredicate : ('a -> bool) -> ('a -> value)
val _ = op bool           : value -> bool
(* [[bool]] is the projection                   *)
(* function, mapping micro-Scheme values into   *)
(* ML Booleans.                                 *)

(* <values ((mlscheme))>=                       *)
fun valueString (NIL)    = "()"
  | valueString (BOOL b) = if b then "#t" else "#f"
  | valueString (NUM n)  = String.map (fn #"~" => #"-" | c => c) (Int.toString n
                                                                               )
  | valueString (SYM v)  = v
  | valueString (PAIR(car, cdr))  = 
      let fun tail (PAIR(car, cdr)) = " " ^ valueString car ^ tail cdr
            | tail NIL = ")"
            | tail v = " . " ^ valueString v ^ ")"
      in  "(" ^ valueString car ^ tail cdr
      end
  | valueString (CLOSURE   _) = "<procedure>"
  | valueString (PRIMITIVE _) = "<procedure>"
(*    Printing                                  *)
(*                                              *)
(* We render an S-expression as a string.       *)
(* <boxed values 21>=                           *)
val _ = op valueString : value -> string
(* [[Int.toString]] indicates the               *)
(* [[toString]] function from the standard      *)
(* module [[Int]]. This function, which is part *)
(* of ML's Standard Basis Library, converts an  *)
(* integer to a string. We use another standard *)
(* function, [[String.map]], to change the      *)
(* minus sign from the ML convention ([[ ]]) to *)
(* the Scheme convention ([[-]]).               *)



(*****************************************************************)
(*                                                               *)
(*   TYPE CHECKING FOR {\TUSCHEME}                               *)
(*                                                               *)
(*****************************************************************)

(* <type checking for {\tuscheme}>=             *)
fun kindof(tau, Delta) =
  let (* A type variable is looked up in the          *)
      (* environment. \usetyKindIntroVar As in the    *)
      (* concrete syntax, we distinguish a type       *)
      (* variable from a type constructor by putting  *)
      (* a quote mark in front of its name.           *)
      (* <internal function [[kind]]>=                *)
      fun kind (TYVAR a) =
            (find("'" ^ a, Delta)
             handle NotFound _ => raise TypeError ("unknown type variable '" ^ a
                                                                              ))
      (* A type constructor is also looked up.        *)
      (* \usetyKindIntroCon                           *)
      (* <internal function [[kind]]>=                *)
        | kind (TYCON c) =
            (find(c, Delta)
             handle NotFound _ => raise TypeError ("unknown type constructor " ^
                                                                             c))
      (* The tuple type constructor may be used to    *)
      (* combine any number of types. \usetyKindTuple *)
      (* <internal function [[kind]]>=                *)
        | kind (CONAPP (TYCON "tuple", actuals)) =
            if List.all (fn tau => kind tau = TYPE) actuals then
                TYPE
            else
                raise TypeError "tuple formed from non-types"
      (* [[List.all]]                                 *)
      (* corresponds to micro-Scheme's function       *)
      (* [[all?]].                                    *)

      (* Every type constructor other than [[tuple]]  *)
      (* must be applied in a way that is consistent  *)
      (* with its kind. \usetyKindApp                 *)
      (* <internal function [[kind]]>=                *)
        | kind (CONAPP (tau, actuals)) =
            (case kind tau
               of ARROW (formals, result) =>
                    if formals = map kind actuals then
                        result
                    else
                        raise TypeError ("type constructor " ^ typeString tau ^
                                         " applied to the wrong arguments")
                | TYPE =>
                    raise TypeError ("tried to apply type " ^ typeString tau ^
                                     " as type constructor"))
      (* A quantified type must always have           *)
      (* kind \ktype. \usetyKindAll The quantified    *)
      (* variables \ldotsnalpha may be used in tau,   *)
      (* so we insert them into Delta before checking *)
      (* the kind of tau.                             *)
      (* <internal function [[kind]]>=                *)
        | kind (FORALL (alphas, tau)) =
            let val Delta' =
                  foldl (fn (a, Delta) => bind("'" ^ a, TYPE, Delta)) Delta
                                                                          alphas
            in  case kindof(tau, Delta')
                  of TYPE    => TYPE
                   | ARROW _ => raise TypeError
                                      "quantifed a non-nullary type constructor"
            end
(*  Checking a kind                             *)
(*                                              *)
(* The function [[kindof]] implements the       *)
(* kinding judgment \kindistau\kind, which says *)
(* that given kind environment Delta,           *)
(* type-level expression tau has kind \kind.    *)
(* Given Delta and tau, [[kindof(]]tau[[,       *)
(* ]]Delta[[)]] returns a \kind such that       *)
(* \kindistau\kind, or if no such kind exists,  *)
(* it raises the exception [[TypeError]].       *)
(* <boxed values 2>=                            *)
val _ = op kindof : tyex * kind env -> kind
val _ = op kind   : tyex            -> kind
  in  kind tau
  end
(* <type checking for {\tuscheme}>=             *)
fun asType (ty, Delta) =
  case kindof(ty, Delta)
    of TYPE    => ty
     | ARROW _ => raise TypeError ("used type constructor `" ^ typeString ty ^
                                   "' as a type")
(* A type-level expression used to describe a   *)
(* variable or parameter must have kind         *)
(* [[TYPE]]. The function [[asType]] ensures    *)
(* it.                                          *)
(* <boxed values 3>=                            *)
val _ = op asType : tyex * kind env -> tyex
(* <type checking for {\tuscheme}>=             *)
fun tysubst (tau, varenv) =
  let fun subst (TYVAR a) = (find(a, varenv) handle NotFound _ => TYVAR a)
        | subst (TYCON c) = (TYCON c)
        | subst (CONAPP (tau, taus)) = CONAPP (subst tau, map subst taus)
        | subst (FORALL (alphas, tau)) =
            FORALL (alphas, tysubst(tau, bindList(alphas, map TYVAR alphas,
                                                                       varenv)))
(* The function [[tysubst(]]tau[[, varenv)]]    *)
(* changes type tau by substituting for type    *)
(* variables as specified by the environment    *)
(* [[varenv]]. Environment [[varenv]] does not  *)
(* necessarily substitute for every type        *)
(* variable in tau; some type variables may be  *)
(* left alone. The inner function [[subst]]     *)
(* walks the type, leaving [[varenv]] alone.    *)
(* <boxed values 4>=                            *)
val _ = op tysubst : tyex * tyex env -> tyex
val _ = op subst   : tyex            -> tyex
  in  subst tau
  end
(* <type checking for {\tuscheme}>=             *)
fun instantiate (FORALL (formals, tau), actuals, Delta) =
      (case List.find (fn t => kindof(t, Delta) <> TYPE) actuals
         of SOME t => raise TypeError ("instantiated at type constructor `" ^
                                       typeString t ^ "', which is not a type")
          | NONE =>
              (tysubst(tau, bindList (formals, actuals, emptyEnv))
               handle BindListLength =>
                 raise TypeError
                   "instantiated polymorphic term at wrong number of arguments")
                                                                               )
  | instantiate _ = raise TypeError
                             "tried to instantiate term of non-polymorphic type"
(* Instantiation is a matter of substituting    *)
(* for type variables. Most of the code is      *)
(* error checking. We must instantiate only at  *)
(* type-level expressions of kind type, we must *)
(* instantiate with the right number of types,  *)
(* and we must instantiate only polymorphic     *)
(* types.                                       *)
(* <boxed values 5>=                            *)
val _ = op instantiate : tyex * tyex list * kind env -> tyex
val _ = List.find : ('a -> bool) -> 'a list -> 'a option
(* [[List.find]] takes a                        *)
(* predicate and searches a list for an element *)
(* satisfying that predicate.                   *)

(* <type checking for {\tuscheme}>=             *)
fun eqType (TYCON c, TYCON c') = c = c'
  | eqType (TYVAR a, TYVAR a') = a = a'
  | eqType (CONAPP (tau, taus), CONAPP (tau', taus')) =
      eqType (tau, tau') andalso eqTypes (taus, taus')
  | eqType (FORALL (alphas, tau), FORALL (alphas', tau')) =
      (eqType (tau, tysubst (tau', bindList (alphas', map TYVAR alphas, emptyEnv
                                                                             )))
       handle BindListLength => false)
  | eqType _ = false
and eqTypes (t::taus, t'::taus') = eqType(t, t') andalso eqTypes(taus, taus')
  | eqTypes ([], []) = true
  | eqTypes _ = false
(*  Type equality                               *)
(*                                              *)
(* As in Typed Impcore, many rules of Typed     *)
(* uScheme require that two types be the same.  *)
(* We can't simply use built-in equality to     *)
(* check, however, because when we compare two  *)
(* quantified types, the names of the type      *)
(* variables should be irrelevant. For example, *)
(* the two types \/alpha\alldotlistalpha-->int  *)
(* and \/beta\alldotlistbeta-->int should be    *)
(* considered to be equal. We implement the     *)
(* check by substituting one set of names for   *)
(* the other.                                   *)
(* <boxed values 6>=                            *)
val _ = op eqType  : tyex      * tyex      -> bool
val _ = op eqTypes : tyex list * tyex list -> bool
(* A \xvalrec binding, by contrast, is          *)
(* recursive, and it requires an explicit type. *)
(* The explicit type must be the type of the    *)
(* right-hand side. \tyrule ValRec \ptypeis[{x  *)
(* |->tau}] e tau \topt\xvalrec(x, tau, e)      *)
(* -->Gamma{x |->tau} The bound name x is       *)
(* visible during the elaboration of the        *)
(* right-hand side e, but for safety, x must    *)
(* not be evaluated within e until after e's    *)
(* definition has been stored in x. In          *)
(* micro-Scheme, the way to keep something from *)
(* being evaluated is to protect it under a     *)
(* \xlambda, [In a lazy language like Haskell,  *)
(* a right-hand side is never evaluated until   *)
(* its value is needed, so a definition like    *)
(* [[(val-rec int x x)]] is legal, but          *)
(* evaluating~[[x]] produces an infinite loop   *)
(* (sometimes called a ``black hole.'')] so     *)
(* safety can be guaranteed by making sure      *)
(* x does not appear unprotected.[*]            *)
(* <type checking for {\tuscheme}>=             *)
fun appearsUnprotectedIn (x, e) = 
  let fun evaluatesX (LITERAL n) = false
        | evaluatesX (VAR v) = v = x
        | evaluatesX (SET (_, e)) = evaluatesX e
        | evaluatesX (WHILEX (e1, e2)) = evaluatesX e1 orelse evaluatesX e2
        | evaluatesX (APPLY (f, actuals)) =
            evaluatesX f orelse List.exists evaluatesX actuals
        | evaluatesX (LETX (LETSTAR, [], body)) = evaluatesX body
        | evaluatesX (LETX (LETSTAR, (x', e') :: bs, body)) =
            evaluatesX e' orelse
            (x <> x' andalso evaluatesX (LETX (LETSTAR, bs, body)))
        | evaluatesX (LETX (LET, bs, body)) = 
            List.exists (fn (_, e) => evaluatesX e) bs orelse
            (not (List.exists (fn (x', _) => x' = x) bs) andalso evaluatesX body
                                                                               )
        | evaluatesX (IFX (e1, e2, e3)) =
            evaluatesX e1 orelse evaluatesX e2 orelse evaluatesX e3
        | evaluatesX (BEGIN es) = List.exists evaluatesX es
        | evaluatesX (LAMBDA (formals, body)) = false
        | evaluatesX (TYAPPLY (e, args)) = evaluatesX e
        | evaluatesX (TYLAMBDA (alphas, e)) = evaluatesX e
  in  evaluatesX e
  end
(* [[]] as a                                    *)
(* side condition. For details about exactly    *)
(* what happens during evaluation, see          *)
(* page [->] in Appendix [->].                  *)

(* <type checking for {\tuscheme}>=             *)
exception LeftAsExercise
fun topty _ = raise LeftAsExercise
(*  Type checking                               *)
(*                                              *)
(* This book does not provide a type checker    *)
(* for Typed uScheme; its implementation is     *)
(* left as Exercise [->]. Type checking         *)
(* requires a top-level item, a type            *)
(* environment, and a kind environment. Calling *)
(* topty[[(]]t, Gamma, Delta[[)]] should return *)
(* a pair (Gamma', s), where \toptt -->Gamma'   *)
(* and s is a string that represents the type   *)
(* of the top-level item. [*]                   *)
(* <boxed values 8>=                            *)
val _ = op topty : toplevel * tyex env * kind env -> tyex env * string


(*****************************************************************)
(*                                                               *)
(*   PARSING FOR {\TUSCHEME}                                     *)
(*                                                               *)
(*****************************************************************)

(* <parsing for {\tuscheme}>=                   *)
exception SyntaxError of string
fun syntaxErrorAt (source, line) msg =
  raise SyntaxError (source ^ ", line " ^ Int.toString line ^ ": " ^ msg)
fun nodups nameOf msg loc l =
  let fun dup [] = l
        | dup (h::t) = if List.exists (fn x => nameOf x = nameOf h) t then
                         syntaxErrorAt loc (msg h)
                       else
                         dup t
  in  dup l
  end
(* Parsing                                      *)
(*                                              *)
(* This code checks for duplicate identifiers.  *)
(* Because the parser makes a number of such    *)
(* checks, we define one higher-order function  *)
(* to do them all.                              *)
(* <boxed values 12>=                           *)
val _ = op nodups : ('a -> name) -> ('a -> string) -> srcloc -> 'a list -> 'a
                                                                            list
(* We check either single identifiers or pairs. *)
(*                                              *)
(* <parsing for {\tuscheme}>=                   *)
val id  = fn n => n
val fst = fn (n, _) => n
fun defineDups loc f =
  nodups fst (fn x => "formal parameter " ^ fst x ^
                      " appears twice in definition of function " ^ f) loc
fun lambdaDups loc =
  nodups fst (fn x => "formal parameter " ^ fst x ^ " appears twice in lambda")
                                                                             loc
fun tylambdaDups loc =
  nodups id (fn x => "formal parameter " ^ x ^ " appears twice in type-lambda")
                                                                             loc
fun forallDups loc =
  nodups id (fn x => "quantified variable " ^ x ^ " appears twice in forall")
                                                                             loc
fun letDups loc (LETSTAR, bindings) = bindings
  | letDups loc (LET,     bindings) =
      nodups fst (fn x => "name " ^ fst x ^ " appears twice in let") loc
                                                                        bindings
(* We parse concrete syntax into abstract       *)
(* syntax. This operation is relatively         *)
(* uninteresting. The auxiliary functions may   *)
(* help clarify the logic of parsing.           *)
(*                                              *)
(* When parsing a type, we reject anything that *)
(* looks like an expression.                    *)
(* <parsing for {\tuscheme}>=                   *)
fun parseType _ (LIST (loc, NAME "forall"   :: args)) = parseForall   loc args
  | parseType _ (LIST (loc, NAME "function" :: args)) = parseFunction loc args
  | parseType _ (LIST (loc, [NAME "quote", NAME a]))  = TYVAR a
  | parseType _ (LIST (loc, NAME (bad as "quote")  :: args)) =
      syntaxErrorAt loc ("looking for type but found `" ^ bad ^ "'")
  | parseType _ (LIST (loc, NAME (bad as "if")     :: args)) = 
      syntaxErrorAt loc ("looking for type but found `" ^ bad ^ "'")
  | parseType _ (LIST (loc, NAME (bad as "while")  :: args)) = 
      syntaxErrorAt loc ("looking for type but found `" ^ bad ^ "'")
  | parseType _ (LIST (loc, NAME (bad as "set")    :: args)) = 
      syntaxErrorAt loc ("looking for type but found `" ^ bad ^ "'")
  | parseType _ (LIST (loc, NAME (bad as "begin")  :: args)) = 
      syntaxErrorAt loc ("looking for type but found `" ^ bad ^ "'")
  | parseType _ (LIST (loc, NAME (bad as "lambda") :: args)) = 
      syntaxErrorAt loc ("looking for type but found `" ^ bad ^ "'")
  | parseType _ (LIST (loc, NAME (bad as "type-lambda") :: args)) = 
      syntaxErrorAt loc ("looking for type but found `" ^ bad ^ "'")
  | parseType _ (LIST (loc, NAME (bad as "let")    :: args)) = 
      syntaxErrorAt loc ("looking for type but found `" ^ bad ^ "'")
  | parseType _ (LIST (loc, NAME (bad as "let*")   :: args)) = 
      syntaxErrorAt loc ("looking for type but found `" ^ bad ^ "'")
  | parseType _ (LIST (loc, NAME (bad as "@")      :: args)) = 
      syntaxErrorAt loc ("looking for type but found `" ^ bad ^ "'")
  | parseType _ (LIST (loc, tau :: taus)) =
      CONAPP (parseType loc tau, map (parseType loc) taus)
  | parseType _ (LIST (loc, [])) = syntaxErrorAt loc "empty type ()"
  | parseType _ (NAME a) = TYCON a
  | parseType loc (INT n)   = syntaxErrorAt loc "expected type; found integer"
  | parseType loc (SHARP c) = syntaxErrorAt loc
                                          "expected type; found Boolean literal"
and parseForall loc [LIST (floc, formals), tau] =
      FORALL (forallDups floc (map (untyvar floc) formals), parseType loc tau)
  | parseForall loc _ = syntaxErrorAt loc "expected (forall (names) type)"
and parseFunction loc [LIST (aloc, args), tau] =
      funtype (map (parseType aloc) args, parseType loc tau)
  | parseFunction loc _ = syntaxErrorAt loc "expected (function (types) type)"
and untyvar _ (LIST (_, [NAME "quote", NAME n])) = n
  | untyvar loc _ = syntaxErrorAt loc "type parameters must be type variables"
(* When parsing an expression, we reject        *)
(* anything that looks like a type.             *)
(* <parsing for {\tuscheme}>=                   *)
fun parse (NAME s)  = VAR s
  | parse (INT n)   = LITERAL (NUM n)
  | parse (SHARP c) = LITERAL (BOOL (c = #"t"))
  | parse (LIST (loc, NAME "quote"  :: args)) = parseQuote loc args
  | parse (LIST (loc, NAME "if"     :: args)) = parseIf    loc args
  | parse (LIST (loc, NAME "while"  :: args)) = parseWhile loc args
  | parse (LIST (loc, NAME "set"    :: args)) = parseSet   loc args
  | parse (LIST (loc, NAME "begin"  :: args)) = BEGIN (map parse args)
  | parse (LIST (loc, NAME "lambda" :: args)) = parseLambda loc args
  | parse (LIST (loc, NAME "type-lambda" :: args)) = parseTypeLambda loc args
  | parse (LIST (loc, NAME "let"    :: args)) = parseLet loc LET     args
  | parse (LIST (loc, NAME "let*"   :: args)) = parseLet loc LETSTAR args
  | parse (LIST (loc, NAME "@"      :: args)) = parseTyapp loc args
  | parse (LIST (loc, NAME "forall" :: _   )) =
       syntaxErrorAt loc "looking for expression but found `forall'"
  | parse (LIST (loc, NAME "function" :: _   )) =
       syntaxErrorAt loc "looking for expression but found `function"
  | parse (LIST (loc, h::t))  = APPLY (parse h, map parse t)
  | parse (LIST (loc, [])) = syntaxErrorAt loc "empty application ()"
(* <parsing for {\tuscheme}>=                   *)
and parseQuote loc [e] = LITERAL (quote loc e)
  | parseQuote loc _ = syntaxErrorAt loc "expected (quote e)"
and parseIf _ [e1, e2, e3] = IFX (parse e1, parse e2, parse e3)
  | parseIf loc _ = syntaxErrorAt loc "expected (if e1 e2 e3)"
and parseWhile _ [e1, e2] = WHILEX (parse e1, parse e2)
  | parseWhile loc _ = syntaxErrorAt loc "expected (while e1 e2)"
and parseSet _ [NAME v, e] = SET (v, parse e)
  | parseSet loc _ = syntaxErrorAt loc "expected (set v e)"
and parseTyapp loc (exp :: args) = TYAPPLY (parse exp, map (parseType loc) args)
  | parseTyapp loc [] = syntaxErrorAt loc "empty type application (@)"
and parseTypeLambda _ [LIST (loc, formals), body] =
      TYLAMBDA (tylambdaDups loc (map (untyvar loc) formals), parse body)
  | parseTypeLambda loc _ =
      syntaxErrorAt loc "expected (lambda (names) body)"
and parseLambda _ [LIST (loc, formals), body] =
      LAMBDA (lambdaDups loc (map (unformal loc) formals), parse body)
  | parseLambda loc _ =
      syntaxErrorAt loc "expected (lambda (formals) body)"
and parseLet _ kind [LIST (loc, bindings), body] = 
      LETX(kind, letDups loc (kind, map (unBinding loc) bindings), parse body)
  | parseLet loc _ _ = syntaxErrorAt loc "expected (let (bindings) body)"
and unformal _ (LIST (loc, [tau, NAME n])) = (n, parseType loc tau)
  | unformal _ (LIST (loc,_)) = syntaxErrorAt loc
                                          "formal parameter must be (type name)"
  | unformal loc _ = syntaxErrorAt loc "formal parameter must be (type name)"
and unBinding _ (LIST (_, [NAME n, e])) = (n, parse e)
  | unBinding _ (LIST (loc, _)) = syntaxErrorAt loc "expected (v e) in bindings"
  | unBinding loc _ = syntaxErrorAt loc "expected (v e) in bindings"
and quote _ (INT   n) = NUM n
  | quote _ (SHARP c) = BOOL (c = #"t")
  | quote loc (NAME  ".") =
      syntaxErrorAt loc
                      "this interpreter cannot handle . in quoted S-expressions"
  | quote _ (NAME  n) = SYM n
  | quote _ (LIST (loc, l)) = embedList (map (quote loc) l)
(* <parsing for {\tuscheme}>=                   *)
fun toplevel (LIST (loc, NAME "define"  :: args)) = parseDefine loc args
  | toplevel (LIST (loc, NAME "val"     :: args)) = parseVal    loc args
  | toplevel (LIST (loc, NAME "val-rec" :: args)) = parseValRec loc args
  | toplevel (LIST (loc, NAME "global"  :: args)) = parseVal    loc args (*
                                                                       legacy *)
  | toplevel (LIST (loc, NAME "use"     :: args)) = parseUse    loc args
  | toplevel e = EXP (parse e)
and parseDefine loc [ty, NAME name, LIST (floc, formals), body] =
      DEFINE (name, parseType loc ty,
              (defineDups floc name (map (unformal floc) formals), parse body))
  | parseDefine loc _ = syntaxErrorAt loc "expected (define type f (args) body)"
and parseVal _ [NAME x, e] = VAL (x, parse e)
  | parseVal loc _ = syntaxErrorAt loc "expected (val x e)"
and parseValRec loc [tau, NAME x, e] = VALREC (x, parseType loc tau, parse e)
  | parseValRec loc _ = syntaxErrorAt loc "expected (val-rec type x e)"
and parseUse _ [NAME x] = USE x
  | parseUse loc _ = syntaxErrorAt loc "expected (use filename)"


(*****************************************************************)
(*                                                               *)
(*   READERS                                                     *)
(*                                                               *)
(*****************************************************************)

(* Readers                                      *)
(*                                              *)
(*  Line readers                                *)
(*                                              *)
(* Our readers are structured slightly          *)
(* differently from those in C. A reader is a   *)
(* bundle containing three related functions:   *)
(* [[firstline]], which returns the first line  *)
(* read; [[nextline]], which returns all        *)
(* subsequent lines; and [[curloc]], which      *)
(* returns a pair identifying the source        *)
(* location of the most recently returned line. *)
(* Both [[firstline]] and [[nextline]] can      *)
(* raise [[EOF]], (We distinguish [[firstline]] *)
(* from [[nextline]] because in an interactive  *)
(* reader, they may use different prompts.)     *)
(* <readers>=                                   *)
type reader = { firstline : unit -> string
              , nextline  : unit -> string
              , curloc    : unit -> string * int
              }
(* <readers>=                                   *)
fun promptReader (firstPrompt, nextPrompt, reader) =
  let val { firstline = first, nextline = next, curloc = cur } = reader
      fun promptIn get prompt () = ( TextIO.output(TextIO.stdOut, prompt)
                                   ; TextIO.flushOut(TextIO.stdOut)
                                   ; get ()
                                   )
  in  { firstline = promptIn first firstPrompt
      , nextline  = promptIn next  nextPrompt
      , curloc    = cur
      }
  end
(* Given a silent reader, we can make it issue  *)
(* a prompt every time it reads a new line. We  *)
(* use two prompts: one for the first line of   *)
(* an expression and one for later lines.       *)
(* <boxed values 31>=                           *)
val _ = op promptReader : string * string * reader -> reader
(* <readers>=                                   *)
fun filereader (filename, fd)=
  let val lineNum = ref 0
      fun get () =
        let val line = TextIO.inputLine fd
        in  if size line = 0 then raise EOF else line before lineNum := !lineNum
                                                                             + 1
        end
  in  { firstline = get, nextline = get, curloc = (fn () => (filename, !lineNum)
                                                                             ) }
  end

fun stringsreader (what, l) = 
  let val buffer  = ref l
      val lineNum = ref 0
      fun get ()  = case !buffer
                      of [] => raise EOF
                       | h :: t => h before buffer := t before lineNum := !
                                                                     lineNum + 1
  in  { firstline = get, nextline = get, curloc = (fn () => (what, !lineNum)) }
  end              
(* We can read from files or lists of strings.  *)
(* File I/O is provide by the standard module   *)
(* [[TextIO]].                                  *)
(* <boxed values 32>=                           *)
val _ = op filereader    : string * TextIO.instream -> reader
val _ = op stringsreader : string * string list     -> reader
(* [[]], a private mutable cell. When all       *)
(* the strings in the buffer are used up, there *)
(* aren't any more.                             *)

(*  Top-level readers                           *)
(*                                              *)
(* A top-level reader is a buffer of top-level  *)
(* items together with a reader. [*]            *)
(* <readers>=                                   *)
type topreader = { buffer : toplevel list ref, reader : reader }
fun readtop (r as { buffer, reader = { nextline, firstline, curloc } }) =
  case !buffer
    of h::t => h before buffer := t
     | []   => ( buffer := map toplevel (read nextline curloc (firstline()))
               ; readtop r
               )
(* We decide whether to prompt when building a  *)
(* top-level reader.                            *)
(* <readers>=                                   *)
fun topreader (reader, prompt) = 
  let val reader = if prompt then promptReader ("-> ", "   ", reader) else
                                                                          reader
  in  { buffer = ref [], reader = reader }
  end
(* <readers>=                                   *)
fun echoTag f x =
  let val line = f x
      val _ = if (String.substring (line, 0, 2) = ";#" handle _ => false) then
                print line
              else
                ()
  in  line
  end

fun echoReader { nextline=n, firstline=f, curloc=c } =
     { nextline=echoTag n, firstline=echoTag f, curloc=c } 

fun echoTop { buffer = b, reader = r } = { buffer = b, reader = echoReader r }
val topreader = fn args => echoTop (topreader args)
(* As in the C code, we want to print out any   *)
(* line read if it begins with the special      *)
(* string [[;#]]. This string is a formal       *)
(* comment that helps us test the chunks marked *)
(* [[]]. In our ML code, we                     *)
(* arrange it using the higher-order function   *)
(* [[echoReader]].                              *)
(* <boxed values 33>=                           *)
val _ = op echoReader : reader -> reader 


(*****************************************************************)
(*                                                               *)
(*   IMPLEMENTATION OF [[USE]]                                   *)
(*                                                               *)
(*****************************************************************)

(* The implementation of [[use]] is             *)
(* parameterized by [[readEvalPrint]] so we can *)
(* share it with other interpreters. Function   *)
(* [[use]] creates a top-level reader that does *)
(* not prompt, but it uses [[writeln]] to be    *)
(* sure that responses are printed.             *)
(* <implementation of [[use]]>=                 *)
fun use readEvalPrint filename rho =
      let val fd = TextIO.openIn filename
          val treader = topreader (filereader (filename, fd), false)
          fun writeln s = app print [s, "\n"]
      in  readEvalPrint (treader, writeln, writeln) rho
          before TextIO.closeIn fd
      end 
(* [[topreader]] is relegated                   *)
(* to chunk [->] in Appendix [->].              *)



(*****************************************************************)
(*                                                               *)
(*   CHECKING AND EVALUATION FOR {\TUSCHEME}                     *)
(*                                                               *)
(*****************************************************************)

(* <checking and evaluation for {\tuscheme}>=   *)
exception BugInTypeChecking of string
(* <evaluation for {\tuscheme}>=                *)
fun eval(e, rho) =
  let fun ev(LITERAL n) = n
        (* Most of the evaluator for Typed uScheme is   *)
        (* just like the evaluator for micro-Scheme in  *)
        (* Chapter [->]. The code for the two new cases *)
        (* acts as if [[TYAPPLY]] and [[TYLAMBDA]]      *)
        (* aren't there.                                *)
        (* <alternatives for [[ev]] for [[TYAPPLY]] and [[TYLAMBDA]]>= *)
        | ev(TYAPPLY  (e, _)) = ev e
        | ev(TYLAMBDA (_, e)) = ev e
        (* Code for variables is just as in             *)
        (* Chapter [->].                                *)
        (* <more alternatives for [[ev]] for {\tuscheme}>= *)
        | ev(VAR v) = !(find(v, rho))
        | ev(SET (n, e)) = 
            let val v = ev e
            in  find (n, rho) := v;
                v
            end
        (* Code for control flow is just as in          *)
        (* Chapter [->].                                *)
        (* <more alternatives for [[ev]] for {\tuscheme}>= *)
        | ev(IFX (e1, e2, e3)) = ev (if bool (ev e1) then e2 else e3)
        | ev(WHILEX (guard, body)) = 
            if bool (ev guard) then 
              (ev body; ev(WHILEX (guard, body)))
            else
              NIL
        | ev(BEGIN es) =
            let fun b(e::es, lastval) = b(es, ev e)
                  | b(   [], lastval) = lastval
            in  b(es, BOOL false)
            end
        (* Code for a [[lambda]] has to remove the      *)
        (* types from the abstract syntax.              *)
        (* <more alternatives for [[ev]] for {\tuscheme}>= *)
        | ev(LAMBDA (args, body)) = CLOSURE((map (fn (n, ty) => n) args, body),
                                                                            rho)
        (* <more alternatives for [[ev]] for {\tuscheme}>= *)
        | ev(APPLY (f, args))  = 
               (case ev f
                  of PRIMITIVE prim => prim (map ev args)
                   | CLOSURE clo => (* <apply closure [[clo]] to [[args]] ((
                                                                 mlscheme))>= *)
                                    let val ((formals, body), savedrho) = clo
                                        val actuals = map ev args
                                    in  eval(body, bindList(formals, map ref
                                                             actuals, savedrho))
                                        handle BindListLength => 
                                            raise RuntimeError (
                                      "Wrong number of arguments to closure; " ^
                                                                "expected (" ^
                                                separate("", " ") formals ^ ")")
                                    end
                   | v => raise BugInTypeChecking "applied non-function"
               )
        (* Code for the [[LETX]] family is as in        *)
        (* Chapter [->].                                *)
        (* <more alternatives for [[ev]] for {\tuscheme}>= *)
        | ev(LETX (LET, bs, body)) =
            let val (names, values) = ListPair.unzip bs
            in  eval (body, bindList(names, map (ref o ev) values, rho))
            end
        | ev(LETX (LETSTAR, bs, body)) =
            let fun step ((n, e), rho) = bind(n, ref (eval(e, rho)), rho)
            in  eval (body, foldl step rho bs)
            end
(* Evaluation                                   *)
(*                                              *)
(* The implementation of the evaluator is       *)
(* almost identical to the implementation in    *)
(* Chapter [->]. There are only two significant *)
(* differences: we have to deal with the        *)
(* mismatch in representations between the      *)
(* abstract syntax [[LAMBDA]] and the value     *)
(* [[CLOSURE]], and we have to write cases for  *)
(* the [[TYAPPLY]] and [[TYLAMBDA]]             *)
(* expressions. Another difference is that many *)
(* potential run-time errors should be          *)
(* impossible because the relevant code would   *)
(* be rejected by the type checker. If one of   *)
(* those errors occurs anyway, we raise the     *)
(* exception [[BugInTypeChecking]], not         *)
(* [[RuntimeError]].                            *)
(* <boxed values 13>=                           *)
val _ = op eval : exp * value ref env -> value
val _ = op ev   : exp                 -> value
  in  ev e
  end
(* <evaluation for {\tuscheme}>=                *)
fun topeval (t, rho) =
  case t
    of USE filename => raise RuntimeError
                                       "internal error -- `use' reached topeval"
     | EXP e        => (rho, valueString (eval(e, rho)))
     | DEFINE (name, tau, lambda) => topeval (VALREC (name, tau, LAMBDA lambda),
                                                                            rho)
     | VAL    (name, e)      =>
          let val v   = eval(e, rho)
              val rho = bind(name, ref v, rho)
          in  (rho, showVal name v)
          end
     | VALREC (name, tau, e) => 
          let val rho = bind(name, ref NIL, rho)
              val v   = eval(e, rho)
          in  find(name, rho) := v;
              (rho, showVal name v)
          end
(* [[VALREC]] case, the                         *)
(* interpreter evaluates [[e]] while [[name]]   *)
(* is still bound to [[NIL]]--that is, before   *)
(* the assignment to [[find(name, rho)]].       *)
(* Therefore, as described on page [->],        *)
(* evaluating [[e]] must not evaluate           *)
(* [[name]]--because the mutable cell for       *)
(* [[name]] does not yet contain its correct    *)
(* value.[*]                                    *)

(* Both [[VAL]] and [[VALREC]] show names as    *)
(* follows:                                     *)
(* <evaluation for {\tuscheme}>=                *)
and showVal name v =
      case v
        of CLOSURE   _ => name
         | PRIMITIVE _ => name
         | _ => valueString v
(* Evaluating a top-level item can produce a    *)
(* new environment. The function [[topeval]]    *)
(* also returns a string which, if nonempty,    *)
(* should be printed to show the value of the   *)
(* item. Type soundness requires a change in    *)
(* the evaluation rule for [[VAL]]; as          *)
(* described in Exercise [->] in Chapter [->],  *)
(* [[VAL]] must always create a new binding.    *)
(* <boxed values 14>=                           *)
val _ = op topeval : toplevel * value ref env -> value ref env * string
(* <evaluation for {\tuscheme}>=                *)
fun binaryOp f = (fn [a, b] => f(a, b) | _ => raise BugInTypeChecking "arity 2")
fun unaryOp  f = (fn [a]    => f a     | _ => raise BugInTypeChecking "arity 1")
(* Primitives of Typed uScheme                  *)
(*                                              *)
(* Here are the primitives. As in Chapter [->], *)
(* all are either binary or unary operators.    *)
(* Type checking should guarantee that          *)
(* operators are used with the correct arity.   *)
(* <boxed values 15>=                           *)
val _ = op unaryOp  : (value         -> value) -> (value list -> value)
val _ = op binaryOp : (value * value -> value) -> (value list -> value)
(* <evaluation for {\tuscheme}>=                *)
fun arithOp f =
      binaryOp (fn (NUM n1, NUM n2) => NUM (f(n1, n2)) 
                 | _ => raise BugInTypeChecking "arithmetic on non-numbers")
val arithtype = funtype ([inttype, inttype], inttype)
(* Arithmetic primitives expect and return      *)
(* integers.                                    *)
(* <boxed values 16>=                           *)
val _ = op arithOp   : (int * int -> int) -> (value list -> value)
val _ = op arithtype : tyex
(* [[]]                                         *)
(* cons up all the primitives into one giant    *)
(* list, and we use that list to build the      *)
(* initial environment for the read-eval-print  *)
(* loop. The big difference is that in Typed    *)
(* uScheme, each primitive has a type as well   *)
(* as a value.                                  *)

(* <evaluation for {\tuscheme}>=                *)
fun embedPredicate f args = BOOL (f args)
fun comparison f = binaryOp (embedPredicate f)
fun intcompare f = 
      comparison (fn (NUM n1, NUM n2) => f(n1, n2)
                   | _ => raise BugInTypeChecking "comparing non-numbers")
val comptype = funtype ([inttype, inttype], booltype)
(* Comparisons take two arguments. Most         *)
(* comparisons (except for equality) apply only *)
(* to integers.                                 *)
(* <boxed values 17>=                           *)
val _ = op comparison : (value * value -> bool) -> (value list -> value)
val _ = op intcompare : (int   * int   -> bool) -> (value list -> value)
val _ = op comptype   : tyex
type env_bundle = kind env * tyex env * value ref env
fun topCheckEval (t, envs as (delta, gamma, rho), echo) =
  case t
    of USE filename => use readCheckEvalPrint filename envs
     | _ =>
        let val (gamma, tystring)  = topty   (t, gamma, delta)
            val (rho,   valstring) = topeval (t, rho)
            val _ = if size valstring > 0 then echo (valstring ^ " : " ^
                                                                       tystring)
                    else ()
(*  Top-level processing                        *)
(*                                              *)
(* As in Typed Impcore, we process a top-level  *)
(* item by first elaborating it (which includes *)
(* running the type checker), then evaluating   *)
(* it. The elaborator produces a string that    *)
(* represents a type, and the evaluator         *)
(* produces a string that either is empty or    *)
(* represents a value. If the value string is   *)
(* nonempty, we print both strings.             *)
(* <boxed values 9>=                            *)
val _ = op topCheckEval : toplevel * env_bundle * (string->unit) -> env_bundle
        in  (delta, gamma, rho)
        end
(* <checking and evaluation for {\tuscheme}>=   *)
and readCheckEvalPrint (reader, echo, errmsg) envs =
  let fun loop envs =
        let fun continue msg = (errmsg msg; loop envs)
            fun finish() = envs
        in  loop (topCheckEval (readtop reader, envs, echo))
            handle EOF => finish()
                (* <more read-eval-print handlers>=             *)
                | TypeError         msg => continue ("type error: " ^ msg)
                | BugInTypeChecking msg => continue ("bug in type checking: " ^
                                                                            msg)
                (* The handlers in the first group deal with    *)
                (* problems that arise during I/O, lexical      *)
                (* analysis, and parsing.                       *)
                (* <more read-eval-print handlers>=             *)
                | IO.Io {name, ...} => continue ("I/O error: " ^ name)
                | Paren (_, loc)    => continue ("syntax error: extra ) in " ^
                                                               srclocString loc)
                | PrematureEOF s    => (errmsg ("Missing )? Got EOF reading (" ^
                                                                   s); finish())
                | SyntaxError msg   => continue ("syntax error: " ^ msg)
                (* [[IO.Io]] is part of the                     *)
                (* Standard Basis Library.                      *)

                (* The handlers in the second group deal with   *)
                (* problems that arise during evaluation. They  *)
                (* always continue with the next top-level      *)
                (* item.                                        *)
                (* <more read-eval-print handlers>=             *)
                | Div               => continue "Division by zero"
                | Overflow          => continue "Arithmetic overflow"
                | RuntimeError msg  => continue ("run-time error: " ^ msg)
                | NotFound n        => continue ("variable " ^ n ^ " not found")
                (* [[Div]] and [[Overflow]] are                 *)
                (* predefined.                                  *)

        end
  in  loop envs
  end
(*  The read-eval-print loop                    *)
(*                                              *)
(* The read-eval-print loop is exactly as for   *)
(* Typed Impcore.                               *)
(* <boxed values 10>=                           *)
val _ = op readCheckEvalPrint : topreader * (string->unit) * (string->unit) ->
                                                        env_bundle -> env_bundle


(*****************************************************************)
(*                                                               *)
(*   INITIALIZATION FOR {\TUSCHEME}                              *)
(*                                                               *)
(*****************************************************************)

(* <initialization for {\tuscheme}>=            *)
val initialEnvs =
  let fun addPrim ((name, prim, funty), (tfuns, vfuns)) = 
        ( bind (name, funty, tfuns)
        , bind (name, ref (PRIMITIVE prim), vfuns)
        )
      val (tfuns, vfuns)  = foldl addPrim (emptyEnv, emptyEnv)
                            ((* <primops for {\tuscheme}\ [[::]]>=           *)
                             ("+", arithOp op +,   arithtype) :: 
                             ("-", arithOp op -,   arithtype) :: 
                             ("*", arithOp op *,   arithtype) :: 
                             ("/", arithOp op div, arithtype) ::
                             (* <primops for {\tuscheme}\ [[::]]>=           *)
                             ("<", intcompare op <, comptype) :: 
                             (">", intcompare op >, comptype) ::
                             ("=", comparison (fn (NIL,     NIL    ) => true
                                                | (NUM  n1, NUM  n2) => n1 = n2
                                                | (SYM  v1, SYM  v2) => v1 = v2
                                                | (BOOL b1, BOOL b2) => b1 = b2
                                                |  _                 => false)
                                 , FORALL (["a"], funtype ([tyvarA, tyvarA],
                                                                  booltype))) ::
                             (* The list primitives have polymorphic types.  *)
                             (*                                              *)
                             (* <primops for {\tuscheme}\ [[::]]>=           *)
                             ("null?", unaryOp (embedPredicate (fn (NIL   ) =>
                                                             true | _ => false))
                                 , FORALL (["a"], funtype ([listtype tyvarA],
                                                                  booltype))) ::
                             ("cons", binaryOp (fn (a, b) => PAIR(a, b))
                                 , FORALL (["a"], funtype ([tyvarA, listtype
                                                  tyvarA], listtype tyvarA))) ::
                             ("car",  unaryOp  (fn (PAIR(car, _)) => car 
                                                 | v => raise RuntimeError
                                                                (
                                    "car applied to non-list " ^ valueString v))
                                 , FORALL (["a"], funtype ([listtype tyvarA],
                                                                    tyvarA))) ::
                             ("cdr",  unaryOp  (fn (PAIR(_, cdr)) => cdr 
                                                 | v => raise RuntimeError
                                                                (
                                    "cdr applied to non-list " ^ valueString v))
                                 , FORALL (["a"], funtype ([listtype tyvarA],
                                                           listtype tyvarA))) ::
                             (* The [[print]] primitive also has a           *)
                             (* polymorphic type.                            *)
                             (* <primops for {\tuscheme}\ [[::]]>=           *)
                             ("print", unaryOp (fn x => (print (valueString x^
                                                                     "\n"); x)),
                                 FORALL (["a"], funtype ([tyvarA], unittype)))
                                                                         :: nil)
      fun addKind ((name, kind), kinds) = bind(name, kind, kinds)
      val kinds   = foldl addKind emptyEnv
                    ((* How do we show which type constructors have  *)
                     (* which kinds? In Typed Impcore, we use type   *)
                     (* rules to keep track of this information. For *)
                     (* example, the \rulenameBaseTypes rule shows   *)
                     (* that [[int]] has kind \ktype, and the        *)
                     (* \rulenameArrayFormation rule shows that      *)
                     (* [[array]] has kind \ktype\karrow\ktype. But  *)
                     (* one of the problems we are trying to solve   *)
                     (* is that type rules are not easily            *)
                     (* extensible. In Typed uScheme, type           *)
                     (* constructors are not built in. Instead, we   *)
                     (* keep track of the kind of each constructor   *)
                     (* by using a kind environment. Here is an      *)
                     (* example environment that shows common type   *)
                     (* constructors and their kinds. For clarity,   *)
                     (* we write the binding in a kind environment   *)
                     (* using the :: symbol instead of |->.          *)
                     (*                                              *)
                     (* Delta_0 = { int:: \ktype, bool:: \ktype,     *)
                     (* unit:: \ktype, pair ::                       *)
                     (* \ktype*\ktype\karrow\ktype,                  *)
                     (* sum :: \ktype*\ktype\karrow\ktype, --> ::    *)
                     (* \ktype*\ktype\karrow\ktype, array ::         *)
                     (* \ktype\karrow\ktype, list ::                 *)
                     (* \ktype\karrow\ktype }                        *)
                     (*                                              *)
                     (* All by itself, this environment enables us   *)
                     (* to see how both [[int]] and [[array]] may be *)
                     (* used. Even better, we can add new type       *)
                     (* constructors to a language just by adding    *)
                     (* them to the environment. Here are the type   *)
                     (* constructors that are built into Typed       *)
                     (* uScheme, except those that are left as       *)
                     (* Exercises.                                   *)
                     (* <primitive type constructors for {\tuscheme}\ [[::]]>=
                                                                              *)
                     ("int",      TYPE) ::
                     ("bool",     TYPE) ::
                     ("sym",      TYPE) ::
                     ("unit",     TYPE) ::
                     ("list",     ARROW ([TYPE],       TYPE)) ::
                     ("function", ARROW ([TYPE, TYPE], TYPE)) :: nil)
      val envs    = (kinds, tfuns, vfuns)
      val basis   = (* <ML representation of initial basis>=        *)
                      [ "; tuscheme.nw 573"
                      ,
   "(val list1 (type-lambda ('a) (lambda (('a x)) ((@ cons 'a) x (@ '() 'a)))))"
                      , "(val list2 (type-lambda ('a) (lambda (('a x) ('a y))"
                      ,
            "                               ((@ cons 'a) x ((@ list1 'a) y)))))"
                      ,
                   "(val list3 (type-lambda ('a) (lambda (('a x) ('a y) ('a z))"
                      ,
          "                               ((@ cons 'a) x ((@ list2 'a) y z)))))"
                      , "; tuscheme.nw 593"
                      , "(val o (type-lambda ('a 'b 'c) "
                      , "  (lambda (((function ('b) 'c) f)"
                      , "           ((function ('a) 'b) g))"
                      , "     (lambda (('a x)) (f (g x))))))"
                      , ""
                      , "(val curry (type-lambda ('a 'b 'c)"
                      , "   (lambda (((function ('a 'b) 'c) f)) "
                      , "      (lambda (('a x)) (lambda (('b y)) (f x y))))))"
                      , ""
                      , "(val uncurry (type-lambda ('a 'b 'c)"
                      , "   (lambda (((function ('a) (function ('b) 'c)) f))"
                      , "      (lambda (('a x) ('b y)) ((f x) y)))))"
                      , "; tuscheme.nw 628"
                      , "(val-rec"
                      ,
                  "  (forall ('a) (function ((list 'a)) int))  ; type of length"
                      , "  length                                    ; name"
                      ,
    "  (type-lambda ('a)                         ; value : polymorphic function"
                      , "     (lambda (((list 'a) l))"
                      , "       (if ((@ null? 'a) l) 0"
                      , "          (+ 1 ((@ length 'a) ((@ cdr 'a) l)))))))"
                      , "; tuscheme.nw 639"
                      ,
      "(val-rec (forall ('a) (function ((list 'a) (list 'a)) (list 'a))) revapp"
                      , "  (type-lambda ('a)"
                      , "     (lambda (((list 'a) l1)  ((list 'a) l2))"
                      , "        (if ((@ null? 'a) l1)"
                      , "        l2"
                      ,
  "        ((@ revapp 'a) ((@ cdr 'a) l1) ((@ cons 'a) ((@ car 'a) l1) l2))))))"
                      , "; tuschemea.nw 387"
                      , "(val caar"
                      , "   (type-lambda ('a)"
                      , "      (lambda (((list (list 'a)) l))"
                      , "          ((@ car 'a) ((@ car (list 'a)) l)))))"
                      , "(val cadr"
                      , "   (type-lambda ('a)"
                      , "      (lambda (((list (list 'a)) l))"
                      , "          ((@ car (list 'a)) ((@ cdr (list 'a)) l)))))"
                      , "; tuschemea.nw 398"
                      , "(define bool and ((bool b) (bool c)) (if b  c  b))"
                      , "(define bool or  ((bool b) (bool c)) (if b  b  c))"
                      , "(define bool not ((bool b))          (if b #f #t))"
                      , "; tuschemea.nw 404"
                      ,
      "(val-rec (forall ('a) (function ((list 'a) (list 'a)) (list 'a))) append"
                      , "  (type-lambda ('a)"
                      , "     (lambda (((list 'a) l1)  ((list 'a) l2))"
                      , "       (if ((@ null? 'a) l1)"
                      , "         l2"
                      ,
 "         ((@ cons 'a) ((@ car 'a) l1) ((@ append 'a) ((@ cdr 'a) l1) l2))))))"
                      , "; tuschemea.nw 446"
                      ,
"(val-rec (forall ('a) (function ((function ('a) bool) (list 'a)) (list 'a))) filter"
                      , "   (type-lambda ('a)"
                      ,
                      "      (lambda (((function ('a) bool) p?)  ((list 'a) l))"
                      , "         (if ((@ null? 'a) l)"
                      , "             (@ '() 'a)"
                      , "             (if (p? ((@ car 'a) l))"
                      ,
"                 ((@ cons 'a) ((@ car 'a) l) ((@ filter 'a) p? ((@ cdr 'a) l)))"
                      ,
                       "                 ((@ filter 'a) p? ((@ cdr 'a) l)))))))"
                      , "; missing exists?"
                      , "; missing all?"
                      , "; tuschemea.nw 459"
                      ,
"(val-rec (forall ('a 'b) (function ((function ('a) 'b) (list 'a)) (list 'b))) map"
                      , "   (type-lambda ('a 'b)"
                      , "      (lambda (((function ('a) 'b) f)  ((list 'a) l))"
                      , "         (if ((@ null? 'a) l)"
                      , "             (@ '() 'b)"
                      ,
"             ((@ cons 'b) (f ((@ car 'a) l)) ((@ map 'a 'b) f ((@ cdr 'a) l)))))))"
                      , "; tuschemea.nw 479"
                      , "(define bool <= ((int x) (int y)) (not (> x y)))"
                      , "(define bool >= ((int x) (int y)) (not (< x y)))"
                      ,
     "(val != (type-lambda ('a) (lambda (('a x) ('a y)) (not ((@ = 'a) x y)))))"
                      , "; tuschemea.nw 486"
                      , "(define int max ((int x) (int y)) (if (> x y) x y))"
                      , "(define int min ((int x) (int y)) (if (< x y) x y))"
                      , ""
                      , "(define int mod ((int m) (int n)) (- m (* n (/ m n))))"
                      ,
   "(define int gcd ((int m) (int n)) (if ((@ = int) n 0) m (gcd n (mod m n))))"
                      ,
                      "(define int lcm ((int m) (int n)) (* m (/ n (gcd m n))))"
                      ]
(*  Initializing the interpreter                *)
(*                                              *)
(* To put everything together into a working    *)
(* interpreter, we need an initial kind         *)
(* environment as well as a type environment    *)
(* and a value environment.                     *)
(* <boxed values 11>=                           *)
val _ = op kinds : kind      env
val _ = op tfuns : tyex      env
val _ = op vfuns : value ref env
val _ = op envs  : env_bundle
      val treader = topreader (stringsreader ("initial basis", basis), false)
  in  readCheckEvalPrint (treader, fn _ => (), fn _ => ()) envs
  end
(* The function [[runInterpreter]] takes one    *)
(* argument, which tells it whether to prompt.  *)
(* <initialization for {\tuscheme}>=            *)
fun runInterpreter noisy = 
  let fun writeln s = app print [s, "\n"]
      val treader = topreader (filereader ("standard input", TextIO.stdIn),
                                                                          noisy)
  in  ignore (readCheckEvalPrint (treader, writeln, writeln) initialEnvs)
  end 


(*****************************************************************)
(*                                                               *)
(*   COMMAND LINE                                                *)
(*                                                               *)
(*****************************************************************)

(* Building and exporting a program             *)
(*                                              *)
(* The final step in implementing the           *)
(* interpreter is to create a function that     *)
(* looks at the command line and calls          *)
(* [[runInterpreter]]. With a compiler like     *)
(* Moscow ML or MLton, the module isn't         *)
(* elaborated until run time, so we can simply  *)
(* insert an irrelevant [[val]] binding, the    *)
(* elaboration of which has the side effect of  *)
(* calling [[main]].                            *)
(* [[CommandLine.arguments()]] returns an       *)
(* argument list.                               *)
(* <command line>=                              *)
fun main ["-q"] = runInterpreter false
  | main []     = runInterpreter true
  | main _      =
      TextIO.output(TextIO.stdErr, "Usage: " ^ CommandLine.name() ^ " [-q]\n")
val _ = main (CommandLine.arguments())
