(*   Putting all the pieces together            *)
(*                                              *)
(* We stitch together the parts of the          *)
(* implementation in this order:                *)
(* <ml.sml>=                                    *)


(*****************************************************************)
(*                                                               *)
(*   ENVIRONMENTS                                                *)
(*                                                               *)
(*****************************************************************)

(* <environments>=                              *)
type name = string
type 'a env = (name * 'a) list
val emptyEnv = []

(* lookup and assignment of existing bindings *)
exception NotFound of name
fun find (name, []) = raise NotFound name
  | find (name, (n, v)::tail) = if name = n then v else find(name, tail)

(* adding new bindings *)
exception BindListLength
fun bind(name, v, rho) = (name, v) :: rho
fun bindList(n::vars, v::vals, rho) = bindList(vars, vals, bind(n, v, rho))
  | bindList([], [], rho) = rho
  | bindList _ = raise BindListLength
(* We represent an environment of type [['a     *)
(* env]] as a list of ([[name]], [['a]]) pairs. *)
(* The declarations in the box give the         *)
(* interface to our implementation; through     *)
(* some Noweb hackery, they are checked by the  *)
(* ML compiler.                                 *)
(* <boxed values 21>=                           *)
val _ = op emptyEnv : 'a env
val _ = op find     : name * 'a env -> 'a
val _ = op bind     : name      * 'a      * 'a env -> 'a env
val _ = op bindList : name list * 'a list * 'a env -> 'a env
(* [[error]] procedure,                         *)
(* to indicate when things have gone wrong. The *)
(* exceptions we use are listed in Table [->].  *)



(*****************************************************************)
(*                                                               *)
(*   TYPES                                                       *)
(*                                                               *)
(*****************************************************************)

(* The difference between types and type        *)
(* schemes is reflected in the abstract syntax: *)
(* <types>=                                     *)
datatype ty = TYVAR  of name                (* type variable alpha *)
            | TYCON  of name                (* type constructor mu *)
            | CONAPP of ty * ty list        (* type-level application *)

datatype type_scheme = FORALL of name list * ty
(* <types>=                                     *)
fun tysubst (tau, varenv) =
  let fun subst (TYVAR a) = (find(a, varenv) handle NotFound _ => TYVAR a)
        | subst (TYCON c) = (TYCON c)
        | subst (CONAPP (tau, taus)) = CONAPP (subst tau, map subst taus)
(* We implement substitution using almost       *)
(* exactly the same code as in Chapter [->].    *)
(* <boxed values 1>=                            *)
val _ = op tysubst : ty * ty env -> ty
val _ = op subst   : ty          -> ty
  in  subst tau
  end
(* <types>=                                     *)
type subst = ty -> ty
infix 7 |-->
fun a |--> tau = 
  fn tau' => tysubst(tau', bind(a, tau, emptyEnv))
fun idsubst tau = tau
(* <types>=                                     *)
fun instantiate (FORALL (formals, tau), actuals) =
  tysubst(tau, bindList (formals, actuals, emptyEnv))
          handle BindListLength =>
            let exception ThisCan'tHappen in raise ThisCan'tHappen end
(* <types>=                                     *)
(* <printing types>=                            *)
(* Function [[separate]] helps print a readable *)
(* list of formals for the error message.       *)
(* <definition of [[separate]]>=                *)
fun separate (zero, sep) =  (* useful for printing *)
  let fun s []     = zero
        | s [x]    = x
        | s (h::t) = h ^ sep ^ s t
  in  s
end
local
  (* precedences *)
  val CONp   = 3
  val STARp  = 2
  val ARROWp = 1
  val NONEp  = 0
  
  fun parens s = "(" ^ s ^ ")"
  fun bracket(s, context, prec) = if prec <= context then parens s else s
  fun p(context, CONAPP (TYCON "tuple", l)) = bracket(ptuple l, context, STARp)
    | p(context, CONAPP (TYCON "function", [arg, ret])) = 
          bracket(p(ARROWp, arg) ^ " -> " ^ p(ARROWp, ret), context, ARROWp)
    | p(context, CONAPP (n, []))  = p(context, n)
    | p(context, CONAPP (n, [t])) = p(CONp, t) ^ " " ^ p(CONp, n)
    | p(context, CONAPP (n, l)) = 
          "(" ^ separate("", ", ") (map typeString l) ^ ") " ^ p(CONp, n)
    | p(context, TYCON n) = n
    | p(context, TYVAR v) = "'" ^ v
  and ptuple l = separate("unit", " * ") (map (fn t => p(STARp, t)) l)
  and typeString ty = p(NONEp, ty)
in 
  val typeString = typeString
  val ptuple = ptuple
end
fun typeSchemeString (FORALL (_, ty)) = 
  typeString ty
(* <types>=                                     *)
val inttype  = TYCON "int"
val booltype = TYCON "bool"
val symtype  = TYCON "sym"
val alpha    = TYVAR "a"
val beta     = TYVAR "b"
fun tupletype l     = CONAPP (TYCON "tuple", l)
fun pairtype (x, y) = tupletype [x, y]
val unittype        = tupletype []
fun listtype ty     = CONAPP (TYCON "list", [ty])
fun funtype (args, result) = 
  CONAPP (TYCON "function", [tupletype args, result])
(* To make it easy to create a substitution, we *)
(* define a new infix operator [[|->]]. The     *)
(* expression [[alpha |-> tau]] is the          *)
(* substitution that substitutes [[tau]] for    *)
(* [[alpha]]. We also define [[idsubst]] to be  *)
(* the identity substitution. We could write    *)
(* [[fn ty => ty]] everywhere, but [[idsubst]]  *)
(* is sometimes more convenient. We get all the *)
(* substitutions we need by using these two     *)
(* functions plus the built-in                  *)
(* function-composition function, [[o]].        *)
(* <boxed values 2>=                            *)
type subst = subst
val _ = op |-->    : name * ty -> subst
val _ = op idsubst : subst
(* Instantiation is just as in Chapter [->],    *)
(* except no kind environment is needed.        *)
(* Because the system ensures everything has    *)
(* the right kind, it is an internal error to   *)
(* instantiate with the wrong number of         *)
(* arguments.                                   *)
(* <boxed values 2>=                            *)
val _ = op instantiate : type_scheme * ty list -> ty
(*   Other functions to manipulate types        *)
(*                                              *)
(* Appendix [->] defines a function used to     *)
(* print types. We use the same function to     *)
(* print type schemes, because in ML, it is     *)
(* traditional to omit the list of \/-bound     *)
(* type variables. [This omission is            *)
(* traditional because in any top-level type,   *)
(* \emph{all} type variables are                *)
(* $\forall$-bound.]                            *)
(* <boxed values 2>=                            *)
val _ = op typeString       : ty          -> string
val _ = op typeSchemeString : type_scheme -> string
(* To make it easier to define the primitive    *)
(* operations of uML, we provide convenience    *)
(* functions very much like those from          *)
(* Chapter [->].                                *)
(* <boxed values 2>=                            *)
val _ = op inttype   : ty
val _ = op booltype  : ty
val _ = op symtype   : ty
val _ = op unittype  : ty
val _ = op listtype  : ty -> ty
val _ = op tupletype : ty list -> ty
val _ = op pairtype  : ty * ty -> ty
val _ = op funtype   : ty list * ty -> ty
val _ = op alpha     : ty
val _ = op beta      : ty
(* <types>=                                     *)
type 'a set = 'a list
val emptyset = []
fun member x = 
  List.exists (fn y => y = x)
fun insert (x, l) = 
  if member x l then l else x::l
fun diff  (s1, s2) = 
  List.filter (fn x => not (member x s2)) s1
fun union (s1, s2) = foldl insert s2 s1
(*  Sets of type variables                      *)
(*                                              *)
(* Much of type inference manipulates sets of   *)
(* type variables. We provide a simple          *)
(* implementation using lists. The double prime *)
(* in the type variable [[''a]] means that      *)
(* [[''a]] has to be an ``equality type,''      *)
(* i.e., type variable [[''a]] may be           *)
(* instantiated only at a type that admits      *)
(* equality.                                    *)
(* <boxed values 3>=                            *)
type 'a set = 'a set
val _ = op emptyset : 'a set
val _ = op member   : ''a -> ''a set -> bool
val _ = op insert   : ''a     * ''a set  -> ''a set
val _ = op union    : ''a set * ''a set  -> ''a set
val _ = op diff     : ''a set * ''a set  -> ''a set
(* <types>=                                     *)
fun freevars t =
  let fun f(TYVAR v,          l) = insert (v, l)
        | f(TYCON _,          l) = l
        | f(CONAPP(ty, tys),  l) = foldl f (f (ty, l)) tys
  in  rev (f(t, emptyset))
  end  
(* The function [[freevars]] computes the free  *)
(* type variables of a type. For readability,   *)
(* we ensure that type variables appear in the  *)
(* set in the order of their first appearance   *)
(* in the type, when reading from left to       *)
(* right.                                       *)
(* <boxed values 4>=                            *)
val _ = op freevars : ty -> name set
(* <types>=                                     *)
local
  val n = ref 1
in
  fun freshtyvar _ = TYVAR ("t" ^ Int.toString (!n) before n := !n + 1)
(* In our interpreter, fresh type variables     *)
(* come from the [[freshtyvar]] function. We    *)
(* use a private mutable counter to supply an   *)
(* arbitrary number of type variables of the    *)
(* form t n. Because uML does not permit users  *)
(* to write type variables explicitly, we need  *)
(* not worry about the names colliding with     *)
(* other names.                                 *)
(* <boxed values 5>=                            *)
val _ = op freshtyvar : 'a -> ty
end
(* <types>=                                     *)
fun canonicalize (FORALL (bound, ty)) =
  let fun canonicalTyvarName n =
        if n < 26 then str(chr (ord #"a" + n))
        else "v" ^ Int.toString (n - 25)
      val free = diff (freevars ty, bound)
      fun unusedIndex n =
        if member (canonicalTyvarName n) free then unusedIndex (n+1) else n
      fun newBoundVars (index, [])                = []
        | newBoundVars (index, oldvar :: oldvars) =
            let val n = unusedIndex index
            in  canonicalTyvarName n :: newBoundVars (n+1, oldvars)
            end
      val newBound = newBoundVars (0, bound)
(*  Canonical type schemes                      *)
(*                                              *)
(* Type variables like [['t136]] are not very   *)
(* readable in error messages. If a type        *)
(* variable is \/-bound, its name is            *)
(* irrelevant. For readability, we are better   *)
(* off using the names [['a]], [['b]], etc.,    *)
(* for bound type variables. The function       *)
(* [[canonicalize]] renames bound type          *)
(* variables in a type scheme. We replace each  *)
(* existing bound type variable with a          *)
(* canonically named type variable, being       *)
(* careful not to use the name of any free type *)
(* variable.                                    *)
(* <boxed values 6>=                            *)
val _ = op canonicalize : type_scheme -> type_scheme
val _ = op newBoundVars : int * name list -> name list
  in  FORALL (newBound, tysubst(ty, bindList(bound, map TYVAR newBound, emptyEnv
                                                                             )))
  end
(* [[unusedIndex]] finds a name                 *)
(* for a bound type variable; it ensures that   *)
(* the name is not the name of any free type    *)
(* variable.                                    *)

(* <types>=                                     *)
exception TypeError of string

fun unificationError (t1, t2) =
  case canonicalize (FORALL (union (freevars t1, freevars t2), tupletype [t1, t2
                                                                             ]))
    of FORALL(_, CONAPP(TYCON "tuple", [t1, t2])) => 
         raise TypeError ("Cannot unify " ^ typeString t1 ^ " and " ^ typeString
                                                                             t2)
     | _ => let exception ThisCan'tHappen in raise ThisCan'tHappen end
(* <types>=                                     *)
exception LeftAsExercise
fun unify     (t1, t2) = 
  raise LeftAsExercise
and unifyList (l1, l2) = 
  raise LeftAsExercise
(*  Unification                                 *)
(*                                              *)
(* The fundamental operation in type inference  *)
(* is to unify two types by finding a \subsn    *)
(* such that \subsntau_1 = \subsntau_2. But in  *)
(* an ill-typed program, no such \subsn exists. *)
(* In that case, to issue a readable error      *)
(* message we want to put a pair of types into  *)
(* canonical form. The function                 *)
(* [[unificationError]] does just that, by      *)
(* putting the two types together into a pair   *)
(* type.                                        *)
(* <boxed values 7>=                            *)
val _ = op unificationError : ty * ty -> 'a
(* Our type inference requires unification not  *)
(* only of types, but also of lists of types.   *)
(* The implementations are left as              *)
(* Exercise [->]. [*]                           *)
(* <boxed values 7>=                            *)
val _ = op unify     : ty      * ty      -> subst
val _ = op unifyList : ty list * ty list -> subst
(* Our representation of type environments is   *)
(* therefore as follows:                        *)
(* <types>=                                     *)
type type_env = type_scheme env * subst * name set
(* <types>=                                     *)
val emptyTypeEnv = (emptyEnv, idsubst, emptyset)
(* The empty type environment binds no          *)
(* variables and has no free type variables.    *)
(* When we look up a variable, we lazily apply  *)
(* the identity substitution.                   *)
(* <boxed values 8>=                            *)
val _ = op emptyTypeEnv : type_env
(* <types>=                                     *)
fun freshInstance (FORALL (bound, tau)) =
  instantiate (FORALL (bound, tau), map freshtyvar bound)
fun findtype(v, (Gamma, theta, free)) = theta (freshInstance (find(v, Gamma)))
(* When we look up a variable, we instantiate   *)
(* the variable's type scheme and lazily apply  *)
(* a saved substitution. It is easiest to       *)
(* instantiate first, by substituting fresh     *)
(* type variables for the bound type variables. *)
(* Then we apply the substitution, confident    *)
(* that it will not change the fresh type       *)
(* variables.                                   *)
(* <boxed values 9>=                            *)
val _ = op freshInstance : type_scheme -> ty
val _ = op findtype      : name * type_env -> ty
(* <types>=                                     *)
fun bindtype(v, sigma as FORALL (bound, tau), (Gamma, theta, free)) = 
  (bind(v, sigma, Gamma), theta, union(diff (freevars tau, bound), free))
(* <types>=                                     *)
infixr 2 on
fun theta on (Gamma, oldtheta, free) = 
  let val free = foldl union emptyset (map (freevars o theta o TYVAR) free)
  in  (Gamma, theta o oldtheta, free)
  end
(* To add a new binding, we eagerly update the  *)
(* set of free type variables in Gamma, i.e.,   *)
(* we take the union with the free type         *)
(* variables of the new type scheme sigma.      *)
(* <boxed values 10>=                           *)
val _ = op bindtype : name * type_scheme * type_env -> type_env
(* To apply a substitution [[theta]] to         *)
(* environment [[Gamma]], we use an infix       *)
(* function [[on]], writing [[theta on Gamma]]. *)
(* Because the substitution is applied lazily,  *)
(* we simply compose it with the existing       *)
(* substitution. But because we substitute for  *)
(* the free type variables in [[Gamma]],        *)
(* applying a substitution may also change the  *)
(* set of free type variables. The new set of   *)
(* free type variables must be computed         *)
(* eagerly.                                     *)
(* <boxed values 10>=                           *)
val _ = op on : subst * type_env -> type_env
(* <types>=                                     *)
fun freevarsGamma (_, _, free) = free
(* <types>=                                     *)
fun generalize (tau, Gamma) =
  canonicalize (FORALL (diff (freevars tau, freevarsGamma Gamma), tau))
(* Finally, [[freevarsGamma]] extracts the free *)
(* type variables.                              *)
(* <boxed values 11>=                           *)
val _ = op freevarsGamma : type_env -> name set
(* With this machinery in place, we can         *)
(* implement [[generalize]], which closes over  *)
(* type variables not free in Gamma.            *)
(* <boxed values 11>=                           *)
val _ = op generalize : ty * type_env -> type_scheme


(*****************************************************************)
(*                                                               *)
(*   LEXICAL ANALYSIS                                            *)
(*                                                               *)
(*****************************************************************)

(* A value of type [[par]] represents a list of *)
(* tokens with balanced parentheses. In order   *)
(* to issue better error messages, we keep the  *)
(* source-code location of the each             *)
(* parenthesized list. A source names the       *)
(* source and gives a line number.              *)
(* <lexical analysis ((mlscheme))>=             *)
type srcloc  = string * int
fun srclocString (source, line) = source ^ ", line " ^ Int.toString line
(* <lexical analysis ((mlscheme))>=             *)
datatype par = NAME  of string
             | INT   of int
             | SHARP of char  (* #t or #f *)
             | LIST  of srcloc * par list
(* <lexical analysis ((mlscheme))>=             *)

exception Paren of char list * srcloc
   (* raise if found ) when reading input *)
exception PrematureEOF of string (* shows what we were reading ... *)
exception EOF                 (* raised by reader *)
fun read nextline curloc : string -> par list =
  let (* To help with reading, we have functions to   *)
      (* deal with lexical details. We begin by       *)
      (* defining tests for the delimiters of         *)
      (* symbols:                                     *)
      (* <character-fiddling functions>=              *)
      fun isDelim c = List.exists (fn c' => c' = c) [#"(", #")", #" ", #"\t",
                                                                    #"\n", #";"]
      (* [[]] is like the micro-Scheme                *)
      (* [[exists?]]. We needn't write its            *)
      (* implementation because it is in the initial  *)
      (* basis for Standard ML.                       *)

      (* <character-fiddling functions>=              *)
      fun stringReader []     = NONE
        | stringReader (h::t) = SOME (h, t)
      val strton = Int.scan StringCvt.DEC stringReader
      (* We continue with integer conversion;         *)
      (* [[strton s]] returns an integer, plus        *)
      (* leftover characters from [[s]], or [[NONE]]  *)
      (* if no prefix of [[s]] is an integer literal. *)
      (* It uses [[Int.scan]] from the Standard Basis *)
      (* Library.                                     *)
      (* <boxed values 32>=                           *)
      val _ = op strton : char list -> (int * char list) option
      (* We need to be able to skip whitespace and    *)
      (* comments. [[StringCvt.skipWS]] skips white   *)
      (* space; if we encounter a semicolon, we use   *)
      (* [[StringCvt.dropl]] to skip everything up to *)
      (* the next newline, then resume skipping white *)
      (* space.                                       *)
      (* <character-fiddling functions>=              *)
      fun skip s =
        case StringCvt.skipWS stringReader s
          of #";" :: s => skip (StringCvt.dropl (fn c => c <> #"\n")
                                                                 stringReader s)
           | s => s
      (* [[nonwhite]], which                          *)
      (* summons more input lines as needed.          *)

      (* <character-fiddling functions>=              *)
      fun nonwhite s =
        case skip s
          of [] => nonwhite (explode (nextline()))
           | s  => s
      fun readmap f (result, unconsumed) = (f result, unconsumed)
      (* get is called only when s is guaranteed non-blank, non-comment, non-
                                                                        empty *)
      fun get (#"("::s)  = readmap (fn l => LIST l) (list (curloc(), s, []))
        | get (#")"::s)  = raise (Paren (s, curloc()))
        | get (#"'"::s)  = let val loc = curloc()
                               fun quote p = LIST (loc, [NAME "quote", p])
                           in  readmap quote (get (nonwhite s))
                           end
        | get s          = atom (s, [])
      and atom(c::s, r) = if isDelim c then (mkAtom(rev r), c::s) 
                          else atom(s, c::r)
        | atom([],   r) = (mkAtom(rev r), [])
      and mkAtom [#"#", #"f"] = SHARP #"f"
        | mkAtom [#"#", #"t"] = SHARP #"t"
        | mkAtom s = (case strton s of SOME (n, []) => INT n | _ => NAME (
                                                                     implode s))
      and list (loc, s, l) =
        let val next = nonwhite s
        in  let val (a, u) = get next
            in  list(loc, u, a::l)
            end handle (Paren (u, _)) => ((loc, rev l), skip u)
                     | EOF => raise PrematureEOF (implode next)
        end handle EOF => raise PrematureEOF (implode s)
      fun readline [] = []
        | readline s  = 
             let val(p, u) = get s
             in  p :: readline(skip u)
             end
(* The interface to the [[read]] function       *)
(* includes a [[curloc]] parameter, which can   *)
(* be used to keep track of the source-code     *)
(* location of any token. Of the interpreters   *)
(* in this book, only the uSmalltalk interpter  *)
(* (Appendix [->]) takes advantage of this      *)
(* capability.                                  *)
(*                                              *)
(* Function [[read]] is implemented by a group  *)
(* of local lexing functions. Each such         *)
(* function takes a list of unconsumed          *)
(* characters and returns a pair consisting of  *)
(* the [[par]] read plus unconsumed characters. *)
(* The functions are:                           *)
(*                                              *)
(*   [[get]] --- Read the next input            *)
(*   [[atom]] --- Read a name, integer, or      *)
(*   [[#]] literal                              *)
(*   [[list]] --- Read a list of inputs in      *)
(*   parentheses                                *)
(*                                              *)
(* The [[list]] function takes as an additional *)
(* parameter the source-code location of the    *)
(* start of the list. Internal function         *)
(* [[readline]] puts it all together by reading *)
(* as many inputs are as on the line, until all *)
(* the line is consumed.                        *)
(* <boxed values 31>=                           *)
val _ = op read : (unit -> string) -> (unit -> srcloc) -> string -> par list
val _ = op get  : char list                      -> par                 * char
                                                                            list
val _ = op atom : char list * char list          -> par                 * char
                                                                            list
val _ = op list : srcloc * char list * par list  -> (srcloc * par list) * char
                                                                            list
  in  fn s => readline(skip (explode s)) 
  end


(*****************************************************************)
(*                                                               *)
(*   ABSTRACT SYNTAX AND VALUES                                  *)
(*                                                               *)
(*****************************************************************)

(* Abstract syntax and values of uML            *)
(*                                              *)
(* As in Chapter [->], we define the abstract   *)
(* syntax of uML by presenting type definitions *)
(* from our implementation. It is a subset of   *)
(* the abstract syntax of micro-Scheme.         *)
(* <abstract syntax and values>=                *)
datatype exp = LITERAL of value
             | VAR     of name
             | IFX     of exp * exp * exp
             | BEGIN   of exp list
             | LETX    of let_kind * (name * exp) list * exp
             | LAMBDA  of name list * exp
             | APPLY   of exp * exp list
and let_kind = LET | LETREC | LETSTAR
and (* At the language level, uML and micro-Scheme  *)
    (* have the same values, but we use subtly      *)
    (* different representations in the             *)
    (* interpreters. Because micro-Scheme has       *)
    (* [[set]], environments map names to mutable   *)
    (* cells. In uML, environments map names to     *)
    (* values. Without mutable locations, how are   *)
    (* we to implement recursive functions? The     *)
    (* value of a function is a closure, and to     *)
    (* make a recursive call possible, the closure  *)
    (* must contain the value of the function. To   *)
    (* implement a closure that refers to itself in *)
    (* this way, we need some sort of cyclic        *)
    (* representation. In our implementation of     *)
    (* micro-Scheme, we use mutation to create a    *)
    (* graph with a cycle. In our implementation of *)
    (* uML, we build each closure's environment     *)
    (* lazily; that is, instead of storing an       *)
    (* environment in the closure, we store a       *)
    (* function that produces an environment on     *)
    (* demand.                                      *)
    (* <definition of [[value]]>=                   *)
    value = NIL
          | BOOL      of bool
          | NUM       of int
          | SYM       of name
          | PAIR      of value * value
          | CLOSURE   of lambda * (unit -> value env)
          | PRIMITIVE of primop
    withtype primop = value list -> value (* raises RuntimeError *)
         and lambda = name list * exp
    exception RuntimeError of string (* error message *)
    (* [[RuntimeError]] to signal errors as         *)
    (* in Chapter [->]. Because the values are      *)
    (* nearly the same, we reuse the projection,    *)
    (* embedding, and printing functions from       *)
    (* Chapter [->].                                *)

(* Except for [[VALREC]], top-level items are   *)
(* exactly as in micro-Scheme.                  *)
(* <abstract syntax and values>=                *)
datatype toplevel = EXP    of exp
                  | DEFINE of name * (name list * exp)
                  | VAL    of name * exp
                  | VALREC of name * exp
                  | USE    of name


(*****************************************************************)
(*                                                               *)
(*   VALUES                                                      *)
(*                                                               *)
(*****************************************************************)

(* <values ((mlscheme))>=                       *)



fun embedList []     = NIL
  | embedList (h::t) = PAIR(h, embedList t)
fun embedPredicate f = fn x => BOOL (f x)
fun bool (BOOL b) = b
  | bool _        = true
(*    Embedding and projection                  *)
(*                                              *)
(* An S-expression can represent an integer,    *)
(* Boolean, name, function, list, etc. We may   *)
(* sometimes have an ML boolean, list, or       *)
(* function that we wish to represent as an     *)
(* S-expression, or similarly, an S-expression  *)
(* that we wish to represent as a value of      *)
(* type [[bool]]. Here we define mappings       *)
(* between type [[value]] and some other        *)
(* ML types. Because the set of values          *)
(* representable by an ML value of type         *)
(* [[value]] strictly contains each of the sets *)
(* of values representable by these ML types,   *)
(* these mappings are called embedding and      *)
(* projection. Because the [[value]] type is    *)
(* strictly larger than these ML types, no      *)
(* embedding operation ever fails, but a        *)
(* projection operation might. [This property   *)
(* is a general characteristic of any           *)
(* embedding/projection pair. Mathematical      *)
(* terminology may clarify; an embedding e of   *)
(* S into S' is an injection from S-->S'. The   *)
(* corresponding projection pi_e is a left      *)
(* inverse of the embedding; that is pi_e oe is *)
(* the identity function on S. There is no      *)
(* corresponding guarantee for e opi_e; for     *)
(* example, pi_e may be undefined (_|_) on some *)
(* elements of S', or e(pi_e(x)) may not        *)
(* equal x.] For example, although any ML       *)
(* function of type [[value -> bool]] can be    *)
(* embedded into [[value]] by using the         *)
(* [[PRIMITIVE]] constructor, there are values  *)
(* of type [[value]] that cannot be projected   *)
(* into an ML function of type [[value ->       *)
(* bool]].                                      *)
(*                                              *)
(* Lists and Booleans are straightforward.      *)
(* Function [[embedPredicate]] is not a true    *)
(* embedding; it takes any function returning   *)
(* [[bool]] and returns a corresponding         *)
(* function returning [[value]]. It really      *)
(* embeds the function's result, not the        *)
(* function itself.                             *)
(* <boxed values 22>=                           *)
val _ = op embedList      : value list -> value
val _ = op embedPredicate : ('a -> bool) -> ('a -> value)
val _ = op bool           : value -> bool
(* [[bool]] is the projection                   *)
(* function, mapping micro-Scheme values into   *)
(* ML Booleans.                                 *)

(* <values ((mlscheme))>=                       *)
fun valueString (NIL)    = "()"
  | valueString (BOOL b) = if b then "#t" else "#f"
  | valueString (NUM n)  = String.map (fn #"~" => #"-" | c => c) (Int.toString n
                                                                               )
  | valueString (SYM v)  = v
  | valueString (PAIR(car, cdr))  = 
      let fun tail (PAIR(car, cdr)) = " " ^ valueString car ^ tail cdr
            | tail NIL = ")"
            | tail v = " . " ^ valueString v ^ ")"
      in  "(" ^ valueString car ^ tail cdr
      end
  | valueString (CLOSURE   _) = "<procedure>"
  | valueString (PRIMITIVE _) = "<procedure>"
(*    Printing                                  *)
(*                                              *)
(* We render an S-expression as a string.       *)
(* <boxed values 23>=                           *)
val _ = op valueString : value -> string
(* [[Int.toString]] indicates the               *)
(* [[toString]] function from the standard      *)
(* module [[Int]]. This function, which is part *)
(* of ML's Standard Basis Library, converts an  *)
(* integer to a string. We use another standard *)
(* function, [[String.map]], to change the      *)
(* minus sign from the ML convention ([[ ]]) to *)
(* the Scheme convention ([[-]]).               *)



(*****************************************************************)
(*                                                               *)
(*   PARSING                                                     *)
(*                                                               *)
(*****************************************************************)

(* Parsing                                      *)
(*                                              *)
(* uML can use micro-Scheme's lexical analysis, *)
(* so all we have here is a parser.             *)
(* <parsing>=                                   *)
exception SyntaxError of string
fun syntaxErrorAt (source, line) msg =
  raise SyntaxError (source ^ ", line " ^ Int.toString line ^ ": " ^ msg)
fun parse (NAME s)  = VAR s
  | parse (INT n)   = LITERAL (NUM n)
  | parse (SHARP c) = LITERAL (BOOL (c = #"t"))
  | parse (LIST (loc, NAME "quote"  :: args)) = parseQuote loc args
  | parse (LIST (loc, NAME "if"     :: args)) = parseIf    loc args
  | parse (LIST (loc, NAME "begin"  :: args)) = BEGIN (map parse args)
  | parse (LIST (loc, NAME "lambda" :: args)) = parseLambda loc args
  | parse (LIST (loc, NAME "let"    :: args)) = parseLet loc LET     args
  | parse (LIST (loc, NAME "letrec" :: args)) = parseLet loc LETREC  args
  | parse (LIST (loc, NAME "let*"   :: args)) = parseLet loc LETSTAR args
  | parse (LIST (loc, h::t))  = APPLY (parse h, map parse t)
  | parse (LIST (loc, [])) = raise SyntaxError "empty application ()"
and parseQuote loc [e] = LITERAL (quote e)
  | parseQuote loc _ = syntaxErrorAt loc "expected (quote e)"
and parseIf _ [e1, e2, e3] = IFX (parse e1, parse e2, parse e3)
  | parseIf loc _ = syntaxErrorAt loc "expected (if e1 e2 e3)"
and parseLambda _ [LIST (loc, formals), body] =
      LAMBDA (map (unname loc) formals, parse body)
  | parseLambda loc _ = syntaxErrorAt loc "expected (lambda (names) body)"
and parseLet _ kind [LIST (loc, bindings), body] = 
      LETX (kind, map (unBinding loc) bindings, parse body)
  | parseLet loc _ _ = syntaxErrorAt loc "expected (let (bindings) body)"
and unname _ (NAME n) = n
  | unname loc _ = syntaxErrorAt loc
                                 "formal parameters to function must be symbols"
and unBinding _ (LIST (_, [NAME n, e])) = (n, parse e)
  | unBinding loc _ = syntaxErrorAt loc "expected (v e) in bindings"
and quote (INT   n)     = NUM n
  | quote (SHARP c)     = BOOL (c = #"t")
  | quote (NAME  n)     = SYM n
  | quote (LIST (_, l)) = embedList (map quote l)
(* <parsing>=                                   *)
fun toplevel (LIST (loc, NAME "define" :: args)) = parseDefine loc args
  | toplevel (LIST (loc, NAME "val"    :: args)) = parseVal    loc "val"     VAL
                                                                            args
  | toplevel (LIST (loc, NAME "val-rec":: args)) = parseVal    loc "val-rec"
                                                                     VALREC args
  | toplevel (LIST (loc, NAME "use"    :: args)) = parseUse    loc args
  | toplevel e = EXP (parse e)
and parseDefine _ [NAME name, LIST (loc, formals), body] =
      DEFINE (name, (map (unname loc) formals, parse body))
  | parseDefine loc _ = syntaxErrorAt loc "expected (define f (args) body)"
and parseVal _ concrete abstract [NAME v, e] = abstract (v, parse e)
  | parseVal loc concrete _ _ = syntaxErrorAt loc ("expected (" ^ concrete ^
                                                                        " v e)")
and parseUse _ [NAME v] = USE v
  | parseUse loc _ = syntaxErrorAt loc "expected (use filename)"


(*****************************************************************)
(*                                                               *)
(*   READERS                                                     *)
(*                                                               *)
(*****************************************************************)

(* Readers                                      *)
(*                                              *)
(*  Line readers                                *)
(*                                              *)
(* Our readers are structured slightly          *)
(* differently from those in C. A reader is a   *)
(* bundle containing three related functions:   *)
(* [[firstline]], which returns the first line  *)
(* read; [[nextline]], which returns all        *)
(* subsequent lines; and [[curloc]], which      *)
(* returns a pair identifying the source        *)
(* location of the most recently returned line. *)
(* Both [[firstline]] and [[nextline]] can      *)
(* raise [[EOF]], (We distinguish [[firstline]] *)
(* from [[nextline]] because in an interactive  *)
(* reader, they may use different prompts.)     *)
(* <readers>=                                   *)
type reader = { firstline : unit -> string
              , nextline  : unit -> string
              , curloc    : unit -> string * int
              }
(* <readers>=                                   *)
fun promptReader (firstPrompt, nextPrompt, reader) =
  let val { firstline = first, nextline = next, curloc = cur } = reader
      fun promptIn get prompt () = ( TextIO.output(TextIO.stdOut, prompt)
                                   ; TextIO.flushOut(TextIO.stdOut)
                                   ; get ()
                                   )
  in  { firstline = promptIn first firstPrompt
      , nextline  = promptIn next  nextPrompt
      , curloc    = cur
      }
  end
(* Given a silent reader, we can make it issue  *)
(* a prompt every time it reads a new line. We  *)
(* use two prompts: one for the first line of   *)
(* an expression and one for later lines.       *)
(* <boxed values 33>=                           *)
val _ = op promptReader : string * string * reader -> reader
(* <readers>=                                   *)
fun filereader (filename, fd)=
  let val lineNum = ref 0
      fun get () =
        let val line = TextIO.inputLine fd
        in  if size line = 0 then raise EOF else line before lineNum := !lineNum
                                                                             + 1
        end
  in  { firstline = get, nextline = get, curloc = (fn () => (filename, !lineNum)
                                                                             ) }
  end

fun stringsreader (what, l) = 
  let val buffer  = ref l
      val lineNum = ref 0
      fun get ()  = case !buffer
                      of [] => raise EOF
                       | h :: t => h before buffer := t before lineNum := !
                                                                     lineNum + 1
  in  { firstline = get, nextline = get, curloc = (fn () => (what, !lineNum)) }
  end              
(* We can read from files or lists of strings.  *)
(* File I/O is provide by the standard module   *)
(* [[TextIO]].                                  *)
(* <boxed values 34>=                           *)
val _ = op filereader    : string * TextIO.instream -> reader
val _ = op stringsreader : string * string list     -> reader
(* [[]], a private mutable cell. When all       *)
(* the strings in the buffer are used up, there *)
(* aren't any more.                             *)

(*  Top-level readers                           *)
(*                                              *)
(* A top-level reader is a buffer of top-level  *)
(* items together with a reader. [*]            *)
(* <readers>=                                   *)
type topreader = { buffer : toplevel list ref, reader : reader }
fun readtop (r as { buffer, reader = { nextline, firstline, curloc } }) =
  case !buffer
    of h::t => h before buffer := t
     | []   => ( buffer := map toplevel (read nextline curloc (firstline()))
               ; readtop r
               )
(* We decide whether to prompt when building a  *)
(* top-level reader.                            *)
(* <readers>=                                   *)
fun topreader (reader, prompt) = 
  let val reader = if prompt then promptReader ("-> ", "   ", reader) else
                                                                          reader
  in  { buffer = ref [], reader = reader }
  end
(* <readers>=                                   *)
fun echoTag f x =
  let val line = f x
      val _ = if (String.substring (line, 0, 2) = ";#" handle _ => false) then
                print line
              else
                ()
  in  line
  end

fun echoReader { nextline=n, firstline=f, curloc=c } =
     { nextline=echoTag n, firstline=echoTag f, curloc=c } 

fun echoTop { buffer = b, reader = r } = { buffer = b, reader = echoReader r }
val topreader = fn args => echoTop (topreader args)
(* As in the C code, we want to print out any   *)
(* line read if it begins with the special      *)
(* string [[;#]]. This string is a formal       *)
(* comment that helps us test the chunks marked *)
(* [[]]. In our ML code, we                     *)
(* arrange it using the higher-order function   *)
(* [[echoReader]].                              *)
(* <boxed values 35>=                           *)
val _ = op echoReader : reader -> reader 


(*****************************************************************)
(*                                                               *)
(*   IMPLEMENTATION OF [[USE]]                                   *)
(*                                                               *)
(*****************************************************************)

(* The implementation of [[use]] is             *)
(* parameterized by [[readEvalPrint]] so we can *)
(* share it with other interpreters. Function   *)
(* [[use]] creates a top-level reader that does *)
(* not prompt, but it uses [[writeln]] to be    *)
(* sure that responses are printed.             *)
(* <implementation of [[use]]>=                 *)
fun use readEvalPrint filename rho =
      let val fd = TextIO.openIn filename
          val treader = topreader (filereader (filename, fd), false)
          fun writeln s = app print [s, "\n"]
      in  readEvalPrint (treader, writeln, writeln) rho
          before TextIO.closeIn fd
      end 
(* [[topreader]] is relegated                   *)
(* to chunk [->] in Appendix [->].              *)



(*****************************************************************)
(*                                                               *)
(*   TYPE INFERENCE                                              *)
(*                                                               *)
(*****************************************************************)

(* <type inference>=                            *)
fun typeof (e, Gamma) =
  let (* Calling typesof(<\ldotsne>, Gamma) returns   *)
      (* (<\ldotsntau>, \subsn) satisfying            *)
      (* \typej\subsnGamma e_i tau_i for every i with *)
      (* 1<=i <=n. The base case is trivial; the      *)
      (* induction step uses this rule from           *)
      (* Section [->]: \usetyTypesOf                  *)
      (* <function [[typesof]], to infer the types of a list of expressions>= *)
      fun typesof([], Gamma : type_env) = ([], idsubst)
        | typesof(e::es, Gamma) =
            let val (tau,  theta)  = typeof (e,  Gamma)
                val (taus, theta') = typesof(es, theta on Gamma)
            in  (theta' tau :: taus, theta' o theta)
            end
      (* Computing the type of a literal value is     *)
      (* left as part of Exercise [->].               *)
      (* <function [[literal]], to infer the type of a literal constant>= *)
      fun literal _ = raise LeftAsExercise
      (* To infer the type of a literal value, we     *)
      (* call [[literal]]. The type of a variable is  *)
      (* a most general instance of the variable's    *)
      (* type scheme in Gamma. No substitution is     *)
      (* needed.                                      *)
      (* <function [[ty]], to infer the type of an expression, given [[Gamma]]>=
                                                                              *)
      fun ty(LITERAL n) = literal n
        | ty(VAR v) = (findtype(v, Gamma), idsubst)
        (* Section [->] shows how to rewrite a type     *)
        (* rule to introduce explicit substitutions;    *)
        (* here is the rule for application:            *)
        (* \usetyApply To implement this rule, we let   *)
        (* [[funty]] stand for \tau, [[actualtypes]]    *)
        (* stand for \ldotsntau, and [[rettype]] stand  *)
        (* for alpha. The first premise is implemented  *)
        (* by a call to [[typesof]] and the second by a *)
        (* call to [[unify]].                           *)
        (* <more alternatives for [[ty]]>=              *)
        | ty(APPLY (f, actuals)) = 
             (case typesof (f :: actuals, Gamma)
                of ([], _) => let exception ThisCan'tHappen in raise
                                                             ThisCan'tHappen end
                 | (funty :: actualtypes, theta) =>
                      let val rettype = freshtyvar()
                          val theta' = unify(funty, funtype(actualtypes, rettype
                                                                              ))
                      in  (theta' rettype, theta' o theta)
                      end)
        (* The remaining cases for [[ty]] are left as   *)
        (* exercises, except we provide syntactic sugar *)
        (* for [[LETSTAR]].                             *)
        (* <more alternatives for [[ty]]>=              *)
        | ty(LETX (LETSTAR, [], body)) = ty body
        | ty(LETX (LETSTAR, (b :: bs), body)) = ty (LETX (LET, [b], LETX (
                                                            LETSTAR, bs, body)))
        (* <more alternatives for [[ty]]>=              *)
        | ty(IFX (e1, e2, e3))        = raise LeftAsExercise
        | ty(BEGIN es)                = raise LeftAsExercise
        | ty(LAMBDA (formals, body))  = raise LeftAsExercise 
        | ty(LETX (LET, bs, body))    = raise LeftAsExercise 
        | ty(LETX (LETREC, bs, body)) = raise LeftAsExercise 
(*   Type inference                             *)
(*                                              *)
(*  Type inference for expressions              *)
(*                                              *)
(* Given an expression e and type               *)
(* environment Gamma, typeof(e, Gamma) returns  *)
(* a pair (tau, \subsn) such that               *)
(* \typej\subsnGamma e tau. It uses three       *)
(* auxiliary functions.                         *)
(* <boxed values 12>=                           *)
val _ = op typeof  : exp      * type_env -> ty      * subst
val _ = op typesof : exp list * type_env -> ty list * subst
val _ = op literal : value -> ty * subst
val _ = op ty      : exp   -> ty * subst
  in  ty e
  end
(* <type inference>=                            *)
fun topty (t, Gamma) =
  case t
    of USE filename       => raise RuntimeError 
                                     "internal error -- `use' reached topty"
     | EXP e              => (* The code for [[EXP]] is just like the code   *)
                             (* for [[VAL]] except that we don't change the  *)
                             (* environment.                                 *)
                             (* <infer type for top-level expression [[e]]>= *)
                             let val (tau, _) = typeof (e, Gamma)
                                 val sigma    = generalize (tau, Gamma)
                             in  (Gamma, typeSchemeString sigma)
                             end
     | DEFINE (x, lambda) => topty (VALREC (x, LAMBDA lambda), Gamma)
     | VAL    (x, e)      => (* [[DEFINE]] is syntactic sugar. The cases for *)
                             (* [[EXP]], [[VAL]], and [[VALREC]] are         *)
                             (* similar. We begin with [[VAL]], which        *)
                             (* computes a type and generalizes it.          *)
                             (* <infer and bind type for [[VAL    (x, e)]]>= *)
                             let val (tau, _) = typeof (e, Gamma)
                                 val sigma    = generalize (tau, Gamma)
                             in  (bindtype(x, sigma, Gamma), typeSchemeString
                                                                          sigma)
                             end
     | VALREC (x, e)      => (* The code for [[VALREC]] is a bit more        *)
                             (* complicated. The problem is we need an       *)
                             (* environment that binds x to tau, but we      *)
                             (* don't yet know tau. \usetyValRec The         *)
                             (* solution is to use a fresh type variable to  *)
                             (* stand for tau, then unify it with the type   *)
                             (* of e.                                        *)
                             (* <infer and bind type for [[VALREC (x, e)]]>= *)
                             let val tau           = freshtyvar ()
                                 val Gamma'        = bindtype(x, FORALL([], tau)
                                                                        , Gamma)
                                 val (tau', theta) = typeof(e, Gamma')
                                 val theta'        = unify (tau', theta tau)
                                 val sigma         = generalize(theta' tau',
                                                                          Gamma)
                             in  (bindtype(x, sigma, Gamma), typeSchemeString
                                                                          sigma)
                             end


(*****************************************************************)
(*                                                               *)
(*   CHECKING AND EVALUATION                                     *)
(*                                                               *)
(*****************************************************************)

(* <checking and evaluation>=                   *)
exception BugInTypeInference of string
(* <evaluation>=                                *)
(* Function [[separate]] helps print a readable *)
(* list of formals for the error message.       *)
(* <definition of [[separate]]>=                *)
fun separate (zero, sep) =  (* useful for printing *)
  let fun s []     = zero
        | s [x]    = x
        | s (h::t) = h ^ sep ^ s t
  in  s
end
fun eval(e, rho) =
  let fun ev(LITERAL n)        = n
        | ev(VAR v)            = find (v, rho)
        | ev(IFX (e1, e2, e3)) = ev (if bool (ev e1) then e2 else e3)
        | ev(LAMBDA l)         = CLOSURE(l, fn _ => rho)
        | ev(BEGIN es) =
            let fun b(e::es, lastval) = b(es, ev e)
                  | b(   [], lastval) = lastval
            in  b(es, BOOL false)
            end
        | ev(APPLY (f, args)) = 
           (case ev f
              of PRIMITIVE prim => prim (map ev args)
               | CLOSURE clo => (* To apply a closure, we bind formal
                                                                              *)
                               (* parameters directly to the values of actual
                                                                              *)
                               (* parameters, not to mutable cells.
                                                                              *)
                               (* <apply closure [[clo]] to [[args]]>=
                                                                              *)
                               let val ((formals, body), mkRho) = clo
                                   val actuals = map ev args
                               in  eval(body, bindList(formals, actuals, mkRho()
                                                                              ))
                                   handle BindListLength => 
                                       raise BugInTypeInference
                                          "Wrong number of arguments to closure"
                               end
               | _ => raise BugInTypeInference "Applied non-function"
               )
        (* \xlet evaluates all right-hand sides in rho, *)
        (* then extends rho to evaluate the body.       *)
        (* <more alternatives for [[ev]]>=              *)
        | ev(LETX (LET, bs, body)) =
            let val (names, values) = ListPair.unzip bs
            in  eval (body, bindList(names, map ev values, rho))
            end
        (* \xletstar evaluates pairs in sequence,       *)
        (* adding a binding to rho after each           *)
        (* evaluation.                                  *)
        (* <more alternatives for [[ev]]>=              *)
        | ev(LETX (LETSTAR, bs, body)) =
            let fun step ((n, e), rho) = bind(n, eval(e, rho), rho)
            in  eval (body, foldl step rho bs)
            end
        (* \xletrec is the most interesting case.       *)
        (* Function [[makeRho']] builds an environment  *)
        (* in which each right-hand side stands for a   *)
        (* closure. Each closure's captured environment *)
        (* is the one built by [[makeRho']]. The        *)
        (* recursion is OK because the environment is   *)
        (* built lazily, so [[makeRho']] always         *)
        (* terminates. The right-hand sides must be     *)
        (* lambda abstractions.                         *)
        (* <more alternatives for [[ev]]>=              *)
        | ev(LETX (LETREC, bs, body)) =
            let fun makeRho' () =
                  let fun step ((n, e), rho) =
                            (case e
                               of LAMBDA l => bind(n, CLOSURE (l, makeRho'), rho
                                                                               )
                                | _ => raise RuntimeError "non-lambda in letrec"
                                                                               )
                  in  foldl step rho bs
                  end
            in  eval(body, makeRho'())
            end
  in  ev e
  end
(*  Type inference for top-level items          *)
(*                                              *)
(* Given a top-level item, we infer a type for  *)
(* it, generalize the type, and add the type to *)
(* the environment. We also return a string     *)
(* representation for the type for possible     *)
(* printing in the top-level loop.              *)
(* <boxed values 13>=                           *)
val _ = op topty : toplevel * type_env -> type_env * string
(*   Evaluation                                 *)
(*                                              *)
(*  Evaluation of expressions                   *)
(*                                              *)
(* Because the abstract syntax of uML is a      *)
(* subset of micro-Scheme, the evaluator is     *)
(* almost a subset of the micro-Scheme          *)
(* evaluator. One difference is that because    *)
(* uML doesn't have mutation, environments map  *)
(* names to values, instead of mapping them to  *)
(* mutable cells. Another is that type          *)
(* inference should eliminate most potential    *)
(* errors. If one of those errors occurs        *)
(* anyway, we raise the exception               *)
(* [[BugInTypeInference]].                      *)
(* <boxed values 13>=                           *)
val _ = op eval : exp * value env -> value
(* <evaluation>=                                *)
fun topeval (t, rho) =
  case t
    of USE filename => raise RuntimeError
                                       "internal error -- `use' reached topeval"
     | EXP e        => (rho, valueString (eval(e, rho)))
     | DEFINE (name, lambda) => topeval (VALREC (name, LAMBDA lambda), rho)
     | VAL    (name, e)      =>
          let val v   = eval(e, rho)
              val rho = bind(name, v, rho)
          in  (rho, showVal name v)
          end
     | VALREC (name, LAMBDA lambda) => 
          let fun makeRho' () = bind(name, CLOSURE(lambda, makeRho'), rho)
              val v           = CLOSURE(lambda, makeRho')
          in  (makeRho'(), showVal name v)
          end
     | VALREC _ => raise RuntimeError "expression in val-rec must be lambda"
and showVal name v =
      case v
        of CLOSURE   _ => name
         | PRIMITIVE _ => name
         | _ => valueString v
(*  Top-level evaluation                        *)
(*                                              *)
(* Evaluating a top-level item can produce a    *)
(* new environment. The function [[topeval]]    *)
(* also returns a string which, if nonempty,    *)
(* should be printed to show the value of the   *)
(* item.                                        *)
(* <boxed values 14>=                           *)
val _ = op topeval : toplevel * value env -> value env * string
(* [[VALREC]] works only                        *)
(* for [[LAMBDA]] expressions because these are *)
(* the only expressions for which we can        *)
(* compute the value without having the         *)
(* environment.                                 *)

(* <evaluation>=                                *)
fun binaryOp f = (fn [a, b] => f(a, b) | _ => raise BugInTypeInference "arity 2"
                                                                               )
fun unaryOp  f = (fn [a]    => f a     | _ => raise BugInTypeInference "arity 1"
                                                                               )
(*   Primitives                                 *)
(*                                              *)
(* Here are the primitives. All are either      *)
(* binary or unary operators. Type inference    *)
(* should guarantee that operators are used     *)
(* with the correct arity.                      *)
(* <boxed values 15>=                           *)
val _ = op unaryOp  : (value         -> value) -> (value list -> value)
val _ = op binaryOp : (value * value -> value) -> (value list -> value)
(* <evaluation>=                                *)
fun arithOp f =
      binaryOp (fn (NUM n1, NUM n2) => NUM (f(n1, n2)) 
                 | _ => raise BugInTypeInference "arithmetic on non-numbers")
val arithtype = funtype([inttype, inttype], inttype)
(* Arithmetic primitives expect and return      *)
(* integers. Each primitive operation must be   *)
(* associated with a type scheme in the initial *)
(* environment. It is easier, however, to       *)
(* associate a type with each primitive and to  *)
(* generalize them all at one go when we create *)
(* the initial environment.                     *)
(* <boxed values 16>=                           *)
val _ = op arithOp   : (int * int -> int) -> (value list -> value)
val _ = op arithtype : ty
(* [[<<primops                                  *)
(* [[::]]>>]] to put all the primitives into    *)
(* one giant list, and we use that list to      *)
(* build the initial environment for the        *)
(* read-eval-print loop. The big difference is  *)
(* that in uML, each primitive has a type as    *)
(* well as a value.                             *)

(* <evaluation>=                                *)
fun predOp f     = unaryOp  (embedPredicate f)
fun comparison f = binaryOp (embedPredicate f)
fun intcompare f = 
      comparison (fn (NUM n1, NUM n2) => f(n1, n2)
                   | _ => raise BugInTypeInference "comparing non-numbers")
fun predtype x = funtype([x],    booltype)
fun comptype x = funtype([x, x], booltype)
(* We have two kinds of predicates: ordinary    *)
(* predicates take one argument, and            *)
(* comparisons take two. Some comparisons apply *)
(* only to integers. We get to reuse            *)
(* [[embedPredicate]] for the definitions.      *)
(* <boxed values 17>=                           *)
val _ = op predOp     : (value         -> bool) -> (value list -> value)
val _ = op comparison : (value * value -> bool) -> (value list -> value)
val _ = op intcompare : (int   * int   -> bool) -> (value list -> value)
val _ = op predtype   : ty -> ty
val _ = op comptype   : ty -> ty
type env_bundle = type_env * value env
fun topCheckEval (t, envs as (Gamma, rho), echo) =
  case t
    of USE filename => use readCheckEvalPrint filename envs
     | _ =>
        let val (Gamma, tystring)  = topty   (t, Gamma)
            val (rho,   valstring) = topeval (t, rho)
            val _ = if size valstring > 0 then echo (valstring ^ " : " ^
                                                                       tystring)
                    else ()
(*   Top-level processing: type inference and evaluation *)
(*                                              *)
(* As in Typed Impcore and Typed uScheme, we    *)
(* process a top-level item by first            *)
(* elaborating it (which includes inferring its *)
(* type), then evaluating it. The elaborator    *)
(* and evaluator produce strings that           *)
(* respectively represent type and value. If    *)
(* the value string is nonempty, we print both  *)
(* strings. If [[e]] is not well typed, calling *)
(* [[typeof(e, Gamma)]] raises the              *)
(* [[TypeError]] exception, and we never call   *)
(* [[eval]].                                    *)
(* <boxed values 18>=                           *)
val _ = op topCheckEval : toplevel * env_bundle * (string->unit) -> env_bundle
        in  (Gamma, rho)
        end
(* [[topCheckEval]]                             *)
(* preserves the phase distinction: type        *)
(* inference is independent of [[rho]] and      *)
(* [[topeval]].                                 *)

(* <checking and evaluation>=                   *)
and readCheckEvalPrint (reader, echo, errmsg) envs =
  let fun loop envs =
        let fun continue msg = (errmsg msg; loop envs)
            fun finish() = envs
        in  loop (topCheckEval (readtop reader, envs, echo))
            handle EOF => finish()
                (* <more read-eval-print handlers>=             *)
                | TypeError          msg => continue ("type error: " ^ msg)
                | BugInTypeInference msg => continue ("bug in type inference: "
                                                                          ^ msg)
                (* The handlers in the first group deal with    *)
                (* problems that arise during I/O, lexical      *)
                (* analysis, and parsing.                       *)
                (* <more read-eval-print handlers>=             *)
                | IO.Io {name, ...} => continue ("I/O error: " ^ name)
                | Paren (_, loc)    => continue ("syntax error: extra ) in " ^
                                                               srclocString loc)
                | PrematureEOF s    => (errmsg ("Missing )? Got EOF reading (" ^
                                                                   s); finish())
                | SyntaxError msg   => continue ("syntax error: " ^ msg)
                (* [[IO.Io]] is part of the                     *)
                (* Standard Basis Library.                      *)

                (* The handlers in the second group deal with   *)
                (* problems that arise during evaluation. They  *)
                (* always continue with the next top-level      *)
                (* item.                                        *)
                (* <more read-eval-print handlers>=             *)
                | Div               => continue "Division by zero"
                | Overflow          => continue "Arithmetic overflow"
                | RuntimeError msg  => continue ("run-time error: " ^ msg)
                | NotFound n        => continue ("variable " ^ n ^ " not found")
                (* [[Div]] and [[Overflow]] are                 *)
                (* predefined.                                  *)

        end
  in  loop envs
  end
(*   The read-eval-print loop                   *)
(*                                              *)
(* The read-eval-print loop almost identicial   *)
(* to the read-eval-print loop for Typed        *)
(* uScheme; the only difference is that instead *)
(* of [[BugInTypeChecking]], we have and        *)
(* [[BugInTypeInference]].                      *)
(* <boxed values 19>=                           *)
val _ = op readCheckEvalPrint : topreader * (string->unit) * (string->unit) ->
                                                        env_bundle -> env_bundle


(*****************************************************************)
(*                                                               *)
(*   INITIALIZATION                                              *)
(*                                                               *)
(*****************************************************************)

(*   Initializing the interpreter               *)
(*                                              *)
(* We're now ready to put everything together   *)
(* into a working interpreter.                  *)
(* <initialization>=                            *)
val initialEnvs =
  let fun addPrim ((name, prim, tau), (Gamma, rho)) = 
        ( bindtype(name, generalize(tau, Gamma), Gamma)
        , bind (name, PRIMITIVE prim, rho)
        )
      val envs  = foldl addPrim (emptyTypeEnv, emptyEnv) ((* <primops [[::]]>=
                                                                              *)
                                                          ("+", arithOp op +,
                                                                   arithtype) ::
                                                          ("-", arithOp op -,
                                                                   arithtype) ::
                                                          ("*", arithOp op *,
                                                                   arithtype) ::
                                                          ("/", arithOp op div,
                                                                   arithtype) ::
                                                          (* <primops [[::]]>=
                                                                              *)
                                                          ("<", intcompare op <,
                                                            comptype inttype) ::
                                                          (">", intcompare op >,
                                                            comptype inttype) ::
                                                          ("=", comparison (fn (
                                                       NIL,     NIL)     => true
                                                                             | (
                                                    NUM n1,  NUM n2)  => n1 = n2
                                                                             | (
                                                    SYM v1,  SYM v2)  => v1 = v2
                                                                             | (
                                                    BOOL b1, BOOL b2) => b1 = b2
                                                                             |
                                                     _                 => false)
                                                               , comptype alpha)
                                                                              ::
                                                          ("null?", predOp (fn
                        NIL => true | _ => false), predtype (listtype alpha)) ::
                                                          (* The list primitives
                                                     are easy:                *)
                                                          (*
                                                                              *)
                                                          (* <primops [[::]]>=
                                                                              *)
                                                          ("cons", binaryOp (fn
                                                          (a, b) => PAIR(a, b)),

                            funtype([alpha, listtype alpha], listtype alpha)) ::
                                                          ("car",  unaryOp  (fn
                                                           (PAIR(car, _)) => car
                                                                              |
                           NIL => raise RuntimeError "car applied to empty list"
                                                                              |
                     _   => raise BugInTypeInference "car applied to non-list"),

                                            funtype([listtype alpha], alpha)) ::
                                                          ("cdr",  unaryOp  (fn
                                                           (PAIR(_, cdr)) => cdr
                                                                              |
                           NIL => raise RuntimeError "cdr applied to empty list"
                                                                              |
                     _   => raise BugInTypeInference "cdr applied to non-list"),

                                   funtype([listtype alpha], listtype alpha)) ::
                                                          (* The last two
                                              primitives are [[print]] and    *)
                                                          (* [[error]].
                                                                              *)
                                                          (* <primops [[::]]>=
                                                                              *)
                                                          ("print", unaryOp (fn
                                          x => (print (valueString x^"\n"); x)),
                                                                       funtype([
                                                           alpha], unittype)) ::
                                                          ("error", unaryOp (fn
                                       x => raise RuntimeError (valueString x)),
                                                                       funtype([
                                                               alpha], beta)) ::
                                                          (* [[error]],
                                                                              *)
                                                          (* \/alpha,beta\
                                              alldotalpha-->beta, tells us    *)
                                                          (* something
                                           interesting about its behavior.    *)
                                                          (* The type suggests
                                                   that [[error]] can produce *)
                                                          (* an arbitrary beta
                                                   without ever consuming     *)
                                                          (* one. Such a miracle
                                                     is impossible; what the  *)
                                                          (* type tells us is
                                                  that the [[error]] function *)
                                                          (* never returns
                                               normally. In uML this type     *)
                                                          (* means it must
                                               either halt the interpreter or *)
                                                          (* fail to terminate;
                                                    in full ML, a function of *)
                                                          (* this type could
                                                 also raise an exception.     *)
 nil)
      val basis = (* <ML representation of initial basis>=        *)
                    [ "; ml.nw 2289"
                    , "(define list1 (x) (cons x '()))"
                    , "(define bind (x y alist)"
                    , "  (if (null? alist)"
                    , "    (list1 (pair x y))"
                    , "    (if (= x (fst (car alist)))"
                    , "      (cons (pair x y) (cdr alist))"
                    , "      (cons (car alist) (bind x y (cdr alist))))))"
                    , "; ml.nw 2301"
                    , "(define isbound? (x alist)"
                    , "  (if (null? alist) "
                    , "    #f"
                    , "    (if (= x (fst (car alist)))"
                    , "      #t"
                    , "      (isbound? x (cdr alist)))))"
                    , "(define find (x alist)"
                    , "  (if (null? alist) "
                    , "    (error 'not-found)"
                    , "    (if (= x (fst (car alist)))"
                    , "      (snd (car alist))"
                    , "      (find x (cdr alist)))))"
                    , "; mla.nw 93"
                    , "(define caar (l) (car (car l)))"
                    , "(define cadr (l) (car (cdr l)))"
                    , "(define cdar (l) (cdr (car l)))"
                    , "(define length (l)"
                    , "  (if (null? l) 0"
                    , "    (+ 1 (length (cdr l)))))"
                    , "(define and (b c) (if b  c  b))"
                    , "(define or  (b c) (if b  b  c))"
                    , "(define not (b)   (if b #f #t))"
                    , "; mla.nw 103"
                    , "(define append (l1 l2)"
                    , "  (if (null? l1)"
                    , "     l2"
                    , "     (cons (car l1) (append (cdr l1) l2))))"
                    , "(define revapp (l1 l2)"
                    , "  (if (null? l1)"
                    , "     l2"
                    , "     (revapp (cdr l1) (cons (car l1) l2))))"
                    , "; mla.nw 112"
                    , "(define o (f g) (lambda (x) (f (g x))))"
                    , "(define curry   (f) (lambda (x) (lambda (y) (f x y))))"
                    , "(define uncurry (f) (lambda (x y) ((f x) y)))"
                    , "; mla.nw 116"
                    , "(define filter (p? l)"
                    , "  (if (null? l)"
                    , "    '()"
                    , "    (if (p? (car l))"
                    , "      (cons (car l) (filter p? (cdr l)))"
                    , "      (filter p? (cdr l)))))"
                    , "; mla.nw 123"
                    , "(define map (f l)"
                    , "  (if (null? l)"
                    , "    '()"
                    , "    (cons (f (car l)) (map f (cdr l)))))"
                    , "; mla.nw 128"
                    , "(define exists? (p? l)"
                    , "  (if (null? l)"
                    , "    #f"
                    , "    (if (p? (car l)) "
                    , "      #t"
                    , "      (exists? p? (cdr l)))))"
                    , "(define all? (p? l)"
                    , "  (if (null? l)"
                    , "    #t"
                    , "    (if (p? (car l))"
                    , "      (all? p? (cdr l))"
                    , "      #f)))"
                    , "; mla.nw 141"
                    , "(define foldr (op zero l)"
                    , "  (if (null? l)"
                    , "    zero"
                    , "    (op (car l) (foldr op zero (cdr l)))))"
                    , "(define foldl (op zero l)"
                    , "  (if (null? l)"
                    , "    zero"
                    , "    (foldl op (op (car l) zero) (cdr l))))"
                    , "; mla.nw 150"
                    , "(define <= (x y) (not (> x y)))"
                    , "(define >= (x y) (not (< x y)))"
                    , "(define != (x y) (not (= x y)))"
                    , "; mla.nw 154"
                    , "(define max (x y) (if (> x y) x y))"
                    , "(define min (x y) (if (< x y) x y))"
                    , "(define mod (m n) (- m (* n (/ m n))))"
                    , "(define gcd (m n) (if (= n 0) m (gcd n (mod m n))))"
                    , "(define lcm (m n) (* m (/ n (gcd m n))))"
                    , "; mla.nw 160"
                    , "(define min* (l) (foldr min (car l) (cdr l)))"
                    , "(define max* (l) (foldr max (car l) (cdr l)))"
                    , "(define gcd* (l) (foldr gcd (car l) (cdr l)))"
                    , "(define lcm* (l) (foldr lcm (car l) (cdr l)))"
                    , "; mla.nw 165"
                    , "(define list1 (x)               (cons x '()))"
                    , "(define list2 (x y)             (cons x (list1 y)))"
                    , "(define list3 (x y z)           (cons x (list2 y z)))"
                    , "(define list4 (x y z a)         (cons x (list3 y z a)))"
                    ,
                     "(define list5 (x y z a b)       (cons x (list4 y z a b)))"
                    ,
                   "(define list6 (x y z a b c)     (cons x (list5 y z a b c)))"
                    ,
                 "(define list7 (x y z a b c d)   (cons x (list6 y z a b c d)))"
                    ,
               "(define list8 (x y z a b c d e) (cons x (list7 y z a b c d e)))"
                    , "; mla.nw 174"
                    , "(define takewhile (p? l)"
                    , "  (if (null? l)"
                    , "     '()"
                    , "     (if (p? (car l))"
                    , "         (cons (car l) (takewhile p? (cdr l)))"
                    , "         '())))"
                    , "(define dropwhile (p? l)"
                    , "  (if (null? l)"
                    , "     '()"
                    , "     (if (p? (car l))"
                    , "         (dropwhile p? (cdr l))"
                    , "         l)))"
                    ]
      val treader = topreader (stringsreader ("initial basis", basis), false)
  in  readCheckEvalPrint (treader, fn _ => (), fn _ => ()) envs
  end
(* The function [[runInterpreter]] takes one    *)
(* argument, which tells it whether to prompt.  *)
(* <initialization>=                            *)
fun runInterpreter noisy = 
  let fun writeln s = app print [s, "\n"]
      val treader = topreader (filereader ("standard input", TextIO.stdIn),
                                                                          noisy)
  in  ignore (readCheckEvalPrint (treader, writeln, writeln) initialEnvs)
  end 


(*****************************************************************)
(*                                                               *)
(*   COMMAND LINE                                                *)
(*                                                               *)
(*****************************************************************)

(* Building and exporting a program             *)
(*                                              *)
(* The final step in implementing the           *)
(* interpreter is to create a function that     *)
(* looks at the command line and calls          *)
(* [[runInterpreter]]. With a compiler like     *)
(* Moscow ML or MLton, the module isn't         *)
(* elaborated until run time, so we can simply  *)
(* insert an irrelevant [[val]] binding, the    *)
(* elaboration of which has the side effect of  *)
(* calling [[main]].                            *)
(* [[CommandLine.arguments()]] returns an       *)
(* argument list.                               *)
(* <command line>=                              *)
fun main ["-q"] = runInterpreter false
  | main []     = runInterpreter true
  | main _      =
      TextIO.output(TextIO.stdErr, "Usage: " ^ CommandLine.name() ^ " [-q]\n")
val _ = main (CommandLine.arguments())
