(*   Putting the pieces together                *)
(*                                              *)
(* We stitch together the parts of the          *)
(* implementation in this order:                *)
(* <upr.sml>=                                   *)


(*****************************************************************)
(*                                                               *)
(*   ENVIRONMENTS                                                *)
(*                                                               *)
(*****************************************************************)

(* <environments>=                              *)
type name = string
type 'a env = (name * 'a) list
val emptyEnv = []

(* lookup and assignment of existing bindings *)
exception NotFound of name
fun find (name, []) = raise NotFound name
  | find (name, (n, v)::tail) = if name = n then v else find(name, tail)

(* adding new bindings *)
exception BindListLength
fun bind(name, v, rho) = (name, v) :: rho
fun bindList(n::vars, v::vals, rho) = bindList(vars, vals, bind(n, v, rho))
  | bindList([], [], rho) = rho
  | bindList _ = raise BindListLength
(* We represent an environment of type [['a     *)
(* env]] as a list of ([[name]], [['a]]) pairs. *)
(* The declarations in the box give the         *)
(* interface to our implementation; through     *)
(* some Noweb hackery, they are checked by the  *)
(* ML compiler.                                 *)
(* <boxed values 11>=                           *)
val _ = op emptyEnv : 'a env
val _ = op find     : name * 'a env -> 'a
val _ = op bind     : name      * 'a      * 'a env -> 'a env
val _ = op bindList : name list * 'a list * 'a env -> 'a env
(* [[error]] procedure,                         *)
(* to indicate when things have gone wrong. The *)
(* exceptions we use are listed in Table [->].  *)

(* Tracing code is helpful for debugging.       *)
(*                                              *)
(* <environments>=                              *)
val tracer = ref (app print)
val _ = tracer := (fn _ => ())
fun trace l = !tracer l


(*****************************************************************)
(*                                                               *)
(*   ABSTRACT SYNTAX AND VALUES                                  *)
(*                                                               *)
(*****************************************************************)

(*   Abstract syntax (and no values)            *)
(*                                              *)
(* Prolog has the simplest structure of all the *)
(* languages in this book. Unusually, there is  *)
(* no distinction between ``values'' and        *)
(* ``abstract syntax;'' both are represented as *)
(* terms. A term is a variable, a literal       *)
(* constant, or an application of a functor to  *)
(* a list of terms. [Functors play the same     *)
(* role in \prolog\ that datatype constructors  *)
(* play in~\ml, including participation in      *)
(* pattern matching.] The only kinds of literal *)
(* constants in uProlog are integers.           *)
(* <abstract syntax and values>=                *)
datatype term = VAR     of string
              | LITERAL of int
              | APPLY   of string * term list
(* Goals and applications have identical        *)
(* structure. A term can be a functor applied   *)
(* to a list of terms; a goal is a predicate    *)
(* applied to a list of terms. A ``functor'' or *)
(* a ``predicate'' is uniquely identified by    *)
(* its name.                                    *)
(* <abstract syntax and values>=                *)
type goal = string * term list
(* A clause is a conclusion and a list of       *)
(* premises, all of which are goals. If the     *)
(* list of premises is empty, the clause is a   *)
(* ``fact;'' otherwise it is a ``rule,'' but    *)
(* these distinctions are useful only for       *)
(* thinking about and organizing programs--the  *)
(* underlying meanings are the same. Writing    *)
(* our implementation in ML enables us to use   *)
(* the identifier [[:-]] as a datatype          *)
(* constructor for clauses; we even make it     *)
(* infix.                                       *)
(* <abstract syntax and values>=                *)
datatype clause = :- of goal * goal list
infix 3 :-
(* At the top level, we can do three things:    *)
(* add a clause to the database, ask a query,   *)
(* or include a file. The switch back and forth *)
(* between query mode and rule mode is hidden   *)
(* from the code in this chapter; the details   *)
(* are buried in Section [->].                  *)
(* <abstract syntax and values>=                *)
datatype toplevel
  = ADD_CLAUSE of clause
  | QUERY      of goal list
  | USE        of string
(* <abstract syntax and values>=                *)
(* This Appendix is longer than many others     *)
(* because even Prolog's simple syntax requires *)
(* more code to parse than prefix-parenthesized *)
(* syntax. The modes of the uProlog interpreter *)
(* introduce additional complexity.             *)
(*                                              *)
(* String conversions                           *)
(*                                              *)
(* This code converts terms, goals, and clauses *)
(* to strings.                                  *)
(* <string conversions>=                        *)
fun termString (APPLY ("cons", [car, cdr])) = 
      let fun tail (APPLY("cons", [car, cdr])) = ", " ^ termString car ^ tail
                                                                             cdr
            | tail (APPLY("nil",  []))         = "]"
            | tail x                           = "|" ^ termString x ^ "]"
      in  "[" ^ termString car ^ tail cdr
      end
  | termString (APPLY ("nil", [])) = "[]"
  | termString (APPLY (f, []))     = f
  | termString (APPLY (f, [x, y])) =
      if Char.isAlpha (hd (explode f)) then appString f x [y]
      else String.concat ["(", termString x, " ", f, " ", termString y, ")"]
  | termString (APPLY (f, h::t)) = appString f h t
  | termString (VAR v) = v
  | termString (LITERAL n) = String.map (fn #"~" => #"-" | c => c) (Int.toString
                                                                              n)
and appString f h t =
      String.concat (f :: "(" :: termString h ::
                     foldr (fn (t, tail) => ", " :: termString t :: tail) [")"]
                                                                              t)
(* <string conversions>=                        *)
fun goalString g = termString (APPLY g)
fun clauseString (g :- []) = goalString g
  | clauseString (g :- (h :: t)) =
      String.concat (goalString g :: " :- " :: goalString h ::
                     (foldr (fn (g, tail) => ", " :: goalString g :: tail)) [] t
                                                                               )
(*   There is no ``evaluation.'' [Well, hardly  *)
(* any. The primitive [[is]] does a tiny amount *)
(* of evaluation.] Instead, we have queries.    *)
(* The main features of the implementation are  *)
(* the database, substitution and unification   *)
(* (which are left as Exercises [->] and [->]), *)
(* and the backtracking query engine. To print  *)
(* queries and answers, we use code from        *)
(* Appendix [->].                               *)
(* <boxed values 1>=                            *)
val _ = op termString   : term   -> string
val _ = op goalString   : goal   -> string
val _ = op clauseString : clause -> string


(*****************************************************************)
(*                                                               *)
(*   CLAUSE DATABASE                                             *)
(*                                                               *)
(*****************************************************************)

(* <clause database>=                           *)
type database = clause list
val emptyDatabase = []
fun addClause (r, rs) = rs @ [r] (* must maintain order *)
fun potentialMatches (_, rs) = rs


(*****************************************************************)
(*                                                               *)
(*   SUBSTITUTION AND UNIFICATION                                *)
(*                                                               *)
(*****************************************************************)

(* <substitution and unification>=              *)
(* <substitution on terms>=                     *)
exception LeftAsExercise
type subst = term -> term
infix 7 |-->
fun name |--> term = raise LeftAsExercise
fun idsubst term = term
(* Our implementation uses a simple list. We    *)
(* treat every clause as a potential match.     *)
(* <boxed values 2>=                            *)
type database = database
val _ = op emptyDatabase    : database
val _ = op addClause        : clause * database -> database
val _ = op potentialMatches : goal * database -> clause list
(*  Substitution                                *)
(*                                              *)
(* A substitution \subsn is a function from     *)
(* terms to terms. We present only prototype    *)
(* code for substitutions; the full             *)
(* implementation is left as Exercise [->].     *)
(* <boxed values 2>=                            *)
type subst = subst
val _ = op idsubst : subst
val _ = op |-->    : name * term -> subst
(* <substitution and unification>=              *)
fun lift       theta (f, l)    = (f, map theta l)
fun liftClause theta (c :- ps) = (lift theta c :- map (lift theta) ps)
(* <boxed values 3>=                            *)
val _ = op lift       : subst -> (goal   -> goal)
val _ = op liftClause : subst -> (clause -> clause)
(* <substitution and unification>=              *)
type 'a set = 'a list
val emptyset = []
fun member x = 
  List.exists (fn y => y = x)
fun insert (x, l) = 
  if member x l then l else x::l
fun diff  (s1, s2) = 
  List.filter (fn x => not (member x s2)) s1
fun union (s1, s2) = s1 @ diff (s2, s1)   (* preserves order *)
(*  Sets of variables                           *)
(*                                              *)
(* As in Chapter [->], we use sets of variables *)
(* in unification. We provide a simple          *)
(* implementation using lists. The double prime *)
(* in the type variable [[''a]] means that      *)
(* [[''a]] has to be an ``equality type,''      *)
(* i.e., variable [[''a]] may be instantiated   *)
(* only at types that admit equality.           *)
(* <boxed values 4>=                            *)
type 'a set = 'a set
val _ = op emptyset : 'a set
val _ = op member   : ''a -> ''a set -> bool
val _ = op insert   : ''a     * ''a set  -> ''a set
val _ = op union    : ''a set * ''a set  -> ''a set
val _ = op diff     : ''a set * ''a set  -> ''a set
(* <substitution and unification>=              *)
fun freevars t =
  let fun f(VAR v,     l) = insert (v, l)
        | f(LITERAL _, l) = l
        | f(APPLY(_, args), l) = foldl f l args
  in  rev (f(t, []))
  end  
fun freevarsGoal goal = freevars (APPLY goal)
fun freevarsClause (c :- ps) =
  foldl (fn (p, f) => union (freevarsGoal p, f)) (freevarsGoal c) ps
(* The function [[freevars]] computes the free  *)
(* variables of a term. For readability, we     *)
(* ensure that variables appear in the set in   *)
(* the order of their first appearance in the   *)
(* type, when reading from left to right. We    *)
(* provide similar functions on goals and       *)
(* clauses.                                     *)
(* <boxed values 5>=                            *)
val _ = op freevars       : term   -> name set
val _ = op freevarsGoal   : goal   -> name set
val _ = op freevarsClause : clause -> name set
(* <substitution and unification>=              *)
local
  val n = ref 1
in
  fun freshVar s = VAR ("_" ^ s ^ Int.toString (!n) before n := !n + 1)
(*  ``Freshening''                              *)
(*                                              *)
(* Every time we use a clause, the semantics of *)
(* Prolog require that we rename its variables  *)
(* to fresh variables. To rename a variable, we *)
(* put an underscore in front of its name and a *)
(* unique integer after it. Because the parser  *)
(* in Section [->] does not accept variables    *)
(* whose names begin with an underscore, these  *)
(* names cannot possibly collide with the names *)
(* of variables written by users.               *)
(* <boxed values 6>=                            *)
val _ = op freshVar : string -> term
end
(* <substitution and unification>=              *)
fun freshen c =
  let val free     = freevarsClause c
      val fresh    = map freshVar free
      val renaming =
        ListPair.foldl (fn (free, fresh, theta) => theta o (free |--> fresh))
        idsubst (free, fresh) 
  in  liftClause renaming c
  end
(* <substitution and unification>=              *)
(* <unification>=                               *)
exception Unify
fun unifyTerm (t1, t2) = 
  raise LeftAsExercise
and unifyList (l1, l2) = 
  raise LeftAsExercise
fun unify ((f, args), (f', args')) =
  if f = f' then unifyList (args, args') else raise Unify
(* Function [[freshen]] replaces free variables *)
(* with fresh variables. Value [[renaming]]     *)
(* represents a renaming \renaming, as in       *)
(* Section [->].                                *)
(* <boxed values 7>=                            *)
val _ = op freshen : clause -> clause
(*  Unification                                 *)
(*                                              *)
(* Given goals g and G, unify(g, G) returns a   *)
(* substitution \subsn such that                *)
(* \wh\subsn(g)=\wh\subsn(G), or if no such     *)
(* substitution exists, it raises [[Unify]].    *)
(* The implementation of unification is left as *)
(* Exercise [->]. [*]                           *)
(* <boxed values 7>=                            *)
val _ = op unify     : goal      * goal      -> subst
val _ = op unifyTerm : term      * term      -> subst
val _ = op unifyList : term list * term list -> subst


(*****************************************************************)
(*                                                               *)
(*   LEXICAL ANALYSIS                                            *)
(*                                                               *)
(*****************************************************************)

(* Lexical analysis                             *)
(*                                              *)
(* uProlog has a more complex lexical structure *)
(* than other languages. We have uppercase,     *)
(* lowercase, and symbolic tokens, as well as   *)
(* integers. Finally, it simplifies the parser  *)
(* if we distinguish reserved words and symbols *)
(* using [[RESERVED]].                          *)
(* <lexical analysis>=                          *)
datatype token 
  = UPPER     of string
  | LOWER     of string
  | SYMBOLIC  of string
  | INT_TOKEN of int
  | RESERVED  of string
(* We need to print tokens in error messages.   *)
(*                                              *)
(* <lexical analysis>=                          *)
fun tokenString (UPPER s)     = s
  | tokenString (LOWER s)     = s
  | tokenString (INT_TOKEN n) = Int.toString n
  | tokenString (SYMBOLIC  s) = s
  | tokenString (RESERVED  s) = s
(* Every character is either a symbol, an       *)
(* alphanumeric, a space, or a delimiter.       *)
(* <lexical analysis>=                          *)
val symbols = explode "!%^&*-+:=|~<>/?`$\\"   (* ask Russ Cox why there is a $
                                                                         here *)
fun isSymbol c = List.exists (fn c' => c' = c) symbols
fun isIdent  c = Char.isAlphaNum c orelse c = #"_"
fun isDelim  c = not (isIdent c orelse isSymbol c)
(* Symbolic and lower-case tokens are usually   *)
(* symbolic, but sometimes they are reserved    *)
(* words, and the cut is actually treated as an *)
(* ordinary predicate, because it is nullary,   *)
(* not binary.                                  *)
(* <lexical analysis>=                          *)
fun symbolic ":-" = RESERVED ":-"
  | symbolic "."  = RESERVED "."
  | symbolic "|"  = RESERVED "|"
  | symbolic "!"  = LOWER "!"
  | symbolic s    = SYMBOLIC s
fun lower "is" = RESERVED "is"
  | lower s    = LOWER s
(* A variable consisting of a single underscore *)
(* gets converted to a unique ``anonymous''     *)
(* variable.                                    *)
(* <lexical analysis>=                          *)
fun anonymousVar () =
  case freshVar ""
    of VAR v => UPPER v
     | _ => let exception ThisCan'tHappen in raise ThisCan'tHappen end
(* <lexical analysis>=                          *)
type srcloc = string * int
fun srclocString (source, line) = source ^ ", line " ^ Int.toString line
exception Paren of char list * srcloc (* raise if found ) when reading input *)
exception PrematureEOF of string (* shows what we were reading ... *)
exception EOF                    (* raised by reader *)
exception SyntaxError of string
fun read nextline curloc : string -> token list =
  let fun readmap f (result, unconsumed) = (f result, unconsumed)
      (* get is called only when s is guaranteed non-blank, non-comment, non-
                                                                        empty *)
      (* To help with reading, we have functions to   *)
      (* deal with lexical details. We begin by       *)
      (* defining tests for the delimiters of         *)
      (* symbols:                                     *)
      (* <character-fiddling functions>=              *)
      fun isDelim c = List.exists (fn c' => c' = c) [#"(", #")", #" ", #"\t",
                                                                    #"\n", #";"]
      (* [[]] is like the micro-Scheme                *)
      (* [[exists?]]. We needn't write its            *)
      (* implementation because it is in the initial  *)
      (* basis for Standard ML.                       *)

      (* <character-fiddling functions>=              *)
      fun stringReader []     = NONE
        | stringReader (h::t) = SOME (h, t)
      val strton = Int.scan StringCvt.DEC stringReader
      (* We continue with integer conversion;         *)
      (* [[strton s]] returns an integer, plus        *)
      (* leftover characters from [[s]], or [[NONE]]  *)
      (* if no prefix of [[s]] is an integer literal. *)
      (* It uses [[Int.scan]] from the Standard Basis *)
      (* Library.                                     *)
      (* <boxed values 22>=                           *)
      val _ = op strton : char list -> (int * char list) option
      (* We need to be able to skip whitespace and    *)
      (* comments. [[StringCvt.skipWS]] skips white   *)
      (* space; if we encounter a semicolon, we use   *)
      (* [[StringCvt.dropl]] to skip everything up to *)
      (* the next newline, then resume skipping white *)
      (* space.                                       *)
      (* <character-fiddling functions>=              *)
      fun skip s =
        case StringCvt.skipWS stringReader s
          of #";" :: s => skip (StringCvt.dropl (fn c => c <> #"\n")
                                                                 stringReader s)
           | s => s
      (* [[nonwhite]], which                          *)
      (* summons more input lines as needed.          *)

      (* <character-fiddling functions>=              *)
      fun nonwhite s =
        case skip s
          of [] => nonwhite (explode (nextline()))
           | s  => s
      fun isDelim  c = not (Char.isAlphaNum c orelse isSymbol c)
      exception Empty
      (*arg to get may be empty, but never begins with white space*)
      fun get [] = raise Empty
        | get (#"_"::nil)  = (anonymousVar (), nil)
        | get (chars as #"_"::c::s) = 
            if isIdent c then
              raise SyntaxError ("names may not begin with underscores at " ^
                                                                  implode chars)
            else
              (anonymousVar (), c::s)
        | get (#"/" :: #"*" :: s) = skipComment s
        | get (#"-" :: c :: s) = 
            if Char.isDigit c then
                atom(c::s, [#"-"], implodeInt, Char.isDigit)
            else
                atom(#"-" :: c :: s, [], symbolic o implode, isSymbol)
        | get (c::s) =
            if      isDelim       c then (RESERVED (str c), s)
            else if Char.isLower  c then atom(c::s, [], lower    o implode,
                                                                        isIdent)
            else if Char.isUpper  c then atom(c::s, [], UPPER    o implode,
                                                                        isIdent)
            else if isSymbol      c then atom(c::s, [], symbolic o implode,
                                                                       isSymbol)
            else if Char.isDigit  c then atom(c::s, [], implodeInt,
                                                                   Char.isDigit)
            else raise SyntaxError ("invalid initial character in `" ^ implode (
                                                                    c::s) ^ "'")
      and atom(c::s, r, mkAtom, continue) = 
            if continue c then atom(s, c::r, mkAtom, continue)
            else (mkAtom(rev r), c::s) 
        | atom([],   r, mkAtom, continue) = (mkAtom(rev r), [])
      and implodeInt s =
        case strton s of SOME (n, []) => INT_TOKEN n
                       | _ => raise SyntaxError "this can't happen"
      and skipComment (#"*" :: #"/" :: s) = get (skip s)
        | skipComment (c :: s) = skipComment s
        | skipComment [] = skipComment (explode (nextline()))
      fun readline [] = []
        | readline s  = 
             let val(p, u) = get s
             in  p :: readline(skip u)
             end handle Empty => []
  in  fn s => readline(skip (explode s)) 
  end
(* The lexer converts a string to a sequence of *)
(* tokens. Unlike the other lexers, it calls    *)
(* [[nextline]] only if it is in the middle of  *)
(* skipping a C-style comment.                  *)
(* <boxed values 27>=                           *)
val _ = op read : (unit -> string) -> (unit -> srcloc) -> string -> token list


(*****************************************************************)
(*                                                               *)
(*   PARSING                                                     *)
(*                                                               *)
(*****************************************************************)

(* Syntactically, the concrete syntax used at   *)
(* top level includes a string in brackets, a   *)
(* clause, or a query with multiple goals. The  *)
(* modes are necessary to distinguish           *)
(* [[CLAUSE(g, NONE)]] from [[GOALS [g]]],      *)
(* which are otherwise indistinguishable.       *)
(* <parsing>=                                   *)
datatype concrete
  = BRACKET of string 
  | CLAUSE  of goal * goal list option
  | GOALS   of goal list
(* <parsing>=                                   *)
(* the following definition exists only to make unused code in mlschemea.nw
                                                                      compile *)
fun toplevel _ = let exception Unimp in raise Unimp end
(* <parsing>=                                   *)
fun synerror (t, nt, s) = 
  let val msg = foldr (fn (t, l) => " " :: tokenString t :: l) ("; " :: s) t
      val msg =  "while parsing " :: nt :: " found" :: msg
  in  raise SyntaxError (concat msg)
  end
(* Parsing                                      *)
(*                                              *)
(* [*]                                          *)
(*                                              *)
(* Before we introduce the parser itself, here  *)
(* are a few utility functions. We call         *)
(* [[synerror]] if we see an unexpected token   *)
(* while parsing.                               *)
(* <boxed values 28>=                           *)
val _ = op synerror : token list * string * string list -> 'a
(* <parsing>=                                   *)
local 
  (* To implement suspensions, we use a standard, *)
  (* tricky combination of imperative and         *)
  (* functional code. We also slip in a Standard  *)
  (* ML module, since this is an appendix and we  *)
  (* hope nobody is looking.                      *)
  (* <lazy lists>=                                *)
  structure Susp = struct
    datatype 'a contents = THUNK of unit -> 'a | FORCED of 'a
    type 'a susp = 'a contents ref
    fun delay f = ref (THUNK f)
    fun force cell =
      case !cell
        of THUNK f  => let val v = f() in v before cell := FORCED v end
         | FORCED v => v
    fun valueIn cell =
      case !cell
        of THUNK f  => NONE
         | FORCED v => SOME v
  end
  (* <lazy lists>=                                *)
  datatype 'a llist 
    = ::: of 'a * 'a llist
    | NEED_MORE of 'a llist Susp.susp
  infixr 3 :::
  fun fromLazy (h ::: t) = h :: fromLazy t
    | fromLazy (NEED_MORE l) = fromLazy(valOf(Susp.valueIn l)) handle Option =>
                                                                              []
  fun toLazy more items =
    foldr op ::: (NEED_MORE (Susp.delay (fn () => toLazy more (more())))) items
  (* <lazy lists>=                                *)
  fun lforce f (NEED_MORE t) = lforce f (Susp.force t)
    | lforce f tokens        = f tokens 
  (* <general-purpose parsing combinators>=       *)
  type 'a parser = token llist -> 'a * token llist
  (* A lazy list is never empty--if we run out of *)
  (* tokens, we can always read more input. A     *)
  (* lazy list is either a lazy cons cell         *)
  (* (written with infix [[:::]]) or it is a      *)
  (* suspension holding more list elements. As    *)
  (* noted above, [[toLazy]] and [[fromLazy]] are *)
  (* conversion functions. Function [[fromLazy]]  *)
  (* never applies a thunk; instead it treats a   *)
  (* suspension as the empty list.                *)
  (* <boxed values 30>=                           *)
  val _ = op fromLazy : 'a llist -> 'a list
  val _ = op toLazy   : (unit -> 'a list) -> 'a list -> 'a llist
  (* A value of type [['a parser]] is a function  *)
  (* that takes a token list and returns a value  *)
  (* of type [['a]], plus any leftover tokens.    *)
  (* <boxed values 30>=                           *)
  type 'a parser = 'a parser
  (* <general-purpose parsing combinators>=       *)
  exception E of string * string * token llist
  fun fail nt why toks = raise E(nt, why, toks)

  infix 0 ||
  fun (p1 || p2) toks = p1 toks handle E _ => p2 toks
  (* Internally, each parser uses exception [[E]] *)
  (* to indicate that it can't handle a           *)
  (* particular token. This internal exception is *)
  (* used for control flow; not just for errors.  *)
  (* The parser combinator [[||]] uses [[E]] to   *)
  (* try first one alternative, then another.     *)
  (* <boxed values 31>=                           *)
  val _ = op ||  : 'a parser * 'a parser -> 'a parser
  (* [[]] before raising [[E]] and moving         *)
  (* to [[p2]]. This potential for wasted work    *)
  (* in [[p1]] is the reason combinator-based     *)
  (* parsers can be less efficient than           *)
  (* automatically generated parsers, or even     *)
  (* other kinds of hand-written parsers.         *)

  (* <general-purpose parsing combinators>=       *)
  fun !! p tokens = 
    p tokens handle E (nt, why, tokens) =>
               synerror(fromLazy tokens, nt, [why])
  (* The [[!!]] combinator cuts off the           *)
  (* backtracking, turning the private            *)
  (* exception [[E]] into a syntax error.         *)
  (* <boxed values 32>=                           *)
  val _ = op !!  : 'a parser -> 'a parser
  (* <general-purpose parsing combinators>=       *)
  infix 5 --
  fun (p1 -- p2) tokens = 
    let val (x, tokens) = p1 tokens
        val (y, tokens) = p2 tokens
    in  ((x, y), tokens)
    end
  (* The other basic combinator is [[-]], which   *)
  (* parses two phrases in sequence and returns a *)
  (* pair.                                        *)
  (* <boxed values 33>=                           *)
  val _ = op -- : 'a parser * 'b parser -> ('a * 'b) parser
  (* <general-purpose parsing combinators>=       *)
  fun $a (ts as RESERVED b ::: tokens) = 
         if a = b then (a, tokens) else fail "unknown" ("expected " ^ a) ts
    | $a ts = fail "unknown" ("expected " ^ a) ts
  val $ = lforce o $
  (* <general-purpose parsing combinators>=       *)
  infix 3 >>
  fun (p >> f) tokens = 
    let val (x, tokens) = p tokens
    in  (f x, tokens)
    end
  (* These basic combinators all produce strings  *)
  (* or integers. Quite often it's better to      *)
  (* apply a function to these strings and        *)
  (* integers. The [[>>]] combinator is function  *)
  (* composition for parsers.                     *)
  (* <boxed values 35>=                           *)
  val _ = op >> : 'a parser * ('a -> 'b) -> 'b parser
  (* <general-purpose parsing combinators>=       *)
  infix 6 $-- --$
  fun (a $-- p) = ($a -- !!p    >> #2)
  fun (p --$ a) = (p  -- !!($a) >> #1)
  (* It's typical for reserved words to be used   *)
  (* only for recognition; for example, after we  *)
  (* recognize a token such as [[RESERVED ":-"]], *)
  (* we probably throw away the string [[":-"]].  *)
  (* Because reserved words are normally used in  *)
  (* sequence with other phrases, it's a nuisance *)
  (* to keep the strings around and ignore them   *)
  (* later. We define two combinators that make   *)
  (* it convenient to ignore reserved words; they *)
  (* have higher precedence than [[-]] so they    *)
  (* can ``pull in'' strings that appear in       *)
  (* sequences.                                   *)
  (* <boxed values 36>=                           *)
  val _ = op $-- : string    * 'a parser -> 'a parser
  val _ = op --$ : 'a parser * string    -> 'a parser
  (* [[--]] combinator cuts off backtracking      *)
  (* after it recognizes the reserved word; the   *)
  (* [[--]] combinator insists on getting the     *)
  (* expected word.                               *)

  (* <general-purpose parsing combinators>=       *)
  fun nothingAs v tokens = (v, tokens)
  fun repeat p tokens = 
    (p -- lforce (repeat p) >> op :: || nothingAs []) tokens
  fun commas p tokens = 
    (p -- ("," $-- lforce (commas p) || nothingAs []) >> op ::) tokens
  (* So far, we can parse only a sequence of      *)
  (* fixed size. Combinator [[repeat]] parses     *)
  (* zero or more repetitions of a phrase, and    *)
  (* [[commas]] parses one or more repetitions    *)
  (* separated by commas. These combinators are   *)
  (* ``greedy;'' that is, they consume as many    *)
  (* tokens as possible; when no more tokens can  *)
  (* be read, they parse nothing (the empty       *)
  (* phrase) using [[nothingAs]].                 *)
  (* <boxed values 37>=                           *)
  val _ = op nothingAs : 'a -> 'a parser
  val _ = op repeat    : 'a parser -> 'a list parser
  val _ = op commas    : 'a parser -> 'a list parser
  (* [[repeat]] and [[commas]]                    *)
  (* terminate. As it stands, if [[p]] parses the *)
  (* empty phrase, [[repeat p]] does not          *)
  (* terminate, so it is an unchecked run-time    *)
  (* error to pass [[repeat]] a parser that       *)
  (* accepts the empty phrase.                    *)

  (* <general-purpose parsing combinators>=       *)
  fun optional p = 
    p >> SOME || nothingAs NONE
  (* Finally, many grammars contain optional      *)
  (* phrases, for which we want to parse zero or  *)
  (* one instances, but no more.                  *)
  (* <boxed values 38>=                           *)
  val _ = op optional : 'a parser -> 'a option parser
  (* <combinators for parsing \uprolog>=          *)
  fun symbol (SYMBOLIC s:::tokens) = (s, tokens)
    | symbol tokens = fail "symbol" "symbol" tokens
  fun upper (UPPER a ::: tokens) = (a, tokens)
    | upper tokens = 
        fail "ident" "Identifier expected" tokens
  fun lower (LOWER a ::: tokens) = (a, tokens)
    | lower tokens = fail "atom" "functor or predicate expected" tokens
  fun int (INT_TOKEN a ::: tokens) = (a, tokens)
    | int tokens = fail "atom" "functor or predicate expected" tokens
  (* Here we rename the basic parsers so they are *)
  (* forced to use nonempty lists.                *)
  (* <combinators for parsing \uprolog>=          *)
  val (lower, upper, int, symbol) = 
    (lforce lower, lforce upper, lforce int, lforce symbol)
  (* The combinators [[||]] and [[-]] take        *)
  (* existing parsers and build more powerful     *)
  (* parsers, but how are we to get a parser at   *)
  (* all? We write one ``parser'' for each kind   *)
  (* of token; it recognizes exactly that token,  *)
  (* no more and no less. Parsers [[symbol]],     *)
  (* [[upper]], [[lower]], and [[int]] represent  *)
  (* symbolic identifiers, upper-case             *)
  (* identifiers, lower-case identifiers, and     *)
  (* integers, respectively. For any              *)
  (* string [[s]], parser [[s]] recognizes [[s]]  *)
  (* as a reserved word.                          *)
  (* <boxed values 34>=                           *)
  val _ = op symbol : string parser
  val _ = op upper  : string parser
  val _ = op lower  : string parser
  val _ = op int    : int    parser
  val _ = op $ : string -> string parser
  (* Each of the basic parsers above must be used *)
  (* only on a nonempty list of tokens. We        *)
  (* therefore provide a higher-order function    *)
  (* [[lforce]], which accepts a function on a    *)
  (* lazy list and ensures it is applied only to  *)
  (* a nonempty lazy list.                        *)
  (* <boxed values 34>=                           *)
  val _ = op lforce   : ('a llist -> 'b) -> ('a llist -> 'b)
  (* We use these combinators to define the       *)
  (* grammar from Figure [->]. We use             *)
  (* [[notSymbol]] to ensure that a term like [[3 *)
  (* + X]] is not followed by another symbol.     *)
  (* This means we don't parse such terms as [[3  *)
  (* + X + Y]].                                   *)
  (* <combinators for parsing \uprolog>=          *)
  fun notSymbol (tokens as SYMBOLIC s ::: _) =
        synerror(fromLazy tokens, "expression",
                 ["arithmetic expressions must be parenthesized"])
    | notSymbol tokens = ((), tokens)
  val notSymbol = lforce notSymbol : unit parser
  (* <combinators for parsing \uprolog>=          *)
  val nilt = nothingAs (APPLY("nil", []))
  fun cons (x, y) = APPLY("cons", [x, y])
  (* Parser [[nilt]] uses the empty list of       *)
  (* tokens to represent the empty list of terms. *)
  (* It needs an explicit type constraint to      *)
  (* avoid falling afoul of the value restriction *)
  (* on polymorphism. Function [[cons]] combines  *)
  (* two terms, which is useful for parsing       *)
  (* lists.                                       *)
  (* <boxed values 39>=                           *)
  val _ = op nilt : term parser
  val _ = op cons : term * term -> term
  (* <combinators for parsing \uprolog>=          *)
  val functr          = lower  (* `functor' is a reserved word *)
  val variable        = upper
  val binaryPredicate = symbol
  fun term tokens = lforce
    (   "[" $-- (commas term -- ("|" $-- !!term || nilt)
                || nothingAs [] -- nilt
                ) --$ "]"           >> (fn (elems, tail) => foldr cons tail
                                                                          elems)
    ||  atom -- "is"  $-- !!term    >> (fn (a, t) => APPLY("is", [a, t]))
    ||  atom -- binaryPredicate -- !!atom -- !!notSymbol 
                                    >> (fn (((l, f), r), _) => APPLY(f, [l, r]))
    ||  atom 
    ) 
    tokens
  and atom tokens = lforce
    (   functr -- ("(" $-- commas (!!term) --$ ")" || nothingAs [])  >> APPLY
    ||  variable >> VAR
    ||  int      >> LITERAL
    ||  "(" $-- term --$ ")"
    ||  fail "" "expected term"
    )
    tokens
  (* The combinators above provide the technology *)
  (* we need to parse uProlog. The grammar is     *)
  (* based on the grammar from Figure [->],       *)
  (* except that we've named an [[atom]]          *)
  (* nonterminal and used some specialized tricks *)
  (* to organize the grammar. Concrete syntax is  *)
  (* not for the faint of heart.                  *)
  (* <boxed values 40>=                           *)
  val _ = op term : term parser
  val _ = op atom : term parser
  (* <combinators for parsing \uprolog>=          *)
  fun toGoal (APPLY g) = g
    | toGoal (VAR v)   = 
        raise SyntaxError ("Variable " ^ v ^ " cannot be a predicate")
    | toGoal (LITERAL n) =
        raise SyntaxError ("Integer " ^ Int.toString n ^
                                                       " cannot be a predicate")
  val goal = (term || fail "" "expected goal") >> toGoal : goal parser
  (* Terms and goals shared the same concrete     *)
  (* syntax but different abstract syntax. Every  *)
  (* goal can be interpreted as a term, but not   *)
  (* every term can be interpreted as a goal.     *)
  (* <boxed values 41>=                           *)
  val _ = op toGoal : term -> goal
  val _ = op goal   : goal parser
  (* <combinators for parsing \uprolog>=          *)
  fun file (tokens as RESERVED "]" ::: _) = 
        fail "filename" "found closing ]" tokens
    | file (t ::: tokens) = (tokenString t, tokens)
    | file (NEED_MORE _) = raise SyntaxError "expected `]' before end of line"
  val file = repeat (lforce file) >> String.concat : string parser

  val concrete 
    =  "[" $-- file --$ "]" --$ "."             >> BRACKET
    ||  goal -- ":-" $-- commas goal --$ "."    >> (fn (g, gs) => CLAUSE (g,
                                                                       SOME gs))
    ||  commas goal --$ "."                     >> GOALS
  val concrete =
    fn tokens => !! (lforce concrete) tokens
                 handle SyntaxError s => synerror(fromLazy tokens, "concrete",
                                            ["\n ... preceding message was ", s]
                                                                               )
  (* A concrete item is either a file in          *)
  (* brackets, a clause, or a list of goals       *)
  (* separated by commas.                         *)
  (* <boxed values 42>=                           *)
  val _ = op concrete : concrete parser
in
  fun getconcrete nextline tokens =
      let fun parse [] = []
            | parse tokens = 
                let val (c, tokens) = concrete (toLazy nextline tokens)
                in  c :: parse (fromLazy tokens)
                end
      in  parse tokens
      end
(* Function [[getconcrete]] is the main parser. *)
(* It calls [[nextline]] if it needs more       *)
(* tokens to finish parsing a [[concrete]]      *)
(* top-level item in progress. The code defined *)
(* in [[]] helps structure the                  *)
(* way we read input. It defines a type [['a    *)
(* llist]], which represents a potentially      *)
(* infinite list of values--in this case, the   *)
(* list of tokens present on standard input.    *)
(* The reason we need a special type, not an    *)
(* ordinary list, is that we don't want to read *)
(* all the tokens at once; in an interactive    *)
(* interpreter, it is essential to read only as *)
(* many lines as are needed to parse the next   *)
(* concrete item. In a batch compiler, where we *)
(* read the entire program at once, laziness is *)
(* unnecessary.                                 *)
(*                                              *)
(* Functions [[toLazy]] and [[fromLazy]]        *)
(* convert between ordinary lists and lazy      *)
(* lists. Lazy lists are used by the code in    *)
(* [[]].                                        *)
(* This code, which is modeled on Paulson's     *)
(* code, provides the actual parsing function   *)
(* [[concrete]].                                *)
(* <boxed values 29>=                           *)
val _ = op getconcrete : (unit -> token list) -> token list -> concrete list
val _ = op concrete    : token llist -> concrete * token llist
val _ = op fromLazy    : 'a llist -> 'a list
val _ = op toLazy      : (unit -> 'a list) -> 'a list -> 'a llist
end
(* <parsing>=                                   *)
fun newparse {firstline, nextline, curloc} = (* a bit bogus on prompting *)
  let fun tokens () = read nextline curloc (firstline ()) 
  in  fn () => getconcrete tokens (tokens())
  end

fun cprint f [] = print "<??empty??>"
  | cprint f (h::t) = (print (f h); app (fn x => app print [", ", f x]) t)

fun printTop (BRACKET s) = app print ["[", s, "].\n"]
  | printTop (CLAUSE (g, rhs)) =
       ( print (goalString g)
       ; case rhs
           of NONE => ()
            | SOME gs => (print " :- "; cprint goalString gs)
       ; print ".\n"
       )
  | printTop (GOALS gs) = (cprint goalString gs; print ".\n")


(*****************************************************************)
(*                                                               *)
(*   READERS                                                     *)
(*                                                               *)
(*****************************************************************)

(* Readers                                      *)
(*                                              *)
(*  Line readers                                *)
(*                                              *)
(* Our readers are structured slightly          *)
(* differently from those in C. A reader is a   *)
(* bundle containing three related functions:   *)
(* [[firstline]], which returns the first line  *)
(* read; [[nextline]], which returns all        *)
(* subsequent lines; and [[curloc]], which      *)
(* returns a pair identifying the source        *)
(* location of the most recently returned line. *)
(* Both [[firstline]] and [[nextline]] can      *)
(* raise [[EOF]], (We distinguish [[firstline]] *)
(* from [[nextline]] because in an interactive  *)
(* reader, they may use different prompts.)     *)
(* <readers>=                                   *)
type reader = { firstline : unit -> string
              , nextline  : unit -> string
              , curloc    : unit -> string * int
              }
(* <readers>=                                   *)
fun promptReader (firstPrompt, nextPrompt, reader) =
  let val { firstline = first, nextline = next, curloc = cur } = reader
      fun promptIn get prompt () = ( TextIO.output(TextIO.stdOut, prompt)
                                   ; TextIO.flushOut(TextIO.stdOut)
                                   ; get ()
                                   )
  in  { firstline = promptIn first firstPrompt
      , nextline  = promptIn next  nextPrompt
      , curloc    = cur
      }
  end
(* Given a silent reader, we can make it issue  *)
(* a prompt every time it reads a new line. We  *)
(* use two prompts: one for the first line of   *)
(* an expression and one for later lines.       *)
(* <boxed values 23>=                           *)
val _ = op promptReader : string * string * reader -> reader
(* <readers>=                                   *)
fun filereader (filename, fd)=
  let val lineNum = ref 0
      fun get () =
        let val line = TextIO.inputLine fd
        in  if size line = 0 then raise EOF else line before lineNum := !lineNum
                                                                             + 1
        end
  in  { firstline = get, nextline = get, curloc = (fn () => (filename, !lineNum)
                                                                             ) }
  end

fun stringsreader (what, l) = 
  let val buffer  = ref l
      val lineNum = ref 0
      fun get ()  = case !buffer
                      of [] => raise EOF
                       | h :: t => h before buffer := t before lineNum := !
                                                                     lineNum + 1
  in  { firstline = get, nextline = get, curloc = (fn () => (what, !lineNum)) }
  end              
(* We can read from files or lists of strings.  *)
(* File I/O is provide by the standard module   *)
(* [[TextIO]].                                  *)
(* <boxed values 24>=                           *)
val _ = op filereader    : string * TextIO.instream -> reader
val _ = op stringsreader : string * string list     -> reader
(* [[]], a private mutable cell. When all       *)
(* the strings in the buffer are used up, there *)
(* aren't any more.                             *)

(*  Top-level readers                           *)
(*                                              *)
(* A top-level reader is a buffer of top-level  *)
(* items together with a reader. [*]            *)
(* <readers>=                                   *)
type topreader = { buffer : toplevel list ref, reader : reader }
fun readtop (r as { buffer, reader = { nextline, firstline, curloc } }) =
  case !buffer
    of h::t => h before buffer := t
     | []   => ( buffer := map toplevel (read nextline curloc (firstline()))
               ; readtop r
               )
(* We decide whether to prompt when building a  *)
(* top-level reader.                            *)
(* <readers>=                                   *)
fun topreader (reader, prompt) = 
  let val reader = if prompt then promptReader ("-> ", "   ", reader) else
                                                                          reader
  in  { buffer = ref [], reader = reader }
  end
(* <readers>=                                   *)
fun echoTag f x =
  let val line = f x
      val _ = if (String.substring (line, 0, 2) = ";#" handle _ => false) then
                print line
              else
                ()
  in  line
  end

fun echoReader { nextline=n, firstline=f, curloc=c } =
     { nextline=echoTag n, firstline=echoTag f, curloc=c } 

fun echoTop { buffer = b, reader = r } = { buffer = b, reader = echoReader r }
val topreader = fn args => echoTop (topreader args)
(* As in the C code, we want to print out any   *)
(* line read if it begins with the special      *)
(* string [[;#]]. This string is a formal       *)
(* comment that helps us test the chunks marked *)
(* [[]]. In our ML code, we                     *)
(* arrange it using the higher-order function   *)
(* [[echoReader]].                              *)
(* <boxed values 25>=                           *)
val _ = op echoReader : reader -> reader 
(* Modes and readers                            *)
(*                                              *)
(* [*] The interpreter has two modes-query mode *)
(* and rule mode.                               *)
(* <readers>=                                   *)
datatype mode = QMODE | RMODE
(* A top-level reader keeps a buffer of         *)
(* recently read concrete syntax, functions to  *)
(* get lines, and the current mode. The mode is *)
(* used not only to convert [[concrete]] to     *)
(* [[toplevel]] but also to issue a             *)
(* mode-specific prompt. For this reason, we    *)
(* cannot reuse the [[reader]] type used in     *)
(* other interpreters.                          *)
(* <readers>=                                   *)
type topreader = { buffer    : concrete list ref
                 , mode      : mode ref
                 , nextline  : unit -> string
                 , firstline : mode -> string
                 , curloc    : unit -> srcloc
                 }
(* <readers>=                                   *)
fun readtop (r as { buffer, nextline, firstline, curloc, mode }) : toplevel =
  let fun tokens () = read nextline curloc (firstline (!mode))
      fun item () =
        case !buffer
          of h::t => h before buffer := t
           | []   => ( buffer := getconcrete tokens (tokens())
                     ; item ()
                     )
  in  case (item(), !mode)
        of (BRACKET "rule",   _) => (mode := RMODE; readtop r)
         | (BRACKET "fact",   _) => (mode := RMODE; readtop r)
         | (BRACKET "user",   _) => (mode := RMODE; readtop r)
         | (BRACKET "clause", _) => (mode := RMODE; readtop r)
         | (BRACKET "query",  _) => (mode := QMODE; readtop r)
         | (BRACKET s, _)        => USE s
         | (CLAUSE (g, ps),     RMODE) => ADD_CLAUSE (g :- getOpt(ps, []))
         | (CLAUSE (g, NONE),   QMODE) => QUERY [g]
         | (CLAUSE (_, SOME _), QMODE) =>
              raise SyntaxError ("You cannot enter a new clause in query mode; "
                                                                               ^
                                 "type `[rule].' to change modes")
         | (GOALS l,   QMODE) => QUERY l
         | (GOALS [g], RMODE) => ADD_CLAUSE (g :- [])
         | (GOALS l,   RMODE) => 
              raise SyntaxError ("You cannot enter a query in clause mode; " ^
                                 "type `[query].' to change modes")
  end            
(* Function [[readtop]] converts [[concrete]]   *)
(* to [[toplevel]] tracking the mode in the     *)
(* process.                                     *)
(* <boxed values 26>=                           *)
val _ = op readtop : topreader -> toplevel
(* The top-level reader is different from the   *)
(* others because it has to manage the two      *)
(* modes and therefore also two sets of         *)
(* prompts. The [[echoTag]] function is a       *)
(* special hack that enables us to test the     *)
(* chunks labeled [[]] in this                  *)
(* book.                                        *)
(* <readers>=                                   *)
fun topreader' mode ({firstline, nextline, curloc}, prompt) = 
  let fun promptIn getline prompt () = ( TextIO.output(TextIO.stdOut, prompt)
                                       ; TextIO.flushOut(TextIO.stdOut)
                                       ; getline ()
                                       )
      fun mprompt RMODE = promptIn firstline "-> " ()
        | mprompt QMODE = promptIn firstline "?- " ()
  in  if prompt then
        { buffer = ref [], mode = ref mode, curloc = curloc,
          nextline = echoTag (promptIn nextline "   "), firstline = echoTag
                                                                       mprompt }
      else 
        { buffer = ref [], mode = ref mode, curloc = curloc,
          nextline = echoTag nextline, firstline = echoTag (fn _ => firstline ()
                                                                             ) }
  end
(* A top-level reader is in rule mode by        *)
(* default.                                     *)
(* <readers>=                                   *)
fun topreader x = topreader' RMODE x


(*****************************************************************)
(*                                                               *)
(*   IMPLEMENTATION OF [[USE]]                                   *)
(*                                                               *)
(*****************************************************************)

(* The implementation of [[use]] is             *)
(* parameterized by [[readEvalPrint]] so we can *)
(* share it with other interpreters. Function   *)
(* [[use]] creates a top-level reader that does *)
(* not prompt, but it uses [[writeln]] to be    *)
(* sure that responses are printed.             *)
(* <implementation of [[use]]>=                 *)
fun use readEvalPrint filename rho =
      let val fd = TextIO.openIn filename
          val treader = topreader (filereader (filename, fd), false)
          fun writeln s = app print [s, "\n"]
      in  readEvalPrint (treader, writeln, writeln) rho
          before TextIO.closeIn fd
      end 
(* [[topreader]] is relegated                   *)
(* to chunk [->] in Appendix [->].              *)



(*****************************************************************)
(*                                                               *)
(*   PRIMITIVES                                                  *)
(*                                                               *)
(*****************************************************************)

(* <primitives>=                                *)
exception RuntimeError of string
fun eval (LITERAL n) = n
  | eval (APPLY ("+", [x, y])) = eval x  +  eval y
  | eval (APPLY ("*", [x, y])) = eval x  *  eval y
  | eval (APPLY ("-", [x, y])) = eval x  -  eval y
  | eval (APPLY ("/", [x, y])) = eval x div eval y
  | eval (APPLY ("-", [x]))    = ~ (eval x)
  | eval (APPLY (f, _))        = 
      raise RuntimeError (f ^ " is not an arithmetic predicate " ^
                          "or is used with wrong arity")
  | eval (VAR v) = raise RuntimeError ("Used uninstantiated variable " ^ v ^
                                       " in arithmetic expression")
(* To implement the primitive predicate [[is]], *)
(* we need a very small evaluator. Because it   *)
(* works only with integers, never with         *)
(* variables, there is no environment; it's     *)
(* just a tree walk.                            *)
(* <boxed values 10>=                           *)
val _ = op eval : term -> int
(* We implement t_1[[ is ]]t_2 by evaluating    *)
(* term t_2 as an integer expression and        *)
(* unifying t_1 with the result.                *)
(* <primitives>=                                *)
fun is [x, e] succ fail = (succ (unifyTerm (x, LITERAL (eval e))) fail
                           handle Unify => fail())
  | is _      _    fail = fail ()
(* The parser should arrange that each          *)
(* comparison is applied to exactly two         *)
(* arguments. If these arguments aren't         *)
(* integers, it's a run-time error.             *)
(* <primitives>=                                *)
fun compare name cmp [LITERAL n, LITERAL m] succ fail =
      if cmp (n, m) then succ idsubst fail else fail ()
  | compare name _ [_, _] _ _ =
      raise RuntimeError ("Used comparison " ^ name ^ " on non-integer term")
  | compare name _ _ _ _ =
      raise RuntimeError ("this can't happen---non-binary comparison?!")


(*****************************************************************)
(*                                                               *)
(*   TRACING FUNCTIONS                                           *)
(*                                                               *)
(*****************************************************************)

(*   Create a tracing version of the            *)
(* interpreter that logs every entry to and     *)
(* exit from a Byrd box. Use the following      *)
(* functions:                                   *)
(* <tracing functions>=                         *)
fun logSucc goal succ rho resume =
  ( app print ["SUCC: ", goalString goal, " becomes ", goalString (lift rho goal
                                                                        ), "\n"]
  ; succ rho resume
  )
fun logFail goal fail () = 
  ( app print ["FAIL: ", goalString goal, "\n"]
  ; fail ()
  )
fun logResume goal resume () = 
  ( app print ["REDO: ", goalString goal, "\n"]
  ; resume ()
  )
fun logSolve solve goal succ fail = 
  ( app print ["START: ", goalString goal, "\n"]
  ; solve goal succ fail
  )


(*****************************************************************)
(*                                                               *)
(*   SEARCH                                                      *)
(*                                                               *)
(*****************************************************************)

(* <search>=                                    *)
fun 'a query database =
  let val builtins = foldl (fn ((n, p), rho) => bind (n, p, rho))
                     emptyEnv ((* Printing a term always succeeds and produces
                                                                              *)
                               (* the identity substitution.
                                                                              *)
                               (* <primops ::>=
                                                                              *)
                               ("print", fn args => fn succ => fn fail =>
                                           ( app (fn x => (print (termString x);
                                                                print " ")) args
                                           ; print "\n"
                                           ; succ (fn x => x) fail
                                           )) ::
                               (* <primops ::>=
                                                                              *)
                               ("is", is) ::
                               (* <primops ::>=
                                                                              *)
                               ("<",  compare "<"  op < ) ::
                               (">",  compare ">"  op > ) ::
                               ("=<", compare "=<" op <= ) ::
                               (">=", compare ">=" op >= ) ::
                               (* Two more predicates, [[!]] and [[not]],
                                                                              *)
                               (* cannot be implemented using this technique;
                                                                              *)
                               (* they have to be added directly to the
                                                                              *)
                               (* interpreter (Exercises [->] and [->]). Here
                                                                              *)
                               (* we ensure that they can't be used by
                                                                              *)
                               (* mistake.
                                                                              *)
                               (* <primops ::>=
                                                                              *)
                               ("!",   fn _ => raise RuntimeError
                              "The cut (!) must be added to the interpreter") ::
                               ("not", fn _ => raise RuntimeError
                  "The predicate `not' must be added to the interpreter") :: [])
      fun solveOne (goal as (func, args)) succ fail =
            find(func, builtins) args succ fail
            handle NotFound _ =>
              let fun search [] = fail ()
                    | search (clause :: clauses) =  
                        let fun resume() = search clauses
                            val conclusion :- premises = freshen clause
                            val theta = unify (goal, conclusion)
                        in  solveMany (map (lift theta) premises) theta succ
                                                                          resume
                        end
                        handle Unify => search clauses
(* Here is the code:                            *)
(*                                              *)
(* <boxed values 8>=                            *)
val _ = op query  : database -> goal list -> (subst -> (unit->'a) -> 'a) -> (
                                                                 unit->'a) -> 'a
val _ = op solveOne  : goal               -> (subst -> (unit->'a) -> 'a) -> (
                                                                 unit->'a) -> 'a
val _ = op solveMany : goal list -> subst -> (subst -> (unit->'a) -> 'a) -> (
                                                                 unit->'a) -> 'a
val _ = op search    : clause list -> 'a
              in  search (potentialMatches (goal, database))
              end
      and solveMany []            theta succ fail = succ theta fail
        | solveMany (goal::goals) theta succ fail =
            solveOne goal
            (fn theta' => fn resume =>
               solveMany (map (lift theta') goals) (theta' o theta) succ resume)
            fail
  in  fn gs => solveMany gs (fn x => x)
  end
(* [[builtins]] holds the                       *)
(* built-in predicates. Because the             *)
(* implementations of these predicates are      *)
(* polymorphic ML functions, ML's ``value       *)
(* restriction'' prevents us from defining them *)
(* at top level. We therefore have to build the *)
(* [[builtins]] environment dynamically, but as *)
(* long as we do this only once per query, the  *)
(* cost should be acceptable.                   *)



(*****************************************************************)
(*                                                               *)
(*   INTERACTION                                                 *)
(*                                                               *)
(*****************************************************************)

(* <interaction>=                               *)
fun showAndContinue prompt theta gs =
  let fun showVar v = app print [v, " = ", termString (theta (VAR v))]
      val vars = foldr union emptyset (map freevarsGoal gs)
  in  case vars
        of [] => false
         | h :: t => ( showVar h
                     ; app (fn v => (print "\n"; showVar v)) t
                     ; if prompt then
                         case explode (TextIO.inputLine TextIO.stdIn)
                           of #";" :: _ => (print "\n"; true)
                            | _ => false
                       else
                         (print "\n"; false)
                     )
  end


(*****************************************************************)
(*                                                               *)
(*   EVALUATION                                                  *)
(*                                                               *)
(*****************************************************************)

(* <evaluation>=                                *)
fun topeval prompt (t, database, echo) =
  case t
    of USE filename => (* To read from a file, we read without         *)
                       (* prompting. If we fail to open the file, we   *)
                       (* try adding ``[[.P]]'' to the name; this is   *)
                       (* the convention used by XSB Prolog, a free    *)
                       (* Prolog interpreter available from the State  *)
                       (* University of New York at Stony Brook. If    *)
                       (* this too fails, we try adding ``[[.pl]]'' to *)
                       (* the name; this is the convention used by GNU *)
                       (* Prolog.                                      *)
                       (* <read from file [[filename]]>=               *)
                       let fun try filename = use (readEvalPrint false) filename
                                                                        database
                       in  try filename          handle IO.Io _ => 
                           try (filename ^ ".P") handle IO.Io _ => 
                           try (filename ^ ".pl")
                       end
     | ADD_CLAUSE c => addClause (c, database)
     | QUERY gs     => ((* To issue a query, we provide success and     *)
                        (* failure continuations. The success           *)
                        (* continuation uses [[showAndContinue]] to     *)
                        (* decide if it should resume the search,       *)
                        (* looking for another solution, or just        *)
                        (* declare victory and stop.                    *)
                        (* <query goals [[gs]] against [[database]]>=   *)
                        query database gs
                          (fn theta => fn resume =>
                             if showAndContinue prompt theta gs then resume()
                                                             else print "yes\n")
                          (fn () => print "no\n"); database)
(* <evaluation>=                                *)
and readEvalPrint prompt (reader, echo, errmsg) database =
  let fun loop database =
        let fun continue msg = (errmsg msg; loop database)
            fun finish() = database
        in  loop (topeval prompt (readtop reader, database, echo))
            handle EOF => finish()
                (* The handlers in the first group deal with    *)
                (* problems that arise during I/O, lexical      *)
                (* analysis, and parsing.                       *)
                (* <more read-eval-print handlers>=             *)
                | IO.Io {name, ...} => continue ("I/O error: " ^ name)
                | Paren (_, loc)    => continue ("syntax error: extra ) in " ^
                                                               srclocString loc)
                | PrematureEOF s    => (errmsg ("Missing )? Got EOF reading (" ^
                                                                   s); finish())
                | SyntaxError msg   => continue ("syntax error: " ^ msg)
                (* [[IO.Io]] is part of the                     *)
                (* Standard Basis Library.                      *)

                (* The handlers in the second group deal with   *)
                (* problems that arise during evaluation. They  *)
                (* always continue with the next top-level      *)
                (* item.                                        *)
                (* <more read-eval-print handlers>=             *)
                | Div               => continue "Division by zero"
                | Overflow          => continue "Arithmetic overflow"
                | RuntimeError msg  => continue ("run-time error: " ^ msg)
                | NotFound n        => continue ("variable " ^ n ^ " not found")
                (* [[Div]] and [[Overflow]] are                 *)
                (* predefined.                                  *)

        end
  in  loop database
  end
(*   Top-level processing                       *)
(*                                              *)
(* Where another interpreter might use an       *)
(* environment, the uProlog interpreter uses a  *)
(* database of clauses.                         *)
(* <boxed values 9>=                            *)
val _ = op topeval : bool -> toplevel * database * (string->unit) -> database
(* To show a solution, we apply the             *)
(* substitution to the free varables of the     *)
(* query. If we're prompting, we wait for a     *)
(* line of input. If the line begins with a     *)
(* semicolon, we continue; otherwise we quit.   *)
(* If we're not prompting, we're in batch mode, *)
(* and we produce at most one solution.         *)
(* <boxed values 9>=                            *)
val _ = op showAndContinue : bool -> subst -> goal list -> bool
(*   The read-eval-print loop                   *)
(*                                              *)
(* The read-eval-print loop is much the same as *)
(* in other interpreters, except it takes an    *)
(* additional argument that tells whether to    *)
(* prompt. You may have noticed this argument   *)
(* in [[]],                                     *)
(* which arranges not to prompt inside [[USE]]. *)
(* <boxed values 9>=                            *)
val _ = op readEvalPrint : bool -> topreader * (string->unit) * (string->unit)
                                                         -> database -> database


(*****************************************************************)
(*                                                               *)
(*   PROLOG COMMAND LINE                                         *)
(*                                                               *)
(*****************************************************************)

(* Command line                                 *)
(*                                              *)
(* uProlog's command-line differs from our      *)
(* other interpreters, because it has to deal   *)
(* with modes. In [[noisy]] state it starts     *)
(* waiting for a query, but in quiet state it   *)
(* waits for a rule.                            *)
(* <Prolog command line>=                       *)
fun runInterpreter noisy = 
  let fun writeln s = app print [s, "\n"]
      val mode    = if noisy then QMODE else RMODE
      val treader = topreader' mode (filereader ("standard input", TextIO.stdIn)
                                                                        , noisy)
  in  ignore (readEvalPrint noisy (treader, writeln, writeln) emptyDatabase)
  end 
(* <Prolog command line>=                       *)
fun copy file =
  let val parse = newparse (filereader file)
      fun go () = (app printTop (parse()); print "/**----**/\n"; go())
  in  go () handle EOF => ()
                 | SyntaxError s => app print ["Syntax error: ", s, "\n"]
  end
(* The [[-q]] option is as in other             *)
(* interpreters, and the [[-trace]] option      *)
(* turns on tracing. The [[-copy]] option is    *)
(* undocumented.                                *)
(* <Prolog command line>=                       *)
fun runmain ["-q"]          = runInterpreter false
  | runmain []              = runInterpreter true
  | runmain ("-trace" :: t) = (tracer := app print; runmain t)
  | runmain ["-copy"]       = copy ("<stdin>", TextIO.stdIn) (* secret copy
                                                                     function *)
  | runmain _  =
      TextIO.output(TextIO.stdErr, "Usage: " ^ CommandLine.name() ^ " [-q]\n")
(* <Prolog command line>=                       *)
val _ = runmain (CommandLine.arguments())
