/* The Interpreter                              */
/*                                              */
/* [*]                                          */
/*                                              */
/* This section presents the code for an        */
/* interpreter written in C. As in the Impcore  */
/* interpreter, we break the program into       */
/* modules with well-defined interfaces. We     */
/* reuse the name routines, the lexer, the      */
/* input readers, and the printer from the      */
/* implementation of Impcore.                   */
/*                                              */
/*   Interfaces                                 */
/*                                              */
/* As in Impcore, we gather all the interfaces  */
/* into a single C header file.                 */
/* <{\Tt all.h} for \uscheme>=                  */
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <setjmp.h>
#include <ctype.h>

/* For real implementations, it is convenient   */
/* to build names from strings. Unlike          */
/* C strings, names are immutable, and they can */
/* be compared using pointer equality.          */
/* <shared type definitions>=                   */
typedef struct Name Name;
typedef struct Namelist Namelist;   /* list of Name* */
/*    Reading characters                        */
/*                                              */
/* A [[Reader]] encapsulates a source of        */
/* characters: a string or a file. The          */
/* [[readline]] function prints a prompt, reads */
/* the next line of input from the source, and  */
/* returns a pointer to the line. The caller    */
/* needn't worry about how long the line is;    */
/* [[readline]] always allocates enough memory  */
/* to hold it. Because [[readline]] reuses the  */
/* same memory to hold successive lines, it is  */
/* an unchecked run-time error to retain a      */
/* pointer returned by [[readline]] after a     */
/* subsequent call to [[readline]]. A client    */
/* that needs to save input characters must     */
/* copy the result of [[readline]] before       */
/* calling [[readline]] again.                  */
/* <shared type definitions>=                   */
typedef struct Reader Reader;
/*    Reading top-level items                   */
/*                                              */
/* A [[Topreader]] encapsulates a source of     */
/* top-level items. The [[readtop]] function    */
/* returns the next top-level item. It may also */
/* return [[NULL]], if there are no more        */
/* top-level items in the source, or call       */
/* [[error]] (page [->]), if there is some      */
/* problem converting the source to abstract    */
/* syntax.                                      */
/* <shared type definitions>=                   */
typedef struct Topreader Topreader;
/* <shared type definitions>=                   */
typedef void Printer(va_list*);
/* <shared type definitions>=                   */
typedef struct Parlist Parlist; /* list of Par* */
/* <shared type definitions>=                   */
typedef struct Par Par;
/* <type definitions for \uscheme>=             */
typedef struct Lambda Lambda;
typedef struct Value Value;
/* <type definitions for \uscheme>=             */
typedef struct Toplevel Toplevel;
typedef struct Exp Exp;
/* <type definitions for \uscheme>=             */
typedef struct Root Root;
/* <type definitions for \uscheme>=             */
typedef enum Letkeyword { LET, LETSTAR, LETREC } Letkeyword;
typedef struct Explist Explist; /* list of Exp* */
/* <type definitions for \uscheme>=             */
typedef struct Valuelist Valuelist;     /* list of Value */
typedef Value Primitive(Exp *e, int tag, Valuelist *vl);
/* [[Primitive]] calls for                      */
/* explanation, since it would be simpler to    */
/* define a primitive function as one that      */
/* accepts a [[Valuelist]] and returns a        */
/* [[Value]]. We pass the abstract syntax [[Exp */
/* *e]] so that if the primitive fails, it can  */
/* issue a suitable error message. We pass the  */
/* integer [[tag]] so that we can write a       */
/* single C function that implements multiple   */
/* primitives. For example, the arithmetic      */
/* primitives are implemented by a single       */
/* function, which makes it possible for them   */
/* to share the code that ensures both          */
/* arguments are numbers. Implementations of    */
/* the primitives appear in Section [->].       */

/*  The Environment and the Store               */
/*                                              */
/* [*] In the operational semantics, the        */
/* store sigma is intended to model the         */
/* machine's memory. It should come as no       */
/* surprise, then, that we use C pointers (of   */
/* type [[Value *]]) as locations and the       */
/* machine's memory as the store. An            */
/* environment [[Env]] maps names to pointers;  */
/* find(x, rho) returns rho(x) if x in dom rho; */
/* otherwise it returns [[NULL]].               */
/* <type definitions for \uscheme>=             */
typedef struct Env Env;
/* An input reader is just a line-reading       */
/* function and some extra state.               */
/* <shared structure definitions>=              */
struct Reader {
    char *buf;               /* holds the last line read */
    int nbuf;                /* size of buf */
    int line;                /* current line number */
    const char *readername;  /* identifies this reader */

    FILE *fin;               /* filereader */
    const char *s;           /* stringreader */
};
/* <shared structure definitions>=              */
typedef enum { ATOM, LIST } Party;

struct Par {
    Party ty;
    union {
        Name *atom;
        Parlist *list;
    } u;
};

/* <structure definitions for \uscheme>=        */
struct Lambda {
    Namelist *formals;
    Exp *body;
};
typedef enum {
    NIL, BOOL, NUM, SYM, PAIR, CLOSURE, PRIMITIVE, FORWARD, INVALID
} Valuety;

struct Value {
    Valuety ty;
    union {
        int bool;
        int num;
        Name *sym;
        struct {  Value *car; Value *cdr; } pair;
        struct {  Lambda lambda; Env *env; } closure;
        struct {  int tag; Primitive *function; } primitive;
        Value  *forward;
        const char  *invalid;
    } u;
};

/* <structure definitions for \uscheme>=        */
typedef enum { EXP, DEFINE, VAL, USE } Toplevelty;

struct Toplevel {
    Toplevelty ty;
    union {
        Exp *exp;
        struct {  Name *name; Lambda lambda; } define;
        struct {  Name *name; Exp *exp; } val;
        Name *use;
    } u;
};

typedef enum { LITERAL, VAR, SET, IFX, WHILEX, BEGIN, LETX, LAMBDAX, APPLY }
                                                                          Expty;

struct Exp {
    Expty ty;
    union {
        Value literal;
        Name *var;
        struct {  Name *name; Exp *exp; } set;
        struct {  Exp *cond; Exp *true; Exp *false; } ifx;
        struct {  Exp *cond; Exp *body; } whilex;
        Explist *begin;
        struct {  Letkeyword let; Namelist *nl; Explist *el; Exp *body; } letx;
        Lambda lambdax;
        struct {  Exp *fn; Explist *actuals; } apply;
    } u;
};

/* <structure definitions for \uscheme>=        */
struct Explist {
    Exp *hd;
    Explist *tl;
};

struct Parlist {
    Par *hd;
    Parlist *tl;
};

struct Valuelist {
    Value hd;
    Valuelist *tl;
};

struct Namelist {
    Name *hd;
    Namelist *tl;
};

/* <structure definitions for \uscheme>=        */
typedef enum {
    STACKVALUEROOT,
    HEAPLOCROOT,
    ENVROOT,
    EXPROOT,
    EXPLISTROOT,
    VALUELISTROOT,
    TOPLEVELROOT
} Rootty;

struct Root {
    Rootty ty;
    union {
        Value *stackvalueroot;
        Value* *heaplocroot;
        Env* *envroot;
        Exp* *exproot;
        Explist* *explistroot;
        Valuelist* *valuelistroot;
        Toplevel* *toplevelroot;
    } u;
};

/* <structure definitions for \uscheme>=        */
struct Env {
    Name *name;
    Value *loc;
    Env *tl;
};
/* Pointer comparison is built into C, but we   */
/* provide two other operations on names.       */
/* <shared function prototypes>=                */
Name *strtoname(const char *s);
const char *nametostr(Name *n);
/* [[]]                                         */
/*   [[strcmp(s, t) == 0]] if and only if       */
/*   [[strtoname(s) == strtoname(t)]]           */
/*                                              */
/* Informally, the first law says if you build  */
/* a name from a string, [[nametostr]] returns  */
/* a copy of your original string. The second   */
/* law says you can compare names using pointer */
/* equality.                                    */

/* Function [[printname]] prints names.         */
/*                                              */
/* <shared function prototypes>=                */
Printer printname;
/* <shared function prototypes>=                */
char *readline(Reader *r, char *prompt);
/* To create a reader, we need a string or a    */
/* file. A reader also gets a name, which we    */
/* can use in error messages. [*] [*]           */
/* <shared function prototypes>=                */
Reader *stringreader(const char *stringname, const char *s);
Reader *filereader  (const char *filename,   FILE *fin);
/* <shared function prototypes>=                */
Toplevel *readtop(Topreader *r);
/* <shared function prototypes>=                */
Topreader *topreader(Reader *r, int doprompt);
/* [[doprompt]] flag controls whether the       */
/* resulting [[Topreader]] prompts when reading */
/* input.                                       */

/*  Printing                                    */
/*                                              */
/* [*] The standard C function [[printf]] and   */
/* its siblings [[fprintf]] and [[vfprintf]]    */
/* convert internal C values to sequences of    */
/* characters in an output file. Unfortunately, */
/* they convert only numbers, characters, and   */
/* strings. It would be a lot more convenient   */
/* if [[printf]] also converted [[Name]]s,      */
/* [[Exp]]s, and so on. We can't change         */
/* [[printf]], so instead we define two new     */
/* functions: [[print]] and [[vprint]], with    */
/* similar interfaces.                          */
/* <shared function prototypes>=                */
void print (const char *fmt, ...);
void vprint(const char *fmt, va_list va);
/* [[print]] and                                */
/* [[vprint]] is that they are extensible; we   */
/* can add new printing functions to print new  */
/* kinds of values. Such functions have type    */
/* [[Printer]], and we add them using           */
/* [[installprinter]]. [*]                      */

/* <shared function prototypes>=                */
void installprinter(unsigned char c, Printer *fmt);
/* The printer interface provides printers to   */
/* format percent signs, strings, and decimal   */
/* integers.                                    */
/* <shared function prototypes>=                */
Printer printpercent, printstring, printdecimal;
/*  Error handling                              */
/*                                              */
/* Any module can signal an error by calling    */
/* [[error]]. When called, [[error]] acts just  */
/* like [[print]] except that it does not       */
/* return after printing; instead it            */
/* [[longjmp]]s to [[errorjmp]]. It is an       */
/* unchecked run-time error to call [[error]]   */
/* except while a suitable [[setjmp]] is active */
/* on the C call stack. [*]                     */
/* <shared function prototypes>=                */
void error(const char *fmt, ...);
extern jmp_buf errorjmp;        /* longjmp here on error */
/* Function [[checkargc]] is used to check the  */
/* number of arguments to both user-defined and */
/* primitive functions. The first argument is   */
/* an abstract-syntax tree representing the     */
/* application being checked; if                */
/* [[expected]] != [[actual]], [[checkargc]]    */
/* calls [[error]], passing a message that      */
/* contains [[e]].                              */
/* <shared function prototypes>=                */
void checkargc(Exp *e, int expected, int actual);
/* The [[duplicatename]] function finds a       */
/* duplicate name on a [[Namelist]] if such a   */
/* name exists. It is used to check that formal */
/* parameters to user-defined functions all     */
/* have different names. If the name list       */
/* [[names]] contains a duplicate occurrence of */
/* any name, the function returns such a name;  */
/* otherwise it returns [[NULL]].               */
/* <shared function prototypes>=                */
Name *duplicatename(Namelist *names);
/* <shared function prototypes>=                */
/* <shared function prototypes>=                */
Printer printpar;
/* The [[readparlist]] function reads and       */
/* returns a list of parenthesized expressions  */
/* from the given [[Reader]], stopping when an  */
/* end-of-expression and end-of-line coincide.  */
/* The [[doquote]] flag specifies that an       */
/* expression like [['(1 2 3)]] should be       */
/* turned into [[(quote (1 2 3))]]; it is used  */
/* to implement micro-Scheme. The [[doprompt]]  */
/* flag specifies that prompts should be        */
/* printed when reading input lines.            */
/* <shared function prototypes>=                */
Parlist *readparlist(Reader *r, int doquote, int doprompt);
/* <shared function prototypes>=                */
Par *mkAtom(Name*);
Par *mkList(Parlist*);
/* <function prototypes for \uscheme>=          */
Lambda mkLambda(Namelist *formals, Exp *body);
Value mkNil(void);
Value mkBool(int);
Value mkNum(int);
Value mkSym(Name*);
Value mkPair(Value *car, Value *cdr);
Value mkClosure(Lambda lambda, Env *env);
Value mkPrimitive(int tag, Primitive *function);
Value mkForward(Value *);
Value mkInvalid(const char *);
/* <function prototypes for \uscheme>=          */
Toplevel *mkExp(Exp*);
Toplevel *mkDefine(Name *name, Lambda lambda);
Toplevel *mkVal(Name *name, Exp *exp);
Toplevel *mkUse(Name*);
Exp *mkLiteral(Value);
Exp *mkVar(Name*);
Exp *mkSet(Name *name, Exp *exp);
Exp *mkIfx(Exp *cond, Exp *true, Exp *false);
Exp *mkWhilex(Exp *cond, Exp *body);
Exp *mkBegin(Explist*);
Exp *mkLetx(Letkeyword let, Namelist *nl, Explist *el, Exp *body);
Exp *mkLambdax(Lambda);
Exp *mkApply(Exp *fn, Explist *actuals);
/* <function prototypes for \uscheme>=          */
int      lengthEL (Explist *l);
Exp*     nthEL    (Explist *l, unsigned n);
Explist *mkEL     (Exp *hd, Explist *tl);
Printer  printexplist;
/* <function prototypes for \uscheme>=          */
int      lengthPL (Parlist *l);
Par*     nthPL    (Parlist *l, unsigned n);
Parlist *mkPL     (Par *hd, Parlist *tl);
Printer  printparlist;
/* <function prototypes for \uscheme>=          */
int        lengthVL (Valuelist *l);
Value      nthVL    (Valuelist *l, unsigned n);
Valuelist *mkVL     (Value hd, Valuelist *tl);
Printer    printvaluelist;
/* <function prototypes for \uscheme>=          */
int       lengthNL (Namelist *l);
Name*     nthNL    (Namelist *l, unsigned n);
Namelist *mkNL     (Name *hd, Namelist *tl);
Printer   printnamelist;
/* <function prototypes for \uscheme>=          */
Root mkStackvalueroot(Value*);
Root mkHeaplocroot(Value**);
Root mkEnvroot(Env**);
Root mkExproot(Exp**);
Root mkExplistroot(Explist**);
Root mkValuelistroot(Valuelist**);
Root mkToplevelroot(Toplevel**);
/* <function prototypes for \uscheme>=          */
Value *find(Name *name, Env *env);
/* The function [[bindalloc]] binds a name to a */
/* freshly allocated location, and it puts a    */
/* value in that location. Formally, when       */
/* called with store sigma, bindalloc(x, v,     */
/* rho) chooses an l\notindom sigma, updates    */
/* the store to be sigma{l|->v}, and returns    */
/* the extended environment rho{x |->l}.        */
/* <function prototypes for \uscheme>=          */
Env *bindalloc    (Name *name,   Value v,       Env *env);
Env *bindalloclist(Namelist *nl, Valuelist *vl, Env *env);
/* [[bindalloclist]]                            */
/* initializes to values \ldotsnv.              */

/*  Allocation                                  */
/*                                              */
/* The fresh locations created by [[bindalloc]] */
/* and [[bindalloclist]] come from              */
/* [[allocate]]. Calling [[allocate(v)]] finds  */
/* a location l\notindom sigma, stores [[v]]    */
/* in l (thereby updating sigma), and           */
/* returns l.                                   */
/* <function prototypes for \uscheme>=          */
Value *allocate(Value v);
/* Clients must call [[initallocate]] before    */
/* calling [[allocate]]. For reasons that are   */
/* disclosed in Chapter [->], [[initallocate]]  */
/* requires [[envp]], a pointer to the global   */
/* environment, and it sets [[*envp = NULL]].   */
/* <function prototypes for \uscheme>=          */
void initallocate(Env **envp);
/* Finally, if an error occurs during program   */
/* execution, clients must call                 */
/* [[resetallocate]], also passing a pointer to */
/* the global environment.                      */
/* <function prototypes for \uscheme>=          */
void resetallocate(Env **envp);
/*  Values                                      */
/*                                              */
/* [*] The representation of values appears in  */
/* chunk [->] in Section [->]. The value        */
/* interface also exports predefined values     */
/* [[true]] and [[false]].                      */
/* <function prototypes for \uscheme>=          */
Value true, false;
/* [[initvalue]] before                         */
/* executing any code that refers to [[true]]   */
/* or [[false]].                                */

/* <function prototypes for \uscheme>=          */
void initvalue(void);
/* Function [[istrue]] takes a value and        */
/* returns [[1]] if the value should be         */
/* regarded as true (i.e., is not [[#f]]) and   */
/* [[0]] otherwise.                             */
/* <function prototypes for \uscheme>=          */
int istrue(Value v);
/* For example, eval(e, rho), when evaluated    */
/* with store sigma, finds a v and a sigma'     */
/* such that \evale ==>\evalr[']v, it updates   */
/* the store to be sigma', and it returns v.    */
/* <function prototypes for \uscheme>=          */
Value eval(Exp *e, Env *rho);
Env *topeval(Toplevel *t, Env *rho, int echo);
/* [[echo]] is nonzero, it                      */
/* also prints the name or value of whatever    */
/* expression is evaluated or added to rho.     */

/* In our implementation of micro-Scheme,       */
/* [[readevalprint]] has side effects on an     */
/* environment pointer that is passed by        */
/* reference. This technique ensures that in an */
/* interactive session, errors don't destroy    */
/* the results of definitions and [[val]]       */
/* bindings that precede the error. If          */
/* [[readevalprint]] simply returned a new      */
/* environment, any partially created           */
/* environment would be lost after an error.    */
/* <function prototypes for \uscheme>=          */
void readevalprint(Topreader *r, Env **envp, int echo);
/*  Primitives                                  */
/*                                              */
/* Compared to Impcore, micro-Scheme has many   */
/* primitives. The function [[primenv]] returns */
/* an environment binding all the primitive     */
/* operations.                                  */
/* <function prototypes for \uscheme>=          */
Env *primenv(void);
/* Here are some of the printing functions      */
/* used.                                        */
/* <function prototypes for \uscheme>=          */
void printenv     (va_list*);
void printvalue   (va_list*);
void printclosure (va_list*);
void printexp     (va_list*);
void printtoplevel(va_list*);
void printlambda  (va_list*);
/* Because potential roots include local        */
/* variables of C functions such as [[eval]],   */
/* the set of roots changes as the interpreter  */
/* executes. In a compiled system, the garbage  */
/* collector is in cahoots with the compiler,   */
/* which knows where all the local variables    */
/* are. In fact, the compiler knows everything  */
/* about the layout of the activation records   */
/* and the call stack. A compiler can therefore */
/* arrange for a garbage collector to be able   */
/* to find roots easily, with no added overhead */
/* during ``real'' computation. Since our       */
/* garbage collector is not in cahoots with the */
/* C compiler, we need some other way of        */
/* finding roots. What we do is maintain an     */
/* explicit root stack, in addition to the      */
/* C call stack, and we put extra code in our   */
/* C procedures to push roots onto and pop      */
/* roots off of the root stack.                 */
/* <function prototypes for \uscheme>=          */
void pushroot(Root r);
void poproot (Root r);
/* [[poproot]] doesn't match the                */
/* corresponding [[pushroot]].                  */

/* We make the allocator independent of         */
/* initialization; chunk [->] shows how to use  */
/* [[allocloc]] to implement [[allocate]].      */
/* <function prototypes for \uscheme>=          */
Value *allocloc(void);
/* [[]] only when all objects that              */
/* could lead to live values have been pushed   */
/* on the root stack. The copying collector's   */
/* implementation of [[allocloc]] also requires */
/* that, when called, all pointers to allocated */
/* values must be reachable from the root       */
/* stack, so that they can be updated when the  */
/* values move. Section [->] explains the       */
/* programming techniques we use to ensure that */
/* these preconditions are satisfied.           */

/* There are a few more operations that affect  */
/* the root stack. Clients may use              */
/* [[resetroots]] to clear the root stack,      */
/* e.g., in case an error unwinds the C stack.  */
/* <function prototypes for \uscheme>=          */
void resetrootstack(void);
/* Collectors need to get at the roots on the   */
/* stack. Rather than define a procedural       */
/* interface, we expose the representation. The */
/* invariant is that for all 0 <=i <            */
/* rootstacksize, [[rootstack[i]]] is a live    */
/* root. The representation is read only; it is */
/* an unchecked run-time error for a client to  */
/* modify [[rootstack]] or [[rootstacksize]].   */
/* <function prototypes for \uscheme>=          */
extern Root *rootstack;
extern int rootstacksize;
/* Finally, the interface provides a routine to */
/* dump the root information; [[printroots]]    */
/* may help you debug code for the Exercises.   */
/* <function prototypes for \uscheme>=          */
void printroots(void);
/* <function prototypes for \uscheme>=          */
int gammadesired(int defaultval, int minimum);
/* To debug a collector, it can help to wrap    */
/* values in calls to [[validate]]. Calling     */
/* [[validate(v)]] returns [[v]], unless        */
/* [[v]] is invalid, in which case it causes an */
/* assertion failure.                           */
/* <function prototypes for \uscheme>=          */
Value validate(Value); 
