(* We build the full interpreter by             *)
(* concatenating the parts in this order:       *)
(* <mlscheme.sml>=                              *)


(*****************************************************************)
(*                                                               *)
(*   ENVIRONMENTS                                                *)
(*                                                               *)
(*****************************************************************)

(* <environments>=                              *)
type name = string
type 'a env = (name * 'a) list
val emptyEnv = []

(* lookup and assignment of existing bindings *)
exception NotFound of name
fun find (name, []) = raise NotFound name
  | find (name, (n, v)::tail) = if name = n then v else find(name, tail)

(* adding new bindings *)
exception BindListLength
fun bind(name, v, rho) = (name, v) :: rho
fun bindList(n::vars, v::vals, rho) = bindList(vars, vals, bind(n, v, rho))
  | bindList([], [], rho) = rho
  | bindList _ = raise BindListLength
(* We represent an environment of type [['a     *)
(* env]] as a list of ([[name]], [['a]]) pairs. *)
(* The declarations in the box give the         *)
(* interface to our implementation; through     *)
(* some Noweb hackery, they are checked by the  *)
(* ML compiler.                                 *)
(* <boxed values 6>=                            *)
val _ = op emptyEnv : 'a env
val _ = op find     : name * 'a env -> 'a
val _ = op bind     : name      * 'a      * 'a env -> 'a env
val _ = op bindList : name list * 'a list * 'a env -> 'a env
(* [[error]] procedure,                         *)
(* to indicate when things have gone wrong. The *)
(* exceptions we use are listed in Table [->].  *)



(*****************************************************************)
(*                                                               *)
(*   LEXICAL ANALYSIS                                            *)
(*                                                               *)
(*****************************************************************)

(* A value of type [[par]] represents a list of *)
(* tokens with balanced parentheses. In order   *)
(* to issue better error messages, we keep the  *)
(* source-code location of the each             *)
(* parenthesized list. A source names the       *)
(* source and gives a line number.              *)
(* <lexical analysis>=                          *)
type srcloc  = string * int
fun srclocString (source, line) = source ^ ", line " ^ Int.toString line
(* <lexical analysis>=                          *)
datatype par = NAME  of string
             | INT   of int
             | SHARP of char  (* #t or #f *)
             | LIST  of srcloc * par list
(* <lexical analysis>=                          *)

exception Paren of char list * srcloc
   (* raise if found ) when reading input *)
exception PrematureEOF of string (* shows what we were reading ... *)
exception EOF                 (* raised by reader *)
fun read nextline curloc : string -> par list =
  let (* To help with reading, we have functions to   *)
      (* deal with lexical details. We begin by       *)
      (* defining tests for the delimiters of         *)
      (* symbols:                                     *)
      (* <character-fiddling functions>=              *)
      fun isDelim c = List.exists (fn c' => c' = c) [#"(", #")", #" ", #"\t",
                                                                    #"\n", #";"]
      (* [[]] is like the micro-Scheme                *)
      (* [[exists?]]. We needn't write its            *)
      (* implementation because it is in the initial  *)
      (* basis for Standard ML.                       *)

      (* <character-fiddling functions>=              *)
      fun stringReader []     = NONE
        | stringReader (h::t) = SOME (h, t)
      val strton = Int.scan StringCvt.DEC stringReader
      (* We continue with integer conversion;         *)
      (* [[strton s]] returns an integer, plus        *)
      (* leftover characters from [[s]], or [[NONE]]  *)
      (* if no prefix of [[s]] is an integer literal. *)
      (* It uses [[Int.scan]] from the Standard Basis *)
      (* Library.                                     *)
      (* <boxed values 2>=                            *)
      val _ = op strton : char list -> (int * char list) option
      (* We need to be able to skip whitespace and    *)
      (* comments. [[StringCvt.skipWS]] skips white   *)
      (* space; if we encounter a semicolon, we use   *)
      (* [[StringCvt.dropl]] to skip everything up to *)
      (* the next newline, then resume skipping white *)
      (* space.                                       *)
      (* <character-fiddling functions>=              *)
      fun skip s =
        case StringCvt.skipWS stringReader s
          of #";" :: s => skip (StringCvt.dropl (fn c => c <> #"\n")
                                                                 stringReader s)
           | s => s
      (* [[nonwhite]], which                          *)
      (* summons more input lines as needed.          *)

      (* <character-fiddling functions>=              *)
      fun nonwhite s =
        case skip s
          of [] => nonwhite (explode (nextline()))
           | s  => s
      fun readmap f (result, unconsumed) = (f result, unconsumed)
      (* get is called only when s is guaranteed non-blank, non-comment, non-
                                                                        empty *)
      fun get (#"("::s)  = readmap (fn l => LIST l) (list (curloc(), s, []))
        | get (#")"::s)  = raise (Paren (s, curloc()))
        | get (#"'"::s)  = let val loc = curloc()
                               fun quote p = LIST (loc, [NAME "quote", p])
                           in  readmap quote (get (nonwhite s))
                           end
        | get s          = atom (s, [])
      and atom(c::s, r) = if isDelim c then (mkAtom(rev r), c::s) 
                          else atom(s, c::r)
        | atom([],   r) = (mkAtom(rev r), [])
      and mkAtom [#"#", #"f"] = SHARP #"f"
        | mkAtom [#"#", #"t"] = SHARP #"t"
        | mkAtom s = (case strton s of SOME (n, []) => INT n | _ => NAME (
                                                                     implode s))
      and list (loc, s, l) =
        let val next = nonwhite s
        in  let val (a, u) = get next
            in  list(loc, u, a::l)
            end handle (Paren (u, _)) => ((loc, rev l), skip u)
                     | EOF => raise PrematureEOF (implode next)
        end handle EOF => raise PrematureEOF (implode s)
      fun readline [] = []
        | readline s  = 
             let val(p, u) = get s
             in  p :: readline(skip u)
             end
(* The interface to the [[read]] function       *)
(* includes a [[curloc]] parameter, which can   *)
(* be used to keep track of the source-code     *)
(* location of any token. Of the interpreters   *)
(* in this book, only the uSmalltalk interpter  *)
(* (Appendix [->]) takes advantage of this      *)
(* capability.                                  *)
(*                                              *)
(* Function [[read]] is implemented by a group  *)
(* of local lexing functions. Each such         *)
(* function takes a list of unconsumed          *)
(* characters and returns a pair consisting of  *)
(* the [[par]] read plus unconsumed characters. *)
(* The functions are:                           *)
(*                                              *)
(*   [[get]] --- Read the next input            *)
(*   [[atom]] --- Read a name, integer, or      *)
(*   [[#]] literal                              *)
(*   [[list]] --- Read a list of inputs in      *)
(*   parentheses                                *)
(*                                              *)
(* The [[list]] function takes as an additional *)
(* parameter the source-code location of the    *)
(* start of the list. Internal function         *)
(* [[readline]] puts it all together by reading *)
(* as many inputs are as on the line, until all *)
(* the line is consumed.                        *)
(* <boxed values 1>=                            *)
val _ = op read : (unit -> string) -> (unit -> srcloc) -> string -> par list
val _ = op get  : char list                      -> par                 * char
                                                                            list
val _ = op atom : char list * char list          -> par                 * char
                                                                            list
val _ = op list : srcloc * char list * par list  -> (srcloc * par list) * char
                                                                            list
  in  fn s => readline(skip (explode s)) 
  end


(*****************************************************************)
(*                                                               *)
(*   ABSTRACT SYNTAX AND VALUES                                  *)
(*                                                               *)
(*****************************************************************)

(*   _______________________________________    *)
(*                                              *)
(* Exceptions raised at parse time              *)
(* Paren We found an extra right parenthesis    *)
(* when reading input.                          *)
(* PrematureEOF The input ended while we were   *)
(* still trying to read characters (probably    *)
(* because of a missing right parenthesis).     *)
(* SyntaxError Something else went wrong during *)
(* parsing, i.e., during the execution of       *)
(* [[readtop]].                                 *)
(*                                              *)
(* Exceptions raised at run time                *)
(* NotFound A name was looked up in an          *)
(* environment but not found there.             *)
(* BindListLength A call to [[bindList]] tried  *)
(* to extend an environment, but it passed two  *)
(* lists (names and values) of different        *)
(* lengths.                                     *)
(* RuntimeError Something else went wrong       *)
(* during evaluation, i.e., during the          *)
(* execution of [[eval]].                       *)
(*                                              *)
(* Exceptions defined especially for this       *)
(* interpreter [*]                              *)
(*   _______________________________________    *)
(*                                              *)
(* Abstract syntax and values                   *)
(*                                              *)
(* An abstract-syntax tree can contain a        *)
(* literal value. A value, if it is a closure,  *)
(* can contain an abstract-syntax tree. These   *)
(* two types are therefore mutually recursive,  *)
(* and in ML, two mutually recursive types have *)
(* to be declared together, using [[and]]. [ML  *)
(* is more restrictive than C, in which we can  *)
(* use incomplete structures to distribute the  *)
(* definitions of mutually recursive types over *)
(* distant source locations, even in different  *)
(* files. Similar results can be achieved in ML *)
(* by using ``two-level types''                 *)
(* [cite sheard:generic], but the details are   *)
(* beyond the scope of this book.]              *)
(* <abstract syntax and values>=                *)
datatype       exp = LITERAL of value
                   | VAR     of name
                   | SET     of name * exp
                   | IFX     of exp * exp * exp
                   | WHILEX  of exp * exp
                   | BEGIN   of exp list
                   | LETX    of let_kind * (name * exp) list * exp
                   | LAMBDA  of lambda
                   | APPLY   of exp * exp list
and       let_kind = LET | LETREC | LETSTAR
and          value = NIL
                   | BOOL      of bool   
                   | NUM       of int
                   | SYM       of name
                   | PAIR      of value * value
                   | CLOSURE   of lambda * value ref env
                   | PRIMITIVE of primitive
withtype primitive = value list -> value (* raises RuntimeError *)
     and lambda    = name list * exp

exception RuntimeError of string (* error message *)
(* [[value list ->                              *)
(* value]]. This simple representation requires *)
(* no names or tags, and it is suggested by a   *)
(* principle of functional programming: don't   *)
(* encode a function as data if you can use the *)
(* function itself. When we apply a primitive   *)
(* function and something goes wrong, we raise  *)
(* the [[RuntimeError]] exception.              *)

(* The top-level items are straightforward.     *)
(*                                              *)
(* <abstract syntax and values>=                *)
datatype toplevel = EXP    of exp
                  | DEFINE of name * lambda
                  | VAL    of name * exp
                  | USE    of name


(*****************************************************************)
(*                                                               *)
(*   VALUES                                                      *)
(*                                                               *)
(*****************************************************************)

(* <values>=                                    *)



fun embedList []     = NIL
  | embedList (h::t) = PAIR(h, embedList t)
fun embedPredicate f = fn x => BOOL (f x)
fun bool (BOOL b) = b
  | bool _        = true
(*    Embedding and projection                  *)
(*                                              *)
(* An S-expression can represent an integer,    *)
(* Boolean, name, function, list, etc. We may   *)
(* sometimes have an ML boolean, list, or       *)
(* function that we wish to represent as an     *)
(* S-expression, or similarly, an S-expression  *)
(* that we wish to represent as a value of      *)
(* type [[bool]]. Here we define mappings       *)
(* between type [[value]] and some other        *)
(* ML types. Because the set of values          *)
(* representable by an ML value of type         *)
(* [[value]] strictly contains each of the sets *)
(* of values representable by these ML types,   *)
(* these mappings are called embedding and      *)
(* projection. Because the [[value]] type is    *)
(* strictly larger than these ML types, no      *)
(* embedding operation ever fails, but a        *)
(* projection operation might. [This property   *)
(* is a general characteristic of any           *)
(* embedding/projection pair. Mathematical      *)
(* terminology may clarify; an embedding e of   *)
(* S into S' is an injection from S-->S'. The   *)
(* corresponding projection pi_e is a left      *)
(* inverse of the embedding; that is pi_e oe is *)
(* the identity function on S. There is no      *)
(* corresponding guarantee for e opi_e; for     *)
(* example, pi_e may be undefined (_|_) on some *)
(* elements of S', or e(pi_e(x)) may not        *)
(* equal x.] For example, although any ML       *)
(* function of type [[value -> bool]] can be    *)
(* embedded into [[value]] by using the         *)
(* [[PRIMITIVE]] constructor, there are values  *)
(* of type [[value]] that cannot be projected   *)
(* into an ML function of type [[value ->       *)
(* bool]].                                      *)
(*                                              *)
(* Lists and Booleans are straightforward.      *)
(* Function [[embedPredicate]] is not a true    *)
(* embedding; it takes any function returning   *)
(* [[bool]] and returns a corresponding         *)
(* function returning [[value]]. It really      *)
(* embeds the function's result, not the        *)
(* function itself.                             *)
(* <boxed values 7>=                            *)
val _ = op embedList      : value list -> value
val _ = op embedPredicate : ('a -> bool) -> ('a -> value)
val _ = op bool           : value -> bool
(* [[bool]] is the projection                   *)
(* function, mapping micro-Scheme values into   *)
(* ML Booleans.                                 *)

(* <values>=                                    *)
fun valueString (NIL)    = "()"
  | valueString (BOOL b) = if b then "#t" else "#f"
  | valueString (NUM n)  = String.map (fn #"~" => #"-" | c => c) (Int.toString n
                                                                               )
  | valueString (SYM v)  = v
  | valueString (PAIR(car, cdr))  = 
      let fun tail (PAIR(car, cdr)) = " " ^ valueString car ^ tail cdr
            | tail NIL = ")"
            | tail v = " . " ^ valueString v ^ ")"
      in  "(" ^ valueString car ^ tail cdr
      end
  | valueString (CLOSURE   _) = "<procedure>"
  | valueString (PRIMITIVE _) = "<procedure>"
(*    Printing                                  *)
(*                                              *)
(* We render an S-expression as a string.       *)
(* <boxed values 8>=                            *)
val _ = op valueString : value -> string
(* [[Int.toString]] indicates the               *)
(* [[toString]] function from the standard      *)
(* module [[Int]]. This function, which is part *)
(* of ML's Standard Basis Library, converts an  *)
(* integer to a string. We use another standard *)
(* function, [[String.map]], to change the      *)
(* minus sign from the ML convention ([[ ]]) to *)
(* the Scheme convention ([[-]]).               *)



(*****************************************************************)
(*                                                               *)
(*   PARSING                                                     *)
(*                                                               *)
(*****************************************************************)

(* Parsing                                      *)
(*                                              *)
(* Parsing begins with some fairly elaborate    *)
(* error detection, which is intended to        *)
(* duplicate the error messages used in         *)
(* Appendix [->].                               *)
(* <parsing>=                                   *)
exception SyntaxError of string
fun syntaxErrorAt loc msg =
  raise SyntaxError (srclocString loc ^ ": " ^ msg)
fun nodups msg loc l =
  let fun dup [] = l
        | dup (h::t) = if List.exists (fn x => x = h) t then
                         syntaxErrorAt loc (msg h)
                       else
                         dup t
  in  dup l
  end
val lambdaDups   =
  nodups (fn x => "formal parameter " ^ x ^ " appears twice in lambda")
fun defineDups loc f =
  nodups (fn x => "formal parameter " ^ x ^ " appears twice in " ^
                  "definition of function " ^ f) loc
fun letDups loc (LETSTAR, bindings) = bindings
  | letDups loc (kind,    bindings) =
      let val names    = map (fn (n, _) => n) bindings
          val kindName = case kind of LET => "let" | LETREC => "letrec" | _ =>
                                                                            "??"
          fun msg x = "name " ^ x ^ " appears twice in " ^ kindName
      in  nodups msg loc names;
          bindings
      end
(* We parse concrete syntax into abstract       *)
(* syntax. This operation is relatively         *)
(* uninteresting. The auxiliary functions may   *)
(* help clarify the logic of parsing.           *)
(* <parsing>=                                   *)
fun parse (NAME s)  = VAR s
  | parse (INT n)   = LITERAL (NUM n)
  | parse (SHARP c) = LITERAL (BOOL (c = #"t"))
  | parse (LIST (loc, NAME "quote"  :: args)) = parseQuote  loc args
  | parse (LIST (loc, NAME "if"     :: args)) = parseIf     loc args
  | parse (LIST (loc, NAME "while"  :: args)) = parseWhile  loc args
  | parse (LIST (loc, NAME "set"    :: args)) = parseSet    loc args
  | parse (LIST (loc, NAME "begin"  :: args)) = BEGIN (map parse args)
  | parse (LIST (loc, NAME "lambda" :: args)) = parseLambda loc args
  | parse (LIST (loc, NAME "let"    :: args)) = parseLet loc LET     args
  | parse (LIST (loc, NAME "letrec" :: args)) = parseLet loc LETREC  args
  | parse (LIST (loc, NAME "let*"   :: args)) = parseLet loc LETSTAR args
  | parse (LIST (loc, h::t))  = APPLY (parse h, map parse t)
  | parse (LIST (loc, [])) = raise SyntaxError "empty application ()"
and parseQuote loc [e] = LITERAL (quote loc e)
  | parseQuote loc _ = syntaxErrorAt loc "expected (quote e)"
and parseIf _ [e1, e2, e3] = IFX (parse e1, parse e2, parse e3)
  | parseIf loc _ = syntaxErrorAt loc "expected (if e1 e2 e3)"
and parseWhile _ [e1, e2] = WHILEX (parse e1, parse e2)
  | parseWhile loc _ = syntaxErrorAt loc "expected (while e1 e2)"
and parseSet _ [NAME v, e] = SET (v, parse e)
  | parseSet loc _ = syntaxErrorAt loc "expected (set v e)"
and parseLambda _ [LIST (loc, formals), body] =
      LAMBDA (lambdaDups loc (map (unname loc) formals), parse body)
  | parseLambda loc _ =
      syntaxErrorAt loc "expected (lambda (names) body)"
and parseLet _ kind [LIST (loc, bindings), body] = 
      LETX(kind, letDups loc (kind, map (unBinding loc) bindings), parse body)
  | parseLet loc _ _ = syntaxErrorAt loc "expected (let (bindings) body)"
and unname _ (NAME n) = n
  | unname loc _ = syntaxErrorAt loc
                                 "formal parameters to function must be symbols"
and unBinding _ (LIST (_, [NAME n, e])) = (n, parse e)
  | unBinding loc _ = syntaxErrorAt loc "expected (v e) in bindings"
and quote loc (INT   n) = NUM n
  | quote loc (SHARP c) = BOOL (c = #"t")
  | quote loc (NAME  ".") =
      syntaxErrorAt loc
                      "this interpreter cannot handle . in quoted S-expressions"
  | quote loc (NAME  n) = SYM n
  | quote _ (LIST (loc, l)) = embedList (map (quote loc) l)
(* <parsing>=                                   *)
fun toplevel (LIST (loc, NAME "define" :: args)) = parseDefine loc args
  | toplevel (LIST (loc, NAME "val"    :: args)) = parseVal    loc args
  | toplevel (LIST (loc, NAME "global" :: args)) = parseVal    loc args (*
                                                                       legacy *)
  | toplevel (LIST (loc, NAME "use"    :: args)) = parseUse    loc args
  | toplevel e = EXP (parse e)
and parseDefine _ [NAME name, LIST (loc, formals), body] =
      DEFINE (name, (defineDups loc name (map (unname loc) formals), parse body)
                                                                               )
  | parseDefine loc _ = syntaxErrorAt loc "expected (define f (args) body)"
and parseVal _ [NAME v, e] = VAL (v, parse e)
  | parseVal loc _ = syntaxErrorAt loc "expected (val v e)"
and parseUse _ [NAME v] = USE v
  | parseUse loc _ = syntaxErrorAt loc "expected (use filename)"


(*****************************************************************)
(*                                                               *)
(*   READERS                                                     *)
(*                                                               *)
(*****************************************************************)

(* Readers                                      *)
(*                                              *)
(*  Line readers                                *)
(*                                              *)
(* Our readers are structured slightly          *)
(* differently from those in C. A reader is a   *)
(* bundle containing three related functions:   *)
(* [[firstline]], which returns the first line  *)
(* read; [[nextline]], which returns all        *)
(* subsequent lines; and [[curloc]], which      *)
(* returns a pair identifying the source        *)
(* location of the most recently returned line. *)
(* Both [[firstline]] and [[nextline]] can      *)
(* raise [[EOF]], (We distinguish [[firstline]] *)
(* from [[nextline]] because in an interactive  *)
(* reader, they may use different prompts.)     *)
(* <readers>=                                   *)
type reader = { firstline : unit -> string
              , nextline  : unit -> string
              , curloc    : unit -> string * int
              }
(* <readers>=                                   *)
fun promptReader (firstPrompt, nextPrompt, reader) =
  let val { firstline = first, nextline = next, curloc = cur } = reader
      fun promptIn get prompt () = ( TextIO.output(TextIO.stdOut, prompt)
                                   ; TextIO.flushOut(TextIO.stdOut)
                                   ; get ()
                                   )
  in  { firstline = promptIn first firstPrompt
      , nextline  = promptIn next  nextPrompt
      , curloc    = cur
      }
  end
(* Given a silent reader, we can make it issue  *)
(* a prompt every time it reads a new line. We  *)
(* use two prompts: one for the first line of   *)
(* an expression and one for later lines.       *)
(* <boxed values 3>=                            *)
val _ = op promptReader : string * string * reader -> reader
(* <readers>=                                   *)
fun filereader (filename, fd)=
  let val lineNum = ref 0
      fun get () =
        let val line = TextIO.inputLine fd
        in  if size line = 0 then raise EOF else line before lineNum := !lineNum
                                                                             + 1
        end
  in  { firstline = get, nextline = get, curloc = (fn () => (filename, !lineNum)
                                                                             ) }
  end

fun stringsreader (what, l) = 
  let val buffer  = ref l
      val lineNum = ref 0
      fun get ()  = case !buffer
                      of [] => raise EOF
                       | h :: t => h before buffer := t before lineNum := !
                                                                     lineNum + 1
  in  { firstline = get, nextline = get, curloc = (fn () => (what, !lineNum)) }
  end              
(* We can read from files or lists of strings.  *)
(* File I/O is provide by the standard module   *)
(* [[TextIO]].                                  *)
(* <boxed values 4>=                            *)
val _ = op filereader    : string * TextIO.instream -> reader
val _ = op stringsreader : string * string list     -> reader
(* [[]], a private mutable cell. When all       *)
(* the strings in the buffer are used up, there *)
(* aren't any more.                             *)

(*  Top-level readers                           *)
(*                                              *)
(* A top-level reader is a buffer of top-level  *)
(* items together with a reader. [*]            *)
(* <readers>=                                   *)
type topreader = { buffer : toplevel list ref, reader : reader }
fun readtop (r as { buffer, reader = { nextline, firstline, curloc } }) =
  case !buffer
    of h::t => h before buffer := t
     | []   => ( buffer := map toplevel (read nextline curloc (firstline()))
               ; readtop r
               )
(* We decide whether to prompt when building a  *)
(* top-level reader.                            *)
(* <readers>=                                   *)
fun topreader (reader, prompt) = 
  let val reader = if prompt then promptReader ("-> ", "   ", reader) else
                                                                          reader
  in  { buffer = ref [], reader = reader }
  end
(* <readers>=                                   *)
fun echoTag f x =
  let val line = f x
      val _ = if (String.substring (line, 0, 2) = ";#" handle _ => false) then
                print line
              else
                ()
  in  line
  end

fun echoReader { nextline=n, firstline=f, curloc=c } =
     { nextline=echoTag n, firstline=echoTag f, curloc=c } 

fun echoTop { buffer = b, reader = r } = { buffer = b, reader = echoReader r }
val topreader = fn args => echoTop (topreader args)
(* As in the C code, we want to print out any   *)
(* line read if it begins with the special      *)
(* string [[;#]]. This string is a formal       *)
(* comment that helps us test the chunks marked *)
(* [[]]. In our ML code, we                     *)
(* arrange it using the higher-order function   *)
(* [[echoReader]].                              *)
(* <boxed values 5>=                            *)
val _ = op echoReader : reader -> reader 


(*****************************************************************)
(*                                                               *)
(*   IMPLEMENTATION OF [[USE]]                                   *)
(*                                                               *)
(*****************************************************************)

(* The implementation of [[use]] is             *)
(* parameterized by [[readEvalPrint]] so we can *)
(* share it with other interpreters. Function   *)
(* [[use]] creates a top-level reader that does *)
(* not prompt, but it uses [[writeln]] to be    *)
(* sure that responses are printed.             *)
(* <implementation of [[use]]>=                 *)
fun use readEvalPrint filename rho =
      let val fd = TextIO.openIn filename
          val treader = topreader (filereader (filename, fd), false)
          fun writeln s = app print [s, "\n"]
      in  readEvalPrint (treader, writeln, writeln) rho
          before TextIO.closeIn fd
      end 
(* [[topreader]] is relegated                   *)
(* to chunk [->] in Appendix [->].              *)



(*****************************************************************)
(*                                                               *)
(*   EVALUATION                                                  *)
(*                                                               *)
(*****************************************************************)

(* <evaluation>=                                *)
(* Function [[separate]] helps print a readable *)
(* list of formals for the error message.       *)
(* <definition of [[separate]]>=                *)
fun separate (zero, sep) =  (* useful for printing *)
  let fun s []     = zero
        | s [x]    = x
        | s (h::t) = h ^ sep ^ s t
  in  s
end
fun eval(e, rho) =
  let fun ev(LITERAL n) = n
        (* Because [[rho]] binds names to mutable ref   *)
        (* cells, not to values directly, we need the   *)
        (* ML functions [[!]] and [[:=]] to read and    *)
        (* change the contents of the ref cells. The    *)
        (* right-hand side of [[SET]] is evaluated in   *)
        (* the same environment as the [[SET]], so we   *)
        (* use [[ev]].                                  *)
        (* <more alternatives for [[ev]]>=              *)
        | ev(VAR v) = !(find(v, rho))
        | ev(SET (n, e)) = 
            let val v = ev e
            in  find (n, rho) := v;
                v
            end
        (* Using the projection function [[bool]] makes *)
        (* it easy to map micro-Scheme control flow     *)
        (* onto ML control-flow constructs.             *)
        (* <more alternatives for [[ev]]>=              *)
        | ev(IFX (e1, e2, e3)) = ev (if bool (ev e1) then e2 else e3)
        | ev(WHILEX (guard, body)) = 
            if bool (ev guard) then 
              (ev body; ev(WHILEX (guard, body)))
            else
              NIL
        | ev(BEGIN es) =
            let fun b(e::es, lastval) = b(es, ev e)
                  | b(   [], lastval) = lastval
            in  b(es, BOOL false)
            end
        (* Capturing a closure is as simple as in C.    *)
        (* Applying a primitive function is even        *)
        (* simpler; we just evaluate the arguments and  *)
        (* apply the primitive to the results.          *)
        (* <more alternatives for [[ev]]>=              *)
        | ev(LAMBDA l) = CLOSURE(l, rho)
        | ev(APPLY (f, args))  = 
               (case ev f
                  of PRIMITIVE prim => prim (map ev args)
                   | CLOSURE clo    => (* <apply closure [[clo]] to [[args]]>=
                                                                              *)
                                       let val ((formals, body), savedrho) = clo
                                           val actuals = map ev args
                                       in  eval(body, bindList(formals, map ref
                                                             actuals, savedrho))
                                           handle BindListLength => 
                                               raise RuntimeError (
                                      "Wrong number of arguments to closure; " ^
                                                                   "expected ("
                                              ^ separate("", " ") formals ^ ")")
                                       end
                   | v => raise RuntimeError ("Applied non-function " ^
                                                                  valueString v)
               )
        (* [[allocate]] for this                        *)
        (* purpose; in ML, the built-in function        *)
        (* [[ref]] does the same thing: create a new    *)
        (* location and initialize its contents with a  *)
        (* value.                                       *)

        (* <more alternatives for [[ev]]>=              *)
        | ev(LETX (LET, bs, body)) =
            let val (names, values) = ListPair.unzip bs
        (* To interpret [[let]], it is easiest to unzip *)
        (* the list of pairs [[bs]] into a pair of      *)
        (* lists [[(names, values)]]. For [[let*]],     *)
        (* however, it is easier to walk the bindings   *)
        (* one pair at a time. The implementation of    *)
        (* [[letrec]] does both.                        *)
        (*                                              *)
        (* The function [[ListPair.unzip]] is from the  *)
        (* [[ListPair]] module in the Standard Basis    *)
        (* Library.                                     *)
        (* <boxed values 10>=                           *)
        val _ = ListPair.unzip : ('a * 'b) list -> 'a list * 'b list
            in  eval (body, bindList(names, map (ref o ev) values, rho))
            end
        | ev(LETX (LETSTAR, bs, body)) =
            let fun step ((n, e), rho) = bind(n, ref (eval(e, rho)), rho)
            in  eval (body, foldl step rho bs)
            end
        | ev(LETX (LETREC, bs, body)) =
            let val (names, values) = ListPair.unzip bs
                val rho' = bindList(names, map (fn _ => ref NIL) values, rho)
                val bs = map (fn (n, e) => (n, eval(e, rho'))) bs
            in  List.app (fn (n, v) => find(n, rho') := v) bs; 
                eval(body, rho')
            end
  in  ev e
  end
(* Evaluation                                   *)
(*                                              *)
(* The machinery above is enough to write the   *)
(* evaluator, which takes an expression and an  *)
(* environment and produces a value. To make    *)
(* the evaluator easy to write, we do most of   *)
(* the work of evaluation in the nested         *)
(* function [[ev]], which inherits the          *)
(* environment [[rho]] from the outer function  *)
(* [[eval]]. Because most AST nodes are         *)
(* evaluated in the same environment as their   *)
(* parents, we can evaluate most of them by     *)
(* calling [[ev]], which lets [[rho]] be        *)
(* implicit.                                    *)
(* <boxed values 9>=                            *)
val _ = op eval : exp * value ref env -> value
(* <evaluation>=                                *)
fun arityError n args =
  raise RuntimeError ("primitive function expected " ^ Int.toString n ^
                      "arguments; got " ^ Int.toString (length args))
fun binaryOp f = (fn [a, b] => f(a, b) | args => arityError 2 args)
fun unaryOp  f = (fn [a]    => f a     | args => arityError 1 args)
(* Primitives                                   *)
(*                                              *)
(* Here are the primitives. All are either      *)
(* binary or unary operators. As in C, we would *)
(* like to reuse the code that does the arity   *)
(* checks. In ML, it's easy; we use             *)
(* higher-order functions to write the arity    *)
(* checks just once. If a check fails, function *)
(* [[arityError]] raises [[RuntimeError]] with  *)
(* a suitable message.                          *)
(* <boxed values 11>=                           *)
val _ = op unaryOp  : (value         -> value) -> (value list -> value)
val _ = op binaryOp : (value * value -> value) -> (value list -> value)
(* <evaluation>=                                *)
fun arithOp f = binaryOp (fn (NUM n1, NUM n2) => NUM (f(n1, n2)) 
                           | _ => raise RuntimeError "integers expected")
(* Arithmetic primitives expect and return      *)
(* integers. As in C, we reuse the code that    *)
(* projects two arguments to integers, but as   *)
(* above, we do it using higher-order           *)
(* functions.                                   *)
(* <boxed values 12>=                           *)
val _ = op arithOp: (int * int -> int) -> (value list -> value)
(* [[]]                                         *)
(* names into one giant list. We use that list  *)
(* to build the initial environment for the     *)
(* read-eval-print loop; it plays the same role *)
(* as the unspeakable [[xx]] macros in file     *)
(* prim.h in the C implementation (chunk [->]). *)

(* <evaluation>=                                *)
fun predOp f     = unaryOp  (embedPredicate f)
fun comparison f = binaryOp (embedPredicate f)
fun intcompare f = comparison (fn (NUM n1, NUM n2) => f(n1, n2)
                                | _ => raise RuntimeError "integers expected")
(* We have two kinds of predicate: an ordinary  *)
(* predicate takes one argument and a           *)
(* comparison takes two. Some comparisons apply *)
(* only to integers. We reuse                   *)
(* [[embedPredicate]] for the definitions.      *)
(* <boxed values 13>=                           *)
val _ = op predOp     : (value         -> bool) -> (value list -> value)
val _ = op comparison : (value * value -> bool) -> (value list -> value)
val _ = op intcompare : (int   * int   -> bool) -> (value list -> value)
(* <evaluation>=                                *)
fun topeval (t, rho, echo) =
  case t
    of USE filename => use readEvalPrint filename rho
     | EXP e        => (echo (valueString (eval(e, rho))); rho)
     | DEFINE (name, lambda) => 
         topeval (VAL (name, LAMBDA lambda), rho, echo)
     | VAL (name, e) =>
         let val rho = (find(name, rho); rho)
                       handle NotFound _ => bind (name, ref NIL, rho)
             val v   = eval(e, rho)
         in  find(name, rho) := v;
             echo (showVal name e v);
             rho
         end
and showVal name (LAMBDA _) _ = name
  | showVal name _          v = valueString v
(* <evaluation>=                                *)
and readEvalPrint (reader, echo, errmsg) rho =
  let fun loop rho =
        let fun continue msg = (errmsg msg; loop rho)
            fun finish () = rho
        in  loop (topeval (readtop reader, rho, echo))
            handle EOF => finish()
                   (* The handlers in the first group deal with    *)
                   (* problems that arise during I/O, lexical      *)
                   (* analysis, and parsing.                       *)
                   (* <more read-eval-print handlers>=             *)
                   | IO.Io {name, ...} => continue ("I/O error: " ^ name)
                   | Paren (_, loc)    => continue ("syntax error: extra ) in "
                                                             ^ srclocString loc)
                   | PrematureEOF s    => (errmsg (
                                  "Missing )? Got EOF reading (" ^ s); finish())
                   | SyntaxError msg   => continue ("syntax error: " ^ msg)
                   (* [[IO.Io]] is part of the                     *)
                   (* Standard Basis Library.                      *)

                   (* The handlers in the second group deal with   *)
                   (* problems that arise during evaluation. They  *)
                   (* always continue with the next top-level      *)
                   (* item.                                        *)
                   (* <more read-eval-print handlers>=             *)
                   | Div               => continue "Division by zero"
                   | Overflow          => continue "Arithmetic overflow"
                   | RuntimeError msg  => continue ("run-time error: " ^ msg)
                   | NotFound n        => continue ("variable " ^ n ^
                                                                   " not found")
                   (* [[Div]] and [[Overflow]] are                 *)
                   (* predefined.                                  *)

        end
  in  loop rho
  end
(* Top-level evaluation                         *)
(*                                              *)
(* As in Chapter [->], the implementation of    *)
(* the top-level rules is straightforward.      *)
(* Function [[topeval]] takes a top-level item  *)
(* and an environment, and it returns a new     *)
(* environment. It also takes [[echo]], which   *)
(* is not a flag but a function used to print   *)
(* responses. When we get an expression, we     *)
(* print its value. When we get a [[val]]       *)
(* binding, we print the result of [[showVal]], *)
(* which is either the value of the right-hand  *)
(* side or, in the case of a binding to a       *)
(* function, the name of the function. As       *)
(* usual, [[define]] is syntactic sugar.        *)
(* <boxed values 14>=                           *)
val _ = op topeval : toplevel * value ref env * (string->unit) -> value ref env
(* The read-eval-print loop                     *)
(*                                              *)
(* The read-eval-print loop is built around the *)
(* inner function [[loop]], which repeatedly    *)
(* calls [[readtop reader]] for a top-level     *)
(* item and [[topeval]] to get a new            *)
(* environment. Function [[loop]] takes one     *)
(* argument [[rho]], which represents the       *)
(* current environment. As in [[topeval]],      *)
(* [[echo]] is not a flag but a function, which *)
(* takes a string and may print it or do        *)
(* nothing. The argument [[errmsg]] is similar  *)
(* to [[echo]], but it is used to issue error   *)
(* messages.                                    *)
(* <boxed values 14>=                           *)
val _ = op readEvalPrint : topreader * (string->unit) * (string->unit) -> value
                                                        ref env -> value ref env
(* [[topeval]] and [[readEvalPrint]]            *)
(* are mutually recursive, so we use [[and]]    *)
(* instead of [[fun]].                          *)

(* [[loop]] runs until something                *)
(* raises an exception. The executions of       *)
(* [[readtop]] and [[topevel]] are protected by *)
(* a large collection of exception handlers,    *)
(* which may do one of two things:              *)
(*   * If the input is not exhausted, call      *)
(*     [[continue]] to print an error message   *)
(*     and continue looping.                    *)
(*   * If the input is exhausted, call          *)
(*     [[finish]] to return the last good       *)
(*     environment.                             *)
(*                                              *)
(* We reuse the same exception handlers in      *)
(* later interpreters, sometimes with different *)
(* implementations of [[finish]] and            *)
(* [[continue]].                                *)



(*****************************************************************)
(*                                                               *)
(*   INITIALIZATION                                              *)
(*                                                               *)
(*****************************************************************)

(* Initializing and running the interpreter     *)
(*                                              *)
(* To make a working interpreter, we need an    *)
(* initial environment. We create the           *)
(* environment by starting with the empty       *)
(* environment, binding the primitive           *)
(* operators, then adding the initial basis.    *)
(* <initialization>=                            *)
fun initialEnv() =
  let val rho = foldl (fn ((name, prim), rho) => bind(name, ref (PRIMITIVE prim)
                                                                         , rho))
                      emptyEnv ((* <primitives [[::]]>=
                                                                              *)
                                ("+", arithOp op +  ) :: 
                                ("-", arithOp op -  ) :: 
                                ("*", arithOp op *  ) :: 
                                ("/", arithOp op div) ::
                                (* [[op]] makes it possible to
                                                                              *)
                                (* use an infix identifier as an ordinary
                                                                              *)
                                (* value, so [[arithOp op +]] passes the
                                                                              *)
                                (* value [[+]] (a binary function) to the
                                                                              *)
                                (* function [[arithOp]].
                                                                              *)

                                (* <primitives [[::]]>=
                                                                              *)
                                ("<", intcompare op <) :: 
                                (">", intcompare op >) ::
                                ("=", comparison (fn (NIL,     NIL    ) => true
                                                   | (NUM  n1, NUM  n2) => n1 =
                                                                              n2
                                                   | (SYM  v1, SYM  v2) => v1 =
                                                                              v2
                                                   | (BOOL b1, BOOL b2) => b1 =
                                                                              b2
                                                   |  _                 => false
                                                                           )) ::
                                ("null?",    predOp (fn (NIL   ) => true | _ =>
                                                                      false)) ::
                                ("boolean?", predOp (fn (BOOL _) => true | _ =>
                                                                      false)) ::
                                ("number?",  predOp (fn (NUM  _) => true | _ =>
                                                                      false)) ::
                                ("symbol?",  predOp (fn (SYM  _) => true | _ =>
                                                                      false)) ::
                                ("pair?",    predOp (fn (PAIR _) => true | _ =>
                                                                      false)) ::
                                ("procedure?",
                                             predOp (fn (PRIMITIVE _) => true |
                                          (CLOSURE  _) => true | _ => false)) ::
                                (* The list primitives are easy:
                                                                              *)
                                (*
                                                                              *)
                                (* <primitives [[::]]>=
                                                                              *)
                                ("cons", binaryOp (fn (a, b) => PAIR(a, b))) ::
                                ("car",  unaryOp  (fn (PAIR(car, _)) => car 
                                                    | v => raise RuntimeError
                                                                   (
                                "car applied to non-list " ^ valueString v))) ::
                                ("cdr",  unaryOp  (fn (PAIR(_, cdr)) => cdr 
                                                    | v => raise RuntimeError
                                                                   (
                                "cdr applied to non-list " ^ valueString v))) ::
                                (* Finally, the only remaining primitives are
                                                                              *)
                                (* [[print]] and [[error]]:
                                                                              *)
                                (* <primitives [[::]]>=
                                                                              *)
                                ("print", unaryOp (fn v => (print (valueString v
                                                              ^"\n"); v)))    ::
                                ("error", unaryOp (fn v => raise RuntimeError (
                                                        valueString v))) :: nil)
      val basis  = (* <ML representation of initial basis>=        *)
                     [ "; scheme.nw 444"
                     , "(define caar (l) (car (car l)))"
                     , "(define cadr (l) (car (cdr l)))"
                     , "(define cdar (l) (cdr (car l)))"
                     , "; scheme.nw 464"
                     , "(define list1 (x)     (cons x '()))"
                     , "(define list2 (x y)   (cons x (list1 y)))"
                     , "(define list3 (x y z) (cons x (list2 y z)))"
                     , "; scheme.nw 485"
                     , "(define length (l)"
                     , "  (if (null? l) 0"
                     , "    (+ 1 (length (cdr l)))))"
                     , "; scheme.nw 3712"
                     , "(define and (b c) (if b  c  b))"
                     , "(define or  (b c) (if b  b  c))"
                     , "(define not (b)   (if b #f #t))"
                     , "; scheme.nw 528"
                     ,
"(define atom? (x) (or (number? x) (or (symbol? x) (or (boolean? x) (null? x)))))"
                     , "(define equal? (s1 s2)"
                     , "  (if (or (atom? s1) (atom? s2))"
                     , "    (= s1 s2)"
                     ,
             "    (and (equal? (car s1) (car s2)) (equal? (cdr s1) (cdr s2)))))"
                     , "; scheme.nw 548"
                     , "(define append (l1 l2)"
                     , "  (if (null? l1)"
                     , "     l2"
                     , "     (cons (car l1) (append (cdr l1) l2))))"
                     , "; scheme.nw 580"
                     , "(define revapp (l1 l2)"
                     , "  (if (null? l1)"
                     , "     l2"
                     , "     (revapp (cdr l1) (cons (car l1) l2))))"
                     , "; scheme.nw 737"
                     , "(define bind (x y alist)"
                     , "  (if (null? alist)"
                     , "    (list1 (list2 x y))"
                     , "    (if (equal? x (caar alist))"
                     , "      (cons (list2 x y) (cdr alist))"
                     , "      (cons (car alist) (bind x y (cdr alist))))))"
                     , "(define find (x alist)"
                     , "  (if (null? alist) '()"
                     , "    (if (equal? x (caar alist))"
                     , "      (car (cdar alist))"
                     , "      (find x (cdr alist)))))"
                     , "; scheme.nw 1636"
                     , "(define o (f g) (lambda (x) (f (g x))))"
                     , "; scheme.nw 1706"
                     , "(define curry   (f) (lambda (x) (lambda (y) (f x y))))"
                     , "(define uncurry (f) (lambda (x y) ((f x) y)))"
                     , "; scheme.nw 2014"
                     , "(define filter (p? l)"
                     , "  (if (null? l)"
                     , "    '()"
                     , "    (if (p? (car l))"
                     , "      (cons (car l) (filter p? (cdr l)))"
                     , "      (filter p? (cdr l)))))"
                     , "; scheme.nw 2026"
                     , "(define map (f l)"
                     , "  (if (null? l)"
                     , "    '()"
                     , "    (cons (f (car l)) (map f (cdr l)))))"
                     , "; scheme.nw 2036"
                     , "(define exists? (p? l)"
                     , "  (if (null? l)"
                     , "    #f"
                     , "    (if (p? (car l)) "
                     , "      #t"
                     , "      (exists? p? (cdr l)))))"
                     , "(define all? (p? l)"
                     , "  (if (null? l)"
                     , "    #t"
                     , "    (if (p? (car l))"
                     , "      (all? p? (cdr l))"
                     , "      #f)))"
                     , "; scheme.nw 2067"
                     , "(define foldr (op zero l)"
                     , "  (if (null? l)"
                     , "    zero"
                     , "    (op (car l) (foldr op zero (cdr l)))))"
                     , "(define foldl (op zero l)"
                     , "  (if (null? l)"
                     , "    zero"
                     , "    (foldl op (op (car l) zero) (cdr l))))"
                     , "; scheme.nw 3730"
                     , "(define <= (x y) (not (> x y)))"
                     , "(define >= (x y) (not (< x y)))"
                     , "(define != (x y) (not (= x y)))"
                     , "; scheme.nw 3736"
                     , "(define max (x y) (if (> x y) x y))"
                     , "(define min (x y) (if (< x y) x y))"
                     , "; scheme.nw 3743"
                     , "(define mod (m n) (- m (* n (/ m n))))"
                     , "(define gcd (m n) (if (= n 0) m (gcd n (mod m n))))"
                     , "(define lcm (m n) (* m (/ n (gcd m n))))"
                     , "; scheme.nw 3755"
                     , "(define caar (x) (car (car x)))"
                     , "(define cdar (x) (cdr (car x)))"
                     , "(define cadr (x) (car (cdr x)))"
                     , "(define cddr (x) (cdr (cdr x)))"
                     , "(define caaar (x) (car (caar x)))"
                     , "(define cdaar (x) (cdr (caar x)))"
                     , "(define caadr (x) (car (cadr x)))"
                     , "(define cdadr (x) (cdr (cadr x)))"
                     , "(define cadar (x) (car (cdar x)))"
                     , "(define cddar (x) (cdr (cdar x)))"
                     , "(define caddr (x) (car (cddr x)))"
                     , "(define cdddr (x) (cdr (cddr x)))"
                     , "; scheme.nw 3771"
                     , "(define list1 (x)               (cons x '()))"
                     , "(define list2 (x y)             (cons x (list1 y)))"
                     , "(define list3 (x y z)           (cons x (list2 y z)))"
                     , "(define list4 (x y z a)         (cons x (list3 y z a)))"
                     ,
                     "(define list5 (x y z a b)       (cons x (list4 y z a b)))"
                     ,
                   "(define list6 (x y z a b c)     (cons x (list5 y z a b c)))"
                     ,
                 "(define list7 (x y z a b c d)   (cons x (list6 y z a b c d)))"
                     ,
               "(define list8 (x y z a b c d e) (cons x (list7 y z a b c d e)))"
                     ]
      val treader = topreader (stringsreader ("initial basis", basis), false)
  in  readEvalPrint (treader, fn _ => (), fn _ => ()) rho
  end
(* <initialization>=                            *)
fun runInterpreter noisy = 
  let val rho = initialEnv()
      fun writeln s = app print [s, "\n"]
      val treader = topreader (filereader ("standard input", TextIO.stdIn),
                                                                          noisy)
  in  ignore (readEvalPrint (treader, writeln, writeln) rho) handle EOF => ()
  end 
(* The function [[runInterpreter]] takes one    *)
(* argument, which tells it whether to prompt.  *)
(* It reads from standard input.                *)
(* <boxed values 15>=                           *)
val _ = op runInterpreter : bool -> unit
(* [[writeln]]                                  *)
(* acts as both the [[echo]] and the [[errmsg]] *)
(* functions.                                   *)



(*****************************************************************)
(*                                                               *)
(*   COMMAND LINE                                                *)
(*                                                               *)
(*****************************************************************)

(* Building and exporting a program             *)
(*                                              *)
(* The final step in implementing the           *)
(* interpreter is to create a function that     *)
(* looks at the command line and calls          *)
(* [[runInterpreter]]. With a compiler like     *)
(* Moscow ML or MLton, the module isn't         *)
(* elaborated until run time, so we can simply  *)
(* insert an irrelevant [[val]] binding, the    *)
(* elaboration of which has the side effect of  *)
(* calling [[main]].                            *)
(* [[CommandLine.arguments()]] returns an       *)
(* argument list.                               *)
(* <command line>=                              *)
fun main ["-q"] = runInterpreter false
  | main []     = runInterpreter true
  | main _      =
      TextIO.output(TextIO.stdErr, "Usage: " ^ CommandLine.name() ^ " [-q]\n")
val _ = main (CommandLine.arguments())
