/*   Implementations of the primitives          */
/*                                              */
/* [*] We associate each primitive with a       */
/* unique tag, which identifies the primitive.  */
/* We also associate the primitive with a       */
/* function, which implements the primitive.    */
/* The tags make it easy for one function to    */
/* implement more than one primitive, which     */
/* makes it easy for similar primitives to      */
/* share code. We use the following functions   */
/* to implement primitives:                     */
/*                                              */
/* [[arith]] Arithmetic functions, which expect */
/* integers as arguments                        */
/* [[binary]] Non-arithmetic functions of two   */
/* arguments                                    */
/* [[unary]] Functions of one argument          */
/* <prim.c>=                                    */
#include "all.h"

static Primitive arith, binary, unary;
/* To define the primitives and create these    */
/* associations, we resort to macro madness.    */
/* Each primitive appears in file prim.h as a   */
/* macro [[xx(]]name[[, ]]tag[[,                */
/* ]]function[[)]]. We use the same macros with */
/* two different definitions of [[xx]]: one to  */
/* create an enumeration with distinct tags,    */
/* and one to install the primitives in an      */
/* empty environment. There are other           */
/* initialization techniques that don't require */
/* macros, but this technique guarantees that   */
/* the enumeration type is consistent with the  */
/* initialization code.                         */
/* <prim.c>=                                    */
enum {
  #define xx(NAME, TAG, FUNCTION) TAG,
  #include "prim.h"
  #undef xx
  UNUSED_TAG
};
/* In [[primenv]], the [[xx]] macros create the */
/* initial environment.                         */
/* <prim.c>=                                    */
Env*
primenv(void)
{
    Env *env;

    env = NULL;
    #define xx(NAME, TAG, FUNCTION) \
        env = bindalloc(strtoname(NAME), mkPrimitive(TAG, FUNCTION), env);
    #include "prim.h"
    #undef xx
    return env;
}
/* Each arithmetic primitive expects two        */
/* integer arguments, so we need to be able to  */
/* project micro-Scheme values into C integers. */
/* The projection function [[projectint]] takes */
/* not only a value but also an expression, so  */
/* it can issue an informative error message.   */
/* <prim.c>=                                    */
static int
projectint(Exp *e, Value v)
{
    if (v.ty != NUM)
        error("in %e, expected an integer, but got %v", e, v);
    return v.u.num;
}
/* Some of the arithmetic primitives,           */
/* like [[<]], return Booleans. Because we      */
/* implement these primitives using             */
/* C expressions, we need a function to take a  */
/* C Boolean and embed it as a micro-Scheme     */
/* value.                                       */
/* <prim.c>=                                    */
static Value
bool(int i)
{
    return i ? true : false;
}
/* We need special support for division,        */
/* because while micro-Scheme requires that     */
/* division round toward minus infinity,        */
/* C guarantees only that dividing positive     */
/* operands rounds toward zero.                 */
/* <prim.c>=                                    */
static int
divide(int n, int m)
{
    if (n >= 0)
        if (m >= 0)
            return n / m;
        else
            return -(( n - m - 1) / -m);
    else
        if (m >= 0)
            return -((-n + m - 1) /  m);
        else
            return -n / -m;
}
/* Given the functions above, we can write      */
/* [[arith]] as some common argument processing */
/* followed by a [[switch]] statement.          */
/* <prim.c>=                                    */
static Value
arith(Exp *e, int tag, Valuelist *args)
{
    int n, m;

    checkargc(e, 2, lengthVL(args));
    n = projectint(e, nthVL(args, 0));
    m = projectint(e, nthVL(args, 1));

    switch (tag) {
    case PLUS:
        return mkNum(n + m);
    case MINUS:
        return mkNum(n - m);
    case TIMES:
        return mkNum(n * m);
    case DIV:
        if (m==0)
            error("division by zero");
        return mkNum(divide(n, m));
    case LT:
        return bool(n < m);
    case GT:
        return bool(n > m);
    default:
        assert(0);
        return false;
    }
}
/* We implement them with the function          */
/* [[binary]].                                  */
/* <prim.c>=                                    */
static Value
binary(Exp *e, int tag, Valuelist *args)
{
    Value v, w;

    checkargc(e, 2, lengthVL(args));
    v = nthVL(args, 0);
    w = nthVL(args, 1);

    switch (tag) {
    case CONS:
        /* Because S-expressions are a recursive type,  */
        /* the representation of a cons cell has to     */
        /* contain pointers to S-expressions, not       */
        /* S-expressions themselves. Every [[cons]]     */
        /* must therefore allocate fresh locations for  */
        /* the pointers. This behavior makes [[cons]] a */
        /* major source of allocation in Scheme         */
        /* programs.                                    */
        /* <return cons cell containing [[v]] and [[w]]>= */
        return mkPair(allocate(v), allocate(w));
    case EQ: 
        /* The implementation of equality is not        */
        /* completely trivial. Two values are [[=]]     */
        /* only if they are the same number, the same   */
        /* boolean, the same symbol, or both nil.       */
        /* <return [[(= v w)]]>=                        */
        if (v.ty != w.ty)
            return false;

        switch (v.ty) {
        case NUM:
            return bool(v.u.num  == w.u.num);
        case BOOL:
            return bool(v.u.bool == w.u.bool);
        case SYM:
            return bool(v.u.sym  == w.u.sym);
        case NIL:
            return true;
        default:
            return false;
        }
    }
    assert(0);
    return false; /* not reached */
}
/* Here are the implementations.                */
/*                                              */
/* <prim.c>=                                    */
static Value
unary(Exp *e, int tag, Valuelist *args)
{
    Value v;

    checkargc(e, 1, lengthVL(args));
    v = nthVL(args, 0);
    switch (tag) {
    case NULLP:
        return bool(v.ty == NIL);
    case BOOLEANP:
        return bool(v.ty == BOOL);
    case NUMBERP:
        return bool(v.ty == NUM);
    case SYMBOLP:
        return bool(v.ty == SYM);
    case PAIRP:
        return bool(v.ty == PAIR);
    case PROCEDUREP:
        return bool(v.ty == CLOSURE || v.ty == PRIMITIVE);
    case CAR:
        if (v.ty != PAIR)
            error("car applied to non-pair %v in %e", v, e);
        return *v.u.pair.car;
    case CDR:
        if (v.ty != PAIR)
            error("cdr applied to non-pair %v in %e", v, e);
        return *v.u.pair.cdr;
    case PRINT:
        print("%v\n", v);
        return v;
    case ERROR:
        error("%v", v);
        return v;
    default:
        assert(0);
        return false; /* not reached */
    }
}
