/*   Implementation of the top-level evaluator  */
/*                                              */
/* The function [[topeval]] evaluates a         */
/* top-level item, updates the store, and       */
/* returns a new environment. If [[echo]] is    */
/* nonzero, [[topeval]] also prints.            */
/* <topeval.c>=                                 */
#include "all.h"
/* <topeval.c>=                                 */
Env*
topeval(Toplevel *t, Env *env, int echo)
{
    switch (t->ty) {
    case EXP:
        /* As in Impcore, top-level expressions go      */
        /* straight to [[eval]]. Evaluating a top-level */
        /* expression doesn't change the environment.   */
        /* <evaluate top-level expression and return [[env]]>= */
        if (echo)
            print("%v\n", eval(t->u.exp, env));
        else
            eval(t->u.exp, env);
        return env;
    case DEFINE:
        /* We rewrite \xdefine to \xval.                */
        /*                                              */
        /* <evaluate function definition and return new environment>= */
        /* <if [[t->u.define.lambda.formals]] contains a duplicate, call
                                                                  [[error]]>= */
        if (duplicatename(t->u.define.lambda.formals) != NULL)
            error(
               "formal parameter %n appears twice in definition of function %n",
                  duplicatename(t->u.define.lambda.formals), t->u.define.name);
        return topeval(mkVal(t->u.define.name, mkLambdax(t->u.define.lambda)),
                       env, echo);
    case VAL:
        /* According to the operational semantics, the  */
        /* right-hand side of a [[val]] binding must be */
        /* evaluated in an environment in which the     */
        /* name [[t->u.val.name]] is bound. If the      */
        /* binding is not already present, we bind the  */
        /* name to nil.                                 */
        /* <evaluate [[val]] binding and return new environment>= */
        {
            Value v;

            if (find(t->u.val.name, env) == NULL)
                env = bindalloc(t->u.val.name, mkNil(), env);
            v = eval(t->u.val.exp, env);
            *find(t->u.val.name, env) = v;
            if (echo) {
                if (t->u.val.exp->ty == LAMBDAX)
                    print("%n\n", t->u.val.name);
                else
                    print("%v\n", v);
            }
            return env;
        }
    case USE:
        /* Reading a file is just as in Impcore, except */
        /* we have to be sure to mutate the correct     */
        /* environment, as discussed below.             */
        /* <read in a file and return new environment>= */
        {
            FILE *fin;
            const char *filename;

            filename = nametostr(t->u.use);
            fin = fopen(filename, "r");
            if (fin == NULL)
                error("cannot open file \"%s\"", filename);
            readevalprint(topreader(filereader(filename, fin), 0), &env, 1);
            fclose(fin);
            return env;
        }
    default:
        assert(0);
        return NULL;
    }
}
/* The [[readevalprint]] function works         */
/* approximately as in Impcore, except that     */
/* because a micro-Scheme environment is not a  */
/* mutable type, [[readevalprint]] updates the  */
/* environment by assignment after every        */
/* successful call to [[topeval]]. This update  */
/* ensures that the result of a successful call */
/* to [[topeval]] is not lost after a later     */
/* call to [[error]]. No such update is needed  */
/* in the Impcore interpreter because the       */
/* environments in that interpreter are         */
/* mutable.                                     */
/* <topeval.c>=                                 */
void
readevalprint(Topreader *reader, Env **envp, int echo)
{
    Toplevel *t;

    while ((t = readtop(reader)))
        *envp = topeval(t, *envp, echo);
}
