(* The structure of the interpreter is          *)
(* therefore as follows:                        *)
(* <usm.sml>=                                   *)


(*****************************************************************)
(*                                                               *)
(*   ENVIRONMENTS                                                *)
(*                                                               *)
(*****************************************************************)

(* <environments>=                              *)
type name = string
type 'a env = (name * 'a) list
val emptyEnv = []

(* lookup and assignment of existing bindings *)
exception NotFound of name
fun find (name, []) = raise NotFound name
  | find (name, (n, v)::tail) = if name = n then v else find(name, tail)

(* adding new bindings *)
exception BindListLength
fun bind(name, v, rho) = (name, v) :: rho
fun bindList(n::vars, v::vals, rho) = bindList(vars, vals, bind(n, v, rho))
  | bindList([], [], rho) = rho
  | bindList _ = raise BindListLength
(* We represent an environment of type [['a     *)
(* env]] as a list of ([[name]], [['a]]) pairs. *)
(* The declarations in the box give the         *)
(* interface to our implementation; through     *)
(* some Noweb hackery, they are checked by the  *)
(* ML compiler.                                 *)
(* <boxed values 23>=                           *)
val _ = op emptyEnv : 'a env
val _ = op find     : name * 'a env -> 'a
val _ = op bind     : name      * 'a      * 'a env -> 'a env
val _ = op bindList : name list * 'a list * 'a env -> 'a env
(* [[error]] procedure,                         *)
(* to indicate when things have gone wrong. The *)
(* exceptions we use are listed in Table [->].  *)



(*****************************************************************)
(*                                                               *)
(*   LEXICAL ANALYSIS                                            *)
(*                                                               *)
(*****************************************************************)

(* A source-code location includes a name for   *)
(* the source, plus line number.                *)
(* <lexical analysis>=                          *)
type srcloc = string * int
val nullsrc = ("internally generated SEND node", 1)
fun srclocString (source, line) = source ^ ", line " ^ Int.toString line
(* The representation of a token stream with    *)
(* balanced parentheses is almost the same as   *)
(* in micro-Scheme. The differences are that a  *)
(* [[#]] character does not introduce a         *)
(* Boolean, and we attach a source-code         *)
(* location to every list of tokens in          *)
(* parentheses.                                 *)
(* <lexical analysis>=                          *)
datatype par = INT   of int
             | NAME  of name
             | SHARP of par
             | LIST  of srcloc * par list
(* <lexical analysis>=                          *)

fun srclocString (source, line) = source ^ ", line " ^ Int.toString line
exception Paren of char list * srcloc
   (* raise if found ) when reading input *)
exception PrematureEOF of string (* shows what we were reading ... *)
exception EOF                    (* raised by reader *)
fun read nextline curloc : string -> par list =
  let (* To help with reading, we have functions to   *)
      (* deal with lexical details. We begin by       *)
      (* defining tests for the delimiters of         *)
      (* symbols:                                     *)
      (* <character-fiddling functions>=              *)
      fun isDelim c = List.exists (fn c' => c' = c) [#"(", #")", #" ", #"\t",
                                                                    #"\n", #";"]
      (* [[]] is like the micro-Scheme                *)
      (* [[exists?]]. We needn't write its            *)
      (* implementation because it is in the initial  *)
      (* basis for Standard ML.                       *)

      (* <character-fiddling functions>=              *)
      fun stringReader []     = NONE
        | stringReader (h::t) = SOME (h, t)
      val strton = Int.scan StringCvt.DEC stringReader
      (* We continue with integer conversion;         *)
      (* [[strton s]] returns an integer, plus        *)
      (* leftover characters from [[s]], or [[NONE]]  *)
      (* if no prefix of [[s]] is an integer literal. *)
      (* It uses [[Int.scan]] from the Standard Basis *)
      (* Library.                                     *)
      (* <boxed values 34>=                           *)
      val _ = op strton : char list -> (int * char list) option
      (* We need to be able to skip whitespace and    *)
      (* comments. [[StringCvt.skipWS]] skips white   *)
      (* space; if we encounter a semicolon, we use   *)
      (* [[StringCvt.dropl]] to skip everything up to *)
      (* the next newline, then resume skipping white *)
      (* space.                                       *)
      (* <character-fiddling functions>=              *)
      fun skip s =
        case StringCvt.skipWS stringReader s
          of #";" :: s => skip (StringCvt.dropl (fn c => c <> #"\n")
                                                                 stringReader s)
           | s => s
      (* [[nonwhite]], which                          *)
      (* summons more input lines as needed.          *)

      (* <character-fiddling functions>=              *)
      fun nonwhite s =
        case skip s
          of [] => nonwhite (explode (nextline()))
           | s  => s
      val isDelim = fn c => isDelim c orelse c = #"[" orelse c = #"]"
      fun readmap f (result, unconsumed) = (f result, unconsumed)
      (* get called only when guaranteed non-blank, non-comment, non-empty *)
      fun get (#"("::s)  = readmap LIST (list (curloc(), s, []))
        | get (#")"::s)  = raise (Paren (s, curloc()))
        | get (#"["::s)  = let fun wrap (loc, l) =
                                 LIST (loc, NAME "block" :: LIST (loc, []) :: l)
                           in  readmap wrap (list (curloc(), s, []))
                           end
        | get (#"]"::s)  = raise (Paren (s, curloc()))
        | get (#"#"::s)  = readmap SHARP (get (nonwhite s))
        | get s          = atom (s, [])
      and atom(x::s, r) = if isDelim x then (mkAtom(rev r), x::s) 
                          else atom(s, x::r)
        | atom([],   r) = (mkAtom(rev r), [])
      and mkAtom s = (case strton s of SOME (n, []) => INT n | _ => NAME (
                                                                     implode s))
      and list (loc, s, l) =
        let val next = nonwhite s
        in  let val (x, u) = get next
            in  list(loc, u, x::l)
            end handle (Paren (u, _)) => ((loc, rev l), skip u)
                     | EOF => raise PrematureEOF (implode next)
        end handle EOF => raise PrematureEOF (implode s)
      fun readline [] = []
        | readline s  = 
             let val(p, u) = get s
             in  p :: readline(skip u)
             end
  in  fn s => readline(skip (explode s)) 
  end
(* The lexer turns square brackets into         *)
(* [[(block () ...)]]. Parameter [[loc]] to the *)
(* internal [[list]] function gives the         *)
(* source-code location of the start of the     *)
(* list.                                        *)
(* <boxed values 20>=                           *)
val _ = op read : (unit -> string) -> (unit -> string * int) -> string -> par
                                                                            list
(* [[<<character-fiddling                       *)
(* functions>>]] from the micro-Scheme          *)
(* interpreter, but we have to redefine         *)
(* [[isDelim]] to make square brackets          *)
(* delimiters.                                  *)



(*****************************************************************)
(*                                                               *)
(*   ABSTRACT SYNTAX AND VALUES                                  *)
(*                                                               *)
(*****************************************************************)

(*   Abstract syntax and values                 *)
(*                                              *)
(* The most important kinds of abstract syntax  *)
(* are [[VAR]], [[SET]], [[SEND]], [[BEGIN]],   *)
(* and [[BLOCK]]. In order to bootstrap the     *)
(* interpreter, we handle literals somewhat     *)
(* differently than in other interpreters. The  *)
(* [[LITERAL]] node contains a literal          *)
(* representation of an integer or symbol; it   *)
(* must be associated with a class before it    *)
(* becomes a full-fledged value. The [[VALUE]]  *)
(* node contains a value that includes a class; *)
(* we use it internally to help bootstrap the   *)
(* interpreter and to handle array literals.    *)
(* Finally, the [[SUPER]] node makes it easy to *)
(* recognize messages to [[super]] and give     *)
(* them the proper semantics.                   *)
(* <abstract syntax and values>=                *)
datatype exp = VAR     of name
             | SET     of name * exp
             | SEND    of srcloc * name * exp * exp list
             | BEGIN   of exp list
             | BLOCK   of name list * exp list
             | LITERAL of rep
             | VALUE   of value
             | SUPER
(* A top-level item is a block definition, a    *)
(* class definition, an expression, or one of   *)
(* our old friends [[VAL]] and [[USE]]. Only    *)
(* the class definitions have any complexity.   *)
(* <abstract syntax and values>=                *)
and toplevel 
  = DEFINE of name * name list * exp
  | CLASSD of { name    : string
              , super   : string
              , ivars   : string list     (* instance variables *)
              , methods : (method_kind * name * method_impl) list
              }
  | EXP of exp
  | VAL of name * exp
  | USE of name
and method_kind = IMETHOD 
                | CMETHOD
and method_impl = USER_IMPL of name list * name list * exp
                | PRIM_IMPL of name
(* In a compiled or bytecoded implementation of *)
(* Smalltalk, the representation of an object   *)
(* is a sequence of machine words. The methods, *)
(* and in particular the primitive methods,     *)
(* ``know'' which words stand for integers,     *)
(* which words are slots for instance           *)
(* variables, etc. We can't emulate this        *)
(* technique in ML, because it's an unsafe      *)
(* technique and ML is a safe language. What we *)
(* do instead is define a [[rep]] type for the  *)
(* representation, which is either an           *)
(* environment giving the locations of instance *)
(* variables, for user objects, or some         *)
(* interesting primitive representation, for    *)
(* primitive objects.                           *)
(* <abstract syntax and values>=                *)
and    rep = USER     of value ref env (* collection of named instance variables
                                                                              *)
           | ARRAY    of value Array.array
           | NUM      of int
           | SYM      of name
           | CLOSURE  of name list * exp list * value ref env * class
           | CLASSREP of class
(* [[CLOSURE]] is a representation of a         *)
(* block; it captures not only the environment  *)
(* used to bind variables, but also the class   *)
(* used to interpret messages to [[super]].     *)

(* The representation of a class includes its   *)
(* superclass, instance variables, methods, and *)
(* a unique identifier.                         *)
(* <abstract syntax and values>=                *)
and class  = CLASS of { name    : name            (* name of the class *)
                      , super   : class option    (* superclass, if any *)
                      , ivars   : string list     (* instance variables *)
                      , methods : method env      (* both exported and private
                                                                              *)
                      , id      : int             (* uniquely identifies class
                                                                              *)
                      }
(* [[Object]]. A                                *)
(* class's [[ivars]] and [[methods]] lists      *)
(* include only the instance variables and      *)
(* methods defined in that class, not those of  *)
(* its superclass.                              *)

(* uSmalltalk has two kinds of methods.         *)
(* Primitive methods are functions, while       *)
(* user-defined methods include parameters,     *)
(* local variables, and a body. We also store   *)
(* in each user-defined method the superclass   *)
(* of the class in which the method is defined, *)
(* which we use only to interpret messages sent *)
(* to [[super]] from within that method.        *)
(* <abstract syntax and values>=                *)
and method
  = PRIM_METHOD of name * (value * value list -> value)
  | USER_METHOD of { name : name, formals : name list, locals : name list, body
                                                                           : exp
                   , superclass : class (* used to send messages to super *)
                   }
(* Finally, a value is a combination of class   *)
(* and representation. The representation is    *)
(* inherently either a primitive representation *)
(* or a collection of instance variables, not a *)
(* combination of both; it is therefore not     *)
(* useful to inherit from a class with a        *)
(* primitive representation.                    *)
(* <abstract syntax and values>=                *)
withtype value = class * rep
(* As in the implementation of micro-Scheme, a  *)
(* primitive method might raise the             *)
(* [[RuntimeError]] exception. Because the      *)
(* uSmalltalk interpreter uses more mutable     *)
(* state than other interpreters and is         *)
(* therefore harder to get right, we also       *)
(* define the exception [[InternalError]],      *)
(* which indicates a bug in the interpreter.    *)
(* Raising [[InternalError]] is the moral       *)
(* equivalent of an assertion failure in a      *)
(* language like C.                             *)
(* <abstract syntax and values>=                *)
exception RuntimeError  of string (* error caused by user *)
exception InternalError of string (* bug in the interpreter *)
(* <abstract syntax and values>=                *)
fun className (CLASS {name, ...}) = name
fun classId   (CLASS {id,   ...}) = id
(*   Creating the primitive classes and values  *)
(*                                              *)
(*  Utilities for manipulating classes          *)
(*                                              *)
(* Because a class can point to its superclass, *)
(* the type [[class]] has to be a recursive     *)
(* type implemented as an ML [[datatype]]. To   *)
(* get access to information about a class, we  *)
(* have to write a pattern match. Pattern       *)
(* matching is fairly heavy notation when all   *)
(* we want is a class's name or its unique      *)
(* identifier, so we provide two convenience    *)
(* functions.                                   *)
(* <boxed values 1>=                            *)
val _ = op className : class -> name
val _ = op classId   : class -> int
(* <abstract syntax and values>=                *)
fun methodName (PRIM_METHOD (n, _)) = n
  | methodName (USER_METHOD { name, ... }) = name
fun renameMethod (n, PRIM_METHOD (_, f)) = PRIM_METHOD (n, f)
  | renameMethod _ = raise InternalError "renamed user method"
fun methods l = foldl (fn (m, rho) => bind(methodName m, m, rho)) emptyEnv l
(* We extract a name from a method using        *)
(* another convenience function,                *)
(* [[methodName]]. Other manipulations of       *)
(* methods include [[renameMethod]], which is   *)
(* used when a user class wants to use a        *)
(* primitive method with a name other than the  *)
(* one we built in, and [[methods]], which      *)
(* builds an environment suitable for use in a  *)
(* class.                                       *)
(* <boxed values 2>=                            *)
val _ = op methodName   : method -> name
val _ = op methods      : method list -> method env
val _ = op renameMethod : name * method -> method
(* <abstract syntax and values>=                *)
local 
  val n = ref 10 (* new classes start here *)
  fun uid () = !n before n := !n + 1
in
  fun mkClass name super ivars ms =
    CLASS { name = name, super = SOME super, ivars = ivars , methods = methods
                                                                             ms,
            id = uid () }
end
(* To build a class, we keep a private supply   *)
(* of unique identifiers. Identifiers below 10  *)
(* are reserved for built-in classes not        *)
(* created with [[mkClass]]; the current        *)
(* implementation uses only id 1, for           *)
(* [[Object]], and id 0, for a bootstrapping    *)
(* class. No class ever has a negative          *)
(* identifier.                                  *)
(* <boxed values 3>=                            *)
val _ = op mkClass : name -> class -> name list -> method list -> class


(*****************************************************************)
(*                                                               *)
(*   BOOTSTRAPPING SUPPORT                                       *)
(*                                                               *)
(*****************************************************************)

(* <bootstrapping support>=                     *)
local 
  val intClass    = ref NONE : class option ref
  val symbolClass = ref NONE : class option ref
  val arrayClass  = ref NONE : class option ref
  fun badlit what = 
    raise InternalError
        ("(bootstrapping) -- cannot " ^ what ^ " anywhere in initial basis")
in
  fun mkInteger n = (valOf (!intClass), NUM n)
    handle Option => badlit "evaluate integer literal or use array literal"
  
  fun mkSymbol s = (valOf (!symbolClass), SYM s)
    handle Option => badlit "evaluate symbol literal or use array literal"
  
  fun mkArray a = (valOf (!arrayClass), ARRAY (Array.fromList a))
    handle Option => badlit "use array literal"
(* We break the circularity by defining a       *)
(* function [[mkInteger]]. If called before the *)
(* interpreter is bootstrapped, [[mkInteger]]   *)
(* fails. If called afterward, [[mkInteger]]    *)
(* creates a value of class [[SmallInteger]].   *)
(* We treat symbols and arrays in similar       *)
(* fashion.                                     *)
(* <boxed values 4>=                            *)
val _ = op mkInteger : int        -> value
val _ = op mkSymbol  : string     -> value
val _ = op mkArray   : value list -> value
(* [[valOf]] and exception [[Option]]           *)
(* are part of the initial basis of Standard    *)
(* ML.                                          *)

(* <bootstrapping support>=                     *)
  fun findInitialClass (name, xi) =
        case !(find(name, xi))
          of (_, CLASSREP c) => c
           | _ => raise InternalError (name ^
                                         " is not a class in the initial basis")
  fun closeLiteralsCycle xi =
    ( intClass    := SOME (findInitialClass ("SmallInteger", xi))
    ; symbolClass := SOME (findInitialClass ("Symbol",       xi))
    ; arrayClass  := SOME (findInitialClass ("Array",        xi))
    )
end
(* Once the initial basis has been read, the    *)
(* initialization code calls                    *)
(* [[closeLiteralsCycle]] to assign the         *)
(* appropriate classes to the ref cells. The    *)
(* environment containing the initial basis is  *)
(* passed as the parameter [[xi]]. [Xi is the   *)
(* Greek letter xi, pronounced ``ksee.'']       *)
(* <boxed values 5>=                            *)
val _ = op findInitialClass : string * value ref env -> class
(* <bootstrapping support>=                     *)
local
  val trueValue  = ref NONE : value option ref
  val falseValue = ref NONE : value option ref
in
  fun mkBoolean b = valOf(!(if b then trueValue else falseValue))
    handle Option => 
        raise InternalError 
            "Bad booleanClass; evaluated boolean expression in initial basis?"
  fun closeBooleansCycle xi =
    ( trueValue  := SOME (!(find ("true",  xi)))
    ; falseValue := SOME (!(find ("false", xi)))
    )
end
(*    Booleans                                  *)
(*                                              *)
(* We use the same technique for Booleans,      *)
(* except instead of saving classes, we save    *)
(* values.                                      *)
(* <boxed values 6>=                            *)
val _ = op mkBoolean : bool -> value
(* <bootstrapping support>=                     *)
local
  val blockClass = ref NONE : class option ref
in
  fun mkBlock c = (valOf (!blockClass), CLOSURE c)
    handle Option => 
        raise InternalError 
            "Bad blockClass; evaluated block expression in initial basis?"
  fun closeBlocksCycle xi =
    blockClass := SOME (findInitialClass ("Block", xi))
end
(*    Blocks                                    *)
(*                                              *)
(* We use the technique again for blocks. It is *)
(* not actually required for blocks, but by     *)
(* declaring [[Block]] and [[Boolean]]          *)
(* together, we help clarify the relationship   *)
(* between them, especially the implementations *)
(* of the [[whileTrue:]] and [[whileFalse:]]    *)
(* methods.                                     *)
(* <boxed values 7>=                            *)
val _ = op mkBlock : name list * exp list * value ref env * class -> value


(*****************************************************************)
(*                                                               *)
(*   PARSING                                                     *)
(*                                                               *)
(*****************************************************************)

(*   Parsing                                    *)
(*                                              *)
(* We parse concrete syntax into abstract       *)
(* syntax. We keep source-code locations only   *)
(* for [[SEND]] nodes; otherwise, this          *)
(* operation is relatively uninteresting. The   *)
(* auxiliary functions may help clarify the     *)
(* logic.                                       *)
(* <parsing>=                                   *)
exception SyntaxError of string
val unknown = ("unknown source", ~99)
fun syntaxErrorAt (source, line) msg =
      if (source, line) = unknown then
        raise SyntaxError msg
      else
        raise SyntaxError (source ^ ", line " ^ Int.toString line ^ ": " ^ msg)
(* Smalltalk has simple rules for computing the *)
(* arity of a message based on the message's    *)
(* name: if the name is symbolic, the message   *)
(* is binary (one receiver, one argument); if   *)
(* the name is alphanumeric, the number of      *)
(* arguments is the number of colons.           *)
(* Unfortunately, in uSmalltalk we can write a  *)
(* message name that is a mix of alphanumerics  *)
(* and symbols. We use the first character of   *)
(* the message name to decide the issue.        *)
(* <parsing>=                                   *)
fun arity "if"    = 2
  | arity "while" = 1
  | arity name =
      let val cs = explode name
      in  if Char.isAlpha (hd cs) then
            length (List.filter (fn c => c = #":") cs)
          else
            1
      end

fun arityOk "value" _ = true
  | arityOk name args = arity name = length args

fun arityErrorAt loc what msgname args =
  let fun argn n = if n = 1 then "1 argument" else Int.toString n ^ " arguments"
  in  syntaxErrorAt loc ("in " ^ what ^ ", message " ^ msgname ^ " expects " ^
                         argn (arity msgname) ^ ", but gets " ^
                         argn (length args))
  end
(* <parsing>=                                   *)
fun parse loc (INT n) = LITERAL (NUM n)
  | parse loc (NAME "super") = SUPER
  | parse loc (NAME n) = VAR n
  | parse loc (SHARP (NAME n)) = LITERAL (SYM n)
  | parse loc (SHARP (INT  n)) = LITERAL (SYM (Int.toString n))
  | parse loc (SHARP p) = VALUE (sharp loc p) (* better not happen in initial
                                                                        basis *)
  | parse loc (LIST (_, NAME "set"    :: args)) = parseSet   loc args
  | parse loc (LIST (_, NAME "begin"  :: args)) = BEGIN (map (parse loc) args)
  | parse loc (LIST (_, NAME "block"  :: args)) = parseBlock loc args
  | parse loc (LIST (_, NAME "locals" :: args)) =
      syntaxErrorAt loc "expected and expression but found '(locals ...)'"
  | parse _   (LIST (loc, msg::receiver::args)) =
     let val msgname = unname loc "method names" msg
     in  if arityOk msgname args then
           SEND (loc, msgname, parse loc receiver, map (parse loc) args)
         else
           arityErrorAt loc "message send" msgname args
     end
  | parse _ (LIST (loc, [NAME x])) =
      syntaxErrorAt loc ("sent message " ^ x ^ " to no object")
  | parse _ (LIST (loc, [x])) = syntaxErrorAt loc "message send to no object"
  | parse _ (LIST (loc, []))  = syntaxErrorAt loc "empty message send ()"
and parseSet loc [NAME v, e] = SET (canSet loc v, parse loc e)
  | parseSet loc _ = syntaxErrorAt loc "expected (set v e)"
and canSet loc n =
  if List.exists (fn n' => n' = n) ["true", "false", "nil", "self", "super"]
                                                                            then
    syntaxErrorAt loc ("you cannot set or val-bind pseudovariable " ^ n)
  else
    n
(* The remaining parser functions are mostly    *)
(* straightforward. The [[sharp]] function may  *)
(* call [[mkSymbol]], [[mkInteger]], or         *)
(* [[mkArray]], which must not be called until  *)
(* after the initial basis is read in.          *)
(* <parsing>=                                   *)
and parseBlock loc (LIST (floc, formals) :: body) =
      BLOCK (map (unname floc "block parameters") formals, map (parse loc) body)
  | parseBlock loc _ = 
      syntaxErrorAt loc "block must have formal-parameter list"
and unname loc ? (NAME n) = if String.sub(n, 0) = #"#" then badName loc ? else n
  | unname loc ? _        = badName loc ?
and badName loc ? = syntaxErrorAt loc (? ^ " must be ordinary names")
and sharp loc (NAME n)        = mkSymbol n
  | sharp loc (INT n)         = mkInteger n
  | sharp _   (LIST (loc, l)) = mkArray (map (sharp loc) l)
  | sharp loc (SHARP _)       = syntaxErrorAt loc ("# within # is not legal")
(* The top-level parser recognizes [[method]]   *)
(* and [[classMethod]], because this helps      *)
(* detect premature right parentheses.          *)
(* <parsing>=                                   *)
fun toplevel (LIST (loc, NAME "define"      :: args)) = parseDefine   loc args
  | toplevel (LIST (loc, NAME "class"       :: args)) = parseClassDef loc args
  | toplevel (LIST (loc, NAME "val"         :: args)) = parseVal      loc args
  | toplevel (LIST (loc, NAME "use"         :: args)) = parseUse      loc args
  | toplevel (LIST (loc, NAME "method"      :: args)) = methodError loc "method"
                                                                            args
  | toplevel (LIST (loc, NAME "classMethod" :: args)) =
                                                   methodError loc "classMethod"
                                                                            args
  | toplevel e = EXP (parse unknown e)
and methodError loc s args = 
  syntaxErrorAt loc
    ("unexpected `(" ^ s ^ " " ^ mName args ^ "...'; " ^
     "did a class definition end prematurely?")
and mName (NAME n :: _) = n ^ " "
  | mName _ = ""
(* <parsing>=                                   *)
and parseDefine loc l = DEFINE (parseDefinition loc l)
and parseDefinition loc [NAME name, LIST (floc, formals), body] =
      (name, map (unname floc "formal parameters") formals, parse loc body)
  | parseDefinition loc _ = syntaxErrorAt loc "expected (define f (args) body)"
(* The class definition is by far the most      *)
(* complex construct to parse.                  *)
(* <parsing>=                                   *)
and parseClassDef loc (NAME name :: NAME super :: LIST (rloc, rep) :: methods) =
      CLASSD    { name = name, super = super
                , ivars = map (unname rloc "instance variables of class") rep
                , methods = map (parseMethod loc) methods
                }
  | parseClassDef loc _ = syntaxErrorAt loc
                                     "expected (class name super (rep) methods)"
and parseMethod _ (LIST (loc, NAME keyword :: NAME name :: impl)) =
      (parseMethodKind loc keyword, name, parseMethodImpl loc keyword name impl)
  | parseMethod loc _ = syntaxErrorAt loc "expected (method f (args) body)"
and parseMethodKind loc "method" = IMETHOD
  | parseMethodKind loc "classMethod" = CMETHOD
  | parseMethodKind loc n = syntaxErrorAt loc ("keyword `" ^ n ^
                                                     "' cannot define a method")
and parseMethodImpl loc _ _ [NAME "primitive", NAME name] = PRIM_IMPL name
  | parseMethodImpl loc key name (LIST (_, formals) ::
                                  LIST (_, NAME "locals" :: locals) :: body) =
      if arityOk name formals then
        USER_IMPL (parms loc "parameters" key name formals,
                   parms loc "locals" key name locals, 
                   BEGIN (map (parse loc) body))
      else
        arityErrorAt loc (key ^ " definition") name formals
  | parseMethodImpl loc key name (LIST (_, formals) :: body) =
      if arityOk name formals then
        USER_IMPL (parms loc "parameters" key name formals, [],
                   BEGIN (map (parse loc) body))
      else
        arityErrorAt loc (key ^ " definition") name formals
  | parseMethodImpl loc key name _ =
      syntaxErrorAt loc ("expected method definition (" ^
                       key ^ " " ^ name ^ " (args) body)")
and parms loc what key name = map (unname loc (what ^ " of " ^ key ^ " " ^ name)
                                                                               )
(* <parsing>=                                   *)
and parseVal loc [NAME v, e] = VAL (canSet loc v, parse loc e)
  | parseVal loc _ = syntaxErrorAt loc "expected (val v e)"
(* <parsing>=                                   *)
and parseUse loc [NAME v] = USE v
  | parseUse loc _ = syntaxErrorAt loc "expected (use filename)"


(*****************************************************************)
(*                                                               *)
(*   READERS                                                     *)
(*                                                               *)
(*****************************************************************)

(* Readers                                      *)
(*                                              *)
(*  Line readers                                *)
(*                                              *)
(* Our readers are structured slightly          *)
(* differently from those in C. A reader is a   *)
(* bundle containing three related functions:   *)
(* [[firstline]], which returns the first line  *)
(* read; [[nextline]], which returns all        *)
(* subsequent lines; and [[curloc]], which      *)
(* returns a pair identifying the source        *)
(* location of the most recently returned line. *)
(* Both [[firstline]] and [[nextline]] can      *)
(* raise [[EOF]], (We distinguish [[firstline]] *)
(* from [[nextline]] because in an interactive  *)
(* reader, they may use different prompts.)     *)
(* <readers>=                                   *)
type reader = { firstline : unit -> string
              , nextline  : unit -> string
              , curloc    : unit -> string * int
              }
(* <readers>=                                   *)
fun promptReader (firstPrompt, nextPrompt, reader) =
  let val { firstline = first, nextline = next, curloc = cur } = reader
      fun promptIn get prompt () = ( TextIO.output(TextIO.stdOut, prompt)
                                   ; TextIO.flushOut(TextIO.stdOut)
                                   ; get ()
                                   )
  in  { firstline = promptIn first firstPrompt
      , nextline  = promptIn next  nextPrompt
      , curloc    = cur
      }
  end
(* Given a silent reader, we can make it issue  *)
(* a prompt every time it reads a new line. We  *)
(* use two prompts: one for the first line of   *)
(* an expression and one for later lines.       *)
(* <boxed values 35>=                           *)
val _ = op promptReader : string * string * reader -> reader
(* <readers>=                                   *)
fun filereader (filename, fd)=
  let val lineNum = ref 0
      fun get () =
        let val line = TextIO.inputLine fd
        in  if size line = 0 then raise EOF else line before lineNum := !lineNum
                                                                             + 1
        end
  in  { firstline = get, nextline = get, curloc = (fn () => (filename, !lineNum)
                                                                             ) }
  end

fun stringsreader (what, l) = 
  let val buffer  = ref l
      val lineNum = ref 0
      fun get ()  = case !buffer
                      of [] => raise EOF
                       | h :: t => h before buffer := t before lineNum := !
                                                                     lineNum + 1
  in  { firstline = get, nextline = get, curloc = (fn () => (what, !lineNum)) }
  end              
(* We can read from files or lists of strings.  *)
(* File I/O is provide by the standard module   *)
(* [[TextIO]].                                  *)
(* <boxed values 36>=                           *)
val _ = op filereader    : string * TextIO.instream -> reader
val _ = op stringsreader : string * string list     -> reader
(* [[]], a private mutable cell. When all       *)
(* the strings in the buffer are used up, there *)
(* aren't any more.                             *)

(*  Top-level readers                           *)
(*                                              *)
(* A top-level reader is a buffer of top-level  *)
(* items together with a reader. [*]            *)
(* <readers>=                                   *)
type topreader = { buffer : toplevel list ref, reader : reader }
fun readtop (r as { buffer, reader = { nextline, firstline, curloc } }) =
  case !buffer
    of h::t => h before buffer := t
     | []   => ( buffer := map toplevel (read nextline curloc (firstline()))
               ; readtop r
               )
(* We decide whether to prompt when building a  *)
(* top-level reader.                            *)
(* <readers>=                                   *)
fun topreader (reader, prompt) = 
  let val reader = if prompt then promptReader ("-> ", "   ", reader) else
                                                                          reader
  in  { buffer = ref [], reader = reader }
  end
(* <readers>=                                   *)
fun echoTag f x =
  let val line = f x
      val _ = if (String.substring (line, 0, 2) = ";#" handle _ => false) then
                print line
              else
                ()
  in  line
  end

fun echoReader { nextline=n, firstline=f, curloc=c } =
     { nextline=echoTag n, firstline=echoTag f, curloc=c } 

fun echoTop { buffer = b, reader = r } = { buffer = b, reader = echoReader r }
val topreader = fn args => echoTop (topreader args)
(* As in the C code, we want to print out any   *)
(* line read if it begins with the special      *)
(* string [[;#]]. This string is a formal       *)
(* comment that helps us test the chunks marked *)
(* [[]]. In our ML code, we                     *)
(* arrange it using the higher-order function   *)
(* [[echoReader]].                              *)
(* <boxed values 37>=                           *)
val _ = op echoReader : reader -> reader 


(*****************************************************************)
(*                                                               *)
(*   PRIMITIVE BASICS                                            *)
(*                                                               *)
(*****************************************************************)

(* <primitive basics>=                          *)
type primitive = value * value list -> value
fun arityError n (receiver, args) =
  raise RuntimeError ("primitive method expected " ^ Int.toString n ^
                      " arguments; got " ^ Int.toString (length args))
fun unaryPrim  f = (fn (a, [])  => f a     | ra => arityError 0 ra)
fun binaryPrim f = (fn (a, [b]) => f(a, b) | ra => arityError 1 ra)
fun primMethod name f = PRIM_METHOD (name, f)
(*   Primitives                                 *)
(*                                              *)
(*  Utilities for creating primitives           *)
(*                                              *)
(* We create most primitives directly from      *)
(* functions written in ML. Here we turn unary  *)
(* and binary functions into primitives, then   *)
(* turn primitives into methods.                *)
(* <boxed values 8>=                            *)
val _ = op unaryPrim  : (value         -> value) -> primitive
val _ = op binaryPrim : (value * value -> value) -> primitive
val _ = op primMethod : name -> primitive -> method
(* <primitive basics>=                          *)
fun userMethod name formals locals body = 
  let val super = CLASS { name = "should never be used", super = NONE, ivars = [
                                                                              ],
                          methods = [], id = 0 }
  in  USER_METHOD { name = name, formals = formals, locals = locals, body = exp
                                                                           body,
                    superclass = super }
  end
and exp s = 
  let fun nextline () = raise PrematureEOF "parsing internal exp"
      fun curloc   () = ("(internal expression)", 1)
  in  parse (curloc()) (hd (read nextline curloc s))
  end
(* There are a few primitives that are more     *)
(* easily created as user methods. For this     *)
(* purpose, we define function [[userMethod]].  *)
(* The only tricky bit is the [[superclass]]    *)
(* field of the user method. Because this class *)
(* is used only to define the meaning of        *)
(* messages to [[super]], and because none of   *)
(* our predefined user methods sends messages   *)
(* to [[super]], we can get away with a bogus   *)
(* class that understands no messages.          *)
(*                                              *)
(* Function [[exp]] is an auxiliary function    *)
(* used to parse a string into abstract syntax. *)
(* <boxed values 9>=                            *)
val _ = op exp : string -> exp


(*****************************************************************)
(*                                                               *)
(*   PRIMITIVE METHODS FOR [[OBJECT]] AND [[UNDEFINEDOBJECT]]    *)
(*                                                               *)
(*****************************************************************)

(* <primitive methods for [[Object]] and [[UndefinedObject]]>= *)
fun eqRep ((cx, x), (cy, y)) = 
  classId cx = classId cy andalso
  case (x, y)
    of (ARRAY x,    ARRAY    y) => x = y
     | (NUM   x,    NUM      y) => x = y
     | (SYM   x,    SYM      y) => x = y
     | (USER  x,    USER     y) => find("self", x) = find("self", y)
     | (CLOSURE  x, CLOSURE  y) => false
     | (CLASSREP x, CLASSREP y) => classId x = classId y
     | _ => false
(*  Object primitives                           *)
(*                                              *)
(*    Equality                                  *)
(*                                              *)
(* The primitive equality compares              *)
(* representations directly. Because ML arrays  *)
(* and ref cells are generative, both the       *)
(* [[ARRAY]] and [[USER]] cases provide a true  *)
(* object identity. We can safely use the ML    *)
(* identity on numbers and symbols because      *)
(* these types are immutable. There's no        *)
(* identity on blocks, which always compare     *)
(* different.                                   *)
(* <boxed values 10>=                           *)
val _ = op eqRep : value * value -> bool
(*    Printing                                  *)
(*                                              *)
(* By default, an object prints as its class    *)
(* name in angle brackets.                      *)
(* <primitive methods for [[Object]] and [[UndefinedObject]]>= *)
fun defaultPrint (self as (c, _)) = ( app print ["<", className c, ">"]; self )
(*    Class membership                          *)
(*                                              *)
(* For [[memberOf]], the class [[c]] of         *)
(* [[self]] has to be the same as the           *)
(* class [[c']] of the argument. For            *)
(* [[kindOf]], it just has to be a subclass.    *)
(* <primitive methods for [[Object]] and [[UndefinedObject]]>= *)
fun memberOf ((c, _), (_, CLASSREP c')) = mkBoolean (classId c = classId c')
  | memberOf _ = raise RuntimeError "argument of isMemberOf: must be a class"

fun kindOf ((c, _), (_, CLASSREP (CLASS {id=u', ...}))) =
      let fun subclassOfClassU' (CLASS {super, id=u, ... }) =
            u = u' orelse (case super of NONE => false | SOME c =>
                                                            subclassOfClassU' c)
      in  mkBoolean (subclassOfClassU' c)
      end
  | kindOf _ = raise RuntimeError "argument of isKindOf: must be a class"
(* The [[error:]] primitive raises              *)
(* [[RuntimeError]].                            *)
(* <primitive methods for [[Object]] and [[UndefinedObject]]>= *)
fun error (_, (_, SYM s)) = raise RuntimeError s
  | error (_, (c, _    )) =
      raise RuntimeError ("error: got class " ^ className c ^
                                                            "; expected Symbol")


(*****************************************************************)
(*                                                               *)
(*   BUILT-IN CLASSES [[OBJECT]] AND [[UNDEFINEDOBJECT]]         *)
(*                                                               *)
(*****************************************************************)

(*    The distinguished root class              *)
(*                                              *)
(* [[Object]] is the ultimate superclass. By    *)
(* putting [[self]] in the representation, we   *)
(* ensure that every object has an instance     *)
(* variable called [[self]]. We also make sure  *)
(* that every object responds to messages in    *)
(* the [[Object]] protocol that appears in      *)
(* Figure [->] on page [->].                    *)
(* <built-in classes [[Object]] and [[UndefinedObject]]>= *)
val objectClass = 
  CLASS { name = "Object", super = NONE, ivars = ["self"], id = 1
        , methods = methods
            [ primMethod "print"   (unaryPrim defaultPrint)
            , userMethod "println" [] []
                                     "(begin (print self) (print newline) self)"
            , primMethod "isNil"   (unaryPrim (fn _ => mkBoolean false))
            , primMethod "notNil"  (unaryPrim (fn _ => mkBoolean true))
            , primMethod "error:"  (binaryPrim error)
            , primMethod "="       (binaryPrim (mkBoolean o eqRep))
            , userMethod "!=" ["x"] [] "(not (= self x))"
            , primMethod "isKindOf:"    (binaryPrim kindOf)
            , primMethod "isMemberOf:"  (binaryPrim memberOf)
            , primMethod "subclassResponsibility"
               (unaryPrim
                  (fn _ => raise RuntimeError

           "subclass failed to implement a method that was its responsibility"))
            ]
        }
(*    The undefined object                      *)
(*                                              *)
(* Class [[UndefinedObject]], whose sole        *)
(* instance is [[nil]], redefines [[isNil]],    *)
(* [[notNil]], and [[print]].                   *)
(* <built-in classes [[Object]] and [[UndefinedObject]]>= *)
val nilClass = 
  mkClass "UndefinedObject" objectClass []
    [ primMethod "isNil"  (unaryPrim (fn _ => mkBoolean true))
    , primMethod "notNil" (unaryPrim (fn _ => mkBoolean false))
    , primMethod "print"  (unaryPrim (fn x => (print "nil"; x)))
    ]
(* To create the [[nil]] value, we have to bind *)
(* [[self]]; otherwise [[println]] won't work   *)
(* on [[nil]].                                  *)
(* <built-in classes [[Object]] and [[UndefinedObject]]>= *)
val nilValue = 
  let val nilCell  = ref (nilClass, USER []) : value ref
      val nilValue = (nilClass, USER (bind("self", nilCell, emptyEnv)))
      val _        = nilCell := nilValue
  in  nilValue
  end


(*****************************************************************)
(*                                                               *)
(*   PRIMITIVE METHODS FOR REMAINING CLASSES                     *)
(*                                                               *)
(*****************************************************************)

(*  Integer primitives                          *)
(*                                              *)
(* We can't simply use [[Int.toString]] to      *)
(* print integers, because ML and uSmalltalk    *)
(* use different unary minus signs. [*]         *)
(* <primitive methods for remaining classes>=   *)
fun printInt (self as (_, NUM n)) =
      ( print (String.map (fn #"~" => #"-" | c => c) (Int.toString n))
      ; self
      )
  | printInt _ =
      raise RuntimeError ("cannot print when object inherits from Int")
(* <primitive methods for remaining classes>=   *)
fun binaryInt mk $ ((_, NUM n), (_, NUM m)) = mk ($ (n, m))
  | binaryInt _ _  ((_, NUM n), (c, _)) =
      raise RuntimeError ("numeric primitive expected numeric argument, got <"
                          ^ className c ^ ">")
  | binaryInt _ _  ((c, _), _) =
      raise RuntimeError ("numeric primitive method defined on <" ^ className c
                                                                          ^ ">")
fun arithop    $ = binaryPrim (binaryInt mkInteger $)
fun intcompare $ = binaryPrim (binaryInt mkBoolean $)
(* To create a new integer, you must pass an    *)
(* argument that is represented by an integer.  *)
(* <primitive methods for remaining classes>=   *)
fun newInteger ((_, CLASSREP c), (_, NUM n)) = (c, NUM n)
  | newInteger _ = raise RuntimeError (
                                   "made new integer with non-int or non-class")
(*  Symbol primitives                           *)
(*                                              *)
(* A symbol prints as its name, with no         *)
(* leading [[#]].                               *)
(* <primitive methods for remaining classes>=   *)
fun printSymbol (self as (_, SYM s)) = (print s; self)
  | printSymbol _ = raise RuntimeError
                                 "cannot print when object inherits from Symbol"
(* To create a new symbol, you must pass an     *)
(* argument that is represented by an symbol.   *)
(* <primitive methods for remaining classes>=   *)
fun newSymbol ((_, CLASSREP c), (_, SYM s)) = (c, SYM s)
  | newSymbol _ = raise RuntimeError (
                                 "made new symbol with non-symbol or non-class")
(* A new array contains all [[nil]]. [*]        *)
(*                                              *)
(* <primitive methods for remaining classes>=   *)
fun newArray ((_, CLASSREP c), (_, NUM n)) = (c, ARRAY (Array.array (n, nilValue
                                                                             )))
  | newArray _ = raise RuntimeError
                                "Array new sent to non-class or got non-integer"
(* <primitive methods for remaining classes>=   *)
fun arrayPrimitive f ((_, ARRAY a), l) = f (a, l)
  | arrayPrimitive f _ = raise RuntimeError "Array primitive used on non-array"

fun arraySize (a, []) = mkInteger (Array.length a)
  | arraySize ra      = arityError 0 ra
(* Arrays in uSmalltalk are indexed from 1, but *)
(* arrays in ML are indexed from 0.             *)
(* <primitive methods for remaining classes>=   *)
fun arrayAt (a, [(_, NUM n)]) = Array.sub(a, n - 1)  (* convert to 0-indexed *)
  | arrayAt (_, [_])  = raise RuntimeError "Non-integer used as array subscript"
  | arrayAt ra        = arityError 1 ra

fun arrayAtPut (a, [(_, NUM n), x]) = (Array.update(a, n-1, x); x)
  | arrayAtPut (_, [_, _]) = raise RuntimeError
                                           "Non-integer used as array subscript"
  | arrayAtPut ra          = arityError 2 ra
(* <primitive methods for remaining classes>=   *)
fun classPrimitive f = 
  unaryPrim (fn (meta, CLASSREP c) => f (meta, c)
              | _ => raise RuntimeError "class primitive sent to non-class")
(* A binary integer operation created with      *)
(* [[arith]] expects as arguments two integers  *)
(* [[m]] and [[n]]; it applies an               *)
(* operation [[]] to them and uses a creator    *)
(* function [[mk]] to convert the result to a   *)
(* value. We use [[binaryInt]] to build         *)
(* arithmetic and comparison.                   *)
(* <boxed values 11>=                           *)
val _ = op binaryInt  : ('a -> value) -> (int * int -> 'a)   -> value * value ->
                                                                           value
val _ = op arithop    :                  (int * int -> int)  -> primitive
val _ = op intcompare :                  (int * int -> bool) -> primitive
(* We use [[arrayPrimitive]] to create          *)
(* primitives that expect [[self]] to be an     *)
(* array.                                       *)
(* <boxed values 11>=                           *)
val _ = op arrayPrimitive : (value array * value list -> value) -> primitive
(*  Class primitives                            *)
(*                                              *)
(* The class primitives take both the metaclass *)
(* and the class as arguments.                  *)
(* <boxed values 11>=                           *)
val _ = op classPrimitive : (class * class -> value) -> primitive
(* <primitive methods for remaining classes>=   *)
local
  fun mkIvars (CLASS { ivars, super, ... }) =
    let val supervars = case super of NONE => emptyEnv | SOME c => mkIvars c
    in  foldl (fn (n, rho) => bind(n, ref nilValue, rho)) supervars ivars
    end
in
  fun newUserObject (_, c) =
        let val ivars = mkIvars c
            val self = (c, USER ivars)
        in  (find("self", ivars) := self; self)
        end
(*    Object creation                           *)
(*                                              *)
(* The most important primitive defined on      *)
(* classes is [[new]]. To create a new object,  *)
(* we have got to give values to all the        *)
(* instance variables, including those          *)
(* inherited from superclasses. Every instance  *)
(* variable is initially [[nil]]; then [[self]] *)
(* is bound to the object itself.               *)
(* <boxed values 12>=                           *)
val _ = op mkIvars       : class -> value ref env
val _ = op newUserObject : class * class -> value
end
(*    Showing protocols                         *)
(*                                              *)
(* The [[showProtocol]] function helps          *)
(* implement the [[protocol]] and               *)
(* [[localProtocol]] primitives, which are      *)
(* defined on class [[Class]]. Its              *)
(* implementation is not very interesting.      *)
(* Function [[insert]] helps implement an       *)
(* insertion sort, which we use to present      *)
(* methods in alphabetical order by name.       *)
(* <primitive methods for remaining classes>=   *)
(* Function [[separate]] helps print a readable *)
(* list of formals for the error message.       *)
(* <definition of [[separate]]>=                *)
fun separate (zero, sep) =  (* useful for printing *)
  let fun s []     = zero
        | s [x]    = x
        | s (h::t) = h ^ sep ^ s t
  in  s
end
local
  fun showProtocol doSuper kind c =
    let fun member x l = List.exists (fn x' => x' = x) l
        fun insert (x, []) = [x]
          | insert (x, (h::t)) =
              case compare x h
                of LESS    => x :: h :: t
                 | EQUAL   => x :: t (* replace *)
                 | GREATER => h :: insert (x, t)
        and compare (name, _) (name', _) = String.compare (name, name')
        fun methods (CLASS { super, methods = ms, name, ... }) =
              if not doSuper orelse (kind = "classMethod" andalso name = "Class"
                                                                          ) then
                foldl insert [] ms
              else
                foldl insert (case super of NONE => [] | SOME c => methods c) ms
        fun show (name, PRIM_METHOD _) =
              app print ["(", kind, " ", name, " primitive ...)\n"]
          | show (name, USER_METHOD { formals, ... }) =
              app print ["(", kind, " ", name,
                         " (", separate ("", " ") formals, ") ...)\n"]
    in  app show (methods c)
    end
in
  fun protocols all (meta, c) =
    ( showProtocol all "classMethod" meta
    ; showProtocol all "method" c
    ; (meta, CLASSREP c)
    )
end
(* The implementations of the primitives are    *)
(* easy; we try to build a block containing the *)
(* result, but if the computation overflows, we *)
(* answer the overflow block instead.           *)
(* <primitive methods for remaining classes>=   *)
fun withOverflow binop ((_, NUM n), [(_, NUM m), ovflw]) = 
      (mkBlock ([], [VALUE (mkInteger (binop (n, m)))], emptyEnv, objectClass)
       handle Overflow => ovflw)
  | withOverflow _ (_, [_, _]) =
      raise RuntimeError "numeric primitive with overflow expects numbers"
  | withOverflow _ _ =
      raise RuntimeError
                     "numeric primitive with overflow expects receiver + 2 args"


(*****************************************************************)
(*                                                               *)
(*   REMAINING BUILT-IN CLASSES                                  *)
(*                                                               *)
(*****************************************************************)

(*    Class [[Class]]                           *)
(*                                              *)
(* Class [[Class]] is in the interpreter so     *)
(* metaclasses can inherit from it. The methods *)
(* that are defined on class [[Class]], and     *)
(* therefore defined on all class objects, are  *)
(* [[new]], [[protocol]], and                   *)
(* [[localProtocol]], as explained in           *)
(* Figure [->] on page [->].                    *)
(* <remaining built-in classes>=                *)
val classClass =
  mkClass "Class" objectClass []
    [ primMethod "new"           (classPrimitive newUserObject) 
    , primMethod "protocol"      (classPrimitive (protocols true))
    , primMethod "localProtocol" (classPrimitive (protocols false))
    ]


(*****************************************************************)
(*                                                               *)
(*   IMPLEMENTATION OF [[USE]], WITH BOOLEAN [[ECHO]]            *)
(*                                                               *)
(*****************************************************************)

(* Miscellaneous                                *)
(*                                              *)
(* We have a different version of [[use]] than  *)
(* in micro-Scheme; [[echo]] is a Boolean, not  *)
(* a function.                                  *)
(* <implementation of [[use]], with Boolean [[echo]]>= *)
fun use readEvalPrint filename rho =
      let val fd = TextIO.openIn filename
          val treader = topreader (filereader (filename, fd), false)
          fun writeln s = app print [s, "\n"]
      in  readEvalPrint (treader, true, writeln) rho
          before TextIO.closeIn fd
      end 


(*****************************************************************)
(*                                                               *)
(*   EVALUATION                                                  *)
(*                                                               *)
(*****************************************************************)

(* <evaluation>=                                *)
(* Support for tracing                          *)
(*                                              *)
(* Tracing support is divided into three parts: *)
(* support for printing indented messages,      *)
(* which is conditioned on the value of the     *)
(* variable [[ --- trace]]; support for         *)
(* maintaining a stack of source-code           *)
(* locations, which is used to provide          *)
(* information when an error occurs; and        *)
(* exposed tracing functions, which are used in *)
(* the main part of the interpreter. To keep    *)
(* the details hidden from the rest of the      *)
(* interpreter, the first two parts are made    *)
(* [[local]].                                   *)
(* <tracing>=                                   *)
local
  (* <private state and functions for printing indented traces>= *)
  fun traceMe xi =
    let val count = find("&trace", xi)
    in  case !count
          of (c, NUM n) =>
              if n = 0 then false
              else ( count := (c, NUM (n - 1))
                   ; if n = 1 then (print "<trace ends>\n"; false) else true
                   )
           | _ => false
    end handle NotFound _ => false
  (* The [[traceMe]] function is used internally  *)
  (* to decide whether to trace; it not only      *)
  (* returns a Boolean but also decrements [[ --- *)
  (* trace]] if needed.                           *)
  (* <boxed values 21>=                           *)
  val _ = op traceMe : value ref env -> bool
  (* The local variable [[tindent]] maintains the *)
  (* current trace state; [[indent]] uses it to   *)
  (* print an indentation string.                 *)
  (* <private state and functions for printing indented traces>= *)
  val tindent = ref 0
  fun indent 0 = ()
    | indent n = (print "  "; indent (n-1))
  (* Any actual printing is done by               *)
  (* [[tracePrint]], conditional on [[traceMe]]   *)
  (* returning [[true]]. The argument             *)
  (* [[direction]] of type [[indentation]]        *)
  (* controls the adjustment of [[indent]]. For   *)
  (* consistency, we outdent from the previous    *)
  (* level before printing a message; we indent   *)
  (* from the current level after printing a      *)
  (* message.                                     *)
  (* <private state and functions for printing indented traces>= *)
  datatype indentation = INDENT_AFTER | OUTDENT_BEFORE

  fun tracePrint direction xi f =
      if traceMe xi then
        let val msg = f () (* could change tindent *)
        in  ( if direction = OUTDENT_BEFORE then tindent := !tindent - 1 else ()
            ; indent (!tindent)
            ; app print msg
            ; print "\n"
            ; if direction = INDENT_AFTER   then tindent := !tindent + 1 else ()
            )
        end
      else
          ()    
  (* Printing of trace messages is conditional,   *)
  (* but we always maintain a stack of            *)
  (* source-code locations. The stack is          *)
  (* displayed when an error occurs.              *)
  (* <private state and functions for mainting a stack of source-code locations
                                                                           >= *)
  val locationStack = ref [] : (string * srcloc) list ref
  fun push srcloc = locationStack := srcloc :: !locationStack
  fun pop () = case !locationStack
                 of []     => raise InternalError "tracing stack underflows"
                  | h :: t => locationStack := t
in
  (* <exposed tracing functions>=                 *)
  fun resetTrace ()       = (locationStack := []; tindent := 0)
  fun traceIndent what xi = (push what; tracePrint INDENT_AFTER   xi)
  fun outdentTrace     xi = (pop ();    tracePrint OUTDENT_BEFORE xi)
  fun showStackTrace () =
    let fun show (msg, (file, n)) =
          app print ["  Sent '", msg, "' in ", file, ", line ", Int.toString n,
                                                                           "\n"]
    in  case !locationStack
          of [] => ()
           | l  => ( print "Method-stack traceback:\n"; app show (!locationStack
                                                                             ) )
    end
  (* Here are the tracing-related functions that  *)
  (* are exposed to the rest of the interpreter.  *)
  (* The interpreter uses [[traceIndent]] to      *)
  (* trace sends, [[outdentTrace]] to trace       *)
  (* answers, and [[resetTrace]] to reset         *)
  (* indentation.                                 *)
  (* <boxed values 22>=                           *)
  val _ = op resetTrace     : unit -> unit
  val _ = op traceIndent    : string * srcloc -> value ref env -> (unit ->
                                                            string list) -> unit
  val _ = op outdentTrace   :                    value ref env -> (unit ->
                                                            string list) -> unit
  val _ = op showStackTrace : unit -> unit
end
(* To avoid confusion, tracing code typically   *)
(* doesn't use [[print]] methods; instead, it   *)
(* uses [[valueString]] to give information     *)
(* about a value.                               *)
(* <tracing>=                                   *)
fun valueString (c, NUM n) = 
      String.map (fn #"~" => #"-" | c => c) (Int.toString n) ^
      valueString(c, USER [])
  | valueString (_, SYM v) = v
  | valueString (c, _) = "<" ^ className c ^ ">"
fun eval(e, rho, superclass, xi) =
  let (* <local helper functions of [[eval]]>=        *)
      fun findMethod (name, class) =
        let fun fm (CLASS { methods, super, ...}) =
              find (name, methods)
              handle NotFound m =>
                case super
                  of SOME c => fm c
                   | NONE   => raise RuntimeError
                                 (className class ^
                                            " does not understand message " ^ m)
        in  fm class
        end
      (* The first function, [[findMethod]],          *)
      (* implements method search. Because we search  *)
      (* sometimes on the class of the receiver, and  *)
      (* sometimes on [[superclass]], the class to    *)
      (* search is a parameter.                       *)
      (* <boxed values 14>=                           *)
      val _ = op findMethod: name * class -> method
      (* <local helper functions of [[eval]]>=        *)
      fun evalMethod (PRIM_METHOD (name, f), receiver, actuals) = f (receiver,
                                                                        actuals)
        | evalMethod (USER_METHOD { name, superclass, formals, locals, body },
                      receiver, actuals) =
            let val rho = instanceVars receiver
                val rho = bindList(formals, map ref actuals, rho)
                val rho = bindList(locals, map (fn _ => ref nilValue) locals,
                                                                            rho)
            in  eval(body, rho, superclass, xi)
            end
              handle BindListLength => 
                  raise RuntimeError
                      ("Wrong number of arguments to method " ^ name ^
                                                                "; expected (" ^
                       separate("", " ") (name :: "self" :: formals) ^ ")")
      and instanceVars (_, USER rep) = rep
        | instanceVars self = bind("self", ref self, emptyEnv)
      (* To evaluate a primitive method, we apply the *)
      (* method's function. To evaluate a             *)
      (* user-defined method, we build a new [[rho]]. *)
      (* As noted above, that [[rho]] binds instance  *)
      (* variables, parameters, and temporaries.      *)
      (* <boxed values 15>=                           *)
      val _ = op evalMethod   : method * value * value list -> value
      val _ = op instanceVars : value -> value ref env
      (* [[self]]                                     *)
      (* is part of the [[rep]]. For every other kind *)
      (* of object, [[instanceVars]] binds only       *)
      (* [[self]].                                    *)

      (* <local helper functions of [[eval]]>=        *)
      fun applyClosure ((formals, body, rho, superclass), actuals) =
        eval(BEGIN body, bindList(formals, map ref actuals, rho), superclass, xi
                                                                               )
        handle BindListLength => 
            raise RuntimeError ("Wrong number of arguments to block; expected "
                                                                               ^
                                "(value <block>" ^ separate("", " ") formals ^
                                                                            ")")
      (* We handle all but one of the primitive       *)
      (* methods in [[evalMethod]], using a           *)
      (* function [[f]] inside a [[PRIM_METHOD]]. But *)
      (* to handle sending the [[value]] message to a *)
      (* block, we need to make a recursive call to   *)
      (* [[eval]]. It's easier to build that ability  *)
      (* into a helper function than to try to        *)
      (* encapsulate it in a [[PRIM_METHOD]].         *)
      (* <boxed values 16>=                           *)
      val _ = op applyClosure : (name list * exp list * value ref env * class) *
                                                             value list -> value
      (* <function [[ev]], the evaluator proper>=     *)
      fun ev(VALUE v) = v
        | ev(LITERAL c) = 
            (case c of NUM n => mkInteger n
                     | SYM n => mkSymbol n
                     | _ => raise InternalError "unexpected literal")
      (* The rules for [[VAR]] and [[SET]] are as we  *)
      (* would expect, given that we have both local  *)
      (* and global environments.                     *)
      (* <function [[ev]], the evaluator proper>=     *)
        | ev(VAR v) = !(find(v, rho) handle NotFound _ => find(v, xi))
        | ev(SET (n, e)) =
            let val v = ev e
                val cell = find(n, rho) handle NotFound _ => find(n, xi)
            in  cell := v; v
            end 
      (* [[SUPER]], when used as an expression, acts  *)
      (* just as [[self]] does.                       *)
      (* <function [[ev]], the evaluator proper>=     *)
        | ev(SUPER) = ev (VAR "self")
      (* The [[BEGIN]] rule is as in micro-Scheme.    *)
      (*                                              *)
      (* <function [[ev]], the evaluator proper>=     *)
        | ev(BEGIN es) =
            let fun b(e::es, lastval) = b(es, ev e)
                  | b(   [], lastval) = lastval
            in  b(es, nilValue)
            end
      (* Evaluating a block means capturing the       *)
      (* current local environment and superclass in  *)
      (* a closure.                                   *)
      (* <function [[ev]], the evaluator proper>=     *)
        | ev(BLOCK (formals, body)) = mkBlock (formals, body, rho, superclass)
      (* The last rule is for message send. First we  *)
      (* evaluate the receiver and the arguments. The *)
      (* message send dispatches on the receiver,     *)
      (* which is used to find the method that        *)
      (* defines [[message]], except when messages    *)
      (* are send to [[super]], in which case we use  *)
      (* the [[superclass]] of the currently running  *)
      (* method. There is one case in which we do not *)
      (* use dynamic dispatch ([[findMethod]]);       *)
      (* sending the [[value]] message to a block is  *)
      (* built directly into the interpreter. Instead *)
      (* of calling [[findMethod]], we call the       *)
      (* evaluator recursively through                *)
      (* [[applyClosure]]. This means that no         *)
      (* subclass of [[Block]] can redefine the       *)
      (* [[value]] method, but creating subclasses of *)
      (* [[Block]] is a useless exercise anyway,      *)
      (* because there is nothing useful to inherit.  *)
      (* [*]                                          *)
      (* <function [[ev]], the evaluator proper>=     *)
        | ev(SEND (srcloc, message, receiver, args))  =
            let val obj as (class, rep) = ev receiver
                val args = map ev args
                val dispatchingClass = case receiver of SUPER => superclass | _
                                                                        => class
                (* We trace every send and answer only if the   *)
                (* uSmalltalk variable [[ --- trace]] is set to *)
                (* a non-zero number. At every trace, we also   *)
                (* decrement [[ --- trace]]. The code that      *)
                (* builds the tracing output is protected by    *)
                (* [[fn () => ]]...; it is executed only if     *)
                (* tracing is turned on.                        *)
                (* <definitions of message-tracing procedures>= *)
                fun traceSend (file, line) =
                  traceIndent (message, (file, line)) xi (fn () =>
                     [file, ", line ", Int.toString line, ": Sending message (",
                      separate ("", " ") (message :: map valueString args), ")",
                      " to object of class ", className dispatchingClass])
                fun traceAnswer (file, line) answer =
                  ( outdentTrace xi (fn () =>
                       [file, ", line ", Int.toString line, ": (",
                        separate  ("", " ") (message :: map valueString (obj ::
                                                                         args)),
                        ") = ", valueString answer])
                  ; answer
                  )
                (* [[traceIndent]], [[outdentTrace]],           *)
                (* and [[valueString]] are defined in           *)
                (* [[]]. Chunk [[]] also                        *)
                (* defines function [[showStackTrace]], which   *)
                (* is called from chunk [->] to show the stack  *)
                (* of active message sends after a run-time     *)
                (* error.                                       *)

                fun performSend () =
                  case (message, rep)
                    of ("value", CLOSURE clo) => applyClosure(clo, args)
                     | _ => evalMethod(findMethod(message, dispatchingClass),
                                                                      obj, args)

            in  ( traceSend srcloc
                ; traceAnswer srcloc (performSend ())
                )
            end
      (* With these helper functions in place, we can *)
      (* write the evaluator. A [[VALUE]] node stands *)
      (* for itself. When we see a [[LITERAL]] node,  *)
      (* we call [[mkInteger]] or [[mkSymbol]] to     *)
      (* build the literal. It is unsafe to call      *)
      (* these functions until we have read the       *)
      (* initial basis and bootstrapped the           *)
      (* interpreter; integer or symbol literals in   *)
      (* the initial basis had better appear only     *)
      (* inside method definitions.                   *)
      (* <boxed values 17>=                           *)
      val _ = op ev : exp -> value
(* The evaluator therefore takes four           *)
(* arguments: an expression to be evaluated; a  *)
(* local environment, which binds instance      *)
(* variables, formal parameters, and            *)
(* temporaries; a class used to send message    *)
(* to [[super]]; and the global environment.    *)
(* <boxed values 13>=                           *)
val _ = op eval: exp * value ref env * class * value ref env -> value
val _ = op ev  : exp -> value
  in  ev e
  end
(*   Function [[addMethodDefn]] takes care of   *)
(* processing each method definition and adding *)
(* it either to the list of class methods or to *)
(* the list of instance methods for the new     *)
(* class. We use [[foldr]] to accumulate these  *)
(* lists and deposit the results in             *)
(* [[imethods]] and [[cmethods]]. [*]           *)
(* <evaluation>=                                *)
val primitiveMethods = methods ((* <primitive methods [[::]]>=
                                                                              *)
                                primMethod "eqObject" (binaryPrim (mkBoolean o
                                                                      eqRep)) ::
                                (* <primitive methods [[::]]>=
                                                                              *)
                                primMethod "print" (unaryPrim defaultPrint) ::
                                (* Here are the primitive operations on small
                                                                              *)
                                (* integers. [*]
                                                                              *)
                                (* <primitive methods [[::]]>=
                                                                              *)
                                primMethod "printSmallInteger" (unaryPrim
                                                                    printInt) ::
                                primMethod "newSmallInteger:"  (binaryPrim
                                                                  newInteger) ::
                                primMethod "+"   (arithop op +  )  ::
                                primMethod "-"   (arithop op -  )  ::
                                primMethod "*"   (arithop op *  )  ::
                                primMethod "div" (arithop op div)  ::
                                primMethod "<"   (intcompare op <) ::
                                primMethod ">"   (intcompare op >) ::
                                (* [[SmallInteger]].
                                                                              *)

                                (* <primitive methods [[::]]>=
                                                                              *)
                                primMethod "printSymbol" (unaryPrim  printSymbol
                                                                            ) ::
                                primMethod "newSymbol"   (binaryPrim newSymbol
                                                                            ) ::
                                (* Here are all the primitive array methods.
                                                                              *)
                                (* [*]
                                                                              *)
                                (* <primitive methods [[::]]>=
                                                                              *)
                                primMethod "arrayNew:"    (binaryPrim newArray)
                                                                              ::
                                primMethod "arraySize"    (arrayPrimitive
                                                                   arraySize) ::
                                primMethod "arrayAt:"     (arrayPrimitive
                                                                     arrayAt) ::
                                primMethod "arrayAt:Put:" (arrayPrimitive
                                                                  arrayAtPut) ::
                                (* [[Array]].
                                                                              *)

                                (*  Block primitives
                                                                              *)
                                (*
                                                                              *)
                                (* Actually, [[value]] is not a primitive; it
                                                                              *)
                                (* is built into [[eval]]. We create a
                                                                              *)
                                (* primitive method called [[value]] anyway,
                                                                              *)
                                (* which we use in the definition of class
                                                                              *)
                                (* [[Block]] (page [->]), so we can get a
                                                                              *)
                                (* suitable error message if there's a bug in
                                                                              *)
                                (* the interpreter.
                                                                              *)
                                (* <primitive methods [[::]]>=
                                                                              *)
                                primMethod "value" (fn _ => raise InternalError
                                                    "got to primitive value") ::
                                (* <primitive methods [[::]]>=
                                                                              *)
                                primMethod "add:withOverflow:" (withOverflow op
                                                                          + ) ::
                                primMethod "sub:withOverflow:" (withOverflow op
                                                                          - ) ::
                                primMethod "mul:withOverflow:" (withOverflow op
                                                                     * ) :: nil)
fun newClassObject {name, super, ivars, methods = ms} xi =
  let val (superMeta, super) = findClass(super, xi)
        handle NotFound s => raise RuntimeError ("Superclass " ^ s ^
                                                                   " not found")
      fun method (kind, name, PRIM_IMPL prim) =
            renameMethod(name, primitiveMethod prim)
        | method (kind, name, USER_IMPL (formals, ls, body)) =
            USER_METHOD { name = name, formals = formals, body = body, locals =
                                                                              ls
                        , superclass = case kind of IMETHOD => super
                                                  | CMETHOD => superMeta
                        }
      fun addMethodDefn (m as (CMETHOD, _, _), (c's, i's)) = (method m :: c's,
                                                                            i's)
        | addMethodDefn (m as (IMETHOD, _, _), (c's, i's)) = (c's, method m ::
                                                                            i's)
      val (cmethods, imethods) = foldr addMethodDefn ([], []) ms
      val metaclassName = "class " ^ name
      val class     = mkClass name          super      ivars imethods
      val metaclass = mkClass metaclassName superMeta  []    cmethods
  in  (metaclass, CLASSREP class)
  end
and primitiveMethod name =
  find(name, primitiveMethods)
  handle NotFound n => raise RuntimeError ("There is no primitive method named "
                                                                            ^ n)
(* <evaluation>=                                *)
and findClass(supername, xi) =
  case !(find(supername, xi))
    of (meta, CLASSREP c) => (meta, c)
     | v => raise RuntimeError ("object " ^ supername ^ " = " ^ valueString v ^
                                " is not a class")
(* <evaluation>=                                *)
fun topeval (t, echo, xi)  =
  case t
    of USE filename => use readEvalPrint filename xi 
     | DEFINE (name, args, body) => 
         topeval (VAL (name, BLOCK (args, [body])), echo, xi)
     | VAL (name, e) => 
         doVal echo (name, eval(e, emptyEnv, objectClass, xi), xi)
     | CLASSD (d as {name, methods, ...}) =>
         doVal echo (name, newClassObject d xi, xi)
     | EXP e => 
         (printValue echo xi (eval(e, emptyEnv, objectClass, xi)); xi)
and doVal echo (name, value, xi)  =
      ((find(name, xi) := value; xi)
       handle NotFound _ => bind (name, ref value, xi)
      )
      before printValue echo xi value
(* We print a value using [[print]] method,     *)
(* which can be redefined by a uSmalltalk       *)
(* programmer. Because a print method writes    *)
(* directly to standard output, we have to      *)
(* define [[echo]] to be a Boolean, not a       *)
(* function as in other interpreters.           *)
(* <evaluation>=                                *)
and printValue echo xi v = 
  if echo then
    ( eval (SEND (nullsrc, "print", VALUE v, []), emptyEnv, objectClass, xi)
    ; print "\n"
    ; ()
    )
  else
    ()
(* [[nullsrc]] identifies                       *)
(* the [[SEND]] as something generated          *)
(* internally, rather than read from a file or  *)
(* a list of strings.                           *)

(*  The read-eval-print loop                    *)
(*                                              *)
(* There are two new things about the           *)
(* read-eval-print loop.                        *)
(*   * If an error occurs, before continuing we *)
(*     call [[showStackTrace]] and              *)
(*     [[resetTrace]].                          *)
(*   * Before each top-level evaluation, we     *)
(*     call [[closeLiteralsCycle]]. This call   *)
(*     updates the definitions (if any) of      *)
(*     [[SmallInteger]], [[Symbol]], and        *)
(*     [[Array]]. The benefit of doing this     *)
(*     update is that a uSmalltalk programmer   *)
(*     can easily change the behavior of these  *)
(*     classes (as is required in Exercises     *)
(*     [->], [->], [->], and [->]).             *)
(*                                              *)
(* Because the calls to [[closeLiteralsCycle]]  *)
(* and [[closeBlocksCycle]] will fail while the *)
(* initial basis is being read, failure is      *)
(* ignored (but see chunk [->]). [*]            *)
(* <evaluation>=                                *)
and readEvalPrint (reader, echo, errmsg) xi =
  let fun loop xi =
        let fun continue msg = (errmsg msg; showStackTrace(); resetTrace(); loop
                                                                             xi)
            fun finish () = xi
            val _ = (closeLiteralsCycle xi; closeBlocksCycle xi)
                    handle NotFound _ => ()
        in  loop (topeval (readtop reader, echo, xi))
            handle EOF => finish () 
                 (* <more read-eval-print handlers>=             *)
                 | InternalError s => continue ("internal interpreter error: " ^
                                                                              s)
                 | Subscript       => continue ("array subscript out of bounds")
                 | Size            => continue ("bad array size")
                 (* The handlers in the first group deal with    *)
                 (* problems that arise during I/O, lexical      *)
                 (* analysis, and parsing.                       *)
                 (* <more read-eval-print handlers>=             *)
                 | IO.Io {name, ...} => continue ("I/O error: " ^ name)
                 | Paren (_, loc)    => continue ("syntax error: extra ) in " ^
                                                               srclocString loc)
                 | PrematureEOF s    => (errmsg ("Missing )? Got EOF reading ("
                                                                 ^ s); finish())
                 | SyntaxError msg   => continue ("syntax error: " ^ msg)
                 (* [[IO.Io]] is part of the                     *)
                 (* Standard Basis Library.                      *)

                 (* The handlers in the second group deal with   *)
                 (* problems that arise during evaluation. They  *)
                 (* always continue with the next top-level      *)
                 (* item.                                        *)
                 (* <more read-eval-print handlers>=             *)
                 | Div               => continue "Division by zero"
                 | Overflow          => continue "Arithmetic overflow"
                 | RuntimeError msg  => continue ("run-time error: " ^ msg)
                 | NotFound n        => continue ("variable " ^ n ^ " not found"
                                                                               )
                 (* [[Div]] and [[Overflow]] are                 *)
                 (* predefined.                                  *)

        end
  in  loop xi
  end
(* The object named as a superclass must in     *)
(* fact represent a class, so its               *)
(* representation must be [[CLASSREP c]], where *)
(* [[c]] is the class it represents. That       *)
(* object is an instance of its metaclass.      *)
(* Function [[findClass]] returns the metaclass *)
(* and the class.                               *)
(* <boxed values 18>=                           *)
val _ = op findClass : name * value ref env -> class * class
(* The top-level evaluator computes a new       *)
(* global environment. We have optimized the    *)
(* \xval binding so that if a previous binding  *)
(* exists, we replace it. It is safe to do this *)
(* because no operation in uSmalltalk makes a   *)
(* copy of the global environment. \xdefine is  *)
(* syntactic sugar for a new block.             *)
(* <boxed values 18>=                           *)
val _ = op topeval : toplevel * bool * value ref env -> value ref env
(* [[]]                                         *)
(* We add handlers for our new exception        *)
(* [[InternalError]] and for the array          *)
(* primitives.                                  *)



(*****************************************************************)
(*                                                               *)
(*   INITIALIZATION                                              *)
(*                                                               *)
(*****************************************************************)

(*   Initializing, bootstrapping, and running the *)
(*   interpreter                                *)
(*                                              *)
(* The first step in creating the initial       *)
(* environment is to add the built-in classes.  *)
(* Each one needs a metaclass to be an instance *)
(* of. To be faithful to Smalltalk, the         *)
(* subclass relationships of the metaclasses    *)
(* should be isomorphic to the subclass         *)
(* relationships of the classes. This is true   *)
(* for the user-defined classes created with    *)
(* [[newClassObject]], but we cheat on the      *)
(* built-in classes, making the metaclasses for *)
(* [[UndefinedObject]] and [[Class]] inherit    *)
(* directly from [[Class]], not from            *)
(* [[Object]]'s metaclass.                      *)
(* <initialization>=                            *)
val initialXi = emptyEnv

fun mkMeta c = mkClass ("class " ^ className c) classClass [] []
fun addClass (c, xi) = bind(className c, ref (mkMeta c, CLASSREP c), xi)
val initialXi = foldl addClass initialXi [ objectClass, nilClass, classClass ]
(* The next step is to read the class           *)
(* definitions in the initial basis. For        *)
(* debugging, it can be helpful to set          *)
(* [[echoBasis]] to [[true]].                   *)
(* <initialization>=                            *)
val echoBasis = false
val initialXi =
  let val rdr = stringsreader ("initial basis", (* <ML representation of initial
                                                               basis>=        *)
                                                  [ "; usm.nw 2929"
                                                  , "(class Boolean Object ()"
                                                  ,
"    (method ifTrue:ifFalse: (trueBlock falseBlock) (subclassResponsibility self))"
                                                  , "  "
                                                  ,
                           "    (method ifFalse:ifTrue: (falseBlock trueBlock) "
                                                  ,
                          "        (ifTrue:ifFalse: self trueBlock falseBlock))"
                                                  ,
        "    (method ifTrue:  (trueBlock)  (ifTrue:ifFalse: self trueBlock []))"
                                                  ,
       "    (method ifFalse: (falseBlock) (ifTrue:ifFalse: self [] falseBlock))"
                                                  , "    "
                                                  ,
   "    (method not ()          (ifTrue:ifFalse: self [false]          [true]))"
                                                  ,
"    (method eqv: (aBoolean) (ifTrue:ifFalse: self [aBoolean]       [(not aBoolean)]))"
                                                  ,
"    (method xor: (aBoolean) (ifTrue:ifFalse: self [(not aBoolean)] [aBoolean]))"
                                                  , ""
                                                  ,
            "    (method & (aBoolean) (ifTrue:ifFalse: self [aBoolean] [self]))"
                                                  ,
            "    (method | (aBoolean) (ifTrue:ifFalse: self [self] [aBoolean]))"
                                                  , "  "
                                                  ,
"    (method and: (alternativeBlock) (ifTrue:ifFalse: self alternativeBlock [self]))"
                                                  ,
"    (method or:  (alternativeBlock) (ifTrue:ifFalse: self [self] alternativeBlock))"
                                                  , ""
                                                  ,
"    (method if (trueBlock falseBlock) (ifTrue:ifFalse: self trueBlock falseBlock))"
                                                  , ")"
                                                  , "; usm.nw 2961"
                                                  , "(class True Boolean ()"
                                                  ,
           "  (method ifTrue:ifFalse: (trueBlock falseBlock) (value trueBlock))"
                                                  , ")"
                                                  , "(class False Boolean ()"
                                                  ,
          "  (method ifTrue:ifFalse: (trueBlock falseBlock) (value falseBlock))"
                                                  , ")"
                                                  , "; usm.nw 3124"
                                                  , "(class Block Object "
                                                  ,
                                              "    () ; internal representation"
                                                  ,
                                            "    (method value primitive value)"
                                                  ,
                                                 "    (method whileTrue: (body)"
                                                  ,
                                         "        (ifTrue:ifFalse: (value self)"
                                                  , "            [(value body)"
                                                  ,
                                          "             (whileTrue: self body)]"
                                                  , "            [nil]))"
                                                  ,
                              "    (method while (body) (whileTrue: self body))"
                                                  ,
                                               "    (method whileFalse: (body) "
                                                  ,
                                        "         (ifTrue:ifFalse: (value self)"
                                                  ,
                                                "             [nil]            "
                                                  ,
                                                   "             [(value body) "
                                                  ,
                                      "              (whileFalse: self body)]))"
                                                  , ")"
                                                  , "; usm.nw 5913"
                                                  , "(class Symbol Object"
                                                  ,
                                              "    () ; internal representation"
                                                  ,
              "    (classMethod new () (error: self #can't-send-new-to-Symbol))"
                                                  ,
                                  "    (classMethod new:   primitive newSymbol)"
                                                  ,
                                "    (method      print  primitive printSymbol)"
                                                  , ")"
                                                  , "; usm.nw 4106"
                                                  , "(class Magnitude Object "
                                                  , "    () ; abstract class"
                                                  ,
"    (method =  (x) (subclassResponsibility self)) ; may not inherit from Object"
                                                  ,
                             "    (method <  (x) (subclassResponsibility self))"
                                                  ,
                                                "    (method >  (y) (< y self))"
                                                  ,
                                          "    (method <= (x) (not (> self x)))"
                                                  ,
                                          "    (method >= (x) (not (< self x)))"
                                                  ,
   "    (method min: (aMagnitude) (if (< self aMagnitude) [self] [aMagnitude]))"
                                                  ,
   "    (method max: (aMagnitude) (if (> self aMagnitude) [self] [aMagnitude]))"
                                                  , ")"
                                                  , "; usm.nw 4120"
                                                  , "(class Number Magnitude"
                                                  , "    ()  ; abstract class"
                                                  ,
                                             "    ;;;;;;; basic Number protocol"
                                                  ,
                  "    (method +   (aNumber)     (subclassResponsibility self))"
                                                  ,
                  "    (method *   (aNumber)     (subclassResponsibility self))"
                                                  ,
                  "    (method negated    ()     (subclassResponsibility self))"
                                                  ,
                  "    (method reciprocal ()     (subclassResponsibility self))"
                                                  , "    "
                                                  ,
                  "    (method asInteger  ()     (subclassResponsibility self))"
                                                  ,
                  "    (method asFraction ()     (subclassResponsibility self))"
                                                  ,
                  "    (method asFloat    ()     (subclassResponsibility self))"
                                                  ,
                  "    (method coerce: (aNumber) (subclassResponsibility self))"
                                                  , "    "
                                                  , "; usm.nw 4138"
                                                  ,
                                        "(method -   (y) (+ self (negated  y)))"
                                                  ,
                "(method abs ()  (if (negative self) [(negated  self)] [self]))"
                                                  ,
                                      "(method /   (y) (* self (reciprocal y)))"
                                                  , "; usm.nw 4146"
                                                  ,
                       "(method negative         () (<  self (coerce: self 0)))"
                                                  ,
                       "(method positive         () (>= self (coerce: self 0)))"
                                                  ,
                       "(method strictlyPositive () (>  self (coerce: self 0)))"
                                                  , "; usm.nw 4154"
                                                  ,
                                             "(method squared () (* self self))"
                                                  ,
                                          "(method raisedToInteger: (anInteger)"
                                                  , "    (if (= anInteger 0)"
                                                  , "        [(coerce: self 1)]"
                                                  ,
                                           "        [(if (= anInteger 1) [self]"
                                                  ,
          "            [(* (squared (raisedToInteger: self (div: anInteger 2)))"
                                                  ,
              "                (raisedToInteger: self (mod: anInteger 2)))])]))"
                                                  , "; usm.nw 4181"
                                                  ,
                   "(method sqrt () (sqrtWithin: self (coerce: self (/ 1 10))))"
                                                  ,
                      "(method sqrtWithin: (epsilon) (locals two x_{i-1} x_{i})"
                                                  ,
                             "    ; find square root of receiver within epsilon"
                                                  ,
                                            "    (set two     (coerce: self 2))"
                                                  ,
                                            "    (set x_{i-1} (coerce: self 1))"
                                                  ,
                        "    (set x_{i}   (/ (+ x_{i-1} (/ self x_{i-1})) two))"
                                                  ,
                              "    (while [(> (abs (- x_{i-1} x_{i})) epsilon)]"
                                                  ,
                                               "           [(set x_{i-1} x_{i})"
                                                  ,
                "            (set x_{i} (/ (+ x_{i-1} (/ self x_{i-1})) two))])"
                                                  , "    x_{i})"
                                                  , "; usm.nw 4133"
                                                  , ")"
                                                  , "; usm.nw 4513"
                                                  , "(class Integer Number"
                                                  , "    () ; abstract class"
                                                  ,
                           "    (method div: (n) (subclassResponsibility self))"
                                                  ,
                            "    (method mod: (n) (- self (* n (div: self n))))"
                                                  ,
"    (method gcd: (n) (if (= n (coerce: self 0)) [self] [(gcd: n (mod: self n))]))"
                                                  ,
                         "    (method lcm: (n) (* self (div: n (gcd: self n))))"
                                                  , "    "
                                                  , "; usm.nw 4529"
                                                  ,
                            "(method asFraction () (num:den:  Fraction self 1))"
                                                  ,
                            "(method asFloat    () (mant:exp: Float    self 0))"
                                                  ,
                                                   "(method asInteger  () self)"
                                                  , "; usm.nw 4538"
                                                  ,
                                "(method coerce: (aNumber) (asInteger aNumber))"
                                                  , "; usm.nw 4543"
                                                  ,
                            "(method reciprocal () (num:den: Fraction 1 self)) "
                                                  ,
                            "(method / (aNumber) (/ (asFraction self) aNumber))"
                                                  , "; usm.nw 4549"
                                                  ,
                                  "(method timesRepeat: (aBlock) (locals count)"
                                                  ,
          "    (ifTrue: (negative self) [(error: self #negative-repeat-count)])"
                                                  , "    (set count self)"
                                                  , "    (while [(!= count 0)]"
                                                  , "         [(value aBlock)"
                                                  ,
                                          "          (set count (- count 1))]))"
                                                  , "; usm.nw 4520"
                                                  , ")"
                                                  , "; usm.nw 4576"
                                                  ,
                                                   "(class SmallInteger Integer"
                                                  ,
                                             "    () ; primitive representation"
                                                  ,
                             "    (classMethod new: primitive newSmallInteger:)"
                                                  ,
                                      "    (classMethod new   () (new: self 0))"
                                                  ,
                                         "    (method negated    () (- 0 self))"
                                                  ,
                                "    (method print primitive printSmallInteger)"
                                                  ,
                                                "    (method +     primitive +)"
                                                  ,
                                                "    (method -     primitive -)"
                                                  ,
                                                "    (method *     primitive *)"
                                                  ,
                                              "    (method div:  primitive div)"
                                                  ,
                                         "    (method =     primitive eqObject)"
                                                  ,
                                                "    (method <     primitive <)"
                                                  ,
                                                "    (method >     primitive >)"
                                                  , ")"
                                                  , "; usm.nw 4608"
                                                  , "(class Fraction Number"
                                                  , "    (num den)"
                                                  ,
                "    (classMethod num:den: (a b) (initNum:den: (new self) a b))"
                                                  ,
                                      "    (method initNum:den: (a b) ; private"
                                                  ,
                                                "        (setNum:den: self a b)"
                                                  , "        (signReduce self)"
                                                  , "        (divReduce self))"
                                                  ,
         "    (method setNum:den: (a b) (set num a) (set den b) self) ; private"
                                                  , "    "
                                                  , "; usm.nw 4641"
                                                  ,
                                               "(method signReduce () ; private"
                                                  ,
                                                   "    (ifTrue: (negative den)"
                                                  ,
                    "        [(set num (negated num)) (set den (negated den))])"
                                                  , "    self)"
                                                  , ""
                                                  ,
                                  "(method divReduce () (locals temp) ; private"
                                                  , "    (if (= num 0)"
                                                  , "        [(set den 1)]"
                                                  ,
                                      "        [(set temp (gcd: (abs num) den))"
                                                  ,
                                           "         (set num  (div: num temp))"
                                                  ,
                                         "         (set den  (div: den temp))])"
                                                  , "    self)"
                                                  , "; usm.nw 4660"
                                                  , "(method num () num)"
                                                  , "(method den () den)"
                                                  , "; usm.nw 4669"
                                                  , "(method = (f)"
                                                  ,
                                 "    (and: (= num (num f)) [(= den (den f))]))"
                                                  , "(method < (f)"
                                                  ,
                                      "    (< (* num (den f)) (* (num f) den)))"
                                                  , "; usm.nw 4678"
                                                  ,
            "(method negated () (setNum:den: (new Fraction) (negated num) den))"
                                                  , "; usm.nw 4692"
                                                  , "(method * (f)"
                                                  , "    (divReduce"
                                                  ,
                                           "        (setNum:den: (new Fraction)"
                                                  ,
                                       "                        (* num (num f))"
                                                  ,
                                    "                        (* den (den f)))))"
                                                  , "; usm.nw 4711"
                                                  ,
                                                   "(method + (f) (locals temp)"
                                                  ,
                                             "    (set temp (lcm: den (den f)))"
                                                  , "    (divReduce"
                                                  ,
                                           "        (setNum:den: (new Fraction)"
                                                  ,
"                     (+ (* num (div: temp den)) (* (num f) (div: temp (den f))))"
                                                  ,
                                                  "                     temp)))"
                                                  , "; usm.nw 4725"
                                                  , "(method reciprocal ()"
                                                  ,
                         "   (signReduce (setNum:den: (new Fraction) den num)))"
                                                  , "; usm.nw 4730"
                                                  ,
                     "(method print () (print num) (print #/) (print den) self)"
                                                  , "; usm.nw 4740"
                                                  ,
                                         "(method asInteger  () (div: num den))"
                                                  ,
                        "(method asFloat    () (/ (asFloat num) (asFloat den)))"
                                                  ,
                                                   "(method asFraction () self)"
                                                  ,
                               "(method coerce: (aNumber) (asFraction aNumber))"
                                                  , "; usm.nw 4749"
                                                  ,
                                   "(method negative         () (negative num))"
                                                  ,
                                   "(method positive         () (positive num))"
                                                  ,
                           "(method strictlyPositive () (strictlyPositive num))"
                                                  , "; usm.nw 4617"
                                                  , ")"
                                                  , "; usm.nw 4782"
                                                  , "(class Float Number"
                                                  , "    (mant exp)"
                                                  ,
              "    (classMethod mant:exp: (m e) (initMant:exp: (new self) m e))"
                                                  ,
                                     "    (method initMant:exp: (m e) ; private"
                                                  ,
                            "        (set mant m) (set exp e) (normalize self))"
                                                  ,
                                         "    (method normalize ()    ; private"
                                                  ,
                                         "        (while [(> (abs mant) 32767)]"
                                                  ,
                                     "               [(set mant (div: mant 10))"
                                                  ,
                                         "                (set exp (+ exp 1))])"
                                                  , "        self)"
                                                  , "    "
                                                  , "; usm.nw 4798"
                                                  ,
                                              "(method mant () mant)  ; private"
                                                  ,
                                              "(method exp  () exp)   ; private"
                                                  , "; usm.nw 4806"
                                                  ,
                                          "(method < (x) (negative (- self x)))"
                                                  ,
                                          "(method = (x) (isZero   (- self x)))"
                                                  ,
                                                 "(method isZero () (= mant 0))"
                                                  , "; usm.nw 4813"
                                                  ,
                      "(method negated () (mant:exp: Float (negated mant) exp))"
                                                  , "; usm.nw 4834"
                                                  , "(method + (prime) "
                                                  ,
                                                  "    (if (>= exp (exp prime))"
                                                  ,
"        [(mant:exp: Float (+ (* mant (raisedToInteger: 10 (- exp (exp prime))))"
                                                  ,
                                    "                             (mant prime))"
                                                  ,
                                       "                          (exp prime))]"
                                                  , "        [(+ prime self)]))"
                                                  , "; usm.nw 4846"
                                                  , "(method * (prime) "
                                                  ,
              "    (mant:exp: Float (* mant (mant prime)) (+ exp (exp prime))))"
                                                  , "; usm.nw 4855"
                                                  , "(method reciprocal ()"
                                                  ,
                      "    (mant:exp: Float (div: 1000000000 mant) (- -9 exp)))"
                                                  , "; usm.nw 4861"
                                                  ,
                                  "(method coerce: (aNumber) (asFloat aNumber))"
                                                  , "(method asFloat () self)"
                                                  , "; usm.nw 4867"
                                                  , "(method asInteger ()"
                                                  , "    (if (< exp 0)"
                                                  ,
                     "        [(div: mant (raisedToInteger: 10 (negated exp)))]"
                                                  ,
                             "        [(*    mant (raisedToInteger: 10 exp))]))"
                                                  , "; usm.nw 4875"
                                                  , "(method asFraction ()"
                                                  , "    (if (< exp 0)"
                                                  ,
        "        [(num:den: Fraction mant (raisedToInteger: 10 (negated exp)))]"
                                                  ,
          "        [(num:den: Fraction (* mant (raisedToInteger: 10 exp)) 1)]))"
                                                  , "; usm.nw 4896"
                                                  ,
                                  "(method negative         () (negative mant))"
                                                  ,
                                  "(method positive         () (positive mant))"
                                                  ,
                          "(method strictlyPositive () (strictlyPositive mant))"
                                                  , "; usm.nw 4904"
                                                  , "(method print () "
                                                  ,
                                                   "    (print-normalize self) "
                                                  ,
                                    "    (print mant) (print #x10^) (print exp)"
                                                  , "    self)"
                                                  , ""
                                                  , "(method print-normalize ()"
                                                  ,
                          "    (while [(and: (< exp 0) [(= (mod: mant 10) 0)])]"
                                                  ,
                     "        [(set exp (+ exp 1)) (set mant (div: mant 10))]))"
                                                  , "; usm.nw 4793"
                                                  , ")"
                                                  , "; usm.nw 3187"
                                                  , "(class Collection Object"
                                                  , "  () ; abstract"
                                                  ,
                  "  (method do:     (aBlock)    (subclassResponsibility self))"
                                                  ,
                  "  (method add:    (newObject) (subclassResponsibility self))"
                                                  ,
                               "  (method remove:ifAbsent: (oldObject exnBlock)"
                                                  ,
                  "                              (subclassResponsibility self))"
                                                  ,
                  "  (method species ()          (subclassResponsibility self))"
                                                  , "  "
                                                  , "; usm.nw 3199"
                                                  ,
                                   "(classMethod with: (anObject) (locals temp)"
                                                  , "    (set temp (new self))"
                                                  , "    (add: temp anObject)"
                                                  , "    temp)"
                                                  , "; usm.nw 3211"
                                                  ,
                                                  "(method remove: (oldObject) "
                                                  ,
"    (remove:ifAbsent: self oldObject [(error: self #tried-to-remove-absent-object)]))"
                                                  ,
                                                "(method addAll: (aCollection) "
                                                  ,
                               "    (do: aCollection (block (x) (add: self x)))"
                                                  , "    aCollection)"
                                                  ,
                                             "(method removeAll: (aCollection) "
                                                  ,
                            "    (do: aCollection (block (x) (remove: self x)))"
                                                  , "    aCollection)"
                                                  , "; usm.nw 3231"
                                                  ,
                                         "(method isEmpty () (= (size self) 0))"
                                                  ,
                                                 "(method size () (locals temp)"
                                                  , "    (set temp 0)"
                                                  ,
                              "    (do: self (block (_) (set temp (+ temp 1))))"
                                                  , "    temp)"
                                                  ,
                               "(method occurrencesOf: (anObject) (locals temp)"
                                                  , "    (set temp 0)"
                                                  , "    (do: self (block (x)"
                                                  ,
                     "       (ifTrue: (= x anObject) [(set temp (+ temp 1))])))"
                                                  , "    temp)"
                                                  ,
            "(method includes: (anObject) (< 0 (occurrencesOf: self anObject)))"
                                                  , "(method detect: (aBlock) "
                                                  ,
         "    (detect:ifNone: self aBlock [(error: self #no-object-detected)]))"
                                                  ,
            "(method detect:ifNone: (aBlock exnBlock) (locals answer searching)"
                                                  , "    (set searching true)"
                                                  , "    (do: self (block (x)"
                                                  ,
                          "        (ifTrue: (and: searching [(value aBlock x)])"
                                                  ,
                                           "             [(set searching false)"
                                                  ,
                                              "              (set answer x)])))"
                                                  ,
                                         "    (if searching exnBlock [answer]))"
                                                  , "; usm.nw 3263"
                                                  ,
                                  "(method inject:into: (thisValue binaryBlock)"
                                                  ,
     "   (do: self (block (x) (set thisValue (value binaryBlock x thisValue))))"
                                                  , "   thisValue)"
                                                  , "; usm.nw 3271"
                                                  ,
                                        "(method select: (aBlock) (locals temp)"
                                                  ,
                                            "   (set temp (new (species self)))"
                                                  ,
          "   (do: self (block (x) (ifTrue: (value aBlock x) [(add: temp x)])))"
                                                  , "   temp)"
                                                  , "(method reject: (aBlock)"
                                                  ,
                         "   (select: self (block (x) (not (value aBlock x)))))"
                                                  ,
                                       "(method collect: (aBlock) (locals temp)"
                                                  ,
                                            "   (set temp (new (species self)))"
                                                  ,
                        "   (do: self (block (x) (add: temp (value aBlock x))))"
                                                  , "   temp)"
                                                  , "; usm.nw 3284"
                                                  ,
                                                "(method asSet () (locals temp)"
                                                  , "     (set temp (new Set))"
                                                  ,
                                     "     (do: self (block (x) (add: temp x)))"
                                                  , "     temp)"
                                                  , "; usm.nw 3293"
                                                  , "(method print ()"
                                                  , "    (printName self)"
                                                  , "    (print lparen)"
                                                  ,
                            "    (do: self (block (x) (print space) (print x)))"
                                                  , "    (print space)"
                                                  , "    (print rparen)"
                                                  , "    self)"
                                                  ,
                                     "(method printName () (print #Collection))"
                                                  , "; usm.nw 3195"
                                                  , ")"
                                                  , "; usm.nw 3315"
                                                  , "(class Set Collection"
                                                  ,
                                             "    (members)  ; list of elements"
                                                  ,
                                "    (classMethod new () (initSet (new super)))"
                                                  ,
              "    (method initSet  () (set members (new List)) self) ; private"
                                                  ,
                                "    (method do: (aBlock) (do: members aBlock))"
                                                  , "    (method add: (item)"
                                                  ,
             "        (ifFalse: (includes: members item) [(add: members item)])"
                                                  , "        item)"
                                                  ,
                                 "    (method remove:ifAbsent: (item exnBlock) "
                                                  ,
                             "        (remove:ifAbsent: members item exnBlock))"
                                                  ,
                                                "    (method asSet     () self)"
                                                  ,
                                        "    (method printName () (print #Set))"
                                                  ,
                                                 "    (method species   () Set)"
                                                  , ")"
                                                  , "; usm.nw 3359"
                                                  ,
                                             "(class KeyedCollection Collection"
                                                  , "    ()  ; abstract class"
                                                  ,
          "    (method at:put: (key value)       (subclassResponsibility self))"
                                                  ,
          "    (method associationsDo: (aBlock)  (subclassResponsibility self))"
                                                  , "    "
                                                  , "; usm.nw 3370"
                                                  , "(method do: (aBlock) "
                                                  ,
              "    (associationsDo: self (block (x) (value aBlock (value x)))))"
                                                  , "(method at: (key)    "
                                                  ,
                   "    (at:ifAbsent: self key [(error: self #key-not-found])))"
                                                  ,
                                          "(method at:ifAbsent: (key exnBlock) "
                                                  ,
                                 "    (value (associationAt:ifAbsent: self key "
                                                  ,
             "               [(key:value: Association nil (value exnBlock))])))"
                                                  ,
                                                   "(method includesKey: (key) "
                                                  ,
            "    (isKindOf: (associationAt:ifAbsent: self key []) Association))"
                                                  ,
                                                 "(method associationAt: (key) "
                                                  ,
        "    (associationAt:ifAbsent: self key [(error: self #key-not-found])))"
                                                  ,
           "(method associationAt:ifAbsent: (key exnBlock) (locals finishBlock)"
                                                  ,
                                                "    (set finishBlock exnBlock)"
                                                  ,
                                         "    (associationsDo: self (block (x) "
                                                  ,
                   "        (ifTrue: (= (key x) key) [(set finishBlock [x])])))"
                                                  , "    (value finishBlock))"
                                                  , "; usm.nw 3392"
                                                  ,
                                                  "(method keyAtValue: (value) "
                                                  ,
       "    (keyAtValue:ifAbsent: self value [(error: self #value-not-found)]))"
                                                  ,
              "(method keyAtValue:ifAbsent: (value aBlock) (locals finishBlock)"
                                                  ,
                                                  "    (set finishBlock aBlock)"
                                                  ,
                                         "    (associationsDo: self (block (x) "
                                                  ,
         "        (ifTrue: (= (value x) value) [(set finishBlock [(key x)])])))"
                                                  , "    (value finishBlock))"
                                                  , "; usm.nw 3364"
                                                  , ")"
                                                  , "; usm.nw 3433"
                                                  , "(class Association Object "
                                                  , "   (key value)"
                                                  ,
              "   (classMethod key:value: (a b) (setKey:value: (new self) a b))"
                                                  ,
      "   (method setKey:value: (x y) (set key x) (set value y) self) ; private"
                                                  , "   (method key    ()  key)"
                                                  ,
                                                  "   (method value  ()  value)"
                                                  ,
                                          "   (method key:   (x) (set key   x))"
                                                  ,
                                          "   (method value: (y) (set value y))"
                                                  , ")"
                                                  , "; usm.nw 3446"
                                                  ,
                                             "(class Dictionary KeyedCollection"
                                                  ,
                                                   "    (table) ; list of pairs"
                                                  ,
                   "    (classMethod new ()       (initDictionary (new super)))"
                                                  ,
          "    (method initDictionary () (set table (new List)) self) ; private"
                                                  ,
                            "    (method printName ()      (print #Dictionary))"
                                                  ,
                                     "    (method species ()        Dictionary)"
                                                  , "    "
                                                  , "; usm.nw 3460"
                                                  ,
                          "(method associationsDo: (aBlock) (do: table aBlock))"
                                                  ,
                                "(method at:put: (key value) (locals tempassoc)"
                                                  ,
                     "    (set tempassoc (associationAt:ifAbsent: self key []))"
                                                  , "    (if (isNil tempassoc)"
                                                  ,
                    "         [(add: table (key:value: Association key value))]"
                                                  ,
                                          "         [(value: tempassoc value)])"
                                                  , "    value)"
                                                  , "; usm.nw 3471"
                                                  ,
              "(method add: (_) (error: self #can't-just-add:-to-a-Dictionary))"
                                                  , "; usm.nw 3453"
                                                  , ")"
                                                  , "; usm.nw 3484"
                                                  ,
                                 "(class SequenceableCollection KeyedCollection"
                                                  , "    () ; abstract class"
                                                  ,
                        "    (method firstKey () (subclassResponsibility self))"
                                                  ,
                        "    (method lastKey  () (subclassResponsibility self))"
                                                  ,
                           "    (method last     () (at: self (lastKey  self)))"
                                                  ,
                           "    (method first    () (at: self (firstKey self)))"
                                                  ,
        "    (method at:ifAbsent: (index exnBlock) (locals current resultBlock)"
                                                  ,
                                            "        (set resultBlock exnBlock)"
                                                  ,
                                         "        (set current (firstKey self))"
                                                  ,
                                                  "        (do: self (block (v)"
                                                  ,
               "            (ifTrue: (= current index) [(set resultBlock [v])])"
                                                  ,
                                     "            (set current (+ current 1))))"
                                                  ,
                                                  "        (value resultBlock))"
                                                  , ")"
                                                  , "; usm.nw 3906"
                                                  , "(class Cons Object"
                                                  , "    (car cdr)"
                                                  ,
                                             "    (method car ()           car)"
                                                  ,
                                             "    (method cdr ()           cdr)"
                                                  ,
                         "    (method car: (anObject)  (set car anObject) self)"
                                                  ,
                         "    (method cdr: (anObject)  (set cdr anObject) self)"
                                                  , "    "
                                                  , "; usm.nw 3921"
                                                  ,
                                        "(method deleteAfter () (locals answer)"
                                                  , "    (set answer (car cdr))"
                                                  , "    (set cdr    (cdr cdr))"
                                                  , "    (pred: cdr self)"
                                                  , "    answer)"
                                                  ,
                                               "(method insertAfter: (anObject)"
                                                  ,
                           "    (set cdr (car: (cdr: (new Cons) cdr) anObject))"
                                                  , "    (pred: (cdr cdr) cdr)"
                                                  , "    anObject)"
                                                  , "; usm.nw 3942"
                                                  , "(method do: (aBlock)"
                                                  , "    (value aBlock car)"
                                                  , "    (do: cdr aBlock))"
                                                  , "; usm.nw 3952"
                                                  ,
                   "(method rejectOne:ifAbsent:withPred: (aBlock exnBlock pred)"
                                                  ,
                                                   "    (if (value aBlock self)"
                                                  ,
                                                  "        [(deleteAfter pred)]"
                                                  ,
           "        [(rejectOne:ifAbsent:withPred: cdr aBlock exnBlock self)]))"
                                                  , "; usm.nw 3913"
                                                  ,
                                             "    (method pred: (aCons)    nil)"
                                                  , ")"
                                                  , "; usm.nw 3963"
                                                  , "(class ListSentinel Cons"
                                                  , "    (pred)"
                                                  ,
                                              "    (method do:   (aBlock)  nil)"
                                                  ,
               "    (method rejectOne:ifAbsent:withPred: (aBlock exnBlock pred)"
                                                  ,
                                 "                            (value exnBlock))"
                                                  ,
                                 "    (method pred: (aCons)   (set pred aCons))"
                                                  ,
                                             "    (method pred  ()        pred)"
                                                  ,
                                      "    (classMethod new ()     (locals tmp)"
                                                  ,
                                                 "        (set tmp (new super))"
                                                  , "        (pred: tmp tmp)"
                                                  , "        (cdr:  tmp tmp)"
                                                  , "        tmp))"
                                                  , "; usm.nw 3833"
                                                  ,
                                            "(class List SequenceableCollection"
                                                  , "    (sentinel)"
                                                  ,
   "    (classMethod new  ()        (sentinel: (new super) (new ListSentinel)))"
                                                  ,
              "    (method sentinel: (s)       (set sentinel s) self) ; private"
                                                  ,
                  "    (method isEmpty   ()        (= sentinel (cdr sentinel)))"
                                                  , "    "
                                                  , "; usm.nw 3852"
                                                  ,
               "(method addLast:  (item)   (insertAfter: (pred sentinel) item))"
                                                  ,
                      "(method addFirst: (item)   (insertAfter: sentinel item))"
                                                  ,
                              "(method add: (item)        (addLast: self item))"
                                                  , "; usm.nw 3859"
                                                  ,
                            "(method removeFirst ()     (deleteAfter sentinel))"
                                                  , "; usm.nw 3873"
                                                  ,
                                 "(method remove:ifAbsent: (oldObject exnBlock)"
                                                  ,
                                             "    (rejectOne:ifAbsent:withPred:"
                                                  , "        (cdr sentinel)"
                                                  ,
                                     "        (block (x) (= oldObject (car x)))"
                                                  , "        exnBlock"
                                                  , "        sentinel))"
                                                  , "; usm.nw 3886"
                                                  , "(method firstKey () 1)"
                                                  ,
                                              "(method lastKey  () (size self))"
                                                  ,
                                        "(method at:put: (n value) (locals tmp)"
                                                  ,
                                                  "    (set tmp (cdr sentinel))"
                                                  , "    (whileFalse: [(= n 1)]"
                                                  , "       [(set n (- n 1))"
                                                  ,
                                                 "        (set tmp (cdr tmp))])"
                                                  , "    (car: tmp value))"
                                                  , "; usm.nw 3839"
                                                  ,
                        "    (method last      ()        (car (pred sentinel)))"
                                                  ,
                  "    (method do:       (aBlock)  (do: (cdr sentinel) aBlock))"
                                                  ,
                                         "    (method species   ()        List)"
                                                  ,
                                "    (method printName ()        (print #List))"
                                                  , ")"
                                                  , "; usm.nw 4040"
                                                  ,
                                           "(class Array SequenceableCollection"
                                                  ,
                                          "    () ; representation is primitive"
                                                  ,
                                    "    (classMethod new: primitive arrayNew:)"
                                                  ,
       "    (classMethod new () (error: self #size-of-Array-must-be-specified))"
                                                  ,
                                     "    (method size     primitive arraySize)"
                                                  ,
                                      "    (method at:      primitive arrayAt:)"
                                                  ,
                                  "    (method at:put:  primitive arrayAt:Put:)"
                                                  , "    "
                                                  , "; usm.nw 4057"
                                                  ,
                         "(method add:             (x)   (fixedSizeError self))"
                                                  ,
                         "(method remove:ifAbsent: (x b) (fixedSizeError self))"
                                                  ,
         "(method fixedSizeError   ()    (error: self #arrays-have-fixed-size))"
                                                  , "; usm.nw 4064"
                                                  ,
         "(method select:  (_) (error: self #select-on-arrays-not-implemented))"
                                                  ,
        "(method collect: (_) (error: self #collect-on-arrays-not-implemented))"
                                                  , "; usm.nw 4070"
                                                  , "(method firstKey () 1)"
                                                  ,
                                              "(method lastKey  () (size self))"
                                                  ,
                                           "(method do: (aBlock) (locals index)"
                                                  ,
                                               "    (set index (firstKey self))"
                                                  ,
                                                 "    (timesRepeat: (size self)"
                                                  ,
                                       "       [(value aBlock (at: self index))"
                                                  ,
                                            "        (set index (+ index 1))]))"
                                                  , "; usm.nw 4048"
                                                  ,
                                               "    (method species   () Array)"
                                                  ,
                "    (method printName () nil) ; names of arrays aren't printed"
                                                  , ")"
                                                  ])
  in  readEvalPrint
        ( topreader (rdr, false)
        , echoBasis
        , fn s => app print ["error in initial basis: ", s, "\n"]
        )
        initialXi
  end
(* Before we can close the cycles, we have to   *)
(* create [[VAL]] bindings for [[true]] and     *)
(* [[false]]. We can't do this in uSmalltalk    *)
(* because the parser prevents user code from   *)
(* binding [[true]] and [[false]].              *)
(* <initialization>=                            *)
val initialXi =
      topeval (VAL ("true",  SEND(nullsrc, "new", VAR "True",  [])), false,
                                                                      initialXi)
val initialXi =
      topeval (VAL ("false", SEND(nullsrc, "new", VAR "False", [])), false,
                                                                      initialXi)
(* Once we've read the class definitions, we    *)
(* can close the cycles, update the ref cells,  *)
(* and we're almost ready to go. By this time,  *)
(* all the necessary classes must be defined,   *)
(* so if any cycle fails to close, we halt the  *)
(* interpreter with a fatal error. [*]).        *)
(* <initialization>=                            *)
val _ =
  ( closeLiteralsCycle initialXi
  ; closeBooleansCycle initialXi
  ; closeBlocksCycle   initialXi
  ) handle NotFound n =>
      ( app print ["Fatal error: ", n, " is not defined in the initial basis\n"]
      ; raise InternalError "this can't happen"
      )
  | e => ( print "Error binding basis classes into interpreter\n"; raise e)
(* The last step initialization is to bind in   *)
(* predefined values. The value [[nil]] is      *)
(* bound here because the parser won't let us   *)
(* create a [[val]] binding for it. The other   *)
(* values are symbols that are useful for       *)
(* printing, but that can't be expressed using  *)
(* uSmalltalk's literal symbol notation.        *)
(* <initialization>=                            *)
fun addVal ((name, v), xi) = topeval (VAL (name, VALUE v), false, xi)
val initialXi = foldl addVal initialXi
  [ ("nil", nilValue)
  , ("lparen", mkSymbol "("), ("rparen", mkSymbol ")")
  , ("lbrack", mkSymbol "["), ("rbrack", mkSymbol "]")
  , ("newline", mkSymbol "\n"), ("space",  mkSymbol " ")
  ]
(* <initialization>=                            *)
fun runInterpreter noisy = 
  let val treader = topreader (filereader ("standard input", TextIO.stdIn),
                                                                          noisy)
      fun writeln s = app print [s, "\n"]
  in  ignore (readEvalPrint (treader, true, writeln) initialXi) handle EOF => ()
  end 
(* The function [[runInterpreter]] takes one    *)
(* argument, which tells it whether to prompt.  *)
(* <boxed values 19>=                           *)
val _ = op runInterpreter : bool -> unit


(*****************************************************************)
(*                                                               *)
(*   COMMAND LINE                                                *)
(*                                                               *)
(*****************************************************************)

(* Building and exporting a program             *)
(*                                              *)
(* The final step in implementing the           *)
(* interpreter is to create a function that     *)
(* looks at the command line and calls          *)
(* [[runInterpreter]]. With a compiler like     *)
(* Moscow ML or MLton, the module isn't         *)
(* elaborated until run time, so we can simply  *)
(* insert an irrelevant [[val]] binding, the    *)
(* elaboration of which has the side effect of  *)
(* calling [[main]].                            *)
(* [[CommandLine.arguments()]] returns an       *)
(* argument list.                               *)
(* <command line>=                              *)
fun main ["-q"] = runInterpreter false
  | main []     = runInterpreter true
  | main _      =
      TextIO.output(TextIO.stdErr, "Usage: " ^ CommandLine.name() ^ " [-q]\n")
val _ = main (CommandLine.arguments())
(* [[                                           *)
(* ********************** ) ( * * ) ( *         *)
(* BOOTSTRAPPING SUPPORT * ) ( * * )            *)
(* ( ******************************************* *)
(* ********************** )                     *)
(*                                              *)
(* ]] provides both                             *)
(* the ref cells and the functions used to      *)
(* update them. Primitive methods and built-in  *)
(* objects depend on these ref cells, as does   *)
(* the parser. The evaluator is built on top of *)
(* everything else, and finally [[              *)
(*                                              *)
(* ( ******************************************* *)
(* ********************** ) ( * * ) ( *         *)
(* INITIALIZATION * ) ( * * )                   *)
(* ( ******************************************* *)
(* ********************** )                     *)
(*                                              *)
(* ]] reads the initial                         *)
(* basis, then calls the functions from [[      *)
(*                                              *)
(* ( ******************************************* *)
(* ********************** ) ( * * ) ( *         *)
(* BOOTSTRAPPING SUPPORT * ) ( * * )            *)
(* ( ******************************************* *)
(* ********************** )                     *)
(*                                              *)
(* ]] that close the                            *)
(* cycles.                                      *)

